/*------------------------------------------------------------------------
 * MAGNET Customer Agent Bid Manager components
 * $Id: BidCommand.java,v 1.2.2.1 2002/05/05 19:41:44 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Thu Sep 28 13:04:43 2000  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.bid;

import java.util.Collection;
import java.util.Iterator;
import java.util.HashSet;
import java.util.Date;

/**
 * Instances specify bidding cycles for the Bid Manager.  We assume
 * the Bid Manager has the necessary context in which to run this
 * command.
 * @author John Collins
 */
public class BidCommand extends BidAgendaItem
{
  // Tasks to be bid in this round.  We assume the tasks are already
  // decorated with their time windows.
  private Collection m_taskList = new HashSet();

  // Time parameters for this round of bidding
  private Date m_bidDeadline = null;
  private Date m_earlyAward = null;
  private Date m_awardDeadline = null;
  private Date m_endOfSlot = null;

  // Processing options
  private boolean runScheduler = true;

  /**
   * Default constructor lets you serialize these
   */
  public BidCommand ()
  {
    super();
  }

  /**
   * Normal constructor takes all the data needed to set one up.  Note
   * that the taskList argument is saved by reference, not copied.  If
   * you change the underlying list, this Command object will not
   * notice.
   */
  public BidCommand (Collection taskList,
		     Date bidDeadline,
		     Date earlyAward,
		     Date awardDeadline,
		     Date endOfSlot)
  {
    super();
    m_taskList = taskList;
    m_bidDeadline = bidDeadline;
    m_earlyAward = earlyAward;
    m_awardDeadline = awardDeadline;
    m_endOfSlot = endOfSlot;
  }

  /**
   * This constructor takes a flag to control the RFQ scheduler, in
   * addition to the task list and timing parameters.  Note
   * that the taskList argument is saved by reference, not copied.  If
   * you change the underlying list, this Command object will not
   * notice.
   */
  public BidCommand (Collection taskList,
		     Date bidDeadline,
		     Date earlyAward,
		     Date awardDeadline,
		     Date endOfSlot,
		     boolean runScheduler)
  {
    this(taskList, bidDeadline, earlyAward, awardDeadline, endOfSlot);
    this.runScheduler = runScheduler;
  }

  /**
   * When you execute one of these, it runs the BidManager's bidding
   * process for the specified tasks.
   */
  public void execute (BidManager mgr)
  {
    mgr.doBidCommand(this);
  }

  // -------------- Access to command context data --------------

  /**
   * Returns the task list as a HashSet.
   */
  public Collection getTaskList ()
  {
    return m_taskList;
  }

  /**
   * Returns the Bid Deadline
   */
  public Date getBidDeadline ()
  {
    return m_bidDeadline;
  }

  /**
   * Returns the Early Award time
   */
  public Date getEarlyAward ()
  {
    return m_earlyAward;
  }

  /**
   * Returns the Award Deadline time
   */
  public Date getAwardDeadline ()
  {
    return m_awardDeadline;
  }

  /**
   * Returns the end of the time slot for this action
   */
  public Date getEndOfSlot ()
  {
    return m_endOfSlot;
  }

  /**
   * Returns the "run scheduler" flag.
   */
  public boolean isRunScheduler ()
  {
    return runScheduler;
  }
}
