/*------------------------------------------------------------------------
 * Magnet customer agent components
 * $Id: BidManager.java,v 1.12.2.1 2002/05/05 19:41:45 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Wed May  9 13:56:17 2001  John Collins  #0000016
 *   Provide for external input of BidCommands, don't automatically
 *   search.
 * 
 * Sat May  5 12:53:33 2001  John Collins
 *   Restructure for new RFQStatus and RFQStatusEvent operation
 * 
 * Fri Dec 22 18:56:06 2000  John Collins
 *   Move management of search engine types to BidEvaluator
 * 
 * Wed Dec 13 21:59:06 2000  John Collins
 *   New TaskBidPlan API - switch from Enumerator to Collection
 * 
 * Wed Sep 27 13:31:21 2000  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.bid;

import edu.umn.magnet.protocol.MagnetException;
import edu.umn.magnet.protocol.JobInfo;
import edu.umn.magnet.protocol.InvalidDataException;
import edu.umn.magnet.protocol.cust.RFQ;
import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskPlan;

import edu.umn.magnet.client.eval.BidEvaluator;
import edu.umn.magnet.client.eval.SASearch;
import edu.umn.magnet.client.eval.SystematicSearch;
import edu.umn.magnet.client.eval.TaskBidPlan;
import edu.umn.magnet.client.eval.TaskBid;

import edu.umn.magnet.client.status.PlanStatus;
import edu.umn.magnet.client.status.TaskStatus;
import edu.umn.magnet.client.status.SessionStatus;
import edu.umn.magnet.client.status.RFQStatus;
import edu.umn.magnet.client.status.RFQStatusEvent;
import edu.umn.magnet.client.status.InvalidStateException;

import edu.umn.magnet.util.time.AgentTime;

import org.apache.log4j.Category;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Properties;
import java.util.Vector;
import java.util.Iterator;

/**
 * Instances of BidManager are responsible for managing the bidding
 * process for a single plan, through one or more market Sessions
 * (typically at most one per market).  Basically you create one, give
 * it a plan, time limits, and any run-time configuration information,
 * then send it off to do its thing.  It uses a search engine for
 * evaluation, and returns a TaskBidPlan that maps tasks to accepted
 * bids.</p>
 * <p>
 * In detail, the process goes like this:
 * <ul>
 *  <li>Create a new <tt>BidManager</tt> instance.
 *  <li>Fill in the <tt>sessionStatus</tt> and <tt>props</tt> 
 *      properties. A convenience constructor is provided that accepts
 *      these inputs as parameters.  Typically, the props would
 *      include the type of BidCollector to use.
 *  <li>Invoke the <tt>handleBids</tt> method, which spawns a new thread to
 *      process bids.  This thread will run until awards are let for
 *      all tasks, or until it the Bid Manager determines that the
 *      plan should be abandoned.
 *  <li>At any time, the <tt>getResults</tt> method may be called.
 *      This will return a <tt>TaskBidPlan</tt> that represents the
 *      current state of the bidding process.
 *  <li>The <tt>waitForCompletion</tt> method will block until the
 *      <tt>handleBids</tt> thread has completed its work.  At this
 *      point, the <tt>getResults</tt> method will return the final
 *      state of the bidding process. 
 * </ul></p>
 * <p>
 * <strong>NOTE:</strong> This version only knows how to deal with a
 * single session, and it assume that all tasks in the plan will be
 * handled through that session.  Extension to multi-market bidding
 * situations will require changes here as well as in TaskStatus and
 * probably the search engines.
 * </p>
 * @author John Collins
 */
public class BidManager extends Thread
{
  // init the log category
  static Category log = Category.getInstance(BidManager.class.getName());
  static Category resultLog = Category.getInstance("Result");

  // Configuration data
  private Properties m_props = null;
  private String m_sessionId = "";
  private boolean autopilot = false;
  private double planSlack = 1.2;
  //private TaskDurationAdjuster taskAdjuster = null;
  private PlanTWAdjuster planAdjuster = null;
  private BidScheduler scheduler = null;

  // Instance context
  private List m_bidAgenda = null;
  private TaskPlan m_initialPlan = null;
  private PlanStatus m_planStatus = null;
  private SessionStatus m_sessionStatus = null;
  private TaskBidPlan m_currentPlan = null;
  private int m_nodeCount = 0;
  private boolean m_canProceed = true; // set false to abort

  // --------------- Constructors -----------------

  /**
   * Here's a default constructor in case you want to serialize one of
   * these things or something.  If you call this, you then need to
   * call the {@link #init init} method before anything can happen in
   * here.
   */
  public BidManager ()
  {
    super();
  }

  /**
   * A constructor that sets everything up and, if <tt>start</tt> is
   * true, starts the BidManager thread running.  If you set
   * <tt>start</tt> to false, you have to call {@link #handleBids
   * handleBids} to start the thread.
   */
  public BidManager (SessionStatus session,
		     Properties props,
		     boolean start)
  {
    super();
    this.init(session, props);

    if (start)
      this.handleBids();
  }

  /**
   * Initializes the BidManager.  You only get to do this once, and it
   * gets called automatically if you use the non-default constructor.
   * The following properties are recognized:
   * <dl>
   * <dt><strong><tt>BidManager.PlanSlack</tt></strong></dt>
   *   <dd>The amount of overall slack to add to the plan, expressed
   *     as a ratio.  In other words, if you want 20% overall slack,
   *     you specify <tt>BidManager.PlanSlack</tt> as 1.2.</dd>
   * <dt>BidManager.PlanDeadline</dt>
   *   <dd></dd>
   * <dt>BidManager.Autopilot</dt>
   * </dl>
   */
  public void init (SessionStatus session, Properties props)
  {
    if (m_sessionStatus != null) {
      log.error("Attempt to init BidManager twice");
      return;
    }
    m_sessionStatus = session;
    m_sessionId = session.getSessionId();
    this.setPlanStatus(session.getParent());
    m_props = props;
    
    // Pull out the properties
    try {
      planSlack = (Double.valueOf(m_props.getProperty("PlanSlack", "1.2")))
	  .doubleValue();
      log.info("prop: PlanSlack = " + planSlack);
    }
    catch (NumberFormatException nfe) {
      log.error("Error in property file. Exiting", nfe);
      System.exit(1);
    }
    autopilot
	= (new Boolean(m_props.getProperty("BidManager.Autopilot")))
	.booleanValue();
    log.info("prop: BidManager.Autopilot = " + autopilot);
    
    // Use the static BidScheduler factory to generate our bid scheduler.
    scheduler = BidScheduler.generate(m_props);
    
    // Use the static TaskDurationAdjuster factory to generate the
    // adjustment operator.
    TaskDurationAdjuster taskAdjuster
	= TaskDurationAdjuster.generate(m_props);
    taskAdjuster.setPlanStatus(m_planStatus);
    planAdjuster = PlanTWAdjuster.generate(m_props);
    planAdjuster.setTaskDurationAdjuster(taskAdjuster);
  }

  //---------------- Properties ------------------

  /**
   * Gets the initialPlan property.  Value is a TaskPlan.  You don't
   * get to set it.
   */
  public TaskPlan getInitialPlan ()
  {
    return m_initialPlan;
  }

  /**
   * Gets the sessionStatus property.
   */
  public SessionStatus getSessionStatus ()
  {
    return m_sessionStatus;
  }

  /**
   * Gets the planStatus property.  Value is a PlanStatus.
   */
  public PlanStatus getPlanStatus ()
  {
    return m_planStatus;
  }

  /**
   * Sets the planStatus property.  You only get to do this once, for
   * now.  After that, this method just silently does nothing
   * (except emit a nasty log entry).
   */
  private void setPlanStatus (PlanStatus newPlan)
  {
    if (m_planStatus != null) {
      // This isn't the first time.  Complain.
      log.error("Attempt to re-set the initial plan");
      return;
    }
    m_planStatus = newPlan;
    m_currentPlan = newPlan.getTaskBidPlan();
    m_initialPlan = newPlan.getTaskPlan();
  }

  /**
   * Gets the startAfter property
   */
  public Date getStartAfter ()
  {
    return m_planStatus.getPlanStart();
  }

  /**
   * Gets the finishBefore property
   */
  public Date getFinishBefore ()
  {
    return m_planStatus.getPlanDeadline();
  }

  /**
   * Gets the Properties containing configuration info.
   */
  public Properties getProps ()
  {
    return m_props;
  }

  /**
   * Sets the configuration info.  Note that changes to this property
   * will not be noticed until the next time you call
   * <tt>handleBids</tt>.
   */
  public void setProps (Properties newProps)
  {
    m_props = newProps;
  }

  /**
   * Completes the BidManager's work by setting a new currentPlan and
   * setting the tasks in the current plan to the PENDING state in
   * preparation for generating BidAwards.
   */
  public void setCurrentPlan (TaskBidPlan newPlan)
  {
    m_currentPlan = newPlan;
    m_planStatus.setTaskBidPlan(newPlan);
      
    // Finally, we update the status of the tasks that were bid.  If
    // any of them does not have a bid, then we lack coverage and
    // must abandon the plan.
    if (m_currentPlan != null
	&& m_currentPlan.isFeasible()
	&& m_currentPlan.isCovered()) {
      // The TaskBid instances contain the mapping of tasks to bids.
      // For every task that was offered and for which there is now
      // a bid, we make it pending
      for (Iterator tbs = m_currentPlan.getTaskBids().iterator();
	   tbs.hasNext(); ) {
	TaskBid tb = (TaskBid)tbs.next();
	if (tb.getBid() != null
	    && m_planStatus.getTaskState
	    (tb.getTask()).equals(TaskStatus.OFFERED)) {
	  log.info("Setting task " + tb.getTaskId() + " to PENDING");
	  m_planStatus.setTaskState(tb.getTask(),
				    TaskStatus.PENDING);
	}
      }
    }
  }

  // ------------------- Bid Processing -------------------

  /**
   * Come here to start the Bid Manager off on its work.  It will fire
   * off a thread and return immediately.
   */
  public void handleBids ()
  {
    log.info("Starting Bid Manager");
    this.start();
  }

  /**
   * Come here to block until the Bid Manager has finished its work.
   */
  public void waitForCompletion ()
  {
    try {
      this.join();
    }
    catch (InterruptedException ie) {
      // Guess we don't have to do much here...
    }
    log.info("Bid Manager finished");
  }

  /**
   * This is the body of the Bid Manager thread.  It has to be public
   * to satisfy the Thread protocol, but it should not be called
   * directly.  Use the <tt>handleBids</tt> method for that.
   */
  public void run ()
  {

    // Schedule the bids.  Return value is the agenda.
    m_bidAgenda = scheduler.scheduleBids(m_currentPlan);

    // Process the bid agenda.  We treat the list as a queue, allowing
    // operations that modify it.
    resultLog.info("-------- Start bid manager session "
		   + m_sessionId + " --------");

    // Log the plan statistics to the result log
    m_initialPlan.logPlanStats(resultLog);

    while (!m_bidAgenda.isEmpty() && m_canProceed) {
      BidAgendaItem command = (BidAgendaItem)m_bidAgenda.remove(0);
      command.execute(this);
    }
    // at this point, we have a plan, or we don't.  Evaluate it
    // and report results.
    reportResults(m_currentPlan);
  }

  /**
   * Process a BidCommand.  Currently, we treat the bidding process as
   * synchronous.  In other words, we send out the RFQ and then block
   * until we get back the bids after the bid deadline.
   */
  public void doBidCommand(BidCommand cmd)
  {
    // The BidCommand tells us which tasks are being bid.  Update the
    // task status for those tasks to OFFERED.
    log.debug("doBidCommand");

    RFQ rfq = composeRFQ(cmd);

    // Here, we submit the RFQ to the supply monster.  This will
    // have to be updated to deal with real suppliers and
    // asynchronous bid arrival.
    log.debug("Submitting CFB...");

    RFQStatus rfqStatus = null;
    try {
      rfqStatus = createRFQStatus(rfq);
    }
    catch (MagnetException me) {
      // no need to do anything here, the exception has already been
      // logged.  But we will anyway....
      log.error("Failed to create RFQStatus, bailing: " + me);
    }
    if (rfqStatus != null) {      
      // Now we set up the BidEvaluator so it can intercept the
      // incoming bids, then we activate the RFQ.

      log.info("New search: session ID = " + m_sessionId);
      BidEvaluator evaluator = new BidEvaluator();
      evaluator.setProperties(m_props);
      evaluator.setSessionId(m_sessionId);
      evaluator.setPlanStatus(m_planStatus);
      evaluator.setRFQStatus(rfqStatus);
      // go get the bids, and block until the result is ready
      try {
	rfqStatus.acquireBids();
      }
      catch (InvalidStateException ise) {
	// shouldn't happen...
	log.error("Unexpected problem: " + ise);
	m_canProceed = false;
      }
      TaskBidPlan newPlan = (evaluator.getResult());

      setCurrentPlan(newPlan);
      if (m_currentPlan == null) {
	// This is how we abort the bid manager thread
	m_canProceed = false;
      }
    }
    else {
      log.error("Failed to create RFQStatus instance");
      m_canProceed = false;
    }
  }

  private void summarizeBids (RFQStatus rfqStatus)
  {
    Collection bidv = rfqStatus.getBids();
    int bidCount = bidv.size();
    log.info("Bid count: " + bidCount);
    if (log.isInfoEnabled())
      printBids(bidv);
    int totalSize = 0;
    for (Iterator bids = bidv.iterator();
	 bids.hasNext(); ) {
      Bid theBid = (Bid)bids.next();
      totalSize += theBid.getTaskSet().size();
    }
    resultLog.info("BidCount: " + bidCount + ", avgSize: "
		   + ((double)totalSize/(double)bidCount));
  }

  private void reportResults (TaskBidPlan currentPlan)
  {
    if (currentPlan != null) {
      // we've got something to show.
      log.warn("Solution Found");
      // We'll do our own computation here, since the evaluation done
      // by the BidEvaluator is for its own purposes, and may only
      // include the tasks being bid in the last cycle.
      int cost = 0;
      for (Iterator bids = currentPlan.getBids().iterator();
	   bids.hasNext(); ) {
	cost += ((Bid)bids.next()).getCost();
      }
      long finish = 0L;
      for (Iterator tbs = currentPlan.getTaskBids().iterator();
	   tbs.hasNext(); ) {
	TaskBid tb = (TaskBid)tbs.next();
	long complete = tb.getEarlyStartCPM().getTime()
	                + tb.getDuration();
	if (complete > finish)
	  finish = complete;
      }
      log.info("  FinalEval<" + m_sessionId + ">: cost = " + cost
	       + ", compl = " + (finish - getStartAfter().getTime())
	       + ", nc = " + m_nodeCount);
      log.info(currentPlan.toString());
    }
    else {
      // we don't
      log.warn("  FinalEval<" + m_sessionId
	       + ">: plan abandoned, nc =" + m_nodeCount);
    }
  }

  /**
   * Sets time windows and generates the RFQ instance.
   */
  public RFQ composeRFQ (BidCommand cmd)
  {
    // First, make all the tasks OFFERED
    for (Iterator tasks = cmd.getTaskList().iterator();
	 tasks.hasNext(); ) {
      m_planStatus.setTaskState((SubtaskRequest)tasks.next(),
				TaskStatus.OFFERED);
    }
    
    if (cmd.isRunScheduler()) {
      // Set the time windows - we assume any updates due to earlier
      // bids are already noted in the TaskPlan and SubtaskRequest
      // instances.
      //PlanTWAdjuster planAdjuster = new PlanTWAdjuster();
      planAdjuster.setTimeWindows(m_currentPlan, getStartAfter(),
				  planSlack);
    }

    // Now we need a TaskPlan that has only the tasks we are bidding.
    // That's what we put into the RFQ
    TaskPlan thisPlan = new TaskPlan(m_planStatus.getTaskPlan(),
				     cmd.getTaskList());
    
    if (log.isInfoEnabled()) {
      for (Iterator en = thisPlan.iterator();
	   en.hasNext(); ) {
	SubtaskRequest task = (SubtaskRequest)en.next();
	Date d = task.getEarliestTimeToStart();
	Date d2 = task.getDeadline();
	log.info(task.getId() + ": " + d + " " + d2);
      }
    }

    RFQ cfb = new RFQ();
    try {
      cfb.setId("CFB" + m_sessionId);
      cfb.setAgentId("Experiment");
      cfb.setTaskPlan(thisPlan);
      cfb.setBidDeadline(cmd.getBidDeadline());
      cfb.setEarliestOffer(cmd.getEarlyAward());
      cfb.setEarliestConsider(cmd.getBidDeadline());
    }
    catch (InvalidDataException ide) {
      log.error("Exception filling out RFQ: " + ide.toString());
      return null;
    }
    return cfb;
  }

  /**
   * Creates the RFQStatus instances needed to control the bidding
   * process for a single RFQ.
   * @exception InvalidDataException thrown if the RFQ is not valid in
   *   the current environment.
   */
  public RFQStatus createRFQStatus (RFQ rfq)
    throws InvalidDataException, InvalidStateException
  {

    RFQStatus rfqStatus
	= m_sessionStatus.createRFQStatus(rfq);
    // this is done in two steps to allow event distribution to happen
    // between them, and to handle errors.
    if (rfqStatus != null) {
      // register for bid arrival and state change events
      // one copy here, one copy to the BidEvaluator
      rfqStatus.addListener("State", new PropertyChangeListener () {
	  public void propertyChange(PropertyChangeEvent pce) {
	    if (pce instanceof RFQStatusEvent
		&& pce.getNewValue().equals(RFQStatus.BIDS_ACQUIRED))
	      summarizeBids((RFQStatus)pce.getSource());
	  }
	});
    }
    return rfqStatus;
  }

  private void printBids (Collection bidv)
  {
    for (Iterator bids = bidv.iterator();
	 bids.hasNext(); ) {
      Bid nextBid = (Bid)bids.next();
      nextBid.logVerbose(log);
    }
  }
}
