/*------------------------------------------------------------------------
 * MAGNET Customer Agent Bid Management components
 * $Id: BidScheduler.java,v 1.4.2.1 2002/05/05 19:41:45 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Thu Jan 11 13:15:00 2001  John Collins
 *   use new logging system
 * 
 * Thu Sep 28 10:25:13 2000  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.bid;

import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.client.eval.TaskBidPlan;

import edu.umn.magnet.util.time.AgentTime;

import org.apache.log4j.Category;

import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.LinkedList;
import java.util.Properties;

/**
 * Instances are responsible for evaluating a bidding situation and
 * producing a set of BidAgendaItem instances for controlling the
 * actual bidding process.  Each subclass implements a different
 * scheduling strategy.  This default implementation just generates a
 * single agenda item that causes the whole plan to be bid in one
 * step.
 * @author John Collins
 */
public class BidScheduler
{
  static Category log = Category.getInstance(BidScheduler.class.getName());

  /**
   * Default constructor doesn't need to do much.
   */
  public BidScheduler ()
  {
    super();
  }

  /**
   * All BidScheduler subclasses must implement an init method that
   * takes a set of Properties.  The default BidScheduler does not use
   * any properties. 
   */
  public void init (Properties props)
  {
  }

  /**
   * This type of scheduler just returns an agenda with a single
   * command to bid out all the tasks at once.
   */
  public List scheduleBids (TaskBidPlan tbPlan)
  {
    LinkedList agenda = new LinkedList();

    // The task list is just the list of tasks in the TaskPlan
    TaskPlan taskPlan = tbPlan.getTaskPlan();
    Collection taskList = taskPlan.getTasks();

    // We punt on the time allocation - eventually this will probably
    // come from a Properties instance or something.
    Date now = AgentTime.getDate();
    //long chunk = 600000L; // 10 minutes
    long chunk = 30000L; // 30 seconds
    Date bidDeadline = new Date(now.getTime() + chunk);
    Date earlyAward = new Date(bidDeadline.getTime() + chunk);
    Date awardDeadline = new Date(earlyAward.getTime() + chunk);
    Date endOfSlot = new Date(awardDeadline.getTime());

    if (taskList.size() > 0) {
      agenda.add(new BidCommand(taskList,
				bidDeadline, earlyAward, awardDeadline,
				endOfSlot));
    }
    return agenda;
  }

  /**
   * Static factory for constructing a BidSchedule of the selected
   * type.  Input is a Properties object that may specify the type of
   * schedule using the BidScheduler property.
   */
  public static BidScheduler generate (Properties props)
  {
    String schedulerType = props.getProperty("BidScheduler.Type", null);
    log.info("prop: Bid Scheduler type = "
	     + schedulerType);
    BidScheduler result = null;
    if (schedulerType == null) {
      log.warn("Scheduler type not specified.  Using default");
      result = new BidScheduler();
    }
    else {      
      try {
	String classname = "edu.umn.magnet.client.bid." + schedulerType;
	Class target = Class.forName(classname);
	result = (BidScheduler)target.newInstance();
      }
      catch (ClassNotFoundException cnf) {
	log.error("specified scheduler not found.  Using default");
	result = new BidScheduler();
      }
      catch (InstantiationException ie) {
	log.error("specified scheduler can't be instantiated.  Using default");
	result = new BidScheduler();
      }
      catch (IllegalAccessException iae) {
	log.error("Illegal access exception.  Using default");
	result = new BidScheduler();
      }
    }
    result.init(props);
    return result;
  }
}




