/*------------------------------------------------------------------------
 * Magnet client evaluation components
 * $Id: BidEvaluator.java,v 1.14.2.3 2002/05/05 19:41:46 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Fri Oct  5 11:27:55 2001  John Collins
 *   Performance improvements
 * 
 * Thu May 10 21:54:37 2001  John Collins  #0000008
 *   Simplify interactive operation by:
 *   - separate generation of coverage and feasibility constraints
 *   - only generate feasibility constraints if requested, by lazy
 *     evaluation.
 *   - add support for a new property BidEval.TimeLimit to limit
 *     search duration.
 *
 * Mon May 8 10:38:13 2001 Paula Greve
 *  Added constants for search types
 *  Allowed m_searchEngineType to be used if the m_requestedSearches was null
 *
 * Sat May  5 12:53:13 2001  John Collins
 *   Adaptations for new RFQStatus setup
 *
 * Fri Jan 12 17:23:45 2001  John Collins
 *   Add singleton and coverage testing, factoring out similar but
 *   less complete functionality from AbstractSearchEngine
 *
 * Fri Dec 29 13:51:31 2000  John Collins
 *   Add compatibility constraints
 *
 * Fri Dec 22 19:53:03 2000  John Collins
 *   Move management of search engine type, including operation of
 *   multiple search engines, into here from BidManager
 *
 * Wed Dec 13 11:07:14 2000  John Collins
 *   Remove SubtaskBids
 *
 * Tue Oct  3 10:07:05 2000  John Collins
 *   Get rid of bogus average price computation.
 *
 * Sun Oct  1 16:18:51 2000  John Collins
 *   Deal with multi-stage bidding.
 *
 * Mon Sep 25 11:11:27 2000  John Collins
 *   Clean up management of search type, including elimination of
 *   evaluateBids2 and related code.
 *
 * Tue Dec  7 11:27:59 1999  John E. Collins
 *   Add session ID property
 *
 * Sat Mar 27 13:15:00 1999  Rashmi Sundareswara
 *   Added a  evaluateBids2 method so that its argument is an instance of the
 *   Eval class
 *
 * Fri Jan 15 23:22:39 1999  John E. Collins
 *   Index Bids by SubtaskBids as they come in.
 *
 * Sun Dec 20 21:44:26 1998  John E. Collins
 *   Initial Entry.  Material migrated from AutoCustomer and SASearch.
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.protocol.cust.BidExaminer;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.InvalidDataException;

import edu.umn.magnet.client.status.AbstractStatus;
import edu.umn.magnet.client.status.PlanStatus;
import edu.umn.magnet.client.status.TaskStatus;
import edu.umn.magnet.client.status.RFQStatus;
import edu.umn.magnet.client.status.RFQStatusEvent;

import edu.umn.magnet.util.Cons;
//import edu.umn.magnet.util.stats.DataCollector;

import org.apache.log4j.Category;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Set;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;
import java.util.Properties;
import java.util.StringTokenizer;

/**
 * A BidEvaluator is the top-level object for bid evaluation.  You
 * create one to handle a given bidding cycle, hand it the bids, and
 * ask it for an evaluation of those bids.  It serves up the "model"
 * to the search and evaluation components that run under its control,
 * by providing a number of methods for querying and evaluating the
 * bids, tasks, and plan.  The actual
 * evaluation search is delegated to a SearchEngine object, of which
 * SASearch is one type.
 * <p>
 * <strong>Future:</strong> This class should be broken into
 * subclasses to implement the different bid-analysis procedures, with
 * a factory that gives you the right one depending on configuration
 * parameters.
 * </p>
 * @author John Collins
 */
public class BidEvaluator extends AbstractStatus
{
  // init the log category
  static Category log = Category.getInstance(BidEvaluator.class.getName());
  static Category resultLog = Category.getInstance("Result");

  // Some constants to help set search engine types
  /** Specifies the systematic search (not a good choice except for
   * very small problems) */
  public static final String SYSTEMATIC = "systematic";

  /** Specifies the IP Search (not a good choice except for very
   * small problems - use the IPCON search instead) */
  public static final String IP = "IP";

  /** Specifies the IP Con Search (our best fully-optimizing
   * search, requires significant preprocessing) */
  public static final String IPCON = "IPCon";

  /** Specifies the SA Search (a good stochastic anytime search
   * method) */
  public static final String SA = "SA";

  /** Specifies the Bidtree-based A* search */
  public static final String ASTAR = "Astar";

  /** Specifies the Bidtree-based IDA* search */
  public static final String IDASTAR = "IDAstar";

  /** Specifies the Bidtree-based Iterative-Deepening fringe search */
  public static final String IDFRINGE = "IDFringe";

  // Event types
  /** Event label: new search result */
  public static String NEW_RESULT = "NewResult";
  /** Event label: search completed */
  public static String SEARCH_COMPLETE = "SearchComplete";
  /** Event label: search aborted */
  public static String SEARCH_ABORTED = "SearchAborted";

  // Local attributes:
  //  Bid data:
  //   The set of bids available
  private ArrayList m_bids = new ArrayList();

  //  Constraint usage
  private boolean m_useCoverage = true;
  private boolean m_useFeasibility = true;
  private boolean m_hasCoverageConstraints = false;

  //   For each bid, the set of bids with which it is pairwise
  //   compatible
  private HashMap m_conflictingBids = new HashMap();
  //   For each bid, sets of size >= 2 of bids that are mutually
  //   incompatible, in which that bid participates.
  private HashSet m_infeasibleBidSets = null;

  // Cost data on bids
  private double m_maxTaskPrice = 0.0;

  //  Task/Plan data:
  private TaskPlan m_taskPlan = null;
  private TaskBidPlan m_initialTBP = null;
  private HashMap m_taskDescriptors = new HashMap();
  private RFQStatus m_rfqStatus = null;

  //   The set of tasks that was offered in the current bid cycle
  private HashSet m_offeredTasks = new HashSet();

  //  Search data:
  private Properties m_props = null;
  private String m_searchEngineType = "SA";
  private boolean m_optimizeCompletion = false;
  private LinkedList m_requestedSearches = new LinkedList();
  private String m_sessionId = "default_session";
  private TaskBidPlan m_bestPlan = null;
  private boolean m_evalComplete = false;
  private boolean m_searchAborted = false;

  // -------------- Setup --------------

  /**
   * Constructs a generic BidEvaluator.  Properties can be set
   * bean-style.
   */
  public BidEvaluator ()
  {
    super();
  }

  /**
   * Sets the PlanStatus, which extracts the current TaskBidPlan, the
   * current set of offered tasks, and creates TaskDescriptors for
   * them.
   */
  public void setPlanStatus (PlanStatus status)
  {
    // extract and copy the current TBP
    m_initialTBP = new TaskBidPlan(status.getTaskBidPlan());
    m_taskPlan = m_initialTBP.getTaskPlan();

    // Create the offered task set.  This is also the initial list of
    // uncovered tasks.
    Vector uncoveredTasks = new Vector();
    for (Iterator tasks = m_taskPlan.iterator();
	 tasks.hasNext(); ) {
      SubtaskRequest task = (SubtaskRequest)tasks.next();
      if (status.getTaskState(task).equals(TaskStatus.OFFERED)) {
	m_offeredTasks.add(task);
      }
    }

    // Create a TaskDescriptor for every offered task
    for (Iterator tasks = getOfferedTasks();
	 tasks.hasNext(); ) {
      SubtaskRequest st= (SubtaskRequest)tasks.next();
      TaskDescriptor newTD = new TaskDescriptor(st);
      m_taskDescriptors.put(st, newTD);
    }
  }

  /**
   * Sets the Session ID for the current session.  This is used to
   * name things, like the data sets that get produced from the search
   * engine.
   */
  public void setSessionId (String newId)
  {
    m_sessionId = newId;
  }

  /**
   * Sets the rfqStatus that is managing the bidding process for this
   * cycle.  We have to process NewBid and State events coming out of
   * it.
   * <p>
   * You only do this if you want this Evaluator to automatically
   * receive bids as then arrive and evaluate them when the
   * BIDS_ACQUIRED state is reached.  Don't call this if you want to
   * use this Evaluator interactively.
   * </p>
   */
  public void setRFQStatus (RFQStatus rs)
  {
    m_rfqStatus = rs;
    rs.addListener("NewBid", new PropertyChangeListener() {
	public void propertyChange (PropertyChangeEvent pce) {
	  if (pce instanceof RFQStatusEvent) {
	    Bid newBid = (Bid)pce.getNewValue();
	    log.debug("New Bid Event: " + newBid);
	    addBid(newBid);
	  }
	}
      });
    rs.addListener("State", new PropertyChangeListener() {
	public void propertyChange (PropertyChangeEvent pce) {
	  if (pce instanceof RFQStatusEvent
	      && pce.getNewValue().equals(RFQStatus.BIDS_ACQUIRED)) {
	    // we have to do this in a thread to avoid blocking the
	    // RFQ event distribution process.
	    startSearch();
	  }
	}
      });
  }

  /**
   * Adds a bid to this Evaluator, checks for bid compatibility, and
   * update the TaskDescriptors affected by this Bid.  Note that this
   * only works correctly if the PlanStatus has already been set.
   */
  public void addBid (Bid theBid)
  {
    log.debug("add bid: " + theBid.getId());
    m_bids.add(theBid);
    for(Iterator taskIds = theBid.getTaskSet().iterator();
	taskIds.hasNext(); ) {
      String taskId = (String)taskIds.next();
      TaskDescriptor td = findTaskDescriptor(taskId);
      if (td != null)
	td.addBid(theBid);
      else
	log.error("Null task descriptor for task "
		  + taskId + " in addBid!");
    }
    double pricePerTask = ((double)theBid.getCost()
			   / (double)theBid.getTaskSet().size());
    if (pricePerTask > m_maxTaskPrice) 
      m_maxTaskPrice = pricePerTask;    
  }

  /**
   * Removes a bid from this Evaluator, and updates all the
   * TaskDescriptors affected by the Bid.
   */
  public void removeBid (Bid theBid)
  {
    log.debug("remove bid: " + theBid.getId());
    m_bids.remove(theBid);
    for(Iterator taskIds = theBid.getTaskSet().iterator();
	taskIds.hasNext(); ) {
      String taskId = (String)taskIds.next();
      TaskDescriptor td = findTaskDescriptor(taskId);
      if (td != null)
	td.removeBid(theBid);
      else
	log.error("Null task descriptor for task "
		  + taskId + " in removeBid!");
    }
  }

  /**
   * Returns the average price/task.  This number is computed as bids
   * arrive, and is not corrected if bids are removed.
   */
  public double getMaxTaskPrice ()
  {
    log.info("Average task price = " + m_maxTaskPrice);
    return m_maxTaskPrice;
  }

  /**
   * Sets the evaluation control properties.
   */
  public void setProperties (Properties props)
  {
    m_props = props;
    String prop;
    // use coverage constraints?
    prop = props.getProperty("Eval.UseCoverageConstraints");
    if (prop != null && prop.equals("false"))
      m_useCoverage = false;

    // use feasibility constraints?
    prop = props.getProperty("Eval.UseFeasibilityConstraints");
    if (prop != null && prop.equals("false"))
      m_useFeasibility = false;

    // optimize completion time?
    prop = props.getProperty("Eval.OptimizeCompletion");
    if (prop != null && prop.equals("true"))
      m_optimizeCompletion = true;

    // process search engine properties
    setSearchProperty();
  }

  /**
   * Sets the search-type property string.  This will eventually be
   * replaced with the XML config stuff.  As a property, it is a
   * string containing a list of space-separated search specs.
   */
  public void setSearchProperty ()
  {
    String prop = m_props.getProperty("Search");
    m_requestedSearches.clear();
    if (prop != null) {
      for (StringTokenizer searchTypes = new StringTokenizer(prop);
	   searchTypes.hasMoreTokens(); ) {
	m_requestedSearches.add(searchTypes.nextToken());
      }
    }
  }

  /**
   * Sets the type of search engine.  If set to one of the class defined
   * strings, then that search will be used, otherwise it will be
   * simulated annealing.
   */
  public void setSearchEngineType (String sType)
  {
    m_searchEngineType = sType;
  }

  // ------------ Bid Evaluation --------------

  /**
   * Evaluates the bids received so far.  This is done in two
   * steps:
   * <ol>
   *  <li> Preprocess the bids to extract the constraints.
   *  <li> For each of the specified search types, we create and
   *       invoke a SearchEngine of that type, reporting results and
   *       keeping the best result found.
   * </ol>
   * <p>
   * We generate two types of constraints:
   * <ol>
   * <li>
   *   Coverage constraints represent 1-of-n choices.  We represent
   *   them pairwise here for convenience, but the basic form of these
   *   constraints is just the set of bids on a given TaskDescriptor.
   * <li>
   *   Feasibility constraints represent sets of two or more bids that
   *   specify infeasible time windows for tasks.  Our method below
   *   does not find all infeasible sets, only the ones on paths
   *   originating in a root task (a task without predecessors).
   * </ol>
   * The reason for distinguishing between these is that we represent
   * them differently.  The hash map <tt>m_conflictingBids</tt>
   * contains, for each bid, the set of other bids that are pairwise
   * incompatible with it.  The hash set
   * <tt>m_infeasibleBidSets</tt> contains the sets of bids that have
   * been found to form infeasible combinations.
   * </p>
   */
  public TaskBidPlan evaluateBids ()
  {
    log.debug("evaluateBids");
    m_searchAborted = false;

    // Set up completion optimization, if requested.
    if (isOptimizingCompletion())
      setupCompletionOptimization();

    // Preprocess the bids to induce constraints.
    // Here we only do the coverage constraints, since the feasibility
    // constraints are much more expensive and aren't needed by all
    // search types.
    if (! checkCoverage()) {
      searchComplete(null);
      return null;
    }

    // Make sure there's at least one search engine selected
    if (m_requestedSearches.size() == 0)
      m_requestedSearches.add(m_searchEngineType);

    // Iterate over search types
    TaskBidPlan bestPlan = null;
    for (Iterator engines = m_requestedSearches.iterator();
	 engines.hasNext(); ) {
      String searchType = (String)engines.next();
      log.info("Search type: " + searchType);
      SearchEngine engine = chooseSearchEngine(searchType);
      resultLog.info("** Session " + m_sessionId
		     + ": search type = " + searchType);
      TaskBidPlan newPlan = runSearch(engine);
      //m_nodeCount += evaluator.getNodeCount();
      if (newPlan != null
	  && (bestPlan == null
	      || newPlan.getEvaluation() < bestPlan.getEvaluation()))
	bestPlan = newPlan;
    }
    //if (bestPlan != null)
    //  bestPlan.clearParent();
    searchComplete(bestPlan);
    return bestPlan;
  }

  /**
   * Starts a search by calling {@link #evaluateBids evaluateBids} in
   * a new thread.
   */
  public void startSearch ()
  {
    Thread evaluator = new Thread() {
	public void run () {
	  evaluateBids();
	}
      };
    evaluator.start();
  }

  /**
   * Retrieves the best plan found so far by the search.  If the search has
   * not completed, this value may be null or less than optimal.  Of
   * course, if you aren't running an optimizing search, it may still
   * be less than optimal even if the search has completed.
   */
  public synchronized TaskBidPlan getBestPlan ()
  {
    return m_bestPlan;
  }

  /**
   * Sets the best-plan value and notifies anyone who is waiting for it.
   */
  private synchronized void searchComplete (TaskBidPlan bestPlan)
  {
    m_bestPlan = bestPlan;
    if (m_searchAborted) {
      log.debug("Search Aborted event");
      m_evalComplete = true;
      sendStatusEvent(new BidEvaluatorEvent(this, SEARCH_ABORTED,
					    null, bestPlan));
    }
    else {
      log.debug("Search Complete event");
      m_evalComplete = true;
      sendStatusEvent(new BidEvaluatorEvent(this, SEARCH_COMPLETE,
					    null, bestPlan));
      log.debug("sent Search Complete event");
    }
    notifyAll();
  }

  /**
   * Waits for the evaluation result, then returns it.
   */
  public synchronized TaskBidPlan getResult ()
  {
    try {
      while (m_evalComplete == false)
	wait();
    }
    catch (InterruptedException ie) {
      log.warn("getResult was interrupted");
      return null;
    }
    log.debug("Result returned");
    return m_bestPlan;
  }

  /**
   * Used by search engines to send notification of a new search
   * result.
   */
  void notifyNewResult (TaskBidPlan result)
  {
    m_bestPlan = result;
    sendStatusEvent(new BidEvaluatorEvent(this, NEW_RESULT, null, result));
  }

  /**
   * Runs a single search, timing it and generating the necessary log
   * output.
   */
  public TaskBidPlan runSearch (SearchEngine engine)
  {
    long runTime = 0L;
    engine.setInitialNode(m_initialTBP);
    runTime = System.currentTimeMillis();
    TaskBidPlan newPlan = engine.search(runTime);
    runTime = System.currentTimeMillis() - runTime;
    log.info("Search result:");
    String resultString = ((newPlan == null) ? "failed"
			   : newPlan.toString());
    log.info(resultString);
    resultLog.info("  Result = " + resultString
		   + ", time = " + runTime);
    // Put out the first plan, if any
    TaskBidPlan first = engine.getFirstPlan();
    if (first != null)
      resultLog.info("  First = " + first.toString());
    return newPlan;
  }

  /**
   * Aborts the currently-running search.
   */
  public void abortSearch ()
  {
    m_searchAborted = true;
  }

  /**
   * Tests the <tt>searchAborted</tt> flag.
   */
  public boolean isSearchAborted ()
  {
    return m_searchAborted;
  }

  /**
   * Chooses a search engine based on the characteristics of the
   * problem.  Actually, this version just uses an "engineType" string
   * to make the choice.
   */
  public SearchEngine chooseSearchEngine (String engineType)
  {
    SearchEngine result = null;
    if (engineType.equals(SYSTEMATIC))
      result = new SystematicSearch();
    else if (engineType.equals(IP))
      result = new IPSearch();
    else if (engineType.equals(IPCON))
      result = new IPConSearch();
    else if (engineType.equals(ASTAR))
      result = new BidtreeSearch();
    else if (engineType.equals(IDASTAR))
      result = new BidtreeIDAstar();
    else if (engineType.equals(IDFRINGE))
      result = new BidtreeFringeSearch();
    else
      result = new SASearch();
    result.setBidEvaluator(this);
    return result;
  }

  // ---------- Completion Optimization ----------
  /**
   * Sets up data structures needed for completion optimization.  This
   * includes a dummy leaf task, and a dummy bid for that task.  We
   * take care to work only with tasks that are OFFERED; if no leaf
   * tasks of the plan are OFFERED, then nothing happens.  Therefore,
   * the dummy leaf task is a successor to all the OFFERED leaf tasks.
   */
  protected void setupCompletionOptimization ()
  {

  }

  // ---------- Constraint processing -----------

  /**
   * Generates a map of coverage constraints in m_conflictingBids,
   * indexed by bid, of the form {b1 b2 ...} where the bs are the bids
   * that conflict with the bid at the given index.
   */
  public void generateCoverageConstraints ()
  {
    log.debug("generateCoverageConstraints()");

    if (!m_hasCoverageConstraints) {
      // Before doing anything with constraints, we must initialize the
      // compatibleBids map.
      for (Iterator bids = m_bids.iterator();
	   bids.hasNext(); ) {
	m_conflictingBids.put(bids.next(), new HashSet());
      }
      
      // Now, we simply map through all the bids and generate the
      // conflict set.  Each time we discover a conflict between two
      // bids a and b, we could record both a conflicts-with b and b
      // conflicts-with a.  We don't.  Since we are going to visit all
      // the bids anyway, and there's no way to do this in a triangular
      // fashion, we just pick up the second constraint when we visit
      // the second bid.
      for (Iterator bids = m_bids.iterator();
	   bids.hasNext(); ) {
	Bid theBid = (Bid)bids.next();
	HashSet conflictSet = (HashSet)m_conflictingBids.get(theBid);
	for (Iterator btasks = theBid.getTaskSet().iterator();
	     btasks.hasNext(); ) {
	  TaskDescriptor td = findTaskDescriptor((String)btasks.next());
	  for (Iterator otherBids = td.getBids();
	       otherBids.hasNext(); ) {
	    Bid otherBid = (Bid)otherBids.next();
	    if (otherBid != theBid)
	      conflictSet.add(otherBid);
	  }
	}
      }
      m_hasCoverageConstraints = true;
    } // if (!m_hasCoverageConstraints)
  }

  /**
   * Returns true just in case we have a set of bids that cover all
   * the tasks.  If there is any task that is not covered by a bid, we
   * return false.  Otherwise, we find all "singleton" bids
   * (bids that are the only bid to cover some task) and remove any
   * other bids that conflict with them (in the sense of coverage
   * conflict) and check again for uncovered tasks and further
   * singletons.  If we survive this process, we have coverage and
   * return true.
   */
  public boolean checkCoverage ()
  {
    log.debug("checkCoverage");
    generateCoverageConstraints(); // just make sure...

    boolean tryAgain = true;
    while (tryAgain) {
      // only go one more time unless something changes
      tryAgain = false;
      for (Iterator e = getTaskDescriptors().iterator();
	   e.hasNext(); ) {
	TaskDescriptor aTD = (TaskDescriptor)e.next();
	if (aTD.getBids() == null) {
	  log.debug("No bid for task " + aTD.getTaskId());
	  resultLog.info("No coverage");
	  return false;
	}
	if (aTD.getBidCount() == 1) {
	  // we have a singleton here
	  Bid theBid = (Bid)aTD.getBids().next();
	  log.debug("Singleton bid: " + theBid.getId());
	  for (Iterator conflicts = conflictingBids(theBid).iterator();
	       conflicts.hasNext(); ) {
	    Bid toRemove = (Bid)conflicts.next();
	    removeBid(toRemove);
	    conflicts.remove();
	    m_conflictingBids.remove(toRemove);
	    tryAgain = true;
	  }
	}
      }
    }
    // if we fall through to here, we have coverage.
    return true;
  }

  /**
   * Generates temporal feasibility constraints.  A constraint is
   * simply a set of bids that are incompatible with each other.  The
   * approach is to choose a bid, find its leaf tasks in the task
   * successor relationships, choose bids that cover those tasks, and
   * check for compatibility.  Any bid that roots a set that can't be
   * extended to the leaves of the task precedence network can be
   * discarded, although we currently don't check for this.
   */
  protected void generateFeasibilityConstraints ()
  {
    long runTime = System.currentTimeMillis();
    // The basic process is to generate a "fringe" of tasks that we
    // wish to cover with bids.  We extend the fringe if the bids we
    // choose are compatible with the bid/task mappings we have
    // already chosen, otherwise we mark the incompatible sets we have
    // found.
    Fringe fringe = new Fringe();

    // Each task in the fringe is decorated with a time which is the
    // earliest time it can start.  If we choose a bid that can't
    // start at least that late, we have found a problem.  Note that
    // roots that have no successors are not used here.
    for (Iterator roots = m_taskPlan.getSuccessorRoots();
	 roots.hasNext(); ) {
      SubtaskRequest root = (SubtaskRequest)roots.next();
      Iterator succs = m_taskPlan.getSuccessors(root);
      if (succs.hasNext())
	fringe.add(new FringeData(root.getId()));
    }

    // Now we extend the fringe recursively
    extendFringe(fringe, new HashSet(m_bids));

    runTime = System.currentTimeMillis() - runTime;
    resultLog.info("Feasibility Constraints: time = " + runTime);
  }

  /**
   * Extends the fringe of the task precedence network, looking for
   * feasible task mappings.
   */
  protected void extendFringe (Fringe fringe, Set compatibleBids)
  {
    if (log.isDebugEnabled())
      log.debug("extendFringe: " + fringe.toString());
    // if the fringe set is empty, there's nothing to do.  This stops
    // the recursion.
    if (fringe.size() == 0)
      return;

    // First, generate a list of bids that cover one or more tasks in
    // fringe.
    HashSet fringeBids = new HashSet();
    for (Iterator fds = fringe.iterator();
	 fds.hasNext(); ) {
      FringeData fd = (FringeData)fds.next();
      String taskId = fd.taskId;
      TaskDescriptor td = findTaskDescriptor(taskId);
      for (Iterator bids = td.getBids();
	   bids.hasNext(); ) {
	Object bid = bids.next();
	if (compatibleBids.contains(bid))
	  fringeBids.add(bid);
      }
    }

    // Map through all the bids that cover a fringe task and are
    // (at least pairwise) compatible with the bids already assigned.
    for (Iterator bids = fringeBids.iterator();
	 bids.hasNext(); ) {
      Bid bid = (Bid)bids.next();
      // check for feasibility
      //  question - do we do this just for fd, or for every task
      //  in the fringe covered by bid?
      //  answer - every covered task, because otherwise we can't
      //  record bid compatibility?
      //    -- or only taskd, because we'll catch the others later,
      //    and we're only interested in incompatibilities, not
      //    compatibility.
      Fringe newFringe = updateFringe(fringe, bid);
      if (newFringe == null) {
	// bid is incompatible
	compatibleBids.remove(bid);
	addInfeasibleBidSet(extractSet(fringe.bidPath, bid));
      }
      else {
	// bid is compatible

	// update the compatibility list and call extendFringe
	// recursively.  Don't allow re-use of the bid we came in here
	// with, nor with any of its incompatibles.
	HashSet newCompatibleBids = new HashSet(compatibleBids);
	// this was inserted to prevent cycles.  It also prevents
	// detection of some infeasibilities.
	//newCompatibleBids.remove(bid);
	Set incompatibles = (Set)m_conflictingBids.get(bid);
	for (Iterator ics = incompatibles.iterator();
	     ics.hasNext(); ) {
	  newCompatibleBids.remove(ics.next());
	}
	extendFringe(newFringe, newCompatibleBids);
      }
    }
  }

  /**
   * Updates a fringe with a new bid, assuming the bid is compatible
   * with the given fringe data.  If it is, a new fringe is returned;
   * otherwise, the return value is null.  The new fringe is just the
   * set of tasks that follow bid, and is disjoint with the original
   * fringe (assuming there are no redundant precedence links in the
   * task plan?).
   */
  protected Fringe updateFringe (Fringe fringe, Bid bid)
  {
    if (log.isDebugEnabled())
      log.debug("updateFringe(" + fringe + ", " + bid.getId() + ")");
    // update the fringe...
    // We need to get bids to tell us about early
    // completion times, given a set of fringe entries.  We can
    // do that by telling the bid about early start times for
    // tasks on the old fringe, and letting the bid compute the
    // related early completion times.
    // We also need to collect the set of tasks in the old fringe that
    // are covered by the new bid, to add to the task path of the new
    // fringe.
    HashSet tasksUsed = new HashSet(fringe.tasksUsed);
    Set taskSet = bid.getTaskSet();
    BidExaminer bidEx = bid.createBidExaminer();
    // For each task in taskSet that's in the old fringe, we tell the
    // examiner about its early start times.
    for (Iterator ftasks = fringe.iterator();
	 ftasks.hasNext(); ) {
      FringeData fd = (FringeData)ftasks.next();
      // here we need to capture the set of tasks that are being
      // applied to this bid.
      if (taskSet.contains(fd.taskId)) {
	tasksUsed.add(fd.taskId);
	if (fd.earlyStart != null) {
	  // this one needs checking...
	  if (bid.getLateStart(fd.taskId).before(fd.earlyStart)) {
	    // failure
	    return null;
	  }
	  else {
	    try {
	      bidEx.setEarlyStart(fd.taskId, fd.earlyStart);
	    }
	    catch (InvalidDataException ide) {
	      // another form of failure, caused by an infeasibility
	      // detected by the Bid Examiner
	      return null;
	    }
	  }
	}
      }
    }

    // At this point, tasksUsed is the union of the tasksUsed of the
    // original Fringe, and the tasks in the fringe that are covered
    // by the selected bid.

    // then we compute the new fringe, which is the set of tasks in
    // bid that have successors not in bid.  For each one, we use the
    // Examiner to extract the updated early start times.
    //   The new fringe is built as a HashMap so we can avoid
    //   duplication.
    HashMap newFringe = new HashMap();
    Cons bidPath = new Cons(bid, fringe.bidPath);
    for (Iterator bidTasks = taskSet.iterator();
	 bidTasks.hasNext(); ) {
      String bt = (String)bidTasks.next();
      Date ef = bidEx.getEarlyFinish(bt);
      //log.debug("pre fringe bt=" + bt + ", ef=" + ef);
      for (Iterator fringeTasks = m_taskPlan.getSuccessors(bt);
	   fringeTasks.hasNext(); ) {
	String succ = ((SubtaskRequest)fringeTasks.next()).getId();
	// we inspect each successor that is not in the same bid, and
	// has not already been examined in the past along this path
	if (! taskSet.contains(succ)
	    && ! fringe.tasksUsed.contains(succ)) {
	  FringeData fd = (FringeData)newFringe.get(succ);
	  if (fd == null) {
	    fd = new FringeData(succ, ef);
	    newFringe.put(succ, fd);
	    //log.debug("new fringe item: " + fd);
	  }
	  else if (ef.after(fd.earlyStart)) {
	    // here we have to make sure that the ef data is the
	    // latest ef found
	    fd.earlyStart = ef;
	  }
	}
      }
    }
    // all that's left now is to extract the set of FringeData
    // instances from newFringe
    return new Fringe(newFringe.values(), tasksUsed, bidPath);
  }

  /**
   * Returns a Set representation of a set of bids that arrive in the
   * form of a Cons-list of bids plus another bid.
   */
  protected HashSet extractSet (Cons terms, Object other)
  {
    HashSet newSet = new HashSet(terms.size() + 1);
    newSet.add(other);
    for (Cons ls = terms; ls != null; ls = (Cons)ls.cdr()) {
      Bid term = (Bid)ls.car();
      newSet.add(term);
    }
    return newSet;
  }

  /**
   * Adds the specified set to the set of incompatible bid sets
   */
  protected void addInfeasibleBidSet(Set bidSet)
  {
    if (log.isDebugEnabled()) {
      StringBuffer buf = new StringBuffer();
      String sep = "";
      buf.append("Incompatible set: [");
      for (Iterator bids = bidSet.iterator();
	   bids.hasNext(); ) {
	buf.append(sep).append(((Bid)bids.next()).getId());
	sep = " ";
      }
      buf.append("]");
      log.debug(buf);
    }
    m_infeasibleBidSets.add(bidSet);
  }

  // -------------- Model access --------------

  /**
   * Returns a Set containing all the bids received so far
   */
  public List getBids ()
  {
    return m_bids;
  }

  /**
   * Return the TaskPlan being evaluated.
   * @return Value of taskPlan.
   */
  public TaskPlan getTaskPlan ()
  {
    return m_taskPlan;
  }

  /**
   * Returns an Iterator over the offered tasks.  Elements in the
   * iteration are SubtaskRequest instances.
   */
  public Iterator getOfferedTasks ()
  {
    return m_offeredTasks.iterator();
  }

  /**
   * Returns a Collection of TaskDescriptors, in no particular order.
   */
  public Collection getTaskDescriptors ()
  {
    return m_taskDescriptors.values();
  }

  /**
   * Returns the TaskDescriptor for a given subtask
   */
  public TaskDescriptor findTaskDescriptor (SubtaskRequest st)
  {
    return (TaskDescriptor)m_taskDescriptors.get(st);
  }

  /**
   * Returns the TaskDescriptor for a given subtask ID
   */
  public TaskDescriptor findTaskDescriptor (String st)
  {
    return (TaskDescriptor)m_taskDescriptors.get(m_taskPlan.getTaskData(st));
  }

  /**
   * Returns the Session ID for the current session
   */
  public String getSessionId () {
    return m_sessionId;
  }

  /**
   * Returns the set of bids that are pairwise compatible with the
   * specified bid.  A null result means that no compatibliity
   * information is available, so all bids are assumed to be
   * compatible.
   */
  public Set conflictingBids (Bid theBid)
  {
    return (Set)m_conflictingBids.get(theBid);
  }

  /**
   * Returns true just in case the specified set of bids is an
   * incompatible set.
   */
  public boolean isInfeasibleSet (Set bidSet)
  {
    return getInfeasibleSets().contains(bidSet);
  }

  /**
   * Returns the Set of infeasible bid sets.
   */
  public Set getInfeasibleSets ()
  {
    if (m_infeasibleBidSets == null) {
      // lazy instantiation...
      m_infeasibleBidSets = new HashSet();
      generateFeasibilityConstraints();
    }
    return m_infeasibleBidSets;
  }

  /** Returns true just in case we are configured to use coverage
   * constraints.
   */
  public boolean useCoverageConstraints ()
  {
    return m_useCoverage;
  }

  /**
   * Returns true just in case we are configured to use feasibility
   * constraints.
   */
  public boolean useFeasibilityConstraints ()
  {
    return m_useFeasibility;
  }

  /**
   * Returns true just in case we are optimizing completion time.
   */
  public boolean isOptimizingCompletion ()
  {
    return m_optimizeCompletion;
  }

  /**
   * Data holder for a Fringe of tasks, used in inducing feasibility
   * constraints.
   */
  protected class Fringe
  {
    Collection elements = new ArrayList();
    Cons bidPath = null;
    Set tasksUsed = new HashSet();

    Fringe ()
    {
      super();
    }

    Fringe (Collection data, Set tasksUsed, Cons bidPath)
    {
      super();
      this.elements = data;
      this.bidPath = bidPath;
      this.tasksUsed = tasksUsed;
    }

    void add (FringeData fd)
    {
      elements.add(fd);
    }

    Iterator iterator ()
    {
      return elements.iterator();
    }

    int size ()
    {
      return elements.size();
    }

    public String toString ()
    {
      StringBuffer buf = new StringBuffer();
      // task ID and early start
      buf.append("<fringe ").append(elements).append(" ");
      buf.append(tasksUsed);
      buf.append(" (");
      // bid path
      Cons bd = bidPath;
      String sep = "";
      while (bd != null) {
	buf.append(sep).append(((Bid)bd.car()).getId());
	sep = " ";
	bd = (Cons)bd.cdr();
      }
      buf.append(")>");
      return buf.toString();
    }
  }

  /**
   * data holder for elements of a fringe
   */
  protected class FringeData
  {
    String taskId;
    Date earlyStart = null;

    // use this constructor for the root tasks..
    FringeData (String taskId)
    {
      super();
      this.taskId = taskId;
    }

    // this constructor is for the rest of the tasks.
    FringeData (String taskId, Date earlyStart)
    {
      super();
      this.taskId = taskId;
      this.earlyStart = earlyStart;
    }

    public String toString ()
    {
      StringBuffer buf = new StringBuffer();
      // task ID and early start
      buf.append("<fd ").append(taskId).append(" ");
      buf.append(earlyStart).append(">");
      return buf.toString();
    }
  }
}
