/*------------------------------------------------------------------------
 * MAGNET bid evaluation component
 * $Id: Bidtree.java,v 1.13.2.3 2002/05/05 19:41:47 ashutosh Exp $
 * Copyright (c) 2002 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Mon Feb 18 12:36:37 2002  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.protocol.cust.Bid;

import org.apache.log4j.Category;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.TreeSet;
import java.util.Arrays;

/**
 * This is the the basic bid-ordering structure in support of
 * BidtreeSearch. A bidtree is a binary tree with bids at the
 * leaves. Each level of the tree corresponds to a task. Each
 * non-terminal node in the tree has an in-branch, leading to bids
 * that include the task at that level, and an out-branch, leading to
 * bids that do not contain that task. 
 * <p>
 * The basic retrieval operation
 * is {@link #getBids}, in which a mask is used to specify the desired
 * set of bids. The mask is a vector as long as the tree is deep. Each
 * entry in the vector can have one of 3 values:
 * <ol>
 * <li> IN: take the left (in) branch
 * <li> OUT: take the right (out) branch
 * <li> BOTH: take both branches
 * </ol>
 * </p><p>
 * Within a leaf node, bids are sorted by cost. To create the tree,
 * tasks are sorted by bid-count. Therefore, if you start at the top
 * of the tree, you get all the singleton bids by picking (in)
 * branches until you come up with multiple bids. The {@link
 * #getTaskArray} method returns the sorted task set.
 * </p><p>
 * The basic structure comes from Sandholm (see {@link BidtreeSearch})
 * except that each leaf contains a sorted list of bids instead of a
 * single bid.
 * </p>
 * @author John Collins
 * @version $Revision: 1.13.2.3 $
 */
public class Bidtree
{
  static Category log = Category.getInstance(Bidtree.class.getName());

  /** Follow the "in" (left) branch */
  public static final int IN = 1;
  /** Follow the "out" (right) branch */
  public static final int OUT = 2;
  /** Follow both branches */
  public static final int BOTH = 3;

  // local database
  TaskDescriptor[] taskArray = null; // task index -> TaskDescriptor
  BtBidWrapper[][] bidArray = null;  // task index -> sorted bids
  private ArrayList buckets = new ArrayList();
  private BTNode topNode = null;

  // two task comparators are provided, one that sorts by increasing
  // numbers of bids, the other by decreasing numbers

  /** Specify INCREASING to get your tree built narrow at the top */
  static final int INCREASING = 0;
  /** Specify DECREASING to get your tree built wide at the top */
  static final int DECREASING = 1;

  static private Comparator[] comparators = {
    new Comparator () {
	// This implementation of compare is not consistent with
	// equals. We want all the tasks indexed, not just one for
	// each bidCount value.
	public int compare (Object o1, Object o2) {
	  int result = (((TaskDescriptor)o1).getBidCount()
			- ((TaskDescriptor)o2).getBidCount());
	  if (result == 0)
	    result = 1; // not consistent with equals!!
	  return result;
	}
      },
    new Comparator () {
	// This implementation of compare is not consistent with
	// equals. We want all the tasks indexed, not just one for
	// each bidCount value.
	public int compare (Object o1, Object o2) {
	  int result = (((TaskDescriptor)o2).getBidCount()
			- ((TaskDescriptor)o1).getBidCount());
	  if (result == 0)
	    result = 1; // not consistent with equals!!
	  return result;
	}
      }};

  // this is the sorting comparator for returning composite sorted bid
  // lists.
  static private Comparator btComparator =
    new Comparator() {
	public int compare (Object o1, Object o2) {
	  return (int)Math.round(((BtBidWrapper)o1).getCostPerTask()
				 - ((BtBidWrapper)o2).getCostPerTask());
	}
      };

  /**
   * Default constructor provided, but you are supposed to use the
   * static {@link #create} method to manufacture these.
   */
  private Bidtree ()
  {
    super();
  }

  /**
   * (Re)initializes one of these to index the bids in the given
   * BidEvaluator. If you call this on an existing Bidtree, you will
   * lose the data it originally contained.
   */
  public void init (BidEvaluator be, Comparator comparator)
  {
    Collection tds = be.getTaskDescriptors();
    // Create and fill in the task set
    taskArray = new TaskDescriptor[tds.size()];
    Iterator tdi = tds.iterator();
    for (int i = 0; i < taskArray.length; i++) {
      TaskDescriptor td = (TaskDescriptor)tdi.next();
      //if (log.isDebugEnabled())
      //log.debug("TD " + td.getTaskId()
      //	  + ": count = " + td.getBidCount());
      taskArray[i] = td;
    }
    Arrays.sort(taskArray, comparator);
//      if (log.isDebugEnabled()) {
//        StringBuffer sb = new StringBuffer();
//        sb.append("Task order: [");
//        String sep = "";
//        for (int i = 0; i < taskArray.length; i++) {
//  	sb.append(sep).append(taskArray[i].getTaskId());
//  	sep = ",";
//        }
//        sb.append("]");
//        log.debug(sb.toString());
//      }

    // Initialize the task-bid mappings
    ArrayList[] tmpBidArray = new ArrayList[taskArray.length];
    for (int i = 0; i < tmpBidArray.length; i++) {
      tmpBidArray[i] = new ArrayList(taskArray[i].getBidCount());
    }

    // Create a BtBidWrapper for each bid, and add it to the task-bid
    // mapping. Then sort all the task-bid mappings by increasing
    // cost/task.
    List bids = be.getBids();
    BtBidWrapper[] wrappers = new BtBidWrapper[bids.size()];
    Iterator bds = bids.iterator();
    for (int i = 0; i < wrappers.length; i++) {
      BtBidWrapper wrapper = new BtBidWrapper((Bid)bds.next(), this);
      wrappers[i] = wrapper;
      int[] taskIndices = wrapper.getTaskIndices();
      for (int j = 0; j < taskIndices.length; j++) {
	tmpBidArray[taskIndices[j]].add(wrapper);
      } // for j
    } // for i
    bidArray = new BtBidWrapper[tmpBidArray.length][];
    for (int i = 0; i < tmpBidArray.length; i++) {
      Object[] bwx = tmpBidArray[i].toArray();
      Arrays.sort(bwx, btComparator);
      // fix class cast problem, can't cast toArray result to correct
      // type.
      BtBidWrapper[] bws = new BtBidWrapper[bwx.length];
      for (int j = 0; j < bwx.length; j++)
	bws[j] = (BtBidWrapper)bwx[j];
      bidArray[i] = bws;
    }
    
    // Now we finally get around to filling in the bid tree.
    if (taskArray.length > 1)
    {
      // Walk the bid list, filling in the leaf nodes with bids. This
      // is a straightforward n*m method. Note that we assume more than
      // one task here. If you've only got one task, why are you
      // searching?
      topNode = new InteriorNode();
      int[] mask = new int[taskArray.length];
      //for (Iterator bdi = be.getBids().iterator(); bdi.hasNext(); ) {
      for (int bi = 0; bi < wrappers.length; bi++) {
	// For each bid, we create a mask in the form of an array of
	// IN/OUT values to reflect task composition of this bid.
	Arrays.fill(mask, OUT);
	BtBidWrapper bw = wrappers[bi];
	for (int ti = 0; ti < bw.tasks.length; ti++) {
	  mask[bw.tasks[ti]] = IN;
	}

	// Now we walk the mask array to construct and navigate the
	// bidtree. When we get to the leaf, we create or add the bid
	// to the bucket.
	InteriorNode current = (InteriorNode)topNode;
	for (int i = 0; i < mask.length - 1; i++) {
	  if (mask[i] == IN) {
	    if (current.in == null)
	      current.in = new InteriorNode();
	    current = (InteriorNode)current.in;
	  }
	  else { // assume mask[i] == OUT
	    if (current.out == null)
	      current.out = new InteriorNode();
	    current = (InteriorNode)current.out;
	  }
	}
	// children of current are leaf nodes at this point
	int val = mask[mask.length - 1];
	LeafNode leaf = null;
	if (val == IN) {
	  if (current.in == null)
	    current.in = new LeafNode();
	  leaf = (LeafNode)current.in;
	}
	else { // again, assume val == OUT
	  if (current.out == null)
	    current.out = new LeafNode();
	  leaf = (LeafNode)current.out;
	}
	// now leaf is the leaf node that's supposed to hold this
	// bid.
	leaf.addBid(bw);
      }
    }

    // Finally, we need to process the buckets to compute the conflict
    // sets. This is NxN where N is the number of buckets.
    int[] conflicts = new int[buckets.size()];
    for (Iterator bkts = buckets.iterator(); bkts.hasNext(); ) {
      BtBucket bkt = (BtBucket)bkts.next();
      int ci = 0;
      for (int otheri = 0; otheri < buckets.size(); otheri++) {
	BtBucket other = (BtBucket)buckets.get(otheri);
	boolean conflict = false;
	// for each pair of buckets, we look for a task in a that
	// matches a task in b. We only have to find one to have a
	// conflict.
	for (int i = 0; i < bkt.tasks.length && !conflict; i++) {
	  for (int j = 0; j < other.tasks.length && !conflict; j++) {
	    if (bkt.tasks[i] == other.tasks[j]) {
	      conflicts[ci++] = otheri;
	      conflict = true;
	    } // if
	  } // for j
	} // for i
      } // for otheri
      if (ci > 0) {
	int[] cfl = new int[ci];
	for (int i = 0; i < ci; i++)
	  cfl[i] = conflicts[i];
	bkt.conflicts = cfl;
      } // if ci > 0
    } // for bkts
  }

  /**
   * Returns an array of bids for a given mask. The array is sorted by
   * cost/taskCount ratio. If the mask is shorter
   * than the number of tasks, we assume that the mask starts at 0 and
   * that all missing elements are BOTH. If the mask is too long, the
   * tail is ignored.
   */
  public BtBidWrapper[] getBids (int[] mask)
  {
//      if (log.isDebugEnabled()) {
//        StringBuffer sb = new StringBuffer();
//        sb.append("Mask: [");
//        String sep = "";
//        for (int i = 0; i < mask.length; i++) {
//          sb.append(sep).append(Integer.toString(mask[i]));
//          sep = ",";
//        }
//        sb.append("]");
//        log.debug(sb.toString());
//      }
    ArrayList composite = getBidsRecursively(mask, 0,
					     (InteriorNode)topNode);
    if (composite == null)
      return null;

    // now we collect all the bids into a flat array
    int totalSize = 0;
    for (Iterator sli = composite.iterator(); sli.hasNext(); ) {
      totalSize += ((BtBucket)sli.next()).size();
    }
    //log.debug("Total size: " + totalSize);
    BtBidWrapper[] result = new BtBidWrapper[totalSize];
    int index = 0;
    for (Iterator sli = composite.iterator(); sli.hasNext(); ) {
      BtBidWrapper[] sublist = ((BtBucket)sli.next()).getSortedBids();
      for (int i = 0; i < sublist.length; i++ ) {
	result[index++] = sublist[i];
      } // for i
    } // for bid

    // finally we sort the array
    Arrays.sort(result, btComparator);
//      if (log.isDebugEnabled()) {
//        StringBuffer sb = new StringBuffer();
//        sb.append("Next bids: [");
//        String sep = "";
//        for (int i = 0; i < result.length; i++) {
//          sb.append(sep).append(result[i].getId());
//          sep = ",";
//        }
//        sb.append("]");
//        log.debug(sb.toString());
//      }
    return result;
  }

  /**
   * Returns an ArrayList of BtBucket instances. The
   * bids in each sublist cover the same set of tasks. Different
   * sublists represent different task combinations as selected by the
   * mask. Each sublist is sorted by cost.
   */
  public ArrayList getBidCollections (int[] mask)
  {
    return getBidsRecursively(mask, 0, (InteriorNode)topNode);
  }

  /**
   * Generates the bid list for a given mask by doing a depth-first
   * recursive tree walk as directed by the mask. If there are many
   * BOTH nodes, this can generate a lot of garbage. Really makes you
   * wish for Lisp...
   */
  private ArrayList getBidsRecursively (int[] mask,
					int index,
					InteriorNode current)
  {
    ArrayList result = null;
    if (index >= taskArray.length - 1) {
      // we've reached the bottom -- this is where we actually create
      // the result list
      //log.debug("end of recursion");
      result = new ArrayList();
      if (mask.length <= index || mask[index] == BOTH) {
	// return both - Note that if we got here, at least one of
	// these is guaranteed to be non-empty
	//log.debug("both@leaf");
	if (current.in != null)
	  result.add(((LeafNode)current.in).bucket);
	if (current.out != null)
	  result.add(((LeafNode)current.out).bucket);
      }
      else {
	// return one or the other as determined by the mask
	if (mask[index] == IN && current.in != null) {
	  //log.debug("in@leaf");
	  result.add(((LeafNode)current.in).bucket);
	}
	else if (mask[index] == OUT && current.out != null) {
	  //log.debug("out@leaf");
	  result.add(((LeafNode)current.out).bucket);
	}
	if (result.size() == 0)
	  // Don't return an empty list
	  result = null;
      } // if (mask.length...
    }
    else {
      // we're somewhere in the middle
      if (mask.length <= index || mask[index] == BOTH) {
	//log.debug("both@" + index);
	if (current.in != null) {
	  result = getBidsRecursively(mask,
				      index + 1,
				      (InteriorNode)current.in);
	}
	if (current.out != null) {
	  // no result from in branch, get it from out branch
	  if (result == null) {
	    result = getBidsRecursively(mask,
					index + 1,
					(InteriorNode)current.out);
	  }
	  else {
	    ArrayList nxt = getBidsRecursively(mask,
					       index + 1,
					       (InteriorNode)current.out);
	    if (nxt != null)
	      result.addAll(nxt);
	  }
	}
      }
      else {
	// return in or out as determined by mask[index]
	if (mask[index] == IN && current.in != null) {
	  //log.debug("in@" + index);
	  result = getBidsRecursively(mask,
				      index + 1,
				      (InteriorNode)current.in);
	}
	else if (mask[index] == OUT && current.out != null) {
	  //log.debug("out@" + index);
	  result = getBidsRecursively(mask,
				      index + 1,
				      (InteriorNode)current.out);
	}
	else {
	  //log.debug("No such bids");
	  result = null;
	}// if mask[index]
      } // if mask.length
    } // if index >= taskList.size()
    return result;
  }

  public BtBidWrapper[] getBids (int taskIndex)
  {
    return bidArray[taskIndex];
  }

  /**
   * Returns the sorted set of task descriptors.
   */
  public TaskDescriptor[] getTaskArray ()
  {
    return taskArray;
  }

  private void addBucket (BtBucket b)
  {
    b.setIndex(buckets.size());
    buckets.add(b);
  }

  /**
   * Returns the indicated bucket
   */
  public BtBucket getBucket (int index)
  {
    return (BtBucket)buckets.get(index);
  }

  /**
   * Returns the entire list of BtBucket instances.
   */
  public ArrayList getBuckets ()
  {
    return buckets;
  }

  // ------------------- static factory ------------------------------
  /**
   * Creates and returns a new Bidtree containing the bids provided by
   * the given BidEvaluator. You have to do this after all bids are
   * received, because the first step is to sort tasks according to
   * bid count.
   */
  public static Bidtree create (BidEvaluator be)
  {
    return create(be, 0);
  }

  /**
   * Creates and returns a new Bidtree containing the bids provided by
   * the given BidEvaluator. You have to do this after all bids are
   * received, because the first step is to sort tasks according to
   * bid count.
   * @param be The {@link BidEvaluator BidEvaluator} that contains the
   *   bids we want to index.
   * @param comparatorSelector An int that selects the sort direction
   *   for tasks in constructing the tree. The value must be one of
   *   <tt>INCREASING</tt> or <tt>DECREASING</tt>. If the value is
   *   <tt>INCREASING</tt>, then the tasks with the lowest numbers of
   *   bids are at the top of the tree, making the tree narrow at the
   *   top, widening at the bottom. <tt>DECREASING</tt>, of course,
   *   does the opposite.
   */
  public static Bidtree create (BidEvaluator be, int comparatorSelector)
  {
    if (comparatorSelector < INCREASING
	|| comparatorSelector > DECREASING) {
      log.error ("Bad comparator selector " + comparatorSelector
		 + ", using INCREASING");
      comparatorSelector = INCREASING;
    }
    Bidtree instance = new Bidtree();
    instance.init(be, comparators[comparatorSelector]);
    return instance;
  }

  // ---------------------------------------------------------------
  /**
   * This is the tree node structure
   */
  class BTNode
  {
  }

  class InteriorNode extends BTNode
  {
    BTNode in = null;
    BTNode out = null;

    InteriorNode ()
    {
      super();
    }
  }

  class LeafNode extends BTNode
  {
    BtBucket bucket = null; 

    LeafNode ()
    {
      super();
    }
    
    /**
     * Adds the given bid in cost-sorted order. We use a linear
     * insertion here because we expect these bins to be fairly small.
     */
    void addBid (BtBidWrapper b)
    {
      if (bucket == null) {
	// this is the first bid to be added
	bucket = new BtBucket(b.tasks);
	addBucket(bucket);
      }
      bucket.addBid(b);
      b.setBucketIndex(bucket.getIndex());
    }

    /**
     * Returns the sorted bid array from the bucket.
     */
    BtBidWrapper[] getBids ()
    {
      return bucket.getSortedBids();
    }
  }
}
