/*----------------------------------------------------------------------------
 * Magnet Bid Evaluation components
 * $Id: BidtreeFringeSearch.java,v 1.2.2.2 2002/05/05 19:41:47 ashutosh Exp $
 * Copyright (c) 1999 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License. 
 *----------------------------------------------------------------------------
 * Revision History:
 * 
 * Mon Feb 18 11:33:39 2002  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.util.MagnetProperties;
import edu.umn.magnet.util.search.SearchQueue;
import edu.umn.magnet.util.search.SearchNode;

import org.apache.log4j.Category;

import java.util.Iterator;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.LinkedList;


/**
 * Bidtree implementation of the Search Engine abstraction. This
 * method is loosely based on Sandholm's "bidtree" method as described
 * in "Algorithm for Optimal Winner Determination in Combinatorial
 * Auctions", Artificial Intelligence 135, pp 1-54, 2002. There are
 * several possible variations on the basic bidtree search. This one
 * is an attempt to adapt the IDA* method to avoid the extra work, by
 * keeping track of the current fringe in a queue. Preliminary
 * experiments show that it is not as fast as straight A*, and suffers
 * from the same memory exhaustion problems at nearly the same level
 * of complexity. This implementation is also not correct; it
 * occasionally fails to find an optimal solution.
 * <p>
 * Sandholm's bidtree method uses many heuristics and preprocessing
 * steps that are not appropriate for problems with precedence
 * constraints. This adaptation uses the same tree-based search space,
 * and a similar bidtree secondary selection method, except that the
 * leaves of the bidtree are not individual bids but rather sorted
 * collections of bids that cover the same task sets, with possibly
 * different time window specifications. Bids in a leaf are sorted by
 * cost. 
 * </p><p>
 * Domination may be tested, but is a much weaker notion than in
 * Sandholm's formulation. in this case, a bid B is dominated by a bid
 * A just in case both of the following conditions holds:
 * <ol>
 * <li>cost(B) &gt;= cost(A)
 * <li>Time windows in A contain the time windows in B
 * </ol>
 * </p>
 * @author John Collins
 * @version $Revision: 1.2.2.2 $
 */
public class BidtreeFringeSearch
  extends BidtreeSearch
  implements SearchEngine
{
  // Init the log categories
  static Category log
    = Category.getInstance(BidtreeFringeSearch.class.getName());
  static Category result = Category.getInstance("Result");
  
  // Search queue and parameters
  private TaskBidPlan m_initialNode = null;
  private Bidtree m_bidtree = null;
  private double fLimit = 0.0;
  private double m_multiplier = 1.01;
  private SearchQueue m_searchQueue
    = SearchQueue.create(SearchQueue.BEST_FIRST);
  private int maxQueueSize = 20000;
  
  // Statistics
  private int m_nodeCount = 0;  
  private int m_planCount = 0;
  
  private BidtreeNode m_bestPlan = null;  //the best covered and feasible node
  private BidtreeNode bestNode = null;    //the best node so far
  
  // Evaluators and Selectors
  private Eval m_eval;
  
  /**
   * Default constructor.
   */
  public BidtreeFringeSearch ()
  {
    super();
  }

  public void setBidEvaluator (BidEvaluator be)
  {
    super.setBidEvaluator(be);
    readSystemProps();
  }

  /**
   * Run the search.  This is the primary implementation of the
   * SearchEngine interface.  In this case, we just copy some pointers
   * over from the BidEvaluator, set up some randomizers, and take off.
   */
  
  public TaskBidPlan search()
  {
    if (!getBidEvaluator().checkCoverage())
      return null; // no coverage possible
    m_eval = new Eval(bidEval); 
    m_eval.setUsingCoveragePenalty(false);
    
    BidtreeNode result = startSearch();
    if (result == null)
      return null;
    else
      return result.getTaskBidPlan();
  }

  /**
   * Sets the initial search node
   */
  public void setInitialNode (TaskBidPlan node)
  {
    m_initialNode = node;
  }

  public BidtreeNode startSearch()
  {
    log.info("BidtreeFringeSearch.startSearch()");
    m_searchQueue.setMaxSize(maxQueueSize);

    // Create a root node
    TaskBidPlan tbp = m_initialNode;
    BidtreeNode rootNode = null;
    if (tbp == null) {
      tbp = new TaskBidPlan(bidEval.getTaskPlan());
    }
    m_nodeCount++;

    // evaluate the root node
    double rootEval = m_eval.evaluate(tbp);
    rootNode = new BidtreeNode(tbp, this);
    m_searchQueue.insert(rootNode);

    // run the search
    int cycles = 0;
    double newLimit = rootNode.getEvaluation();
    fLimit = newLimit;
    while (m_bestPlan == null && fLimit != Double.MAX_VALUE) {
      // this is the heart of the search
      cycles++;
      log.info("Starting cycle " + cycles
	       + ", Fringe size = " + m_searchQueue.size());
      newLimit = queueProcessor(fLimit);
      if (m_bestPlan == null)
	fLimit = Math.max(newLimit, m_multiplier * fLimit);
    }
    if (fLimit == Double.MAX_VALUE)
      log.info("fLimit reached MAX_VALUE: bailing");
    else
      log.info("Complete in " + cycles + " cycles.");
    
    // record results
    log.info("Complete: Node count:  " + m_nodeCount
	     + ", Feasible and Covered Nodes : " + m_planCount);
    if (m_searchQueue.wasTrimmed())
      log.warn("Waning: queue was trimmed");
    result.info("Node Count = " + m_nodeCount
		+ ", Feasible and Covered Nodes : " + m_planCount
		+ (m_searchQueue.wasTrimmed()?"-opt":""));
    //if (log.isDebugEnabled()) {
    //  if (m_bestPlan == null)
    //log.debug("I have not found a feasible and covered node");
    //else
    //log.debug(m_bestPlan.toString());
    //}
    return m_bestPlan;
  }

  /**
   * As long as the front element of the queue has an evaluation less
   * than fLimit, runs a depth-first expansion on it using
   * <tt>nodeGenerator</tt>. 
   * @returns a double equal to the minimum evaluation of all
   * generated nodes, or <tt>Double.MAX_VALUE</tt> if no valid nodes
   * were generated.
   */
  protected double queueProcessor(double fLimit)
  {
    double nextV = Double.MAX_VALUE;
    if (!m_searchQueue.isEmpty())
      log.debug("Limit = " + fLimit + ", First node eval = "
		+ ((BidtreeNode)m_searchQueue.select()).getEvaluation());
    while (!m_searchQueue.isEmpty()
	   && (((BidtreeNode)m_searchQueue.select()).getEvaluation()
	       <= fLimit)) {
      // Run the depth-first search on this node
      BidtreeNode next = (BidtreeNode)m_searchQueue.extract();
      log.debug("extracted node " + next.getNodeNumber());
      double result = nodeGenerator(next);
      nextV = Math.min(nextV, result);
    }

    // if we didn't find anything in this cycle, the queue may not be
    // empty...
    if ((nextV == Double.MAX_VALUE) && !m_searchQueue.isEmpty())
      nextV = ((BidtreeNode)m_searchQueue.select()).getEvaluation();
    return nextV;
  }
  
  /**
   * Does a depth-first search from <tt>nextNode</tt>, stopping with any
   * node (including <tt>nextNode</tt>) whose evaluation exceeds
   * <tt>fLimit</tt>. Whenever we find a (non-covering) node whose
   * evaluation exceeds <tt>fLimit</tt>, we add it to the fringe
   * queue.
   * @returns the minimum evaluation found among all generated nodes,
   * or <tt>Double.MAX_VALUE</tt> if no valid expansions of
   * <tt>nextNode</tt> were found.
   */
  protected double nodeGenerator(BidtreeNode nextNode)
  {
    double newL = nextNode.value();
    log.debug("nodeGenerator(): fLimit=" + fLimit
	      + ", newL=" + newL);
    
    // Prune when find a higher lower bound
    if (newL > fLimit && m_bestPlan == null) {
      log.debug("inserting node " + nextNode.getNodeNumber());
      m_searchQueue.insert(nextNode);
      return newL;
    }
    
    // If this node is covered & feasible, note the solution and
    // update the limit
    if (nextNode.isCovered()) {
      log.debug("Covered, feasible node found");
      m_planCount++;
      if (newL < fLimit) {
	fLimit = newL;
	m_bestPlan = nextNode;
      }
      return newL;
    }
    
    // Now starts the depth-first loop
    double nextL = Double.MAX_VALUE;
    while (nextNode.hasMoreBids()) {
      // Generate all the children of nextNode
      BidtreeNode result = nextNode.expandNode(m_nodeCount++);
      if (result != null && result.getTaskBidPlan() != null) {
	m_eval.evaluate(result.getTaskBidPlan());
	log.debug("Est. cost: " + result.getEvaluation());
	if (result.isFeasible()) {
	  newL = nodeGenerator(result);
	  nextL = Math.min(nextL, newL);
	}
      }
    }
    if (nextL == Double.MAX_VALUE) {
      // nextNode has an unusable bid combination. We might want to
      // remember this.
    }
    return nextL;
  }    
  
  /**
   * Gets the node count
   */
  public int getNodeCount ()
  {
    return m_nodeCount;
  }

  /**
   * Returns the Bidtree structure used by this Search instance
   */
  public Bidtree getBidtree ()
  {
    if (m_bidtree == null)
      m_bidtree = Bidtree.create(getBidEvaluator());
    return m_bidtree;
  }

  protected void readSystemProps ()
  {
    //String prop;
  }
}				
