/*----------------------------------------------------------------------------
 * Magnet Bid Evaluation components
 * $Id: BidtreeIDAstar.java,v 1.9.2.2 2002/05/05 19:41:47 ashutosh Exp $
 * Copyright (c) 1999 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License. 
 *----------------------------------------------------------------------------
 * Revision History:
 * 
 * Mon Feb 18 11:33:39 2002  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.util.MagnetProperties;
import edu.umn.magnet.util.search.SearchQueue;
import edu.umn.magnet.util.search.SearchNode;

import org.apache.log4j.Category;

import java.util.Iterator;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.LinkedList;


/**
 * Bidtree implementation of the Search Engine abstraction. This
 * method is loosely based on Sandholm's "bidtree" method as described
 * in "Algorithm for Optimal Winner Determination in Combinatorial
 * Auctions", Artificial Intelligence 135, pp 1-54, 2002. This is an
 * IDA* implementation, along the lines of Sandholm's original
 * scheme. It finds solutions in cases where A* runs out of memory,
 * but it does not appear to be competitive with preprocessed IP (see
 * {@link IPConSearch IPConSearch}).
 * <p>
 * Sandholm's bidtree method uses many heuristics and preprocessing
 * steps that are not appropriate for problems with precedence
 * constraints. This adaptation uses the same tree-based search space,
 * and a similar bidtree secondary selection method, except that the
 * leaves of the bidtree are not individual bids but rather sorted
 * collections of bids that cover the same task sets, with possibly
 * different time window specifications. Bids in a leaf are sorted by
 * cost. 
 * </p><p>
 * Domination may be tested, but is a much weaker notion than in
 * Sandholm's formulation. in this case, a bid B is dominated by a bid
 * A just in case both of the following conditions holds:
 * <ol>
 * <li>cost(B) &gt;= cost(A)
 * <li>Time windows in A contain the time windows in B
 * </ol>
 * </p>
 * @author John Collins
 * @version $Revision: 1.9.2.2 $
 */
public class BidtreeIDAstar
  extends BidtreeSearch
  implements SearchEngine
{
  // Init the log categories
  static Category log = Category.getInstance(BidtreeIDAstar.class.getName());
  static Category result = Category.getInstance("Result");
  
  // Search queue and parameters
  private TaskBidPlan m_initialNode = null;
  private Bidtree m_bidtree = null;
  private double fLimit = 0.0;
  private double m_multiplier = 1.05;
  
  // Statistics
  private int m_nodeCount = 0;  
  private BidtreeNode m_bestPlan = null;  //the best covered and feasible node
  
  // Evaluators and Selectors
  private Eval m_eval;
  
  /**
   * Default constructor.
   */
  public BidtreeIDAstar ()
  {
    super();
  }

  public String getSearchType ()
  {
    return "IDA*";
  }

  public void setBidEvaluator (BidEvaluator be)
  {
    super.setBidEvaluator(be);
    readSystemProps();
  }

  /**
   * Run the search.  This is the primary implementation of the
   * SearchEngine interface.  In this case, we just copy some pointers
   * over from the BidEvaluator, set up some randomizers, and take off.
   */
  
  public TaskBidPlan search()
  {
    if (!getBidEvaluator().checkCoverage())
      return null; // no coverage possible
    m_eval = new Eval(bidEval); 
    m_eval.setUsingCoveragePenalty(false);
    
    BidtreeNode result = startSearch();
    if (result == null)
      return null;
    else
      return result.getTaskBidPlan();
  }

  /**
   * Sets the initial search node
   */
  public void setInitialNode (TaskBidPlan node)
  {
    m_initialNode = node;
  }

  public BidtreeNode startSearch()
  {
    log.info("BidtreeIDAstar.startSearch()");

    // Create a root node
    TaskBidPlan tbp = m_initialNode;
    BidtreeNode rootNode = null;
    if (tbp == null) {
      tbp = new TaskBidPlan(bidEval.getTaskPlan());
    }
    m_nodeCount++;

    // evaluate the root node
    double rootEval = m_eval.evaluate(tbp);

    // run the search
    int cycles = 0;
    double newLimit = 0.0;
    while (m_bestPlan == null && fLimit != Double.MAX_VALUE) {
      // this is the heart of the search
      rootNode = new BidtreeNode(tbp, this);
      cycles++;
      //log.debug("Starting cycle " + cycles
      //	+ ", fLimit = " + fLimit);
      newLimit = nodeGenerator(rootNode);
      if (m_bestPlan == null)
	fLimit = Math.max(newLimit, m_multiplier * fLimit);
    }
    if (fLimit == Double.MAX_VALUE)
      log.info("fLimit reached MAX_VALUE: bailing");
    else
      log.info("Complete in " + cycles + " cycles.");

    // record results
    log.info("Complete: Node count:  " + m_nodeCount
	     + ", Feasible and Covered Nodes : " + getPlanCount());
    result.info("Node Count = " + m_nodeCount
		+ ", Feasible and Covered Nodes : " + getPlanCount());
    //if (log.isDebugEnabled()) {
    //  if (m_bestPlan == null)
    //log.debug("I have not found a feasible and covered node");
    //else
    //log.debug(m_bestPlan.toString());
    //}
    return m_bestPlan;
  }
    
  protected double nodeGenerator(BidtreeNode nextNode)
  {
    double newL = nextNode.value();
    //log.debug("nodeGenerator(): fLimit=" + fLimit
    //      + ", newL=" + newL);
    
    // Prune when find a higher lower bound
    if (newL > fLimit)
      return newL;

    // If this node is covered & feasible, note the solution and
    // update the limit
    if (nextNode.isCovered()) {
      //log.debug("Covered node found");
      fLimit = newL;
      m_bestPlan = nextNode;
      return newL;
    }

    // Now starts the depth-first loop
    double nextL = Double.MAX_VALUE;
    while (nextNode.hasMoreBids()) {
      // Generate all the children of nextNode
      BidtreeNode newNode = nextNode.expandNode(m_nodeCount++);
      if (newNode != null && newNode.getTaskBidPlan() != null) {
	m_eval.evaluate(newNode.getTaskBidPlan());
	if (newNode.isFeasible()) {
	  //log.debug("Est. cost: " + newNode.getEvaluation());
	  if (newNode.isCovered())
	    processSolution(newNode.getTaskBidPlan());
	  newL = nodeGenerator(newNode);
	  nextL = Math.min(nextL, newL);
	}
      }
    }
    if (nextL == Double.MAX_VALUE) {
      // nextNode has an unusable bid combination. We might want to
      // remember this.
    }
    return nextL;
  }    

  protected void readSystemProps ()
  {
    super.readSystemProps();

    String prop;
    // Read the multiplier
    prop = MagnetProperties.getProperty("IDA.Multiplier");
    if (prop != null) {
      try {
	double value = Double.valueOf(prop).doubleValue();
	m_multiplier = value;
	log.info("IDA* multiplier = " + value);
      }
      catch (NumberFormatException nfe) {
      log.error("Bad property spec " + prop
		+ " for IDA* multiplier. Using "
		+ m_multiplier);
      } // end of try-catch
    } // end of if ()
  }
}				
