/*----------------------------------------------------------------------------
 * Magnet Bid Evaluation components
 * $Id: BidtreeSearch.java,v 1.13.2.2 2002/05/05 19:41:47 ashutosh Exp $
 * Copyright (c) 1999 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License. 
 *----------------------------------------------------------------------------
 * Revision History:
 * 
 * Mon Feb 18 11:33:39 2002  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.util.MagnetProperties;
import edu.umn.magnet.util.search.SearchQueue;
import edu.umn.magnet.util.search.SearchNode;
import edu.umn.magnet.util.search.Tester;

import org.apache.log4j.Category;

import java.util.Iterator;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.LinkedList;


/**
 * Bidtree implementation of the Search Engine abstraction. This
 * method is loosely based on Sandholm's "bidtree" method as described
 * in "Algorithm for Optimal Winner Determination in Combinatorial
 * Auctions", Artificial Intelligence 135, pp 1-54, 2002. This is an
 * A* adaptation of Sandholm's formulation. It can be made to run out
 * of memory. You can prevent this (at the cost of failing to find a
 * solution) by setting the property BTS.MaxQueueSize to a number that
 * limits the queue size. The default value is 20000. An IDA* version
 * of this algorithm is in {@link BidtreeIDAstar BidtreeIDAstar}.
 * <p>
 * Sandholm's bidtree method uses many heuristics and preprocessing
 * steps that are not appropriate for problems with precedence
 * constraints. This adaptation uses the same tree-based search space,
 * and a similar bidtree secondary selection method, except that the
 * leaves of the bidtree are not individual bids but rather sorted
 * collections of bids that cover the same task sets, with possibly
 * different time window specifications. Bids in a leaf are sorted by
 * cost. 
 * </p><p>
 * Domination may be tested, but is a much weaker notion than in
 * Sandholm's formulation. in this case, a bid B is dominated by a bid
 * A just in case both of the following conditions holds:
 * <ol>
 * <li>cost(B) &gt;= cost(A)
 * <li>Time windows in A contain the time windows in B
 * </ol>
 * </p>
 * @author John Collins
 * @version $Revision: 1.13.2.2 $
 */
public class BidtreeSearch
  extends AbstractSearchEngine
  implements SearchEngine
{
  // Init the log categories
  static Category log = Category.getInstance(BidtreeSearch.class.getName());
  static Category result = Category.getInstance("Result");
  
  // Search queue and parameters
  private TaskBidPlan m_initialNode = null;
  private Bidtree m_bidtree = null;
  private SearchQueue m_searchQueue
    = SearchQueue.create(SearchQueue.BEST_FIRST);
  private int maxQueueSize = 20000;
  private double m_upperBound = Double.MAX_VALUE;
  private int sortDirection = Bidtree.INCREASING;
  private boolean m_useSortedBids = true;
  
  // Statistics
  private int m_nodeCount = 0;  
  private BidtreeNode m_bestPlan = null;  //the best covered and feasible node
  
  // Evaluators and Selectors
  private Eval m_eval;
  
  /**
   * Default constructor.
   */
  public BidtreeSearch ()
  {
    super();
  }

  public String getSearchType ()
  {
    return "A*";
  }

  public void setBidEvaluator (BidEvaluator be)
  {
    super.setBidEvaluator(be);
    readSystemProps();
    m_bidtree = Bidtree.create(be, sortDirection);
  }

  /**
   * Run the search.  This is the primary implementation of the
   * SearchEngine interface.  In this case, we just copy some pointers
   * over from the BidEvaluator, set up some randomizers, and take off.
   */
  
  public TaskBidPlan search()
  {
    if (!getBidEvaluator().checkCoverage())
      return null; // no coverage possible
    m_eval = new Eval(bidEval); 
    m_eval.setUsingCoveragePenalty(false);
    
    BidtreeNode result = startSearch();
    if (result == null)
      return null;
    else
      return result.getTaskBidPlan();
  }

  /**
   * Sets the initial search node
   */
  public void setInitialNode (TaskBidPlan node)
  {
    m_initialNode = node;
  }

  public BidtreeNode startSearch()
  {
    log.info("BidtreeSearch.startSearch()");
    //m_searchQueue.setMaxSize(maxQueueSize);

    // Create a root node
    TaskBidPlan tbp = m_initialNode;
    BidtreeNode node = null;
    if (tbp == null) {
      tbp = new TaskBidPlan(bidEval.getTaskPlan());
    }
    node = new BidtreeNode(tbp, this);

    // evaluate the root node
    double rootEval = m_eval.evaluate(tbp);

    // add the root node
    m_searchQueue.insert((SearchNode)node);
    m_nodeCount++;

    // run the search
    nodeGenerator();    // this is the heart of the search

    // record results
    if (m_searchQueue.wasTrimmed())
      log.warn("queue was trimmed");
    log.info("Complete: Node count:  " + m_nodeCount
	     + ", Feasible and Covered Nodes : " + getPlanCount());
    result.info("Node Count = " + m_nodeCount
		+ ", Feasible and Covered Nodes : " + getPlanCount()
		+ (m_searchQueue.wasTrimmed()?"-opt":""));
    //if (log.isDebugEnabled()) {
    //  if (m_bestPlan == null)
    //log.debug("I have not found a feasible and covered node");
    //else
    //log.debug(m_bestPlan.toString());
    //}
    return m_bestPlan;
  }
    
  protected void nodeGenerator()
  {
    log.debug("nodeGenerator()");
    boolean trimmedOnce = false; // this gets set if we've tried
                                 // trimming the size of the search
                                 // queue. 
    
    while (!m_searchQueue.isEmpty()
	   && m_bestPlan == null) {
      // Get the first element from the queue and expand it. If it
      // returns null, then remove it from the queue and go again. 
      BidtreeNode nextNode = (BidtreeNode)m_searchQueue.select();
      if (nextNode.isCovered() && nextNode.isFeasible()) {
	//log.debug("Covered, feasible node found");
	m_bestPlan = nextNode;
      }
      else {
	if (nextNode.hasMoreBids()) {
	  //log.debug("Trying node " + nextNode.getNodeNumber());
	  BidtreeNode newNode = nextNode.expandNode(m_nodeCount++);
	  
	  // If the returned node is a complete solution, we're
	  // done. Otherwise, add it to the queue.
	  if (newNode == null) {
	    //log.debug("Null result - extracting node "
	    //	      + nextNode.getNodeNumber());
	    m_searchQueue.extract();
	  }
	  else if (newNode.getTaskBidPlan() != null) {
	    m_eval.evaluate(newNode.getTaskBidPlan());
	    if (newNode.isFeasible()
		&& newNode.getEvaluation() < m_upperBound) {
	      //log.debug("Inserting new node "
	      //		+ newNode.getNodeNumber()
	      //		+ ", eval = "
	      //		+ newNode.getEvaluation());
	      m_searchQueue.insert((SearchNode)newNode);
	      if (newNode.isCovered()) {
		processSolution(newNode.getTaskBidPlan());
		m_upperBound = newNode.getEvaluation();
		trimmedOnce = false;
	      }
	      if (m_searchQueue.size() > maxQueueSize) {
		// If an upper bound has been set, we can try trimming
		// back the queue, otherwise we have to abort the
		// search here
		if (m_upperBound < Double.MAX_VALUE
		    && !trimmedOnce) {
		  // we've got a bound, try trimming the queue
		  m_searchQueue.trimByPredicate
		      (new Tester() {
			  public boolean test (SearchNode node) {
			    return node.value() > m_upperBound;
			  }
			});
		  trimmedOnce = true;
		}
		m_searchQueue.clear();
		log.warn("Queue maxed out: Search aborted");
		result.info("Search aborted");
	      }
	    }
	    else {
	      //log.debug("Infeasible node " + newNode.getNodeNumber());
	    } // if (newNode.isCovered() && newNode.isFeasible())
	  }
	  else {
	    //log.debug("Null TBP in node " + newNode.getNodeNumber());
	  } // if (newNode == null)
	}
	else {
	  //log.debug("no more bids in " + nextNode.getNodeNumber());
	  m_searchQueue.extract();
	} // if (nextNode.hasMoreBids())
      } // if next node covered & feasible
    } // while ...
  }

  /**
   * Gets the node count
   */
  public int getNodeCount ()
  {
    return m_nodeCount;
  }

  /**
   * Returns the Bidtree structure used by this Search instance
   */
  public Bidtree getBidtree ()
  {
    if (m_bidtree == null)
      m_bidtree = Bidtree.create(getBidEvaluator());
    return m_bidtree;
  }

  /**
   * Returns the value of the "use sorted bids" flag
   */
  public boolean isUseSortedBids ()
  {
    return m_useSortedBids;
  }

  protected void readSystemProps ()
  {
    String prop;
    // Read the max queue size
    prop = MagnetProperties.getProperty("BTS.MaxQueueSize");
    if (prop != null) {
      try {
	double value = Double.valueOf(prop).doubleValue();
	maxQueueSize = (int)Math.round(value);
	log.info("Max queue size = " + maxQueueSize);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		  + " for max queue size. Using "
		  + maxQueueSize);
      }
    }
    // Read the task-sort criterion
    prop = MagnetProperties.getProperty("BTS.SortDirection");
    if (prop != null) {
      if (prop.equals("Increasing"))
	sortDirection = Bidtree.INCREASING;
      else if (prop.equals("Decreasing"))
	sortDirection = Bidtree.DECREASING;
      else {
	log.error("Bad property spec for " + prop
		  + ", using INCREASING");
	sortDirection = Bidtree.INCREASING;
      }

    // Read the "use sorted bids" property
    prop = MagnetProperties.getProperty("BTS.UseSortedBids");
    if (prop != null) {
      m_useSortedBids = Boolean.valueOf(prop).booleanValue();
      log.info("BTS use sorted bids: " + m_useSortedBids);
    } // end of if ()
    }
  }
}				
