/*------------------------------------------------------------------------
 * MAGNET bid evaluation component
 * $Id: BtBucket.java,v 1.1.2.2 2002/05/05 19:41:47 ashutosh Exp $
 * Copyright (c) 2002 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Sun Mar 10 15:32:24 2002  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.protocol.cust.Bid;
import org.apache.log4j.Category;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;

/**
 * This is a holder for a set of bids that cover the same tasks.
 * @author John Collins
 * @version $Revision: 1.1.2.2 $
 */
public class BtBucket
{
  // This is the list of bids, sorted by price/task.
  private ArrayList bids = new ArrayList();
  BtBidWrapper[] sortedBids = null;

  // All bids in a bucket have the same task sets. This is the set if
  // task indices for this bucket.
  int[] tasks = null;

  // A bucket knows its index within its Bidtree
  int index = -1;

  // This is the list of buckets whose bids conflict with the bids in
  // this bucket.
  int[] conflicts = null;

  // This is the sorting comparator. We create only one for
  // efficiency.
  private static Comparator bidComparator =
    new Comparator() {
	public int compare (Object o1, Object o2) {
	  return (int)Math.round(((BtBidWrapper)o1).getCostPerTask()
				 - ((BtBidWrapper)o2).getCostPerTask());
	}
      };

  public BtBucket (int[] taskIndices)
  {
    super();
    tasks = taskIndices;
  }

  /**
   * Sets the index of this bucket. 
   */
  public void setIndex (int i)
  {
    index = i;
  }

  /**
   * Returns the index of this bucket.
   */
  public int getIndex ()
  {
    return index;
  }

  /**
   * Returns the task indices of the tasks covered by bids in this
   * bucket.
   */
  public int[] getTaskIndices ()
  {
    return tasks;
  }

  /**
   * Adds a bid to this bucket. This resets the sorted bids list.
   */
  public void addBid (BtBidWrapper b)
  {
    bids.add(b);
    sortedBids = null;
  }

  /**
   * Retrieves the BidWrappers from this bucket as a sorted
   * array. Note that the bids are sorted on demand, and will be
   * re-sorted only if more bids have been added.
   */
  public BtBidWrapper[] getSortedBids ()
  {
    if (sortedBids == null) {
      sortedBids = new BtBidWrapper[bids.size()];
      sortedBids = (BtBidWrapper[])bids.toArray(sortedBids);
      Arrays.sort(sortedBids, bidComparator);
      for (int i = 0; i < sortedBids.length; i++) {
	sortedBids[i].setBucketIndex(this.index);
      }
    }
    return sortedBids;
  }

  /**
   * Returns the BidWrapper at the given index, if any. Note that
   * there is no check for overflow here. You are on your own.
   */
  public BtBidWrapper getBid (int i)
  {
    return getSortedBids()[i];
  }

  /**
   * Returns the size of this bucket, which is the number of bids it
   * holds.
   */
  public int size ()
  {
    return getSortedBids().length;
  }

  /**
   * Returns the conflict indices. These are the index values for all
   * the buckets whose bids conflict with the bids in this bucket.
   */
  public int[] getConflicts ()
  {
    return conflicts;
  }

  public String toString ()
  {
    StringBuffer sb = new StringBuffer();
    sb.append("BtBucket[");
    String sep = "";
    for (int i = 0; i < tasks.length; i++) {
      sb.append(sep).append(tasks[i]);
      sep = ",";
    }
    sb.append("][");
    BtBidWrapper[] bids = getSortedBids();
    sep = "";
    for (int i = 0; i < bids.length; i++) {
      sb.append(sep).append(bids[i].getBid().getId());
      sep = ",";
    }
    sb.append("]");
    return sb.toString();
  }
}
