/*------------------------------------------------------------------------
 * Magnet Bid Evaluation components
 * $Id: Eval.java,v 1.4.2.4 2002/06/06 15:33:35 jcollins Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Thu Jun  6 10:28:49 2002  John Collins  #0000207
 *   Correct setting of coverage flag when using coverage penalty
 * 
 * Fri Apr 20 23:35:22 2001  John Collins
 *   No more System.getProperty
 * 
 * Mon Nov 27 21:12:42 2000  John Collins
 *   Update to use log4j logging
 * 
 * Sun Oct  1 16:17:10 2000  John Collins
 *   Evaluate coverage and cost only on tasks that are currently
 *   offered for bid
 * 
 * Sat Jan 16 17:37:10 1999  John E. Collins
 *   Init the evaluators with the model.
 * 
 * Mon Jan  4 21:25:33 1999  John E. Collins
 *   Use System Properties to set coverage and feasibility penalties.
 *   Don't calculate risk if plan is not covered and feasibile.
 * 
 * Tue Sep 29 21:22:44 1998  John E. Collins
 *   Implement the Evaluator interface
 * 
 * Thu Sep 24 17:26:05 1998  John E. Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.util.MagnetProperties;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.Bid;

import org.apache.log4j.Category;

import java.util.Enumeration;
import java.util.Iterator;
import java.util.Properties;

/**
   This class and its subclasses are the evaluators for the Magnet
   client bid-evaluation process.  Each class has an <tt>evaluate</tt>
   method that returns a double value representing the evaluation.
   <p>
   This class and its subclasses are intended to be singletons; you
   create one of these for your search engine, and it takes care of
   the rest.
   @author John Collins
 */
public class Eval implements Evaluator
{
  // Init the log categories
  static Category log = Category.getInstance(Eval.class.getName());

  // These are the evaluation-type symbols for this evaluator
  /** Symbol for overall evaluation */
  static public final String TYPE = "Overall";
  /** Symbol for cost evaluation */
  static public final String COST = "Cost";

  private BidEvaluator m_model;
  private CPMeval m_cpm;
  //private RiskEval m_risk;
  //private CompletionTimeEval m_completion;
  private Properties m_props;

  private double m_coveragePenaltyRatio = 1.0;
  private double m_coveragePenalty = 0.0; // per task
  private boolean m_usingCoveragePenalty = true;
  private double m_feasibilityPenalty = 0.10; // per ms of negative slack
  private double m_deadlinePenalty = 0.0; // up to +/- 10/task

  /**
   * Default constructor to simplify testing. <br>
   * <i>Note</i>: Unfortunately, you can't really do much with one of
   * these unless you somehow set up the model.
   */
  public Eval () 
  {
    super();
    m_cpm = new CPMeval();
    //m_risk = new RiskEval();
    //m_completion = new CompletionTimeEval();
    init();
  }

  /**
   * The constructor needs a pointer to its model
   */
  public Eval (BidEvaluator model) 
  {
    super();
    m_model = model;
    m_cpm = new CPMeval(model);
    //m_risk = new RiskEval(model);
    //m_completion = new CompletionTimeEval(model);
    init();
  }

  /**
   * Read the system properties to set this thing up.
   */
  private void init ()
  {
    // Retrieve coverage and feasibility penalty settings from system
    // properties.
    m_props = MagnetProperties.getProperties();
    String prop;
    prop = m_props.getProperty("CoveragePenalty");
    if (prop != null)
    {
      try
      {
	double value = Double.valueOf(prop).doubleValue();
	setCoveragePenalty(value);
	log.info("prop: CoveragePenalty = " + value);
      }
      catch (NumberFormatException nfe)
      {
	log.error("Bad property spec " + prop + " for Coverage Penalty.");
      }
    }

    prop = m_props.getProperty("Eval.UncoveredCostRatio");
    if (prop != null)
    {
      try
      {
	double value = Double.valueOf(prop).doubleValue();
	m_coveragePenaltyRatio = value;
	log.info("prop: CoveragePenalty = " + value);
      }
      catch (NumberFormatException nfe)
      {
	log.error("Bad property spec " + prop + " for Coverage Penalty.");
      }
    }

    prop = m_props.getProperty("FeasibilityPenalty");
    if (prop != null)
    {
      try
      {
	double value = Double.valueOf(prop).doubleValue();
	setFeasibilityPenalty(value);
	log.info("prop: FeasibilityPenalty = " + value);
      }
      catch (NumberFormatException nfe)
      {
	log.error("Bad property spec " + prop
		  + " for Feasibility Penalty.  Using "
		  + getFeasibilityPenalty());
      }
    }
    else
      log.info("default prop: FeasibilityPenalty = "
	       + getFeasibilityPenalty());

    prop = m_props.getProperty("DeadlinePenalty");
    if (prop != null)
    {
      try
      {
	double value = Double.valueOf(prop).doubleValue();
	setDeadlinePenalty(value);
	log.info("prop: DeadlinePenalty = " + value);
      }
      catch (NumberFormatException nfe)
      {
	log.error("Bad property spec " + prop
		  + " for Deadline Penalty.  Using "
		  + getDeadlinePenalty());
      }
    }
    else
      log.info("default prop: DeadlinePenalty = "
	       + getDeadlinePenalty());
  }

  /**
     Return the evaluation of a node in the search space.  The
     evaluation has four components:
     <ul>
     <li>The actual cost of the plan.  This is defined as the sum of
     the bid prices across the subtasks in the plan.  If a subtask
     is not represented by a bid in the plan, then its average
     price will be used.  It is assumed that evaluation will not
     be attempted for plans that have one or more subtasks without
     bids.
     <li>The expected cost of plan maintenance and repair.
     For each subtask, this is composed of:
     <ul>
     <li>the probability of customer decommitment times
     decommitment cost, integrated over time, plus
     <li>the probability of supplier decommitment, times the cost
     of recovery minus the penalty paid by the supplier,
     integrated over time.
     </ul>
     <li>The expected cost of plan failure.  Factors include total
     outlay for work performed, decommitment costs for withdrawn
     bid awards, and residual value of work performed.
     <li>A penalty for lack of coverage.  For each subtask, this is
     the ratio of <tt>Eval.getCoveragePenalty</tt> and the number
     of bids received for that subtask.
     <li>A penalty for schedule infeasibility.  This is the total
     negative slack times <tt>Eval.getFeasibilityPenalty</tt>.
     </ul>
     As a side effect, the evaluation attributes of the TaskBidPlan
     and its TaskBid objects are updated.
     @return a double representing the final evaluation is returned,
     as well as being stored in the TaskBidPlan.
  */
  public double evaluate (TaskBidPlan plan) 
  {
    double result = 0.0;
    if (plan.evaluationExists(TYPE)) {
      result = plan.findEvaluation(TYPE);
    }
    else {
      // Figure the cost and coverage penalty
      long cost = 0;
      double cov = 0.0;
      // Compute penalties for uncovered tasks
      plan.setCovered(true);
      if (m_usingCoveragePenalty) {
	for (Iterator tbs = m_model.getOfferedTasks();
	     tbs.hasNext(); ) {
	  TaskBid tb = plan.findTaskBid(((SubtaskRequest)tbs.next()).getId());
	  if (tb.getBid() == null) {
	    TaskDescriptor td = m_model.findTaskDescriptor(tb.getTaskId());
	    cov += getCoveragePenalty();
	    cost += td.getSubtaskRequest().getJobInfo().getExpectedPrice();
	    plan.setCovered(false);
	  }
	}
      }
      else {
	// still have to check for coverage, even if there's no
	// penalty
	for (Iterator tbs = m_model.getOfferedTasks();
	     tbs.hasNext(); ) {
	  TaskBid tb = plan.findTaskBid(((SubtaskRequest)tbs.next()).getId());
	  if (tb.getBid() == null) {
	    plan.setCovered(false);
	  }
	}
      } // if (m_usingCoveragePenalty)

      // sum up the costs of the bids
      for (Iterator bids = plan.getBids().iterator();
	   bids.hasNext(); ) {
	Bid theBid = (Bid)bids.next();
	cost += theBid.getCost();
      }
      plan.putEvaluation(COST, cost);
      result = (double)cost + cov;
      
      // Figure the feasibility penalty
      double negslack = m_cpm.evaluate(plan);
      double slackPenalty = (negslack * getFeasibilityPenalty());
      plan.putEvaluation(m_cpm.getEvalClass(), slackPenalty);
      result += slackPenalty;
      plan.setFeasible(negslack == 0.0);
      
      // Figure the risk penalty
      double riskPenalty = 0.0;
      //if (slackPenalty == 0.0 && cov == 0.0) {
      //riskPenalty = m_risk.evaluate(plan);
      //plan.putEvaluation(m_risk.getEvalClass(), riskPenalty);
      //result += riskPenalty;
      //}

      // Figure the deadline penalty if the coefficient is non-zero
      double deadlinePenalty = 0.0;
      //if (Eval.getDeadlinePenalty() != 0.0) {
      //deadlinePenalty = (m_completion.evaluate(plan)
      //		   * Eval.getDeadlinePenalty());
      //plan.putEvaluation(m_completion.getEvalClass(), deadlinePenalty);
      //result += deadlinePenalty;
      //}

      plan.putEvaluation(TYPE, result);
      //if (log.isDebugEnabled())
      //log.debug("Evaluation: cost = " + cost + ", coverage = " + cov
      //	  + ", infeas = " + slackPenalty);
    }
    return result;
  }

  /**
   * Return the evaluation class.
   */
  public String getEvalClass () 
  {
    return TYPE;
  }

  /**
   * True just in case coverage penalties are being applied by this
   * evaluator.
   */
  public boolean isUsingCoveragePenalty ()
  {
    return m_usingCoveragePenalty;
  }

  /**
   * Enables/disables the use of coverage penalties
   */
  public void setUsingCoveragePenalty (boolean newFlag)
  {
    m_usingCoveragePenalty = newFlag;
  }

  /**
   * Returns the objective function cost for a subtask that is not
   * covered by a bid.  This is a lazy-eval method.  If the value has
   * not yet been set, then it is computed from the average price of
   * all tasks and the defaultCoveragePenalty.
   */
  public double getCoveragePenalty () 
  {
    if (m_coveragePenalty == 0.0) {
      m_coveragePenalty = (m_model.getMaxTaskPrice()
			   * m_coveragePenaltyRatio);
    } // end of if ()
    return m_coveragePenalty;
  }

  /**
   * Sets the cost factor for non-covered subtasks.  You only call
   * this if you want to set a fixed value for coverage penalty,
   * rather than computing it as a ratio of average task cost.
   */
  public void setCoveragePenalty (double newValue) 
  {
    m_coveragePenalty = newValue;
  }

  /**
   * Returns the penalty for a millisecond of negative slack in the
   * schedule.
   */
  public double getFeasibilityPenalty () 
  {
    return m_feasibilityPenalty;
  }

  /**
   *  Sets the penalty for a millisecond of negative slack.
   */
  public void setFeasibilityPenalty (double newValue) 
  {
    m_feasibilityPenalty = newValue;
  }

  /**
   * Gets the completion-time penalty
   */
  public double getDeadlinePenalty ()
  {
    return m_deadlinePenalty;
  }

  /**
   * Sets the completion-time penalty
   */
  public void setDeadlinePenalty (double newValue)
  {
    m_deadlinePenalty = newValue;
  }
}
