/*------------------------------------------------------------------------
 * Magnet bid evaluation components
 * $Id: IPSearch.java,v 1.6.2.3 2002/05/05 19:41:48 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Fri Apr 20 23:38:46 2001  John Collins
 *   Remove System.getProperties
 * 
 * Fri Jan 12 17:41:54 2001  John Collins
 *   Remove coverage test - it migrated to BidEvaluator
 * 
 * Fri Nov 17 12:15:26 2000  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.cust.TaskWithPrecedence;

import edu.umn.magnet.util.Cons;
import edu.umn.magnet.util.MagnetProperties;
import edu.umn.magnet.util.time.AgentTime;

import org.apache.log4j.Category;

import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.IOException;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.StringTokenizer;

/**
 * Classes that implement this interface are search engines, capable
 * of generating a mapping between bids and subtasks for some task
 * plan and set of bids.
 * <p>
 * Classes that implement this interface must meet two criteria:
 * <ul>
 * <li> They must provide a no-argument constructor, and
 * <li> They must implement the <tt>search</tt> method as described below.
 * </ul>
 * @author John Collins
 */
public class IPSearch
  extends AbstractSearchEngine
  implements SearchEngine
{
  // init the log category
  static Category log = Category.getInstance(IPSearch.class.getName());
  static Category resultLog = Category.getInstance("Result");

  // The "big number" used to do logic -
  // We have to be careful with number size because the LP system is
  // written in C, and uses something like doubles for numbers.
  static final long M = Long.MAX_VALUE/10000000L;

  // I/O channels - these will be hooked to the lp process
  private PrintWriter out = null;
  private BufferedReader in = null;
  private Process solver = null;
  private StringBuffer outBuffer = new StringBuffer();
  private int eqnCount = 0;

  // instance database
  //BidEvaluator bidEval = null;
  private HashMap bid2index = new HashMap();
  private String[] varNames;   // indexed set of variable names
  private HashMap name2bid = new HashMap();
  private TaskBidPlan finalResult = null;
  private double epsilon = 0.0000000001;

  // these are for the feasibility constraints
  private long[] coefficients; // place to collect coefficients
  private long earliestStart;

  /**
   * The constructor just captures the BidEvaluator reference and sets
   * properties. 
   */
  public IPSearch ()
  {
    super();
  }

  public String getSearchType ()
  {
    return "IP";
  }

  public void setBidEvaluator (BidEvaluator evaluator)
  {
    bidEval = evaluator;

    // Get the epsilon property from the config file
    String prop = MagnetProperties.getProperty("IP.epsilon");
    if (prop != null) {
      try {
	double value = Double.valueOf(prop).doubleValue();
	epsilon = value;
	log.info("prop: IP epsilon = " + value);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		 + " for IP epsilon.  Using " + epsilon);
      }
    }
    else
      log.warn("default prop: IP epsilon = " + epsilon);
  }
    
  /**
   * Find and return a mapping from subtasks to bids.  This engine
   * does it with an external lp solver.
   * @param evaluator a BidEvaluator object that contains the task
   *   plan and the set of bids to be searched.
   * @return a TaskBidPlan that represents the bid->task mapping
   *   resulting from the search.
   */
  public TaskBidPlan search ()
  {
    // 1.0 Start up the lp search process and set up the streams to
    // communicate with it.
    invokeSolver();

    // 2.0 write the problem to the output stream
    //  2.1 build the variable-name:bid table
    assignVarNames();
    //  2.2 write the objective function
    writeObjFn();
    //  2.3 write the coverage constraints
    writeCoverageConstraints();
    //  2.4 write the feasibility constraints
    writeFeasConstraints();
    //  2.5 write the bid compatibility constraints
    //writeCompatibilityConstraints();
    //  2.6 write the {0,1} constraints on the variables
    write01Constraints();

    // 3.0 invoke the lp solver
    out.close();

    // 4.0 initialize the result, then read in the results and update
    // the result
    finalResult = new TaskBidPlan(bidEval.getTaskPlan());
    readResults();
    // evaluate the node and return results
    if (finalResult != null) {
      Eval eval = new Eval(bidEval);
      double evalResult = eval.evaluate(finalResult);
      finalResult.setSearchData("IP", getRunTime());
      log.info("Result: " + evalResult);
    }
    return finalResult;
  }

  // -------------------------------------------------------------

  /**
   * Creates an array of names, and a HashMap that associates bids
   * with indices into the array.  We use this two-step process to
   * simplify the generation of constraints.
   */
  protected void assignVarNames()
  {
    log.debug("assignVarNames");
    // We'll use the ID of the bid for a var name, and we'll decorate
    // it with a suffix if it's not unique.
    // We start by developing an index number for each bid.
    int indexCounter = 0;
    Iterator bids = bidEval.getBids().iterator();
    while (bids.hasNext()) {
      Bid theBid = (Bid)bids.next();
      bid2index.put(theBid, new Integer(indexCounter));
      indexCounter++;
    }

    // Now we initialize the arrays and fill in names
    int suffixCounter = 0;
    varNames = new String[indexCounter];
    coefficients = new long[indexCounter];
    HashSet used = new HashSet(indexCounter);
    for (bids = bid2index.keySet().iterator(); bids.hasNext(); ) {
      Bid theBid = (Bid)bids.next();
      int index = ((Integer)bid2index.get(theBid)).intValue();
      String varName = theBid.getId();
      while (used.contains(varName)) {
	varName = theBid.getId() + "x" + suffixCounter++;
      }
      varNames[index] = varName;
      used.add(varName);
      name2bid.put(varName, theBid);
    }
    // Now you can get the name associated with a Bid as
    // varNames[((Integer)bid2index.get(bid)).intValue()]
  }

  /**
   * Writes out the variable name associated with the given bid.
   */
  private void appendVarName (Object bid)
  {
    outBuffer.append(varNames[((Integer)bid2index.get(bid)).intValue()]);
  }

  /**
   * Writes the objective function to the lp output stream.  For now,
   * the objective function is the sum of the costs of the bids (and
   * subtask bids) multiplied by their selection variables.
   */
  protected void writeObjFn ()
  {
    log.debug("writeObjFn");
    String sep = "";  // this will be " + " after the first term is
                      // written out
    outBuffer.append("min: ");
    Iterator bids = bid2index.keySet().iterator();
    while (bids.hasNext()) {
      // write out the separator, then the coefficient (cost), and
      // finally the var name.
      Bid bid = (Bid)bids.next();
      outBuffer.append(sep);
      outBuffer.append(bid.getCost());
      outBuffer.append(" ");
      appendVarName(bid);
      // after the first time through, we change the separator
      sep = " + ";
    }
    // the objective fn is on one line, terminated with a semicolon
    outBuffer.append(";");
    writeOutput();
  }

  /**
   * Writes out constraints that ensure that each task is included
   * exactly once from among the bids.
   */
  protected void writeCoverageConstraints ()
  {
    log.debug("writeCoverageConstraints");
    // iterate over the tasks.  For each one, write one constraint
    // that makes the sum of the bid variables for that task == 1
    Iterator tasks = bidEval.getOfferedTasks();
    while (tasks.hasNext()) {
      TaskDescriptor theTD
	  = bidEval.findTaskDescriptor((SubtaskRequest)tasks.next());
      String separator = "";
      Iterator bids = theTD.getBids();
      while (bids.hasNext()) {
	outBuffer.append(separator);
	appendVarName(bids.next());
	separator = " + ";
      }
      outBuffer.append(" = 1;");
      writeOutput();
    }
  }

  /**
   * Writes out the feasibility constraints.  See the paper for
   * details on their form.
   */
  protected void writeFeasConstraints ()
  {
    log.debug("writeFeasConstraints");

    // We use the recursive method constraintsForTask to emit
    // constraints for each of the tasks in the plan (or at least the
    // offered ones)
    earliestStart = (AgentTime.getDate()).getTime();
    Iterator tasks = bidEval.getOfferedTasks();
    while (tasks.hasNext()) {
      constraintsForTask((SubtaskRequest)tasks.next(), (Cons)null);
    }
  }

  /**
   * Writes out the set of feasibility constraints for one task.
   */
  private void constraintsForTask (SubtaskRequest task,
				   Cons head)
  {
    if (log.isDebugEnabled())
      log.debug("constraintsForTask("
		+ task.getId() + " " + head + ")");

    TaskPlan plan = bidEval.getTaskPlan();
    TaskDescriptor theTD = bidEval.findTaskDescriptor(task);

    // Note: we are not checking to see that tasks are in the OFFERED
    // state.

    // There are two cases:
    // 1. head is null: If this task has predecessors, then for each
    //    bid, emit the late-start term, then recurse on the
    //    predecessors.
    if (head == null) {
      Iterator preds = plan.getPredecessors(task);
      // This is an Iterator of SubtaskRequest instances
      if (preds.hasNext()) {
	// Now we get to do something.  For each bid, there's a
	// corresponding subtask bid.  We need to do this for both.
	Iterator bids = theTD.getBids();
	while (bids.hasNext()) {
	  Bid theBid = (Bid)bids.next();
	  if (log.isDebugEnabled())
	    log.debug("New constraint: task=" + task.getId()
		      + ", bid=" + theBid.getId());
	  // compute latest possible start time, relative to
	  // earliestStart
	  long f = (theBid.getLateFinish(task.getId()).getTime()
		    - theBid.getDuration(task.getId())
		    - earliestStart);
	  Cons term1 = new Cons(new ConstraintTerm (theBid, (M - f)));
	  preds = plan.getPredecessors(task);
	  while (preds.hasNext()) {
	    SubtaskRequest nextTask = (SubtaskRequest)preds.next();
	    constraintsForTask(nextTask, term1);
	  }
	}
      }
    }

    // 2. head is not null: for each bid, emit the terminal term and
    //    format the output.  Then, if the task has predecessors, for
    //    each bid, emit an intermediate term and recurse on
    //    predecessors.
    else {
      // first, the terminal term
      Iterator bids = theTD.getBids();
      while (bids.hasNext()) {
	Bid theBid = (Bid)bids.next();
	// compute the earliest possible finish time relative to
	// earliestStart
	long e = (theBid.getEarlyStart(task.getId()).getTime()
		  + theBid.getDuration(task.getId())
		  - earliestStart);
	formatFeasibilityConstraint
	    (head.cons(new ConstraintTerm (theBid, e)));
      }

      // then the intermediate term, if there are predecessors
      Iterator preds = plan.getPredecessors(task);
      if (preds.hasNext()) {
	bids = theTD.getBids();
	while (bids.hasNext()) {
	  Bid theBid = (Bid)bids.next();
	  long d = theBid.getDuration(task.getId());
	  int index = 0;
	  Cons term1 = head.cons(new ConstraintTerm(theBid, d));
	  preds = plan.getPredecessors(task);
	  while (preds.hasNext()) {
	    SubtaskRequest nextTask = (SubtaskRequest)preds.next();
	    if (term1 != null)
	      constraintsForTask(nextTask, term1);
	  }
	}
      }
    }
  }

  /**
   * Formats and writes out a single feasiblity constraint.
   */
  private void formatFeasibilityConstraint (Cons list)
  {
    if (log.isDebugEnabled())
      log.debug("formatFeasibilityConstraint(" + list + ")");
    log.assert((list.cdr() != null), "One-element list");

    // The list is a series of ConstraintTerm instances.  There are
    // two steps:
    // 1. Walk the list, fill up the coefficients array.  We simply
    //    add the value of the ConstraintTerm to the indexed cell in
    //    the array.  Keep in mind that the list is
    //    reversed; the head of the list is the last term generated.
    while (list != null) {
      ConstraintTerm term = (ConstraintTerm)list.car();
      int index = ((Integer)bid2index.get(term.bid)).intValue();
      coefficients[index] += term.value;
      list = (Cons)list.cdr();
    }

    // 2. Iterate through the coefficients array, filling up a
    //    StringBuffer with terms.  If there is more than one term,
    //    then add the RHS and emit the result to the output stream.
    //    While we do this, we also clear out the coefficients array.
    StringBuffer buf = new StringBuffer();
    String sep = "";
    int count = 0;
    for (int i = 0; i < coefficients.length; i++) {
      if (coefficients[i] != 0L) {
	buf.append(sep);
	buf.append(coefficients[i]);
	buf.append(" ");
	buf.append(varNames[i]);
	sep = " + "; // the rest of the terms get added
	count++;
	coefficients[i] = 0L;
      }
    }
    if (count > 1) {
      // it's worth writing a constraint.  Maybe it's possible that a
      // bid could be internally infeasible, but that should be caught
      // earlier.
      buf.append(" <= ");
      buf.append(M);
      buf.append(";");
      outBuffer.append(buf);
      writeOutput();
    }
    else {
      // we need to clear out the output buffer if we aren't going to
      // use it.
      outBuffer.setLength(0);
    }
  }

  /**
   * Writes out a constraint for each pair of incompatible bids,
   * ensuring that no incompatible pairs are included in the solution.
   * At the moment, this is probably a waste of time, because if the
   * only source of incompatibility is overlap (2 bids specifying the
   * same task), then the coverage constraints are enough.
   */
  protected void writeCompatibilityConstraints()
  {
    log.debug("writeCompatibilityConstraints");

    // For the moment, what's incompatible are the bids and their
    // subtask bids.  We'll have to rewrite this when we eliminate
    // subtask bids and have supplier id's in the bids.
  }

  /**
   * Writes out a set of constraints that ensure that the bid selector
   * variables are either 0 or 1.
   */
  protected void write01Constraints ()
  {
    log.debug("write01Constraints");

    // First, we make each var be >= 0 and <= 1
    Iterator bids = bid2index.keySet().iterator();
    while (bids.hasNext()) {
      // 2 constraints for each bid, but the solver makes all vars >=
      // 0 already.
      Bid bid = (Bid)bids.next();
      outBuffer.append(varNames[((Integer)bid2index.get(bid)).intValue()]);
      outBuffer.append(" <= 1;");
      writeOutput();
    }
    // Then we make each var an int
    outBuffer.append("int");
    bids = bid2index.keySet().iterator();
    while (bids.hasNext()) {
      // This is just a space-separated list of bid names
      Bid bid = (Bid)bids.next();
      outBuffer.append(" ");
      outBuffer.append(varNames[((Integer)bid2index.get(bid)).intValue()]);
    }
    // the int list is on one line, terminated with a semicolon
    outBuffer.append(";");
    writeOutput();
  }

  /**
   * Writes accumulated output to the solver stream, and also
   * optionally to the debug stream.
   */
  protected void writeOutput ()
  {
    // first, we keep track of the number of formulas we are generating.
    eqnCount++;
    if (eqnCount % 1000 == 0)
      System.out.println("IPSearch: " + eqnCount);
    out.println(outBuffer);
    log.debug(outBuffer);
    outBuffer.setLength(0);
  }

  /**
   * Starts up the solver and hook up its streams
   */
  protected void invokeSolver ()
  {
    log.debug("invokeSolver");
    try {
      solver = Runtime.getRuntime().exec("../../bin/lp_solve -e " + epsilon);
      out = new PrintWriter(solver.getOutputStream());
      in = new BufferedReader(new InputStreamReader(solver.getInputStream()));
    }
    catch (IOException io) {
      log.error("Exception running the solver", io);
    }
  }

  /**
   * Reads the output of the lp solver.  The first line gives the
   * value of the objective function, and the rest contain two fields
   * each: the name of a bid, and either a 0 or 1 to indicate whether
   * the bid is included in the solution.
   */
  protected void readResults ()
  {
    resultLog.info("IPSearch eqn count = " + eqnCount);
    log.debug("readResults");
    String line;
    StringTokenizer tokens;
    String token;
    Method nextMethod = getInputMethod("processFirstLine");
    try {
      while ((line = in.readLine()) != null) {
	// For each line, we invoke the "current" token processing
	// method, and then potentially update the method reference.
	// This allows for a line-oriented state machine.
	log.debug(line);
	tokens = new StringTokenizer(line);
	try {
	  Method res = (Method)nextMethod.invoke(this, new Object[] {tokens});
	  if (res != null) {
	    nextMethod = res;
	  }
	}
	catch (InvocationTargetException ite) {
	  if (ite.getTargetException() instanceof IPParseException) {
	    log.error("Caught IP Parse Exception");
	  }
	  finalResult = null;
	  return;
	}
	catch (Exception ex) {
	  log.error("Unexpected Exception", ex);
	}
      }
    }
    catch (IOException io) {
      log.error("Unexpected Exception", io);
    }
    finally {
      try {
	solver.waitFor();
      }
      catch (InterruptedException ie) {
	// don't do anything - this is OK
      }
    }
    return;
  }
  
  /**
   * Handles the first line of output from the lp solver.  It has the
   * form "Value of objective function: xxxx".  We assume there's only
   * one line of this form, and move on to the Bid lines with the next
   * line. 
   */
  protected Method processFirstLine (StringTokenizer line)
    throws IPParseException
  {
    log.debug("processFirstLine");
    String token = line.nextToken();
    if (token.equals("Value")) {
      while (line.hasMoreTokens()) {
	token = line.nextToken();
      }
      try {
	long obj = Long.parseLong(token);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad Obj Function Value: " + token);
      }
      // switch to the BidLine processor for the next line
      return getInputMethod("processBidLine");
    }
    else if (token.equals("This")) {
      while (line.hasMoreTokens()) {
	token = line.nextToken();
      }
      if (token.equals("infeasible")) {
	log.warn("Infeasible problem");
	throw new IPParseException("infeasible");
      }
      throw new IPParseException("first line match failed: " + token);
    }
    else {
      throw new IPParseException("first line match failed: " + token);
    }
  }

  /**
   * Handles lines of the form "[bid-name] value" where value is 1 or
   * 0.  We assume that once we start processing bid lines, all
   * remaining lines will be of this form, and so we simply return
   * null to indicate that the next line will use the same processing
   * method.
   */
  protected Method processBidLine (StringTokenizer line)
    throws IPParseException
  {
    log.debug("processBidLine");
    // We assume each line has two tokens: the name of a bid, and a 0
    // or 1 to indicate whether it is selected.  Of course, the lp
    // solver works with floats (long doubles actually), and so the
    // values may not be precisely 0 or 1.
    String name = "";
    double val = 0.0;
    try {
      name = line.nextToken();
      val = Double.parseDouble(line.nextToken());
      if (val > 0.9) {
	finalResult.addBid((Bid)name2bid.get(name));
	if (finalResult == null) {
	  throw new IPParseException("Expansion Failure");
	}
      }
    }
    catch (NumberFormatException nfe) {
      throw new IPParseException("Bad Input Line");
    }
    return null;
  }

  /**
   * Returns the method object with the given name and args.
   */
  protected Method getInputMethod (String methodName)
  {
    Class[] args = {StringTokenizer.class};
    try {
      return IPSearch.class.getDeclaredMethod(methodName, args);
    }
    catch (NoSuchMethodException nsm) {
      log.error("Method not found!", nsm);
      return null;
    }
    catch (SecurityException se) {
      log.error("Security problem!", se);
      return null;
    }
  }

  /**
   * data holder class for composing feasibility constraints
   */
  private class ConstraintTerm
  {
    Bid bid = null;
    long value = 0L;

    ConstraintTerm (Bid bid, long value)
    {
      super();
      this.bid = bid;
      this.value = value;
    }

    public String toString ()
    {
      return ("<term " + bid.getId() + " " + value + ">");
    }
  }

  /**
   * Exception thrown when IP input parse fails
   * @author John Collins
   */
  class IPParseException extends Exception
  {
    public IPParseException ()
    {
      super();
    }
    
    public IPParseException (String msg)
    {
      super(msg);
    }
  }
}
