/*----------------------------------------------------------------------------
 * Magnet Bid Evaluation components
 * $Id: SASearch.java,v 1.21.2.4 2002/06/06 15:33:35 jcollins Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *----------------------------------------------------------------------------
 * 
 * Revision History:
 * 
 * Thu Jun  6 10:28:11 2002  John Collins  #0000207
 *   Count solutions correctly
 * 
 * Fri Oct  5 11:27:15 2001  John Collins
 *   Performance improvements
 * 
 * Sat Sep  8 16:03:32 2001  John Collins  #0000159
 *   Make queue length track patience factor
 * 
 * Mon Jul  2 18:03:39 2001  John Collins  #0000156
 *   Mods to support performance evaluation
 * 
 * Fri Apr 20 23:39:28 2001  John Collins
 *   Eliminate use of System properties
 * 
 * Sun Apr 15 14:14:56 2001  John Collins
 *   - Handle cases where singleton processing yields solutions.
 *   - Put search method "SA" and search runtime in solution nodes.
 * 
 * Thu Jan 25 15:01:43 2001  John Collins
 *   Correct problem with one bid covering multiple singletons
 * 
 * Fri Jan 12 09:55:56 2001  John Collins
 *   Add lockAllBids flag.
 * 
 * Thu Dec 14 11:14:43 2000  John Collins
 *   New bid semantics - no breakdowns.
 * 
 * Wed Nov 29 19:24:04 2000  John Collins
 *   When resetting the queue for a new run, we now create new nodes
 *   instead of reusing the old ones.  The result is that the
 *   "expansions tried" list goes away.
 * 
 * Sun Oct  1 20:46:50 2000  John Collins
 *   Make this take an initial node from the evaluator.
 * 
 * Sat Jan 15 08:53:07 2000  John E. Collins
 *   Get control of number of extra cycles in adaptive recycling
 * 
 * Thu Jan 13 10:08:40 2000  John E. Collins
 *   Add adaptive recycling with RecycleRatio parameter
 * 
 * Wed Jan 12 11:25:29 2000  John E. Collins
 *   Get recycling to work correctly, track iteration number of first
 *   solution.
 * 
 * Tue Jan 11 10:00:36 2000  John E. Collins
 *   Add property flags to control whether infeasible nodes get
 *   queued, and whether to restart the unique-nodes list when
 *   recycling.
 * 
 * Wed Jan  5 12:32:36 2000  John E. Collins
 *   Make bid selector come from named property, move Combined
 *   selector behavior out to an actual Selector
 * 
 * Tue Dec  7 11:26:19 1999  John E. Collins
 *   Update for new data collection scheme
 * 
 * Sun Dec  5 18:58:44 1999  John E. Collins
 *   Major updates, including uniqueness testing, and simplified control
 *   of the "combined" search.
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.protocol.cust.Bid;
//import edu.umn.magnet.client.eval.Selectors; // get the selectors compiled

import edu.umn.magnet.util.rvars.RandomVariableFactory;
import edu.umn.magnet.util.rvars.RVFManager;
import edu.umn.magnet.util.rvars.ExponentialVariable;
import edu.umn.magnet.util.stats.DataCollector;
import edu.umn.magnet.util.search.*;
import edu.umn.magnet.util.silk.SilkHook;
import edu.umn.magnet.util.MagnetProperties;

import org.apache.log4j.Category;

import java.util.Random;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Vector;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.NoSuchElementException;

/** 
 * This class is a Simulated Annealing implementation of the Magnet
 * SearchEngine interface.  The process is as follows:
 * <ul>
 * <li> Look through the TaskDescriptors to determine whether coverage
 *      is possible.  If one or more subtasks has no bids, 
 *      stop and return null.
 * <li> Create a root node for the search, and add it to the Node queue.
 * <li> While the search continues,
 *   <ul>
 *   <li> Choose a node from the Node queue, which is ordered by
 *        evaluation score.  Do not remove it.
 *   <li> Choose a bid from the Bid queue, and expand the node with
 *        the chosen bid.  Currently the Bid queue is prioritized by
 *        discount.  Other Bid queues may be incorporated with
 *        different orderings.
 *   </ul>
 * <li> Return the best complete and feasible node found.  If no
 *      complete and feasible nodes were found, return null.
 * </ul>
 */
public class SASearch
  extends AbstractSearchEngine
  implements SearchEngine
{
  // Init the log categories
  static Category log = Category.getInstance(SASearch.class.getName());
  static Category result = Category.getInstance("Result");

  // Search queue and parameters
  private TaskBidPlan m_initialNode = null;
  private SearchQueue m_searchQueue
    = SearchQueue.create(SearchQueue.BEST_FIRST);
  private QueueTest m_tester = new QueueTest();
  private int m_queueSize = 100;
  private double m_patienceFactor = 40.0;
  private int m_nullExpansionFactor = 4;
  private boolean m_testUniqueness = false;
  private boolean m_excludeInfeasibleNodes = false;
  private boolean m_restartUnique = false;
  private boolean m_lockAllBids = false;

  // contains initial contents of the queue, used for recycling
  private LinkedList m_initialQueue = new LinkedList();
  private int m_recycleCount = 0;

  // adaptive recycling
  private double m_recycleRatio = 1.0; // set >1.0 for adaptive behavior
  private double m_recycleRange = 1.0; // set >1.0 for random recycling
  private int m_extraCycles = 1; // one extra cycle after solution found
  private double m_recycleFactor = 1.0; // multiplied by recycleRatio
                                        // every time around.
  private double m_stopRatio = 0.0; // make this >1.0 to increase
                                    // complexity every time an
                                    // improvement is found 

  // Simulated Annealing parameters
  private double m_initialTemperature = 0.1;
  private double m_temperature = 0.1;
  private double m_annealingRatio = 0.8;
  private int m_rate = 0;     // deprecated "annealing rate"
  private double m_asf = 0.0; // annealing "scale factor"
  private double m_asfDefault = 0.2;
  private boolean m_stopOnSolution = false;

  // Statistics
  private int m_nodeCount = 0;  
  private int m_planCount = 0;
  private int m_firstSoln = -1;
  private int m_logInterval = 10; // only log every 10 iterations
  private DataCollector m_successfulExpansions;
  private SearchNode m_bestPlan = null; //the best covered and feasible node
  private SearchNode bestNode = null;   //the best node so far

  // Evaluators and Selectors
  private Eval m_eval;
  private BidSelector m_selector = new RandomBidSelector();
  private double m_scaleFactor = 0.0;

  /**
   * Default constructor only.  These are expected to be created by
   * the factory method in AbstractSearchEngine.
   */
  public SASearch ()
  {
    super();
  }

  public String getSearchType ()
  {
    return "SA";
  }

  public void setBidEvaluator (BidEvaluator be)
  {
    super.setBidEvaluator(be);
    readSystemProps();
  }

  /**
   * Sets the initial node for continuation searches, such as you have
   * when bidding is done in stages.
   */
  public void setInitialNode (TaskBidPlan initialNode)
  {
    m_initialNode = initialNode;
  }
 
  /**
   * Runs the search.  This is the primary implementation of the
   * SearchEngine interface.  In this case, we just copy some pointers
   * over from the BidEvaluator, set up some randomizers, and take off.
   */
  public TaskBidPlan search ()
  {
    if (bidEval == null)
      throw new Error("Bid Evaluator not specified");
    m_eval = new Eval(bidEval); 
    SearchNode.newSearch(bidEval);
    m_successfulExpansions = new DataCollector();
    m_searchQueue.setMaxSize(m_queueSize);
    //try
    //{
    //  Log.startDataCollection(be.getSessionId() + "_EVAL");
    //  Log.addDataCollector("EVAL", new TimedDataCollector());
    //}
    //catch (Exception e) 
    //{ e.printStackTrace(); }
    SearchNode result = startSearch();
    if (result == null)
      return null;
    else
      return result.getTaskBidPlan();
  }

  /**
   * Here's the top of the actual search.  First, make sure we have
   * coverage; otherwise we give up very quickly.  Then create the
   * root node, check for singletons, and start off the node
   * generation loop.
   */
  public SearchNode startSearch() {
    //have to check to see if all the subtasks have been covered
    //if not, then stop right away
    log.info("----- start search -----");

    //Create a root node
    TaskBidPlan tbp = m_initialNode;
    SearchNode node = null;
    if (tbp == null) {
      tbp = new TaskBidPlan(bidEval.getTaskPlan());
      node = new SearchNode(tbp);
    }
    else
      node = new SearchNode(tbp);

    // If we're set up to use coverage constraints, we need to
    // populate the initial SearchNode with the bid set and a
    // backpointer to the BidEvaluator
    if (bidEval.useCoverageConstraints()) {
      log.debug("Init coverage constraints in root node");
      node.setBidEvaluator(bidEval);
      node.setCompatibleBids(bidEval.getBids());
    }

    //evaluate the root node
    double rootEval = m_eval.evaluate(node.getTaskBidPlan());
    //add the root node
    m_initialQueue.add(node);
    m_successfulExpansions.newObservation(1.0);
    m_nodeCount++;
    // Check to see if there are Singletons. Singletons are those tasks
    // who only have one bid covering them
    createSingletons(); // do the easy stuff first, if there is any
    if (! m_initialQueue.isEmpty())
      nodeGenerator();    // this is the heart of the search
    
    result.info("Node Count = " + m_nodeCount
		+ ", Feasible and Covered Nodes : " + m_planCount);
    result.info("First solution at: " + m_firstSoln);
    if (log.isDebugEnabled()) {
      if (m_bestPlan == null)
	log.debug("I have not found a feasible and covered node");
      else
	log.debug(m_bestPlan.toString());
    }
    return m_bestPlan;
  }
  
  protected void createSingletons()
  {
    for (Iterator tDescriptors =
	     bidEval.getTaskDescriptors().iterator();
	 tDescriptors.hasNext(); ) {
      TaskDescriptor aTD =
	  (TaskDescriptor)tDescriptors.next();
      if (aTD.getBidCount() == 1) {
	String taskId = aTD.getTaskId();
	log.info("singleton bid for task " + taskId);
	Bid singleBid = (Bid)aTD.getBids().next();
	createSingletonChildren(singleBid);
      }
    }
  }

  /**
   * Attempts to expand all nodes currently in the queue with both the
   * given Bid and SubtaskBid.  The resulting new nodes, if any, are
   * added back to the queue.
   */
  protected void createSingletonChildren(Bid sBid)
  {
    LinkedList children = new LinkedList();
    while (!m_initialQueue.isEmpty()) {
      SearchNode removeNode = (SearchNode)m_initialQueue.removeFirst();
      if (removeNode.getTaskBidPlan().getBids().contains(sBid)) {
	// Task already covered by a previous singleton bid
	// -- in other words, one bid was a singleton bid for two
	//    or more different tasks
	log.debug("Singleton Shortcut...");
	children.add(removeNode);
      }
      else {
	SearchNode node = removeNode.expandNode(sBid, m_nodeCount);
	
	if (node != null) {
	  m_nodeCount++;
	  double e_score1 = m_eval.evaluate(node.getTaskBidPlan());
	  node.lockBid(sBid);
	  children.add(node);
	  m_successfulExpansions.newObservation(1.0);
	  processSolution(node);
	}
	else {
	  log.info("createSingletonChildren: failed to add "
		   + sBid.getId());
	  m_successfulExpansions.newObservation(0.0);
	}
      }
    }
    // Now take all the nodes we've collected and put them on the
    // queue.  We do this by replacing the initial queue with the new
    // children list.  This only works if they are the same type.
    m_initialQueue = children;
  }

  /**
   * Runs the main search loop.  Generates nodes until one of the stopping
   * conditions is satisfied.  There are many configuration points here:
   * <ul>
   * <li>The annealing temperature is initially set to
   *   initialTemperature.
   * <li>Every annealingRate cycles, the temperature is adjusted by
   *   multiplying it by annealingRatio.
   * <li>If the testUniqueness flag is set, then only unique nodes are
   *     considered successful.
   * <li>If excludeInfeasibleNodes is true, then generated nodes that
   *   prove to be infeasible are not queued for further consideration.
   *   It's probably not a good idea to use this without testUniqueness.
   * <li>If lockAllBids is true, the node expander will not be able to
   *   displace bids from nodes.  This drastically reduces the search
   *   space, but it may make some states harder to reach.
   * <li>The bid selector is configurable.  Its
   *   <tt>setTemperarute</tt> method gets called when the
   *   temperature is adjusted (see {@link #setTemperature(double)
   *   <tt>setTemperature</tt>}), and has an opportunity to switch modes
   *   and go again whenever the node-count stopping condition is
   *   reached and its 
   *   {@link BidSelector#hasMoreChoices() <tt>hasMoreChoices</tt>}
   *   method is called.
   * <li>The stopping condition is the number of nodes generated
   *     without seeing any improvement.  This number is set by the
   *     Scheme hook function <tt>SASearch.scale</tt>, which takes
   *     this SASearch instance as an argument.  If the hook function
   *     doesn't exist, then we use a default scale factor of
   *   <pre>
   *     (ln(bidCount) + ln(taskCount)) * patienceFactor
   *   </pre>
   *     When counting nodes, we count successful expansions as 1, and
   *     unsuccessful expansions as 1/nullExpansionFactor.
   * <li>When the stopping condition is reached, and the bid selector
   *   has no more choices, there is an opportunity to recycle and go
   *   again.  At this point, if a solution has been found, then the
   *   extraCycles count is decremented.  If it is non-negative, then we
   *   recycle.  If no solution has been found, then 
   *   the recycleCount is decremented.  If it
   *   remains non-negative, we recycle.  When recycling,
   *   the scale factor is multiplied by the 
   *   recycleRatio, and the set of unique nodes is discarded if
   *   resetUnique is true.  
   * </ul>
   */
  protected void nodeGenerator() {
    // Start by running the Scheme hook that sets the scale factor.
    SilkHook.runHook("SASearch.scale", this);
    if (m_scaleFactor == 0.0) {
      // hook didn't run
      log.warn("scale factor hook did not run, using default");
      m_scaleFactor = ((Math.log(bidEval.getBids().size())
			+ Math.log(bidEval.getTaskPlan().size()))
		       * m_patienceFactor);
    }
    log.info("scale factor: " + m_scaleFactor);
    double complexity = m_scaleFactor * m_nullExpansionFactor; 
    double annealingRatio = m_annealingRatio;
    int queueSize = (int)(m_queueSize * m_scaleFactor / m_patienceFactor);
    double bestEval = 1e10;

    // Compute the "annealing rate".  The normal way is as a fraction
    // of the scale factor, unless the deprecated property
    // AnnealingRate is set.
    int annealingRate = 0;
    if (m_asf > 0.0)
      // Annealing scale factor is given
      annealingRate = (int)Math.round(m_scaleFactor * m_asf);
    else if (m_rate > 0)
      // Deprecated fixed rate is set
      annealingRate = m_rate;
    else
      // use default
      annealingRate = (int)Math.round(m_scaleFactor * m_asfDefault);
    
    //int logCount = 1;
    RandomVariableFactory factory=
	RVFManager.getRVF("SearchSeed");
    ExponentialVariable EV = factory.createExponentialVariable(1.0);
	
    int nodesWithoutImprovement = 0;
    int nodesSinceRestart = 0;
    initSearchQueue();
    while (nodesWithoutImprovement <  complexity
	   && ! m_searchQueue.isEmpty()
	   && ! bidEval.isSearchAborted()) {
      // Pick a random number to choose a node
      double randomNumber = EV.nextDouble() * m_temperature;
      
      // use the value to pick a node from the queue
      double fv = m_searchQueue.firstValue();
      double lv = m_searchQueue.lastValue();
      double target = fv + randomNumber * (lv - fv);

      // --- Select a node to expand ----------------------
      // --- make sure it has usable bids ---
      m_tester.m_target = target;
      SearchNode nodeToExpand = null;
      nodeToExpand = (SearchNode)
	  m_searchQueue.selectByPredicate(m_tester);
      if (nodeToExpand == null)
	log.warn("Null result from selectByPredicate");
      else if (! nodeToExpand.hasUsableBids()) {
	m_searchQueue.remove(nodeToExpand);
	nodeToExpand = null;
	if (log.isDebugEnabled())
	  log.debug("Queue:" + m_searchQueue.toString());
      }
      else {
	// ------ Select a Bid ------------------------------
	Bid aBid = m_selector.selectBid(nodeToExpand, bidEval);
	
	// --- Attempt the expansion, evaluate the result ---
	if (log.isDebugEnabled())
	  log.debug("Node = " + nodeToExpand + ", bid = " + aBid);
	SearchNode nextNode = nodeToExpand.expandNode(aBid, m_nodeCount);
	
	if (nextNode != null
	    && (!m_testUniqueness
		|| SearchNode.isUnique(nextNode)))
	{
	  // this is a successful expansion
	  m_nodeCount++;
	  nodesSinceRestart++;
	  nodesWithoutImprovement += m_nullExpansionFactor;
	  m_successfulExpansions.newObservation(1.0);
	  
	  // evaluate the node and record results
	  double eval = m_eval.evaluate(nextNode.getTaskBidPlan());
	  
	  if (log.isDebugEnabled()) {
	    log.debug("Node " + nextNode.getNodeNumber() + ": -"
		      + (nextNode.isFeasible() ? "F" : " ")
		      + (nextNode.isCovered()  ? "C" : " ")
		      + "-, eval = " + eval
		      + ", best = " + bestEval
		      + ", temp = " + m_temperature); 
	  }
	  //if (logCount++ % m_logInterval == 0)
	  //  Log.newObservation(m_searchQueue.firstValue());
	  
	  if (m_lockAllBids)
	    nextNode.lockBid(aBid);
	  
	  if (! m_excludeInfeasibleNodes
	      || nextNode.isFeasible())
	  {
	    m_searchQueue.insert(nextNode); 
	    
	    if (eval < bestEval)
	    { // keep track of improvement
	      bestEval = eval;
	      nodesWithoutImprovement = 0;
	      if (m_stopRatio > 0.0) {
		int newcx = (int)(nodesSinceRestart * m_stopRatio);
		if (complexity < newcx) {
		  complexity = newcx;
		  log.debug("Complexity increased to " + complexity);
		}
	      }
	    }

	    processSolution(nextNode);
	  }
	  // run the simulated annealing schedule
	  if (nodesSinceRestart % annealingRate == 0)
	  {
	    setTemperature(m_temperature *= annealingRatio);
	  }
	}
	else
	{
	  m_successfulExpansions.newObservation(0.0);
	  nodesWithoutImprovement++;		
	}
      }
      
      // --------- Stopping conditions -----------
      // Switch among selectors as we go along
      if (nodesWithoutImprovement >= complexity
	  || m_searchQueue.isEmpty()) {
	if (m_selector.hasMoreChoices()
	    && ! m_searchQueue.isEmpty()) {
	  // the selector wants to try again
	  nodesWithoutImprovement = 0;
	}
	else if (--m_recycleCount > 0
		 || m_searchQueue.isEmpty()) {
	  //adaptive recycling
	  // increases the patience factor on each round until a
	  // solution is found, then runs once more.  Limited by the
	  // RecycleCount.
	  if (m_recycleRatio > 1.0) {
	    m_recycleFactor *= m_recycleRatio;
	    log.debug("Recycle Factor increased to " + m_recycleFactor);
	  }
	  // Alternatively, pick a random number, exponentially
	  // distributed, with a minimum of 1.0 and a mean of
	  // m_recycleRange.
	  else if (m_recycleRange > 1.0) {
	    m_recycleFactor = 1.0 + EV.nextDouble() * m_recycleRange;
	    log.debug("Recycle Factor randomly set to " + m_recycleFactor);
	  }
	  complexity = (m_scaleFactor
			* m_nullExpansionFactor
			* m_recycleFactor);
	  annealingRatio = 1.0 - ((1.0 - m_annealingRatio)
				  / m_recycleFactor);
	  m_searchQueue.setMaxSize((int)(queueSize
					 * m_recycleFactor));

	  // if we've found a plan, then initiate stopping condition
	  // if configured to do so.
	  if (m_bestPlan != null
	      && m_stopOnSolution) {
	    m_recycleCount = m_extraCycles--;
	  }
	  if (m_recycleCount > 0) {
	    // start over from the top
	    //System.out.println("Recycle count = " + m_recycleCount
	    //		       + ", Temp = " + m_temperature);
	    if (log.isInfoEnabled()) {
	      log.info("End of cycle, temp:" + m_temperature
		       + ", node count:" + m_nodeCount
		       + ", remaining cycles:" + m_recycleCount);
	    }
	    nodesWithoutImprovement = 0;
	    nodesSinceRestart = 0;
	    bestEval = 1e10;
	    recycleSearch();
	  }
	}
      }
    }
    //System.out.println("Final Temperature: " + m_temperature);
  }

  /**
   * Checks a node for solution, records it if so.  This only works if
   * the node has already been evaluated.  If you wanted to keep track
   * of all solutions in some form, this would be the place to do it.
   */
  private void processSolution (SearchNode node)
  {
    if ((node.isCovered() == true)
	&& (node.isFeasible() == true))
    {
      log.debug("found covered, feasible plan");
      super.processSolution(node.getTaskBidPlan());
      m_planCount += 1;
      if (m_bestPlan == null
	  || m_bestPlan.getEvaluation() > node.getEvaluation()) {
	m_bestPlan = node;
	bidEval.notifyNewResult(m_bestPlan.getTaskBidPlan());
      }
      if (m_firstSoln == -1) {
	m_firstSoln = m_nodeCount;
      }
    }
  }

  private void recycleSearch ()
  {
    m_selector.recycle();
    //clear and initialize the search queue
    initSearchQueue();
    if (m_restartUnique)
      SearchNode.newSearch(bidEval);
  }

  private void initSearchQueue ()
  {
    log.debug("initSearchQueue: queue size = " + m_initialQueue.size());
    m_searchQueue.clear();
    setTemperature(m_initialTemperature);
    int count = 0;
    for (Iterator nodes = m_initialQueue.iterator();
	 nodes.hasNext(); ) {
      SearchNode newNode = new SearchNode((SearchNode)nodes.next(),
					  count++);
      m_eval.evaluate(newNode.getTaskBidPlan());
      m_searchQueue.insert(newNode);
    }
  }

  public double getTemperature ()
  {
    return m_temperature;
  }

  public void setTemperature (double newTemp)
  {
    m_temperature = newTemp;
    m_selector.setTemperature(newTemp);
    log.debug("Reduced temperature to " + m_temperature);
  }

  /**
   * Gets the scale factor that controls the search stopping condition.
   */
  public double getScaleFactor ()
  {
    return m_scaleFactor;
  }

  /**
   * Sets the search scale factor.
   */
  public void setScaleFactor (double newValue)
  {
    m_scaleFactor = newValue;
  }

  /**
   * Returns the patience factor.
   */
  public double getPatienceFactor ()
  {
    return m_patienceFactor;
  }

  /**
   * Gets the node count.  If you haven't yet run a search, it will be zero. 
   */
  public int getNodeCount ()
  {
    return m_nodeCount;
  }

  protected void readSystemProps () {
    String prop;

    // ---- initial annealing Temperature ----
    prop = MagnetProperties.getProperty("SA.InitialTemperature");
    if (prop == null)
      prop = MagnetProperties.getProperty("InitialTemperature");
    if (prop != null) {
      try {
	double value = Double.valueOf(prop).doubleValue();
	m_initialTemperature = value;
	log.info("prop: Initial annealing Temperature = " + value);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		 + " for Initial Temperature.  Using "
		 + m_initialTemperature);
      }
    }
    else
      log.warn("default prop: Initial annealing temperature = "
	       + m_initialTemperature);

    // ---- annealing ratio ----
    prop = MagnetProperties.getProperty("SA.AnnealingRatio");
    if (prop == null)
      prop = MagnetProperties.getProperty("AnnealingRatio");
    if (prop != null) {
      try {
	double value = Double.valueOf(prop).doubleValue();
	m_annealingRatio = value;
	log.info("prop: Initial annealing ratio = " + value);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		  + " for Annealing Ratio.  Using "
		  + m_annealingRatio);
      }
    }
    else
      log.warn("default prop: AnnealingRatio = " + m_annealingRatio);

    // ---- annealing scale factor ----
    prop = MagnetProperties.getProperty("SA.AnnealingScaleFactor");
    if (prop != null) {
      try {
	double value = Double.valueOf(prop).doubleValue();
	m_asf = value;
	log.info("prop: Initial annealing scale factor = " + value);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		  + " for Annealing Factor.  Using "
		  + m_asf);
      }
    }

    // ---- annealing rate ----
    prop = MagnetProperties.getProperty("AnnealingRate");
    if (prop != null) {
      try {
	double value = Double.valueOf(prop).doubleValue();
	m_rate = (int)Math.round(value);
	log.info("prop: Initial annealing rate (deprecated) = " + m_rate);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		 + " for Annealing Rate.  Using "
		 + m_rate);
      }
    }

    // ---- patience factor ----
    prop = MagnetProperties.getProperty("SA.PatienceFactor");
    if (prop == null)
      prop = MagnetProperties.getProperty("PatienceFactor");
    if (prop != null) {
      try {
	m_patienceFactor = Double.valueOf(prop).doubleValue();
	log.info("prop: SA.PatienceFactor = " + m_patienceFactor);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		 + " for Patience Factor.  Using "
		 + m_patienceFactor);
      }
    }
    else
      log.warn("default prop: PatienceFactor = " + m_patienceFactor);

    // ---- recycle count ----
    prop = MagnetProperties.getProperty("RecycleCount");
    if (prop != null) {
      try {
	double value = Double.valueOf(prop).doubleValue();
	m_recycleCount = (int)Math.round(value);
	log.info("prop: RecycleCount = " + m_recycleCount);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		  + " for Recycle Count.  Using "
		  + m_recycleCount);
      }
    }
    else
      log.warn("default prop: RecycleCount = " + m_recycleCount);

    // ---- adaptive recycling ----
    prop = MagnetProperties.getProperty("RecycleRatio");
    if (prop != null) {
      try {
	m_recycleRatio = Double.valueOf(prop).doubleValue();
	log.info("prop: RecycleRatio = " + m_recycleRatio);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		  + " for Recycle Ratio.  Using "
		  + m_recycleRatio);
      }
    }
    else
      log.warn("default prop: RecycleRatio = " + m_recycleRatio);

    // ---- random recycling ----
    prop = MagnetProperties.getProperty("SA.RandomRecycleRange");
    if (prop != null) {
      try {
	m_recycleRange = Double.valueOf(prop).doubleValue();
	log.info("prop: RecycleRange = " + m_recycleRange);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		  + " for Recycle Range.  Using "
		  + m_recycleRange);
      }
    }
    else
      log.warn("default prop: RecycleRange = " + m_recycleRange);

    // ---- adaptive stopping ----
    prop = MagnetProperties.getProperty("SA.StopRatio");
    if (prop != null) {
      try {
	m_stopRatio = Double.valueOf(prop).doubleValue();
	log.info("prop: StopRatio = " + m_stopRatio);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		  + " for Stop Ratio.  Using "
		  + m_stopRatio);
      }
    }
    else
      log.warn("default prop: StopRatio = " + m_stopRatio);

    // ---- extra cycles after solution found ----
    prop = MagnetProperties.getProperty("ExtraCycles");
    if (prop != null) {
      try {
	double value = Double.valueOf(prop).doubleValue();
	m_extraCycles = (int)Math.round(value);
	log.info("prop: ExtraCycles = " + m_extraCycles);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		  + " for extra cycle count.  Using "
		  + m_extraCycles);
      }
    }
    else
      log.warn("default prop: ExtraCycles = " + m_extraCycles);

    // ---- stop on solution or not ----
    prop = MagnetProperties.getProperty("SA.StopOnSolution");
    if (prop != null) {
      m_stopOnSolution = Boolean.valueOf(prop).booleanValue();
      log.info("prop: SA.StopOnSolution = " + m_stopOnSolution);
    }

    // ---- null expansion factor ----
    prop = MagnetProperties.getProperty("NullExpansionFactor");
    if (prop != null) {
      try {
	double value = Double.valueOf(prop).doubleValue();
	m_nullExpansionFactor = (int)Math.round(value);
	log.info("prop: NullExpansionFactor = " + m_nullExpansionFactor);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		  + " for Patience Factor.  Using "
		  + m_nullExpansionFactor);
      }
    }
    else
      log.warn("default prop: NullExpansionFactor = "
		   + m_nullExpansionFactor);

    // ---- feasibility test flag ----
    prop = MagnetProperties.getProperty("ExcludeInfeasibleNodes");
    if (prop != null) {
      if (prop.equals("true")) {
	m_excludeInfeasibleNodes = true;
      }
      else {
	m_excludeInfeasibleNodes = false;
      }
    }
    log.info("prop: ExcludeInfeasibleNodes = " + m_excludeInfeasibleNodes);

    // ---- uniqueness test flag ----
    prop = MagnetProperties.getProperty("TestUniqueness");
    if (prop != null) {
      if (prop.equals("true")) {
	m_testUniqueness = true;
      }
      else {
	m_testUniqueness = false;
      }
    }
    log.info("prop: TestUniqueness = " + m_testUniqueness);

    // ---- restart uniqueness tables on restart ----
    prop = MagnetProperties.getProperty("RestartUnique");
    if (prop != null) {
      if (prop.equals("true")) {
	m_restartUnique = true;
      }
      else {
	m_restartUnique = false;
      }
    }
    log.info("prop: RestartUnique = " + m_restartUnique);

    // ---- lock all bids flag ----
    prop = MagnetProperties.getProperty("LockAllBids");
    if (prop != null) {
      if (prop.equals("true")) {
	m_lockAllBids = true;
      }
      else {
	m_lockAllBids = false;
      }
    }
    log.info("prop: LockAllBids = " + m_lockAllBids);

    // ---- beam width ----
    prop = MagnetProperties.getProperty("BeamWidth");
    if (prop != null) {
      try {
	double value = Double.valueOf(prop).doubleValue();
	m_queueSize = (int)Math.round(value);
	log.info("prop: Beam Width = " + m_queueSize);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		  + " for Beam Width.  Using "
		  + m_queueSize);
      }
    }
    else
      log.warn("default prop: Beam Width = " + m_queueSize);

    // ---- Logging Interval ----
    prop = MagnetProperties.getProperty("LogInterval");
    if (prop != null) {
      try {
	double value = Double.valueOf(prop).doubleValue();
	m_logInterval = (int)Math.round(value);
	log.info("prop: Log Interval = " + m_logInterval);
      }
      catch (NumberFormatException nfe) {
	log.error("Bad property spec " + prop
		  + " for Log Interval.  Using "
		  + m_logInterval);
      }
    }
    else
      log.warn("default prop: Log Interval = " + m_logInterval);

    // ---- Bid Selector Class ----
    prop = MagnetProperties.getProperty("BidSelector");
    if (prop != null) {
      try {
	Class selectorClass = Class.forName("edu.umn.magnet.client.eval."
					    + prop);
	m_selector = (BidSelector)selectorClass.newInstance();
	log.info("prop: BidSelector = " + prop);
      }
      catch (ClassNotFoundException cnf) {
	log.error("Bad property spec " + prop
		  + " for BidSelector class name.  Using "
		  + m_selector.getClass().getName());
      }
      catch (Exception e) {
	log.error("Couldn't create instance of selector class "
		  + prop);
	e.printStackTrace();
      }
    }
  }

  // ============ Tester class for search queue extraction =============
  class QueueTest implements Tester
  {
    // you have to set the target to the value you are testing for
    // before calling the test method.
    double m_target = 0.0;

    public boolean test (edu.umn.magnet.util.search.SearchNode node) {
      return node.value() <= m_target;
    }
  }
}
