/*------------------------------------------------------------------------
 * Magnet bid evaluation components
 * $Id: SearchEngine.java,v 1.3.2.3 2002/05/05 19:41:48 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Tue Dec 29 13:03:53 1998  John E. Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.eval;

/**
 * Classes that implement this interface are search engines, capable
 * of generating a mapping between bids and subtasks for some task
 * plan and set of bids.
 * <p>
 * Classes that implement this interface must meet two criteria:
 * <ul>
 * <li> They must provide a no-argument constructor, and
 * <li> They must implement the <tt>search</tt> method as described below.
 * </ul>
 * @author John Collins
 */
public interface SearchEngine
{
  /**
   * Returns a String giving the name of the search type.
   */
  public String getSearchType ();

  /**
   * Hooks up the BidEvaluator for this engine
   */
  public void setBidEvaluator (BidEvaluator bidEval);

  /**
   * Returns the BidEvaluator for this engine
   */
  public BidEvaluator getBidEvaluator ();

  /**
   * Find and return a mapping from subtasks to bids.
   * @return a SearchNode that represents the bid->task mapping
   *   resulting from the search.
   */
  public TaskBidPlan search ();

  /**
   * Find and return a mapping from subtasks to bids, labeling each
   * solution found with the time required to find it.
   * @param baseTime  a long representing the time at which the search
   *   started.  All recorded times are relative to this time.
   * @return a SearchNode that represents the bid->task mapping
   *   resulting from the search.
   */
  public TaskBidPlan search (long baseTime);

  /**
   * Sets the initial search node for continued searches
   */
  public void setInitialNode (TaskBidPlan node);

  /**
   * Gets the node count for whatever search has been completed so far.
   */
  public int getNodeCount ();

  /**
   * Returns the first plan found by the search, if any. Some searches
   * may not support this, in which case you will just get null for
   * this.
   */
  public TaskBidPlan getFirstPlan ();

  /**
   * Returns the number of covered, feasible plans found by the
   * search. Note that this is only meaningful for searches that can
   * potentially find multiple solutions.
   */
  public int getPlanCount ();
}
