/*----------------------------------------------------------------------------
 * Magnet Bid Evaluation components
 * $Id: TaskBidPlan.java,v 1.7.2.3 2002/05/05 19:41:49 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License. 
 *----------------------------------------------------------------------------
 * Revision History:
 * 
 * Sun Jan 28 20:19:04 2001  John Collins
 *   Refactor into SearchNode and TaskBidPlan, remove from TaskBidPlan
 *   all search-related stuff.
 * 
 * Wed Dec 13 21:57:28 2000  John Collins
 *   Get rid of bid breakdowns, support new bid semantics
 * 
 * Sat Dec  4 08:56:18 1999  John E. Collins
 *   Switch from jgl.HashSet to java.util.HashSet, clean up uniqueness
 *   test.
 * 
 * Fri Nov 19 17:12:49 1999  John Collins
 *   Add uniqueness test to support Systematic Search
 * 
 * Sat Mar 27 17:23:10 1999  Rashmi Sundareswara
 * Method printIds() added which prints out the Generic bid Ids in the 
 * given node 
 * 
 * Sun Jan 10 21:44:37 1999  John E. Collins
 *   Make LockedBidException into an inner class so it can get sucked
 *     into the appropriate jar file.
 *   Add selectorJunk property indexed by selectorType symbol
 *
 * Sat Jan 9 14:06:30 1999 Rashmi Sundareswara
 * Method getBids added 
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskWithPrecedence;

import org.apache.log4j.Category;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.Vector;
import java.util.Stack;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * Instances of TaskBidPlan serve two purposes: 
 * <ul>
 * <li>Nodes in the bid-evaluation search process, where they are
 *     wrapped by SearchNode instances, and</li>
 * <li>The final result of bid evaluation, containing the final set of
 *     bids to be awarded, and the schedule.</li>
 * </ul>
 * Two constructors are provided. One is for creating a root node with
 * no bids, and the other is for creating a deep clone of an existing
 * node in preparation for modifying it to create a new search node.
 * No default constructor is provided.  That could change if these
 * need to be serialized.
 * @author John Collins
 */
public class TaskBidPlan implements java.io.Serializable
{
  // Init the log category
  static Category log = Category.getInstance(TaskBidPlan.class.getName());

  // Basic database
  private TaskPlan m_tp = null;
  private HashMap m_taskBids = null;
  private HashSet m_bids = null;
  private TaskBidPlan m_root = null;

  // Evaluation results
  private boolean m_feasible = false;
  private boolean m_covered  = false;
  private HashMap m_evaluations = new HashMap();
  private long m_time = 0L;     // search time for this solution
  private int m_nodeCount = 0;  // node count for solution
  private String m_method = null; // search method used for this solution

  /**
   * Creates a root node by creating one of these with a TaskPlan and
   * no bids.  TaskBid objects without bids are created to populate the bid
   * table.  These are needed by the evaluator.
   */
  public TaskBidPlan(TaskPlan tPlan)
  {
    super();
    m_tp = tPlan;
    m_root = this;
    m_taskBids = new HashMap();
    m_bids = new HashSet();
    for (Iterator tasks = tPlan.iterator();
	 tasks.hasNext(); ) {
      SubtaskRequest st = (SubtaskRequest)tasks.next();
      m_taskBids.put(st.getId(), new TaskBid(tPlan.getTWP(st)));
    }
  }

  /**
   * Creates a new TaskBidPlan by copying the guts of an old one.  On
   * the assumption that the new one will be modified, the feasibility
   * and coverage flags are set back to false.
   */
  public TaskBidPlan (TaskBidPlan oldTBP)
  {
    super();
    m_tp = oldTBP.getTaskPlan();
    m_root = oldTBP.m_root;
    m_taskBids = (HashMap)oldTBP.m_taskBids.clone();
    m_bids = new HashSet(oldTBP.m_bids);
    //m_feasible = oldTBP.m_feasible;
    //m_covered = oldTBP.m_covered;
    //m_evaluations = oldTBP.m_evaluations;
    //m_time = oldTBP.m_time;
    //m_nodeCount = oldTBP.m_nodeCount;
    //m_method = oldTBP.m_method;
  }
  
  public String toString()
  {
    StringBuffer result = new StringBuffer();
    result.append("TBP");
    if (m_method != null) {
      result.append("(");
      result.append(m_method).append("@").append(m_time);
      result.append(")");
    }
    result.append(": ");
    for (Iterator evals = listEvaluations().iterator();
	 evals.hasNext(); ) {
      String evalType = (String)evals.next();
      result.append(evalType).append("=").append(findEvaluation(evalType));
      result.append(", ");
    }
    result.append("(");
    String sep = "";
    for (Iterator allGB = m_bids.iterator();
	 allGB.hasNext(); ) {
      Bid aGB = (Bid)allGB.next();
      result.append(sep).append(aGB.getId());
      sep = " ";
    }
    result.append(")");
    return result.toString();
  }
  
  // ============== Evaluation results ====================
  //
  //Note that none of these are of any real use until the evaluator has
  //been run.
  
  /**
   * True just in case the instantiated plan is temporally feasible.
   * If the plan is also covered, this means that the set of bids
   * represented by this TaskBidPlan can be accepted to form a
   * feasible set of bids.  If the plan is not covered, it means that
   * the existing subset of bids is temporally consistent with each
   * other and with the original RFQ.
   */
  public boolean isFeasible()
  {
    return m_feasible; 
  }
  
  /**
   * True just in case all subtasks in this plan are covered by bids.
   */
  public boolean isCovered ()
  {
    return m_covered;
  }

  /**
   * Returns the total cost of the bids that are part of this plan.
   */
  public long getCost ()
  {
    return (long)findEvaluation(Eval.COST);
  }

  /**
   * Returns the evaluation of this plan.  This only has meaning after
   * the plan has been evaluated
   */
  public double getEvaluation ()
  {
    return findEvaluation(Eval.TYPE);
  }

  /**
   * Returns a specific evaluation, identified by the EvalClass symbol
   * of one of the evaluators.  Note that the default evaluation value
   * is 0.0.  This is the value returned if no evalution exists under
   * the given EvalClass.  Use evaluationExists to test for presence
   * of a particular evaluation, or listEvaluations to get an
   * Iterator of the eval classes that have been entered.
   */
  public double findEvaluation (String evalClass)
  {
    Double result = (Double)m_evaluations.get(evalClass);
    if (result == null)
      return 0.0;
    else
      return result.doubleValue();
  }

  /**
   * Returns true just in case an evaluation has been entered under
   * the given symbol.
   */
  public boolean evaluationExists (String evalClass)
  {
    return (m_evaluations.get(evalClass) != null);
  }

  /**
   * Returns an Enumeration of the Symbols that label the evaluations
   * of this node.
   */
  public Set listEvaluations ()
  {
    return (m_evaluations.keySet());
  }

  // =============== Evaluator interface =======================	
  /**
   * Used by the evaluator to set the feasibility flag.
   */
  public void setFeasible(boolean value)
  {
    m_feasible = value;
  }

  /**
   * Used by the evaluator to set the coverage flag.
   */
  public void setCovered (boolean flag)
  {
    m_covered = flag;
  }

  /**
   * Add an evaluation to the evaluation set of this node.  This
   * should only be used by an evaluator.
   */
  protected void putEvaluation (String evalClass, double value)
  {
    m_evaluations.put(evalClass, new Double(value));
  }

  /**
   * Return the TaskPlan underlying this plan.
   */
  public TaskPlan getTaskPlan()
  {
    return m_tp;
  }

  // ================== Model access ==========================
  /**
   * Returns an Enumeration of TaskBid instances representing the
   * mapping of tasks to bids in this plan.  There is a TaskBid in the
   * set for each Task in the underlying TaskPlan.  Note that some TaskBid
   * objects represent Bids that cover multiple tasks, some represent
   * SubtaskBids that represent just one subtask, and some represent
   * no bid.
   */
  public Collection getTaskBids()
  {
    return m_taskBids.values();
  }

  /**
   * Finds the requested TaskBid by the ID of its subtask.
   */
  public TaskBid findTaskBid(String taskId)
  {  
    return (TaskBid)m_taskBids.get(taskId);
  }

  /**
   * Return the set of Bids in this node as a HashSet.
   */
  public HashSet getBids()
  {
    return m_bids;
  }

  /**
   * Returns the search time required to produce this solution, in
   * milliseconds.
   */
  public long getSearchTime ()
  {
    return m_time;
  }

  /**
   * Returns the search method used to produce this solution.
   */
  public String getSearchMethod ()
  {
    return m_method;
  }

  /**
   * Returns the node count for this node.  This is only meaningful
   * for node-based search engines like SASearch.
   */
  public int getNodeCount ()
  {
    return m_nodeCount;
  }

  /**
   * Sets the search time and search method.  Note that this method
   * has package visibility only.
   */
  void setSearchData(String method, long time)
  {
    m_method = method;
    m_time = time;
  }

  void setSearchData (String method, long time, int nodeCount)
  {
    m_method = method;
    m_time = time;
    m_nodeCount = nodeCount;
  }

  /**
   * Adds a Bid into this plan.  It has 
   * to remove old TaskBids which might have been covered by another bid, 
   * and then add the new bid.  This should only be called in the
   * context of expanding a node.
   * @return the Set of bids that need to be removed in order to add
   *   this bid.
   */      
  protected Set addBid (Bid bid)
  {
    HashSet bidsToRemove = new HashSet();
    //StringBuffer logBuf = new StringBuffer();
    //if (log.isDebugEnabled())
    //  logBuf.append("  Adding bid " + bid.getId() + ": Tasks {");
    // Make a TaskBid object to represent the mapping of each of this
    // bid's tasks to this bid.  Record any overlapping bids.
    for (Iterator tids = bid.getTaskSet().iterator();
	 tids.hasNext(); ) {
      String taskId = (String)tids.next();
      //if (log.isDebugEnabled())
      //logBuf.append(taskId + " ");
      TaskWithPrecedence aTWP = m_tp.getTWP(taskId);
      TaskBid newTaskBid = new TaskBid(aTWP, bid);
      TaskBid oldTaskBid = (TaskBid)m_taskBids.get(taskId);
      Bid gBidToRemove = oldTaskBid.getBid();
      if (gBidToRemove != null) {
	bidsToRemove.add(gBidToRemove);
      }
      // Add the new TaskBid to the plan.  Note that this displaces
      // any existing TaskBid for the same subtask.  Remember this
      // when looking at removeBid...
      m_taskBids.put(taskId, newTaskBid);
    }
    //if (log.isDebugEnabled()) {
    //  logBuf.append("}");
    //  log.debug(logBuf);
    //}
    // Remove the displaced bids from the plan
    Iterator btr = bidsToRemove.iterator();
    while (btr.hasNext()) {
      removeBid((Bid)btr.next());
    }
    // Add the new bid to the set of bids in this plan,
    // and to the tabu list.
    m_bids.add(bid);
    return bidsToRemove;
  }

  /**
   * Removes a bid from this plan.  This should only be called in the
   * context of adding a bid.  Silently does nothing if the bid is not
   * already part of the plan.  Any orphaned tasks get a reference to
   * the "empty" TaskBid from the root node.
   */
  protected void removeBid (Bid bid)
  {
    for(Iterator tids = bid.getTaskSet().iterator();
	tids.hasNext(); ) {
      String sId = (String)tids.next();
      TaskBid oldTB = (TaskBid)m_taskBids.get(sId);
      if (oldTB == null || (oldTB.getBid() == bid)) {
	TaskBid nullTB = (TaskBid)m_root.m_taskBids.get(sId);
	m_taskBids.put(sId, nullTB);
      }
    }
    m_bids.remove(bid);
  }

  /**
   * Returns the root TBP.  This is where the "empty" TaskBid
   * instances are kept.  This method exists to support testing.
   */
  protected TaskBidPlan getRoot ()
  {
    return m_root;
  }

  //================ TaskPlan delegation ================================
  /*  The following accessors access the methods of TaskPlan to get the 
   *  successors, predecessors, successor roots.  The Iterator returned
   *  is of Subtask ID Strings
   */

  /**
   * Returns the number of tasks in the TaskPlan
   */
  public int size ()
  {
    return m_tp.size();
  }

  /**
   * Returns an Iterator of the ID strings of the subtasks that are
   * successors of the specified subtask.  The Iterator object is
   * returned even in the case where there are no actual successors.
   */
  public Iterator findSuccessors(final SubtaskRequest task) {
    final TaskPlan tp = m_tp;
    return new Iterator () {
	Iterator m_successors = tp.getSuccessors(task);
	public boolean hasNext () {
	  return m_successors.hasNext();
	}
	public Object next () {
	  return
	      ((SubtaskRequest)m_successors.next()).getId();
	}
	public void remove () {
	  throw new UnsupportedOperationException();
	}
    };
  }

   
  /**
   * Returns an Iterator of the ID strings of the subtasks that are
   * successors of the specified subtask.  The Iterator object is
   * returned even in the case where there are no actual successors.
   */
  public Iterator findSuccessors(final String taskId) {
    final TaskPlan tp = m_tp;
    return new Iterator () {
	Iterator m_successors = tp.getSuccessors(taskId);
	public boolean hasNext () {
	  return m_successors.hasNext();
	}
	public Object next () {
	  return
	      ((SubtaskRequest)m_successors.next()).getId();
	}
	public void remove () {
	  throw new UnsupportedOperationException();
	}
    };
  }
   
  /**
   * Returns true just in case <tt>task2</tt> is in the successor list
   * of <tt>task1</tt>.
   */
  public boolean isInSuccessors(SubtaskRequest task1, SubtaskRequest task2) {
    return (m_tp.isInSuccessors(task1, task2));
  }
 
  /**
   * Returns true just in case identified by <tt>taskId2</tt> is in
   * the successor list of the task identified by <tt>taskId1</tt>.
   */
  public boolean isInSuccessors(String taskId1, String taskId2) {
    return (m_tp.isInSuccessors(taskId1, taskId2));
  }


  /**
   * Returns an Iterator of the ID strings of the subtasks that are
   * in the transitive closure of the successors of the specified
   * subtask.  The Iterator object is returned even in the case
   * where there are no actual successors.
   */
  public Iterator findAllSuccessors(final SubtaskRequest task) {
    final TaskPlan tp = m_tp;
    return new Iterator () {
	Iterator m_successors = tp.getAllSuccessors(task);
	public boolean hasNext () {
	  return m_successors.hasNext();
	}
	public Object next () {
	  return
	      ((SubtaskRequest)m_successors.next()).getId();
	}
	public void remove () {
	  throw new UnsupportedOperationException();
	}
    };
  }


  /**
   * Returns an Iterator of the ID strings of the subtasks that are
   * in the transitive closure of the successors of the specified
   * subtask.  The Iterator object is returned even in the case
   * where there are no actual successors.
   */
  public Iterator findAllSuccessors(final String taskId) {
    final TaskPlan tp = m_tp;
    return new Iterator () {
	Iterator m_successors = tp.getAllSuccessors(taskId);
	public boolean hasNext () {
	  return m_successors.hasNext();
	}
	public Object next () {
	  return
	      ((SubtaskRequest)m_successors.next()).getId();
	}
	public void remove () {
	  throw new UnsupportedOperationException();
	}
    };
  }


  /**
   * Returns an Iterator of the ID strings of the subtasks that are
   * predecessors of the specified subtask.  The Iterator object is
   * returned even in the case where there are no actual predecessors.
   */
  public Iterator findPredecessors(final SubtaskRequest task) {
    final TaskPlan tp = m_tp;
    return new Iterator () {
	Iterator m_successors = tp.getPredecessors(task);
	public boolean hasNext () {
	  return m_successors.hasNext();
	}
	public Object next () {
	  return
	      ((SubtaskRequest)m_successors.next()).getId();
	}
	public void remove () {
	  throw new UnsupportedOperationException();
	}
    };
  }

  /**
   * Returns an Iterator of the ID strings of the subtasks that are
   * predecessors of the specified subtask.  The Iterator object is
   * returned even in the case where there are no actual predecessors.
   */
  public Iterator findPredecessors(final String taskId) {
    final TaskPlan tp = m_tp;
    return new Iterator () {
	Iterator m_successors = tp.getPredecessors(taskId);
	public boolean hasNext () {
	  return m_successors.hasNext();
	}
	public Object next () {
	  return
	      ((SubtaskRequest)m_successors.next()).getId();
	}
	public void remove () {
	  throw new UnsupportedOperationException();
	}
    };
  }

  /**
   * Returns an Iterator of the ID strings of the subtasks that are
   * roots of the underlying TaskPlan.
   */
  public Iterator findRoots () {
    final TaskPlan tp = m_tp;
    return new Iterator () {
	Iterator m_successors = tp.getSuccessorRoots();
	public boolean hasNext () {
	  return m_successors.hasNext();
	}
	public Object next () {
	  return
	      ((SubtaskRequest)m_successors.next()).getId();
	}
	public void remove () {
	  throw new UnsupportedOperationException();
	}
    };
  }
}
