/*------------------------------------------------------------------------
 * MAGNET status component
 * $Id: AgentStatus.java,v 1.4.2.1 2002/05/05 19:41:50 ashutosh Exp $
 * Copyright (c) 2001 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Wed May  2 13:02:00 2001  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.status;

import edu.umn.magnet.util.MagnetProperties;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.client.eval.TaskBidPlan;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;

/**
 * Information about the overall status of the Agent.  This includes
 * the currently-active plans, and other items TBD.  There is a
 * singleton instance of this class.  You get a reference to it with
 * the static <tt>getInstance()</tt> method.
 * <p>
 * {@link AgentStatusEvent AgentStatusEvents} will be issued to notify
 * listeners of changes in 
 * AgentStatus, as given below.  You get notified if you first call
 * the method <tt>addListener(PropertyChangeListener)</tt>.
 * </p>
 * <table>
 * <tr>
 *   <th><u>Property Name</u></th><th><u>Event</u></th>
 * </tr>
 * <tr valign=top>
 *   <td><strong>NewPlan</strong></td>
 *   <td>Issued when a new PlanStatus is created for the agent.  The
 *     new PlanStatus instance is given as the value of newValue.
 * </tr>
 * <tr valign=top>
 *   <td><strong>PlanRemoved</strong></td>
 *   <td>Issued when a PlanStatus is removed from this agent's state.
 *     The removed PlanStatus instance is included in the event as the
 *     value of oldValue.
 * </tr>
 * </table>
 * @author John Collins
 * @version $Revision: 1.4.2.1 $
 */
public class AgentStatus extends AbstractStatus
{
  private ArrayList plans = new ArrayList();
  private String agentId = "A";
  private boolean propsRead = false;

  /**
   * The constructor is private.  You have to get the singleton from
   * the class.
   */
  private AgentStatus ()
  {
    super();
  }

  /**
   * Returns the Agent ID of this agent.  Eventually there will be a
   * way to set it. 
   */
  public String getAgentId ()
  {
    ensureProps();
    return agentId;
  }

  /**
   * Creates and returns a new PlanStatus instance that you can use to
   * keep track of a plan.  Several options are available; this one
   * takes a TaskPlan and uses it to create a default TaskBidPlan.  An
   * AgentStatusEvent will be issued.
   * @return the newly-created PlanStatus instance.
   */
  public PlanStatus createPlanStatus(TaskPlan plan)
  {
    PlanStatus newPlan = new PlanStatus(new TaskBidPlan(plan));
    plans.add(newPlan);
    notifyNewPlan(newPlan);
    return newPlan;
  }

  /**
   * Creates and returns a new PlanStatus instance that you can use to
   * keep track of a plan.  This one takes a TaskBidPlan.  An
   * AgentStatusEvent will be issued.
   * @return the newly-created PlanStatus instance.
   */
  public PlanStatus createPlanStatus(TaskBidPlan plan)
  {
    PlanStatus newPlan = new PlanStatus(plan);
    plans.add(newPlan);
    notifyNewPlan(newPlan);
    return newPlan;
  }

  /**
   * Creates and returns a new PlanStatus instance that you can use to
   * keep track of a plan.  This one
   * takes a TaskBidPlan and the start and deadline dates.  An
   * AgentStatusEvent will be issued.
   * @param plan a {@link edu.umn.magnet.client.eval.TaskBidPlan
   *   TaskBidPlan} that represents the initial state of the plan.
   * @param start the intended Start Date of the plan.
   * @param deadline the intended Deadline for the plan.
   * @return the newly-created PlanStatus instance.
   */
  public PlanStatus createPlanStatus(TaskBidPlan plan, Date start, Date deadline)
  {
    PlanStatus newPlan = new PlanStatus(plan, start, deadline);
    plans.add(newPlan);
    notifyNewPlan(newPlan);
    return newPlan;
  }

  /**
   * Returns the list of PlanStatus instances that are active for this
   * agent. 
   */
  public Collection getPlans ()
  {
    return plans;
  }

  private void notifyNewPlan (PlanStatus newPlan)
  {
    sendStatusEvent(new AgentStatusEvent(this, "NewPlan", null, newPlan));
  }

  /**
   * reads the properties to extract the AgentID property.
   */
  private void ensureProps ()
  {
    if (propsRead)
      return;
    agentId = MagnetProperties.getProperty("AgentId", "MagnetAgent");
    propsRead = true;
  }

  // ----------- static singleton management -----------
  private static AgentStatus instance = null;

  public static AgentStatus getInstance ()
  {
    if (instance == null)
      instance = new AgentStatus();
    return instance;
  }
}
