/*------------------------------------------------------------------------
 * MAGNET status component
 * $Id: SessionIdFactory.java,v 1.3.2.2 2002/05/05 19:41:51 ashutosh Exp $
 * Copyright (c) 2001 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Fri May  4 12:41:57 2001  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.status;

import edu.umn.magnet.util.MagnetProperties;

import org.apache.log4j.Category;

/**
 * This class includes both a static factory and default
 * (singleton) implementation of a Session ID factory.  In a
 * standalone experimental situation, sessions are used to distinguish
 * runs.  In a single-market environment, agents can either create
 * their own session ID's, or they can go to the market for one.  In a
 * multiple-market environment, you have to go to the specific market
 * where the session is to be created in order to get a guaranteed
 * unique session ID.  In that case, there may be individual ID
 * factory instances per market.
 * <p>
 * You get a session ID from the instance returned by the factory's
 * getInstance method.  It takes a market ID,
 * which may be null in a single-market environment.  The getSessionId
 * method on the instance takes an agent prefix, to improves the
 * chances of generating a unique ID in a standalone or single-market
 * situation.  So, the incantation for getting a session ID looks like
 * <pre>
 *   id = SessionIdFactory.getInstance(marketId).getSessionId(prefix);
 * </pre>
 * </p><p>
 * The specific subtype of SessionIdFactory you get as the value of
 * the factory object is controlled by the "SessionIdFactory.Class"
 * property.  If that property is not specified, you get a default
 * instance that generates ID numbers locally, restarting the sequence
 * at 0 every time it's restarted.
 * </p>
 * @author John Collins
 * @version $Revision: 1.3.2.2 $
 */
public class SessionIdFactory
{
  private static Category log
    = Category.getInstance(SessionIdFactory.class.getName());

  // -------- instance implementation ----------
  private int idCounter = 0;

  private SessionIdFactory ()
  {
    super();
  }

  public String getSessionId (String prefix)
  {
    return (prefix + (++idCounter));
  }

  // ----------- factory implementation -----------
  static SessionIdFactory instance = null;

  public static SessionIdFactory getInstance (String marketId)
  {
    if (instance == null) {
      // note that once you've created an instance, changing the
      // property won't cause you to create another one...
      String sfClassName
	  = MagnetProperties.getProperty("SessionIdFactory.Class");
      log.info("SessionIdFactory class: " + sfClassName);
      if (sfClassName != null) {
	try {
	  Class sfClass = Class.forName(sfClassName);
	  instance = (SessionIdFactory)sfClass.newInstance();
	}
	catch (ClassNotFoundException cnf) {
	  log.error("Class " + sfClassName + " not found");
	}
	catch (InstantiationException ie) {
	  log.error("Class " + sfClassName + ": Instantion Exception");
	}
	catch (IllegalAccessException iae) {
	  log.error("Class " + sfClassName + ": Illegal Access");
	}
      }
      if (instance == null)
	instance = new SessionIdFactory();
    }
    return instance;
  }
}
