/*------------------------------------------------------------------------
 * MAGNET status component
 * $Id: SessionStatus.java,v 1.8.2.2 2002/05/05 19:41:51 ashutosh Exp $
 * Copyright (c) 2001 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 *
 * Sat May 5  22:43:05 2001 Paula Greve
 *    Added parent being set in constructor without session id
 *
 * Sat May 5  15:09:43 2001 Paula Greve
 *    Added method getRFQs to return a set of the RFQ Statuses with this object
 *
 * Thu May  3 13:28:42 2001  John Collins
 *   Change Session reference to a String
 *
 * Wed May  2 13:02:00 2001  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.status;

import edu.umn.magnet.protocol.InvalidDataException;
import edu.umn.magnet.protocol.cust.RFQ;

import org.apache.log4j.Category;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * Instances of this class keep track of agent status with respect to
 * a given market session.
 * <table>
 * <tr>
 *   <th><u>Property Name</u></th><th><u>Event</u></th>
 * </tr>
 * <tr valign=top>
 *   <td><strong>NewRFQ</strong></td>
 *   <td>Issued when a new RFQ has been created for this
 *     plan.  The event includes the new RFQStatus instance as its
 *     newValue.  You have to go there to interact with the RFQ
 *     itself.</td>
 * </tr>
 * </table>
 * <p>
 * When first initialized, a <tt>SessionStatus</tt> is in the
 * <tt>NONE</tt> state.  After it's associated with a
 * <tt>Session</tt>, it's in the <tt>ACTIVE</tt> state.  You can add
 * RFQ's only when it's <tt>ACTIVE</tt>.  If you cancel a
 * SessionStatus, it moves into the <tt>ABANDONED</tt> state and you
 * can't use it any more.
 * @author John Collins
 * @version $Revision: 1.8.2.2 $
 */
public class SessionStatus extends AbstractStatus
{
  private static Category log
    = Category.getInstance(SessionStatus.class.getName());

  // State constants
  /** State constant: just initialized */
  public static final String NONE = "none";
  /** State constant: active */
  public static final String ACTIVE = "active";
  /** State constant: abandoned */
  public static final String ABANDONED = "abandoned";

  // Here's the list of outstanding RFQStatus instances
  private ArrayList rfqs = new ArrayList();

  // Here's the ID of the Session we're hooked to
  private String theSession = null;

  // PlanStatus parent reference
  private PlanStatus parent = null;

  // state variable
  private String state = NONE;

  /**
   * Usually, you create one of these with a Session.
   */
  protected SessionStatus (PlanStatus parent, String sessionId)
  {
    super();
    this.parent = parent;
    try {
      setSessionId(sessionId);
    }
    catch (InvalidStateException ise) {
      log.error("Very strange. Can't set session at initialization");
    }
  }

  /**
   * You can create it empty and set the Session reference later,
   * though.  But you can't change the Session reference once it's set
   * (even if the server is re-started?  Should it be a handle?).
   */
  protected SessionStatus (PlanStatus parent)
  {
    super();
    this.parent = parent;
  }

  /**
   * Creates and returns a new RFQStatus instance, given an RFQ that
   * has already been created.
   * @exception  throws InvalidDataException if the RFQ not valid in
   *   the current environment.  The reason is given in the message
   *   property of the exception.
   * @exception  throws InvalidStateException if the RFQ has been
   *   abandoned.
   */
  public RFQStatus createRFQStatus (RFQ theRFQ)
    throws InvalidDataException, InvalidStateException
  {
    if (!state.equals(ACTIVE))
      throw new InvalidStateException ("Can't add RFQ - session abandoned");
    try {
      RFQStatus newStat = new RFQStatus(this, theRFQ);
      rfqs.add(newStat);
      notifyNewRFQ(newStat);
      return newStat;
    }
    catch (InvalidDataException ide) {
      log.error("InvalidDataException caught while creating RFQStatus");
      throw ide;
    }
  }

  private void notifyNewRFQ (RFQStatus newStat)
  {
    sendStatusEvent(new SessionStatusEvent(this, "NewRFQ", null, newStat));
  }

  /**
   * Returns the session ID for the associated session
   */
  public String getSessionId ()
  {
    return theSession;
  }

  /**
   * Sets the session ID for this session.  You can only do this if it
   * isn't already set, although it isn't an error to set the same ID
   * twice.
   * @exception throws {@link InvalidStateException} if you are
   *   trying to set a new ID.
   */
  public void setSessionId (String newSession)
    throws InvalidStateException
  {
    if (!state.equals(NONE)) {
      if (newSession.equals(theSession))
	return; // not much to do here
      else
	throw new InvalidStateException("SessionStatus: Can't change session");
    }
    theSession = newSession;
    setState(ACTIVE);
  }

  /**
   * Returns the PlanStatus instance that hosts this Session
   */
  public PlanStatus getParent ()
  {
    return parent;
  }

  /**
   * Returns a collection of the RFQStatus Instances for this
   * session.
   */
  public Collection getRFQs()
  {
    return rfqs;
  }

  /**
   * Returns the state of this Session instance.
   */
  public String getState ()
  {
    return state;
  }

  /**
   * Sets the state.  Somebody want to make this emit an event?
   */
  private void setState (String newState)
  {
    state = newState;
  }

  /**
   * Cancels the session and all its RFQ's.  You'll get state change
   * events from all the RFQSessions as their states change to
   * ABANDONED.  After that, silence.
   */
  public void cancel ()
  {
    // Cancel all the RFQ sessions
    for (Iterator i = rfqs.iterator(); i.hasNext(); ) {
      RFQStatus rs = (RFQStatus)i.next();
      rs.cancel();
    }
    // disconnect references
    rfqs.clear();
    parent = null;

    // set the final state.  From here, there's no way back.
    setState(ABANDONED);
  }
}
