/*------------------------------------------------------------------------
 * MAGNET Customer Agent Plan Management components
 * $Id: TaskStatus.java,v 1.4.2.1 2002/05/05 19:41:51 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Tue Jan  8 17:44:34 2002  John Collins  #0000037
 *   Keep track of bids/task
 * 
 * Sat Jan  6 11:56:29 2001  John Collins
 *   Get rid of tools.Log reference, replace with log4j Category
 * 
 * Fri Sep 29 19:16:50 2000  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.status;

import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.Bid;

import org.apache.log4j.Category;

import java.util.Collection;
import java.util.List;
import java.util.LinkedList;

/**
 * Instances are responsible for tracking the status of an individual
 * task through the bidding and execution process.
 * @author John Collins
 */
public class TaskStatus extends AbstractStatus
{
  private static Category log
    = Category.getInstance(TaskStatus.class.getName());

  // Task Status constants
  public static final String NONE = "n";
  public static final String OFFERED = "o"; // No bids yet associated with task
  public static final String PENDING = "p"; // One or more bids are recommended
  public static final String AWARDED = "w"; // Task is awarded to a bidder
  public static final String ACTIVE = "a";
  public static final String COMPLETE = "c";

  // The underlying task
  private SubtaskRequest task = null;

  // Bid and state data
  // We keep this in a list so we could potentially sort it.
  // If task is awarded, there should be only one bid here.
  private String state = TaskStatus.NONE;
  private List bids = new LinkedList();

  // ----------- Constructors --------------

  /**
   * Default constructor.  You probably don't want to use this.
   */
  public TaskStatus ()
  {
    super();
  }

  /**
   * The normal constructor takes a SubtaskRequest.
   */
  public TaskStatus (SubtaskRequest task)
  {
    super();
    this.task = task;
  }

  // -------------- Property access --------------

  /**
   * Returns the underlying SubtaskRequest.
   */
  public SubtaskRequest getTask ()
  {
    return task;
  }

  /**
   * Sets the underlying SubtaskRequest.  This only works if the task
   * property is currently null; otherwise we complain to the log file
   * and otherwise silently do nothing.
   */
  public void setTask (SubtaskRequest newTask)
  {
    if (task == null)
      task = newTask;
    else
      log.error("Attempt to change task value");
  }

  /**
   * Returns the current state of the task
   */
  public String getState ()
  {
    return state;
  }

  /**
   * Sets the current state of the task.  The value must be one of the
   * state constants, but we don't check.
   */
  public void setState (String newState)
  {
    state = newState;
  }

  /**
   * Returns true just in case the state of this task is one of the
   * listed states.
   */
  public boolean isInStates (Collection states)
  {
    return states.contains(this.getState());
  }

  /**
   * Returns the list of bids.  It's a list, so if you want to add or
   * remove a bid, you can just operate on the list.
   */
  public List getBids ()
  {
    return bids;
  }

  /**
   * Sets the bid list to be a new list.
   */
  public void setBids (List newBids)
  {
    bids = newBids;
  }

  /**
   * Adds a bid to the existing bid list.
   */
  public void addBid (Bid newBid)
  {
    bids.add(newBid);
  }

  /**
   * Removes a bid from the bid list.
   */
  public void removeBid (Bid delBid)
  {
    bids.remove(delBid);
  }
}
