/*------------------------------------------------------------------------
 * MAGNET utility test components
 * $Id: BidEvaluatorTest.java,v 1.4.2.3 2002/05/05 19:41:52 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Sun Jan 14 10:30:15 2001  John Collins
 *   Update for singleton test
 * 
 * Sun Nov 19 13:51:10 2000  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.InvalidDataException;

import edu.umn.magnet.client.SampleTaskPlan;
import edu.umn.magnet.client.status.AgentStatus;
import edu.umn.magnet.client.status.PlanStatus;
import edu.umn.magnet.client.status.TaskStatus;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;

import org.apache.log4j.Category;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.FileAppender;
import org.apache.log4j.Priority;

import java.io.IOException;
import java.util.Date;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.List;
import java.util.Properties;

/**
 * This is a tester for edu.umn.magnet.client.eval.BidEvaluator.
 * There are three aspects to be tested: setup, model access, and the
 * actual bid evaluation.
 * @author John Collins
 */
public class BidEvaluatorTest extends TestCase
{
  static private Category log
    = Category.getInstance(BidEvaluator.class.getName());

  static {
    try {
      PatternLayout layout = new PatternLayout("%5p: %m%n");
      FileAppender file
	  = new FileAppender(layout, "BidEvaluatorTest.log", false);
      log.addAppender(file);
      log.setPriority(Priority.DEBUG);
    }
    catch (IOException ioe) {
      System.out.println("Can't open log file");
    }
  }

  static final long BASE = 1000000000000L;
  static final long HOUR = 3600000L;

  private TaskPlan plan = null;
  private TaskBidPlan tbp = null;
  private PlanStatus status = null;
  private BidEvaluator be = null;

  private HashMap bidMap = null;

  public BidEvaluatorTest (String testName)
  {
    super(testName);
  }

  /**
   * You run the main method to exercise this thing.  There are no
   * command-line args.
   */
  public static void main (String[] args)
  {
    junit.textui.TestRunner.run(suite());
  }

  public static Test suite ()
  {
    return new TestSuite(BidEvaluatorTest.class);
  }

  private void setup1 ()
  {
    plan = SampleTaskPlan.generateTaskPlan();
    tbp = new TaskBidPlan(plan);
    status = AgentStatus.getInstance().createPlanStatus(tbp);
    be = new BidEvaluator();
    be.setSessionId("test");
  }
  
  public void testCreate ()
  {
    log.debug("testCreate()");
    setup1();
    be.setPlanStatus(status);
    int count = 0;
    assertEquals("TaskPlan available", plan, be.getTaskPlan());
    assertEquals("No bids", 0, be.getBids().size());
    for (Iterator ots = be.getOfferedTasks();
	 ots.hasNext(); ) {
      count++;
      ots.next();
    }
    assertEquals("No offered tasks", 0, count);
    assertEquals("No task descriptors", 0, be.getTaskDescriptors().size());
    assertEquals("Correct session ID", "test", be.getSessionId());
  }

  // --- make all the tasks OFFERED and check again ---

  private void setup2 ()
  {
    setup1();
    bidMap = null;
    for (Iterator tasks = plan.iterator();
	 tasks.hasNext(); ) {
      SubtaskRequest task = (SubtaskRequest)tasks.next();
      status.setTaskState(task, TaskStatus.OFFERED);
    }
    be.setPlanStatus(status);
  }

  public void testOffered ()
  {
    log.debug("testOffered()");
    setup2();
    int count = 0;
    for (Iterator ots = be.getOfferedTasks();
	 ots.hasNext(); ) {
      count++;
      ots.next();
    }
    assertEquals("Count offered tasks", 6, count);
    assertEquals("Count task descriptors", 6, be.getTaskDescriptors().size());
  }

  // ----- add some bids & check coverage constraints -----

  private void setup3 ()
  {
    setup2();
    Properties props = new Properties();
    props.put("Eval.UseCoverageConstraints", "true");
    props.put("Eval.UseFeasibilityConstraints", "false");
    be.setProperties(props);

    Bid bid;
    // Create some Bids against the TP
    try {
      bid = new Bid();
      bid.setId("B1");
      bid.addTask("T2");
      bid.addTask("T4");
      bid.addTask("T6");
      be.addBid(bid);
      
      bid = new Bid();
      bid.setId("B2");
      bid.addTask("T1");
      bid.addTask("T3");
      be.addBid(bid);
      
      bid = new Bid();
      bid.setId("B3");
      bid.addTask("T1");
      bid.addTask("T3");
      bid.addTask("T5");
      be.addBid(bid);
      
      bid = new Bid();
      bid.setId("B4");
      bid.addTask("T1");
      bid.addTask("T2");
      bid.addTask("T3");
      be.addBid(bid);
    }
    catch (InvalidDataException ide) {
      fail("InvalidDataException");
    }
  }

  private Bid findBid (String bidId)
  {
    if (bidMap == null) {
      // we need to create the Id -> Bid map
      bidMap = new HashMap();
      for (Iterator bids = be.getBids().iterator();
	   bids.hasNext(); ) {
	Bid bid = (Bid)bids.next();
	bidMap.put(bid.getId(), bid);
      }
    }
    return (Bid)bidMap.get(bidId);
  }

  public void testProps1 ()
  {
    log.debug("testProps1()");
    setup3();
    assert("using coverage constraints", be.useCoverageConstraints());
    assert("not using feasibility constraints",
	   ! be.useFeasibilityConstraints());
  }

  public void testAddBid ()
  {
    log.debug("testAddBid()");
    setup3();
    List theBids = be.getBids();
    assertEquals("Correct number of bids", 4, theBids.size());
    TaskDescriptor td = be.findTaskDescriptor("T1");
    assertEquals("Correct number of bids on T1", 3, td.getBidCount());
  }

  // after constraint processing, we should be left with just bids B1
  // and B3, which do not conflict with each other.
  public void testConflictingBids1 ()
  {
    log.debug("testConflictingBids1()");
    setup3();
    be.checkCoverage();
    assert("B1 remains", be.getBids().contains(findBid("B1")));
    assert("B2 is gone", !be.getBids().contains(findBid("B2")));
    assert("B3 remains", be.getBids().contains(findBid("B3")));
    assert("B4 is gone", !be.getBids().contains(findBid("B4")));
    Set b1c = be.conflictingBids(findBid("B1"));
    assertEquals("Correct number of conflicting bids", 0, b1c.size());
  }

  // --- add some bids with time data & check feasibility constraints ---

  private void setup4 ()
  {
    Bid bid;
    setup2();
    Properties props = new Properties();
    props.put("Eval.UseCoverageConstraints", "true");
    props.put("Eval.UseFeasibilityConstraints", "true");
    be.setProperties(props);

    // Create some Bids against the TP
    try {
      // B1 has 3 tasks, but no milestones
      bid = new Bid();
      bid.setId("B1");
      bid.setEarlyStart(new Date(BASE));
      bid.setLateStart(new Date(BASE + HOUR));
      bid.setDuration(3 * HOUR);
      bid.addTask("T2");
      bid.addTask("T4");
      bid.addTask("T6");
      be.addBid(bid);
      
      // B1a has 3 tasks, all milestones
      bid = new Bid();
      bid.setId("B1a");
      bid.setEarlyStart(new Date(BASE));
      bid.setLateStart(new Date(BASE + HOUR));
      bid.setDuration(3 * HOUR);
      bid.addTask("T2", HOUR);
      bid.addTask("T4", "T2", HOUR);
      bid.addTask("T6", "T4", HOUR);
      be.addBid(bid);
      
      // B2 has two tasks, no milestones
      bid = new Bid();
      bid.setId("B2");
      bid.setEarlyStart(new Date(BASE));
      bid.setLateStart(new Date(BASE + HOUR));
      bid.setDuration(2 * HOUR);
      bid.addTask("T1");
      bid.addTask("T3");
      be.addBid(bid);

      // B2a has two tasks, with a milestone
      bid = new Bid();
      bid.setId("B2a");
      bid.setEarlyStart(new Date(BASE));
      bid.setLateStart(new Date(BASE + HOUR));
      bid.setDuration(2 * HOUR);
      bid.addTask("T1", HOUR);
      bid.addTask("T3", "T1", HOUR);
      be.addBid(bid);
      
      // B3 has three tasks, no milestones
      bid = new Bid();
      bid.setId("B3");
      bid.setEarlyStart(new Date(BASE));
      bid.setLateStart(new Date(BASE + HOUR));
      bid.setDuration(3 * HOUR);
      bid.addTask("T1");
      bid.addTask("T3");
      bid.addTask("T5");
      be.addBid(bid);
      
      // B3a has three tasks, with milestones
      bid = new Bid();
      bid.setId("B3a");
      bid.setEarlyStart(new Date(BASE));
      bid.setLateStart(new Date(BASE + HOUR));
      bid.setDuration(3 * HOUR);
      bid.addTask("T1", HOUR);
      bid.addTask("T3", "T1", HOUR);
      bid.addTask("T5", "T3", HOUR);
      be.addBid(bid);
      
      // B4 has 3 tasks, with milestones that put task2 after task 3
      bid = new Bid();
      bid.setId("B4");
      bid.setEarlyStart(new Date(BASE));
      bid.setLateStart(new Date(BASE + HOUR/2));
      bid.addTask("T1", HOUR);
      bid.addTask("T3", "T1", HOUR/2);
      bid.addTask("T2", "T3", HOUR);
      be.addBid(bid);
      
      // B5 has only T3, which is 1 ms too long to fit between T1 and T6
      bid = new Bid();
      bid.setId("B5");
      bid.setEarlyStart(new Date(BASE + HOUR / 2));
      bid.setLateStart(new Date(BASE + HOUR + 1));
      bid.setDuration(2 * HOUR);
      bid.addTask("T3", HOUR + HOUR + 1);
      be.addBid(bid);
      
      // B6 has only T1
      bid = new Bid();
      bid.setId("B6");
      bid.setEarlyStart(new Date(BASE));
      bid.setLateStart(new Date(BASE + HOUR + 1));
      bid.setDuration(HOUR);
      bid.addTask("T1");
      be.addBid(bid);
      
      // B7 has only T5
      bid = new Bid();
      bid.setId("B7");
      bid.setEarlyStart(new Date(BASE + HOUR));
      bid.setLateStart(new Date(BASE + HOUR * 2 + 1));
      bid.setDuration(HOUR);
      bid.addTask("T5");
      be.addBid(bid);
    }
    catch (InvalidDataException ide) {
      fail("InvalidDataException");
    }
  }

  public void testProps2 ()
  {
    log.debug("testProps2()");
    setup4();
    assert("using coverage constraints", be.useCoverageConstraints());
    assert("using feasibility constraints",
	   be.useFeasibilityConstraints());
  }
  
  public void testAddBid2 ()
  {
    log.debug("testAddBid2()");
    setup4();
    List theBids = be.getBids();
    assertEquals("Correct number of bids", 10, theBids.size());
    TaskDescriptor td = be.findTaskDescriptor("T1");
    assertEquals("Correct number of bids on T1", 6, td.getBidCount());
  }

  public void testConflictingBids3 ()
  {
    log.debug("testConflictingBids3()");
    setup4();
    be.checkCoverage();
    Set b1c = be.conflictingBids(findBid("B1"));
    printBidSet("cb3", b1c);
    assertEquals("Correct number of conflicting bids", 2, b1c.size());
    HashSet comp = new HashSet();
    comp.add(findBid("B1a"));
    comp.add(findBid("B4"));
    assert("Sets match", comp.equals(b1c));
  }

  public void testInfeasibleBids1 ()
  {
    log.debug("testInfeasibleBids1()");
    setup4();
    be.checkCoverage();
    HashSet infeas = new HashSet();
    infeas.add(findBid("B1"));
    infeas.add(findBid("B2"));
    assert("B1, B2 infeasible", be.isInfeasibleSet(infeas));
    infeas.clear();
    infeas.add(findBid("B1"));
    infeas.add(findBid("B2a"));
    assert("B1, B2a infeasible", be.isInfeasibleSet(infeas));
    infeas.clear();
    infeas.add(findBid("B1"));
    infeas.add(findBid("B3"));
    assert("B1, B3 infeasible", be.isInfeasibleSet(infeas));
    infeas.clear();
    infeas.add(findBid("B1"));
    infeas.add(findBid("B3a"));
    assert("B1, B3a infeasible", be.isInfeasibleSet(infeas));
    infeas.clear();
    infeas.add(findBid("B1"));
    infeas.add(findBid("B5"));
    infeas.add(findBid("B6"));
    assert("B1, B5, B6 infeasible", be.isInfeasibleSet(infeas));
    infeas.clear();
  }

  public void testConflictingBids3a ()
  {
    log.debug("testConflictingBids3a()");
    setup4();
    be.checkCoverage();
    Set b1c = be.conflictingBids(findBid("B1a"));
    printBidSet("cb3a", b1c);
    assertEquals("Correct number of conflicting bids", 2, b1c.size());
    HashSet comp = new HashSet();
    comp.add(findBid("B1"));
    comp.add(findBid("B4"));
    assert("Sets match", comp.equals(b1c));
  }
  
  public void testConflictingBids4 ()
  {
    log.debug("testConflictingBids4()");
    setup4();
    be.checkCoverage();
    Set b4c = be.conflictingBids(findBid("B4"));
    printBidSet("cb4", b4c);
    assertEquals("Correct number of conflicting bids", 8, b4c.size());
    HashSet comp = new HashSet();
    comp.add(findBid("B1"));
    comp.add(findBid("B1a"));
    comp.add(findBid("B2"));
    comp.add(findBid("B2a"));
    comp.add(findBid("B3"));
    comp.add(findBid("B3a"));
    comp.add(findBid("B5"));
    comp.add(findBid("B6"));
    assert("Sets match", comp.equals(b4c));
  }
  
  public void testConflictingBids5 ()
  {
    // here we see whether we notice the fact that B5 is too long
    log.debug("testConflictingBids5()");
    setup4();
    be.checkCoverage();
    HashSet comp = new HashSet();
    comp.add(findBid("B6"));
    comp.add(findBid("B5"));
    comp.add(findBid("B7"));
    assert("B6, B5, B7 should be infeasible", be.isInfeasibleSet(comp));
    comp.clear();
    comp.add(findBid("B6"));
    comp.add(findBid("B5"));
    comp.add(findBid("B1a"));
    assert("B6, B5, B1a should be infeasible", be.isInfeasibleSet(comp));
  }

  private void printBidSet (String msg, Set theBids)
  {
    if (log.isDebugEnabled()) {
      StringBuffer buf = new StringBuffer();
      buf.append(msg).append(": ");
      String sep = "";
      for (Iterator bids = theBids.iterator();
	   bids.hasNext(); ) {
	buf.append(sep).append(((Bid)bids.next()).getId());
	sep = " ";
      }
      log.debug(buf);
    }
  }
}
