/*------------------------------------------------------------------------
 * MAGNET utility test components
 * $Id: BidtreeNodeTest.java,v 1.2.2.2 2002/05/05 19:41:52 ashutosh Exp $
 * Copyright (c) 2002 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Mon Mar 11 16:37:33 2002  John Collins
 *   Update for redesigned Bidtree
 * 
 * Wed Feb 20 20:55:06 2002  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.client.SampleTaskPlan;
import edu.umn.magnet.client.status.AgentStatus;
import edu.umn.magnet.client.status.PlanStatus;
import edu.umn.magnet.client.status.TaskStatus;
import edu.umn.magnet.protocol.InvalidDataException;
import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskPlan;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;

import org.apache.log4j.Category;
import org.apache.log4j.FileAppender;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.Priority;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.SortedSet;

/**
 * This is a tester for edu.umn.magnet.client.eval.Bidtree.
 * There are two aspects to be tested: setup and model access.
 * @author John Collins
 */
public class BidtreeNodeTest extends TestCase
{
  static private Category log
    = Category.getInstance(BidtreeNodeTest.class.getName());

  static final double SLOP = 1.0e-6;

//    static {
//      try {
//        PatternLayout layout = new PatternLayout("%5p: %m%n");
//        FileAppender file
//  	  = new FileAppender(layout, "BidtreeNodeTest.log", false);
//        log.addAppender(file);
//        log.setPriority(Priority.DEBUG);
//      }
//      catch (IOException ioe) {
//        System.out.println("Can't open log file");
//      }
//    }

  static final long BASE = 1000000000000L;
  static final long HOUR = 3600000L;

  private TaskPlan plan = null;
  private TaskBidPlan tbp = null;
  private PlanStatus status = null;
  private BidEvaluator be = null;
  private Bidtree bt = null;
  private BidtreeSearch bts = null;
  private BidtreeNode btn = null;

  private HashMap bidMap = null;

  public BidtreeNodeTest (String testName)
  {
    super(testName);
  }

  /**
   * You run the main method to exercise this thing.  There are no
   * command-line args.
   */
  public static void main (String[] args)
  {
    junit.textui.TestRunner.run(suite());
  }

  public static Test suite ()
  {
    return new TestSuite(BidtreeNodeTest.class);
  }

  private void setup_a ()
  {
    plan = SampleTaskPlan.generateTaskPlan();
    tbp = new TaskBidPlan(plan);
    status = AgentStatus.getInstance().createPlanStatus(tbp);
    be = new BidEvaluator();
    be.setSessionId("test");
    bidMap = null;
    for (Iterator tasks = plan.iterator();
	 tasks.hasNext(); ) {
      SubtaskRequest task = (SubtaskRequest)tasks.next();
      status.setTaskState(task, TaskStatus.OFFERED);
    }
    be.setPlanStatus(status);
    Properties props = new Properties();
    props.put("Eval.UseCoverageConstraints", "false");
    props.put("Eval.UseFeasibilityConstraints", "false");
    be.setProperties(props);
  }

  private void addBids1 ()
  {
    Bid bid;
    // Create some Bids against the TP
    try {
      bid = new Bid();
      bid.setId("B1");
      bid.addTask("T2");
      bid.addTask("T4");
      bid.addTask("T6");
      bid.setCost(60L);
      be.addBid(bid);
      
      bid = new Bid();
      bid.setId("B2");
      bid.addTask("T1");
      bid.addTask("T3");
      bid.setCost(80L);
      be.addBid(bid);
      
      bid = new Bid();
      bid.setId("B3");
      bid.addTask("T1");
      bid.addTask("T3");
      bid.addTask("T5");
      bid.setCost(90L);
      be.addBid(bid);
      
      bid = new Bid();
      bid.setId("B4");
      bid.addTask("T1");
      bid.addTask("T2");
      bid.addTask("T3");
      bid.setCost(45L);
      be.addBid(bid);
      
      bid = new Bid();
      bid.setId("B5");
      bid.addTask("T4");
      bid.addTask("T6");
      bid.setCost(36L);
      be.addBid(bid);
    }
    catch (InvalidDataException ide) {
      fail("InvalidDataException");
    }
    // at this point, we have the following task->bid mappings available
    //  T1: B2,B3,B4
    //  T2: B1,B4
    //  T3: B2,B3,B4
    //  T4: B1,B5
    //  T5: B3
    //  T6: B1,B5
  }

  private void setup_b ()
  {
    setup_a();
    addBids1();
    //log.debug("Creating BidtreeSearch");
    bts = new BidtreeSearch();
    //log.debug("Setting BidEvaluator");
    bts.setBidEvaluator(be);
  }
  
  // Make sure we can create one of these things
  public void testCreate ()
  {
    log.debug("testCreate()");
    setup_b();
    btn = new BidtreeNode(tbp, bts);
    assert("BidtreeNode created", (btn != null));
    assert("Correct type", (btn instanceof BidtreeNode));
  }

  // Check initial bid set. Should be one of the singleton bids, B1 or
  // B3
  public void testInitialBidSet ()
  {
    log.debug("testInitialBidSet()");
    setup_b();
    btn = new BidtreeNode(tbp, bts);
    ArrayList initialBids = btn.getNextBidIds();
    assertEquals("Correct number of bids", 1, initialBids.size());
    String bidId = (String)initialBids.get(0);
    assertEquals("Correct initial bid", "B3", bidId);
  }

  private void setup2()
  {
    setup_b();
    //be.generateCoverageConstraints();
    btn = new BidtreeNode(tbp, bts);
  }

  // Test basic model access on first node
  public void testModelAccess_a ()
  {
    log.debug("testModelAccess_a()");
    setup2();
    assert("Should not be feasible", !btn.isFeasible());
    assert("Should not be covered", !btn.isCovered());
    assertEquals("Heuristic value", 111.0, btn.getHeuristic(), SLOP);
    assertEquals("Correct TaskBidPlan", tbp, btn.getTaskBidPlan());
    assertEquals("Correct plan access", plan, btn.getTaskPlan());
    assert("Has more bids", btn.hasMoreBids());
    assertEquals("Correct node number", 0, btn.getNodeNumber());
  }

  // Test first expansion
  public void testExpansion_a ()
  {
    log.debug("testExpansion_a()");
    setup2();
    BidtreeNode next = btn.expandNode(1);
    assert("New node not null", (next != null));
    assert("New node of correct type", (next instanceof BidtreeNode));
    assert("New node not pruned", (next.getTaskBidPlan() != null));
    // At this point, B3 should be mapped to T1, T3, and T5. It
    // conflicts with B2 and B4. Heuristic should now be 60.
    assertEquals("Heuristic value", 56.0, next.getHeuristic(), SLOP);
    // The original node should have no more bids
    assert("No more bids", !btn.hasMoreBids());
    // The new node should have two bids
    assert("New node has bids", next.hasMoreBids());
    assertEquals("Correct node number", 1, next.getNodeNumber());
  }

  // Test second expansion
  public void testExpansion_b ()
  {
    log.debug("testExpansion_b()");
    setup2();
    BidtreeNode next = btn.expandNode(1);
    assert("New node has bids", next.hasMoreBids());
    BidtreeNode n2 = next.expandNode(2);
    assertEquals("Correct node number", 2, n2.getNodeNumber());
    // The first attempt at n2 should have had two bids, B3 and B5,
    // and a heuristic value of 0. It's also a dead end, so it should
    // have been pruned. The second attempt should have bids B3 and
    // B1. It should be covered.
    assert("Original node has no bids", !next.hasMoreBids());
    assert("n2 not pruned", (n2.getTaskBidPlan() != null));
    assertEquals("Correct heuristic for n2", 0.0, n2.getHeuristic(), SLOP);
    assert("n2 has no expansions", !n2.hasMoreBids());
  }
}
