/*------------------------------------------------------------------------
 * MAGNET utility test components
 * $Id: BidtreeTest.java,v 1.4.2.2 2002/05/05 19:41:52 ashutosh Exp $
 * Copyright (c) 2002 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Mon Mar 11 16:37:16 2002  John Collins
 *   Update for redesigned Bidtree
 * 
 * Tue Feb 19 09:02:10 2002  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.eval;

import edu.umn.magnet.client.SampleTaskPlan;
import edu.umn.magnet.client.status.AgentStatus;
import edu.umn.magnet.client.status.PlanStatus;
import edu.umn.magnet.client.status.TaskStatus;
import edu.umn.magnet.protocol.InvalidDataException;
import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskPlan;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;

import org.apache.log4j.Category;
import org.apache.log4j.FileAppender;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.Priority;

import java.io.IOException;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.SortedSet;

/**
 * This is a tester for edu.umn.magnet.client.eval.Bidtree.
 * There are two aspects to be tested: setup and model access.
 * @author John Collins
 */
public class BidtreeTest extends TestCase
{
  static private Category log
    = Category.getInstance(Bidtree.class.getName());

  static {
    try {
      PatternLayout layout = new PatternLayout("%5p: %m%n");
      FileAppender file
	  = new FileAppender(layout, "BidtreeTest.log", false);
      log.addAppender(file);
      log.setPriority(Priority.DEBUG);
    }
    catch (IOException ioe) {
      System.out.println("Can't open log file");
    }
  }

  static final long BASE = 1000000000000L;
  static final long HOUR = 3600000L;

  private TaskPlan plan = null;
  private TaskBidPlan tbp = null;
  private PlanStatus status = null;
  private BidEvaluator be = null;
  private Bidtree bt = null;

  private HashMap bidMap = null;

  public BidtreeTest (String testName)
  {
    super(testName);
  }

  /**
   * You run the main method to exercise this thing.  There are no
   * command-line args.
   */
  public static void main (String[] args)
  {
    junit.textui.TestRunner.run(suite());
  }

  public static Test suite ()
  {
    return new TestSuite(BidtreeTest.class);
  }

  private void setup1 ()
  {
    plan = SampleTaskPlan.generateTaskPlan();
    tbp = new TaskBidPlan(plan);
    status = AgentStatus.getInstance().createPlanStatus(tbp);
    be = new BidEvaluator();
    be.setSessionId("test");
    bidMap = null;
    for (Iterator tasks = plan.iterator();
	 tasks.hasNext(); ) {
      SubtaskRequest task = (SubtaskRequest)tasks.next();
      status.setTaskState(task, TaskStatus.OFFERED);
    }
    be.setPlanStatus(status);
    Properties props = new Properties();
    props.put("Eval.UseCoverageConstraints", "true");
    props.put("Eval.UseFeasibilityConstraints", "false");
    be.setProperties(props);
  }

  private void addBids1 ()
  {
    Bid bid;
    // Create some Bids against the TP
    try {
      bid = new Bid();
      bid.setId("B1");
      bid.addTask("T2");
      bid.addTask("T4");
      bid.addTask("T6");
      bid.setCost(60L);
      be.addBid(bid);
      
      bid = new Bid();
      bid.setId("B2");
      bid.addTask("T1");
      bid.addTask("T3");
      bid.setCost(80L);
      be.addBid(bid);
      
      bid = new Bid();
      bid.setId("B3");
      bid.addTask("T1");
      bid.addTask("T3");
      bid.addTask("T5");
      bid.setCost(90L);
      be.addBid(bid);
      
      bid = new Bid();
      bid.setId("B4");
      bid.addTask("T1");
      bid.addTask("T2");
      bid.addTask("T3");
      bid.setCost(45L);
      be.addBid(bid);
    }
    catch (InvalidDataException ide) {
      fail("InvalidDataException");
    }
    // at this point, we have the following task->bid mappings available
    //  T1: B2,B3,B4
    //  T2: B1,B4
    //  T3: B2,B3,B4
    //  T4: B1
    //  T5: B3
    //  T6: B1
  }

  private Bid findBid (String bidId)
  {
    if (bidMap == null) {
      // we need to create the Id -> Bid map
      bidMap = new HashMap();
      for (Iterator bids = be.getBids().iterator();
	   bids.hasNext(); ) {
	Bid bid = (Bid)bids.next();
	bidMap.put(bid.getId(), bid);
      }
    }
    return (Bid)bidMap.get(bidId);
  }
  
  // Make sure we can create one of these things
  public void testCreate ()
  {
    log.debug("testCreate()");
    setup1();
    addBids1();
    bt = Bidtree.create(be);
    assert("Bidtree created", (bt != null));
    assert("Correct type", (bt instanceof Bidtree));
  }

  private void setup2()
  {
    setup1();
    addBids1();
    bt = Bidtree.create(be);
  }

  // Make sure the tasks are ordered correctly by bid count
  public void testTaskOrder ()
  {
    log.debug("testTaskOrder()");
    setup2();
    TaskDescriptor[] tasks = bt.getTaskArray();
    assertEquals("Correct task count", 6, tasks.length);
    int min = 0;
    for (int i = 0; i < tasks.length; i++) {
      TaskDescriptor td = tasks[i];
      assert("Valid task descriptor", (td != null));
      int count = td.getBidCount();
      assert("No backsliding", (count >= min));
      min = count;
    }
  }

  // Find the single bid B3 associated with task T5
  public void testBidAccess_T5 ()
  {
    log.debug("testBidAccess_T5()");
    setup2();
    TaskDescriptor[] tasks = bt.getTaskArray();
    int[] mask = new int[tasks.length];
    int index = 0;
    for (int i = 0; i < tasks.length; i++) {
      TaskDescriptor td = tasks[i];
      if (td.getTaskId().equals("T5"))
	mask[index] = Bidtree.IN;
      else
	mask[index] = Bidtree.BOTH;
      index++;
    }
    Collection bids = bt.getBidCollections(mask);
    // At this point, we should have a Collection containing 1
    // BtBucket containing bid B3.
    assert("Non-null result", (bids != null));
    assertEquals("Correct size of outer collection", 1, bids.size());
    Iterator bidi = bids.iterator();
    BtBucket bt5 = (BtBucket)bidi.next();
    assertEquals("Correct size of inner collection", 1, bt5.size());
    BtBidWrapper result = bt5.getBid(0);
    assert("Non-null bid", (result != null));
    assertEquals("Correct bid", "B3", result.getBid().getId());
  }

  // Find the single bid B3 associated with task T5, using getBids()
  public void testBidAccess_T5a ()
  {
    log.debug("testBidAccess_T5a()");
    setup2();
    TaskDescriptor[] tasks = bt.getTaskArray();
    int[] mask = new int[tasks.length];
    int index = 0;
    for (int i = 0; i < tasks.length; i++) {
      TaskDescriptor td = tasks[i];
      if (td.getTaskId().equals("T5"))
	mask[index] = Bidtree.IN;
      else
	mask[index] = Bidtree.BOTH;
      index++;
    }
    BtBidWrapper[] bids = bt.getBids(mask);
    // At this point, we should have an array containing bid B3.
    assert("Non-null result", (bids != null));
    assertEquals("Correct size of array", 1, bids.length);
    BtBidWrapper result = bids[0];
    assert("Non-null bid", (result != null));
    assertEquals("Correct bid", "B3", result.getBid().getId());
  }

  // Find the bids that include T1 and do not conflict with B1, using
  // the internal accessor
  public void testBidAccess_T1_B1 ()
  {
    log.debug("testBidAccess_T1_B1()");
    setup2();
    TaskDescriptor[] tasks = bt.getTaskArray();
    int[] mask = new int[tasks.length];
    int index = 0;
    for (int i = 0; i < tasks.length; i++) {
      TaskDescriptor td = tasks[i];
      if (td.getTaskId().equals("T1"))
	mask[index] = Bidtree.IN;
      else if (td.getTaskId().equals("T2")
	       || td.getTaskId().equals("T4")
	       || td.getTaskId().equals("T6"))
	mask[index] = Bidtree.OUT;
      else
	mask[index] = Bidtree.BOTH;
      index++;
    }
    Collection bids = bt.getBidCollections(mask);
    //log.debug("getBids: " + bids.toString());
    // At this point, we should have a Collection containing 2
    // Collections
    assert("Non-null result", (bids != null));
    assertEquals("Correct size of outer collection", 2, bids.size());
    Iterator bidi = bids.iterator();
    BtBucket bidj = (BtBucket)bidi.next();
    assertEquals("Correct size of first inner collection", 1, bidj.size());
    BtBidWrapper result = bidj.getBid(0);
    // first bid should be either B2 or B3
    assert("Non-null bid", (result != null));
    String firstBid = result.getBid().getId();
    assert("Correct first bid", (firstBid.equals("B2")
				 || firstBid.equals("B3")));
    // second bid should be the other one
    String secondBid = "B3";
    if (firstBid.equals("B3"))
      secondBid = "B2";
    log.debug("First " + firstBid + ", second " + secondBid);
    bidj = (BtBucket)bidi.next();
    assertEquals("Correct size of second inner collection", 1, bidj.size());
    result = bidj.getBid(0);
    assert("Non-null second bid", (result != null));
    assertEquals("Correct second bid", secondBid, result.getBid().getId());
  }

  // Find the bids that include T1 and do not conflict with B1
  public void testBidAccess_T1_B1a ()
  {
    log.debug("testBidAccess_T1_B1a()");
    setup2();
    TaskDescriptor[] tasks = bt.getTaskArray();
    int[] mask = new int[tasks.length];
    int index = 0;
    for (int i = 0; i < tasks.length; i++) {
      TaskDescriptor td = tasks[i];
      if (td.getTaskId().equals("T1"))
	mask[index] = Bidtree.IN;
      else if (td.getTaskId().equals("T2")
	       || td.getTaskId().equals("T4")
	       || td.getTaskId().equals("T6"))
	mask[index] = Bidtree.OUT;
      else
	mask[index] = Bidtree.BOTH;
      index++;
    }
    BtBidWrapper[] bids = bt.getBids(mask);
    log.debug("getBids: " + bids.toString());
    // At this point, we should have an array of two bids
    assert("Non-null result", (bids != null));
    assertEquals("Correct size of array", 2, bids.length);
    BtBidWrapper result = bids[0];
    // first bid should be B3
    assert("Non-null bid", (result != null));
    assertEquals("Correct first bid", "B3", result.getBid().getId());
    // second bid should be B2
    result = bids[1];
    assert("Non-null second bid", (result != null));
    assertEquals("Correct second bid", "B2", result.getBid().getId());
  }

  // now add some more bids so some lists have more than one entry.
  private void setup3()
  {
    setup1();
    addBids1();
    Bid bid;
    // Create some additional Bids against the TP
    try {
      bid = new Bid();
      bid.setId("B5");
      bid.addTask("T2");
      bid.addTask("T4");
      bid.setCost(60L);
      be.addBid(bid);
      
      bid = new Bid();
      bid.setId("B6");
      bid.addTask("T1");
      bid.addTask("T3");
      bid.setCost(70L);
      be.addBid(bid);
      
      bid = new Bid();
      bid.setId("B7");
      bid.addTask("T3");
      bid.addTask("T5");
      bid.setCost(90L);
      be.addBid(bid);
      
      bid = new Bid();
      bid.setId("B8");
      bid.addTask("T4");
      bid.addTask("T3");
      bid.setCost(45L);
      be.addBid(bid);
    }
    catch (InvalidDataException ide) {
      fail("InvalidDataException");
    }
    // at this point, we have the following task->bid mappings available
    //  T1: B2,B3,B4,B6
    //  T2: B1,B4,B5
    //  T3: B2,B3,B4,B6,B7,B8
    //  T4: B1,B5,B8
    //  T5: B3,B7
    //  T6: B1
    bt = Bidtree.create(be);
  }

  // Find the bids that include T1 and do not conflict with B1 with
  // expanded bid list, using the internal collections interface
  public void testBidAccess_T1_B1b ()
  {
    log.debug("testBidAccess_T1_B1b()");
    setup3();
    TaskDescriptor[] tasks = bt.getTaskArray();
    int[] mask = new int[tasks.length];
    int index = 0;
    for (int i = 0; i < tasks.length; i++) {
      TaskDescriptor td = tasks[i];
      if (td.getTaskId().equals("T1"))
	mask[index] = Bidtree.IN;
      else if (td.getTaskId().equals("T2")
	       || td.getTaskId().equals("T4")
	       || td.getTaskId().equals("T6"))
	mask[index] = Bidtree.OUT;
      else
	mask[index] = Bidtree.BOTH;
      index++;
    }
    Collection bids = bt.getBidCollections(mask);
    log.debug("getBids: " + bids.toString());

    // At this point, we should have a Collection containing 2
    // BtBuckets
    assert("Non-null result", (bids != null));
    assertEquals("Correct size of outer collection", 2, bids.size());
    Iterator bidi = bids.iterator();
    BtBucket bidj = (BtBucket)bidi.next();
    assert("Non-null first collection", (bidj != null));
    BtBucket bidk = (BtBucket)bidi.next();
    assert("Non-null second collection", (bidk != null));

    // Now we make sure one of these has 1 bid, the other two, and
    // handle the 1-bid set first
    if (bidj.size() > 1) {
      BtBucket x = bidj;
      bidj = bidk;
      bidk = x;
    }

    // Now bidj has one element, B3
    assertEquals("Correct size of first inner collection", 1, bidj.size());
    BtBidWrapper result = (BtBidWrapper)bidj.getBid(0);
    // first bid should be B3
    assert("Non-null bid", (result != null));
    assert("Correct first bid", (result.getBid().getId().equals("B3")));
    
    // And bidk should have two elements, B6 and B2
    assertEquals("Correct size of second inner collection", 2, bidk.size());
    result = (BtBidWrapper)bidk.getBid(0);
    assert("Non-null second bid", (result != null));
    assertEquals("Correct second bid", "B6", result.getBid().getId());
    result = (BtBidWrapper)bidk.getBid(1);
    assert("Non-null third bid", (result != null));
    assertEquals("Correct second bid", "B2", result.getBid().getId());
  }

  // Find the bids that include T1 and do not conflict with B1 with
  // expanded bid list, using the internal collections interface
  public void testBidAccess_T1_B1c ()
  {
    log.debug("testBidAccess_T1_B1c()");
    setup3();
    TaskDescriptor[] tasks = bt.getTaskArray();
    int[] mask = new int[tasks.length];
    int index = 0;
    for (int i = 0; i < tasks.length; i++) {
      TaskDescriptor td = tasks[i];
      if (td.getTaskId().equals("T1"))
	mask[index] = Bidtree.IN;
      else if (td.getTaskId().equals("T2")
	       || td.getTaskId().equals("T4")
	       || td.getTaskId().equals("T6"))
	mask[index] = Bidtree.OUT;
      else
	mask[index] = Bidtree.BOTH;
      index++;
    }
    BtBidWrapper[] bids = bt.getBids(mask);
    log.debug("getBids: " + bids.toString());

    // At this point, we should have 3 bids, B3, B6, and B2
    assert("Non-null result", (bids != null));
    assertEquals("Correct size of array", 3, bids.length);
    assert("Non-null first bid", (bids[0] != null));
    assertEquals("Correct first bid", "B3", bids[0].getBid().getId());
    assert("Non-null second bid", (bids[1] != null));
    assertEquals("Correct second bid", "B6", bids[1].getBid().getId());
    assert("Non-null third bid", (bids[2] != null));
    assertEquals("Correct second bid", "B2", bids[2].getBid().getId());
  }
}
