/*------------------------------------------------------------------------
 * MAGNET utility test components
 * $Id: PlanStatusTest.java,v 1.4.2.1 2002/05/05 19:41:53 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Sat Dec 30 09:24:16 2000  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.status;

import edu.umn.magnet.client.SampleTaskPlan;
import edu.umn.magnet.protocol.MagnetException;
import edu.umn.magnet.protocol.InvalidDataException;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.client.eval.TaskBidPlan;
import edu.umn.magnet.util.time.AgentTime;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Date;

/**
 * This is a tester for edu.umn.magnet.client.Something
 * @author John Collins
 */
public class PlanStatusTest extends TestCase
{
  private static long SECOND = 1000L;

  TaskPlan plan = null;
  TaskBidPlan tbp = null;
  PlanStatus stat = null;
  PlanStatusEvent event = null;
  
  public PlanStatusTest (String testName)
  {
    super(testName);
  }

  /**
   * You run the main method to exercise this thing.  There are no
   * command-line args.
   */
  public static void main (String[] args)
  {
    junit.textui.TestRunner.run(suite());
  }

  public static Test suite ()
  {
    return new TestSuite(PlanStatusTest.class);
  }

  private void setup1 ()
  {
    plan = SampleTaskPlan.generateTaskPlan();
    tbp = new TaskBidPlan(plan);
    AgentStatus as = AgentStatus.getInstance();
    stat = as.createPlanStatus(tbp);
    event = null;
  }
  
  public void testCreate ()
  {
    setup1();
    assertEquals("TBP correct", tbp, stat.getTaskBidPlan());
    assertEquals("TP correct", plan, stat.getTaskPlan());
  }

  public void testTS ()
  {
    setup1();
    TaskStatus ts1 = stat.getTaskStatus(plan.getTaskData("T1"));
    assert("non-null", (ts1 != null));
    assert("correct type", (ts1 instanceof TaskStatus));
  }

  public void testTSIterator ()
  {
    setup1();
    ArrayList tsl = new ArrayList();
    for (Iterator tss = stat.getTaskStatusIterator();
	 tss.hasNext(); ) {
      tsl.add(tss.next());
    }
    assertEquals("Correct task count", 6, tsl.size());
  }

  // test event stuff

  public void testTaskBidPlanEvent ()
  {
    setup1();
    TaskBidPlan tbp1 = new TaskBidPlan(plan);
    stat.addListener(new PropertyChangeListener()
      {
	public void propertyChange(PropertyChangeEvent pse)
	{
	  // keep in mind that Java is a STATIC OO language, not a
	  // dynamic one...
	  if (pse instanceof PlanStatusEvent)
	    event = (PlanStatusEvent)pse;
	  else
	    fail("incorrect event type: " + pse);
	}
      });
    stat.setTaskBidPlan(tbp1);
    assert("event was triggered", event != null);
    assert("correct event type", (event instanceof PlanStatusEvent));
    assertEquals("correct event label",
		 "TaskBidPlan",
		 event.getPropertyName());
    assertEquals("correct old value", tbp, event.getOldValue());
    assertEquals("correct new value", tbp1, event.getNewValue());
  }

  public void testStateEvent ()
  {
    setup1();
    stat.addListener(new PropertyChangeListener()
      {
	public void propertyChange(PropertyChangeEvent pse)
	{
	  if (pse instanceof PlanStatusEvent)
	    event = (PlanStatusEvent)pse;
	  else
	    fail("incorrect event type: " + pse);
	}
      });
    stat.setPlanState(PlanStatus.BIDDING);
    assert("event was triggered", event != null);
    assert("correct event type", (event instanceof PlanStatusEvent));
    assertEquals("correct event label",
		 "State",
		 event.getPropertyName());
    assertEquals("correct old value",
		 PlanStatus.PLANNING,
		 event.getOldValue());
    assertEquals("correct new value",
		 PlanStatus.BIDDING,
		 event.getNewValue());
  }

  // set start time in the past
  public void testStartErr1 ()
  {
    setup1();
    Date now = AgentTime.getDate();
    try {
      stat.setPlanStart(new Date(now.getTime() - 2));
    }
    catch (InvalidDataException ide) {
      return;
    }
    catch (InvalidStateException ise) {
      fail("Wrong exception type");
    }
    fail("Failed to throw exception");
  }

  // set start time in the future, make sure it propagates forward and
  // happens.
  public void testStartEvent ()
  {
    Date now = AgentTime.getDate();
    Date start = new Date(now.getTime() + SECOND/2L);
    setup1();
    stat.addListener(new PropertyChangeListener()
      {
	public void propertyChange(PropertyChangeEvent pse)
	{
	  if (pse instanceof PlanStatusEvent)
	    event = (PlanStatusEvent)pse;
	  else
	    fail("incorrect event type: " + pse);
	}
      });
    try {
      stat.setPlanStart(start);
    }
    catch (MagnetException ide) {
      fail("Failed to set Plan Start Time: " + ide);
      return;
    }
    // Check the TMEntry
    TMEntry tme = stat.getPlanStartTME();
    assert("TME got set", tme != null);
    assertEquals("TME set correctly",
		start.getTime(),
		tme.getExpirationTime());
    // check for set time event
    assert("set start time event was triggered",
	   event != null);
    assert("correct event type",
	   (event instanceof PlanStatusEvent));
    assertEquals("correct event label",
		 "Start",
		 event.getPropertyName());
    assert("correct old value", event.getOldValue() == null);
    assertEquals("correct new value", start, event.getNewValue());
    // check for state change event
    event = null;
    try {
      Thread.sleep(SECOND); // make sure the event has time to happen
    }
    catch (InterruptedException ie) {
      fail("Test interrupted");
      return;
    }
    assert("set state event was triggered", event != null);
    assert("correct event type", (event instanceof PlanStatusEvent));
    assertEquals("correct event label",
		 "State",
		 event.getPropertyName());
    assertEquals("correct old value",
		 PlanStatus.PLANNING,
		 event.getOldValue());
    assertEquals("correct new value",
		 PlanStatus.EXECUTION,
		 event.getNewValue());
  }

  // test incorrect deadline setting
  public void testDeadlineErr1 ()
  {
    setup1();
    Date now = AgentTime.getDate();
    try {
      stat.setPlanDeadline(new Date(now.getTime() - 2));
    }
    catch (InvalidDataException ide) {
      return;
    }
    catch (InvalidStateException ise) {
      fail("Incorrect Exception: " + ise);
    }
    fail("Failed to throw exception");
  }

  // test deadline management
  public void testDeadlineEvent ()
  {
    Date now = AgentTime.getDate();
    Date deadline = new Date(now.getTime() + SECOND/2L);
    setup1();
    stat.addListener(new PropertyChangeListener()
      {
	public void propertyChange(PropertyChangeEvent pse)
	{
	  if (pse instanceof PlanStatusEvent)
	    event = (PlanStatusEvent)pse;
	  else
	    fail("incorrect event type: " + pse);
	}
      });
    try {
      stat.setPlanDeadline(deadline);
    }
    catch (MagnetException ide) {
      fail("Failed to set plan deadline: " + ide);
      return;
    }
    assert("event was triggered", event != null);
    assert("correct event type", (event instanceof PlanStatusEvent));
    assertEquals("correct event label",
		 "Deadline",
		 event.getPropertyName());
    assert("correct old value", event.getOldValue() == null);
    assertEquals("correct new value", deadline, event.getNewValue());
    // check for state change event
    event = null;
    try {
      Thread.sleep(SECOND); // make sure the event has time to happen
    }
    catch (InterruptedException ie) {
      fail("Test interrupted");
      return;
    }
    assert("set state event was triggered", event != null);
    assert("correct event type", (event instanceof PlanStatusEvent));
    assertEquals("correct event label",
		 "State",
		 event.getPropertyName());
    assertEquals("correct old value",
		 PlanStatus.PLANNING,
		 event.getOldValue());
    assertEquals("correct new value",
		 PlanStatus.COMPLETE,
		 event.getNewValue());
  }
}
