/*-----------------------------------------------------------------------
 * Magnet Protocol Element
 * $Id: JobInfoFactory.java,v 1.9 2001/10/17 03:41:07 jcollins Exp $
 * Copyright (c) 1998 University of Minesota
 *-----------------------------------------------------------------------
 * Revision History:
 *
 * Mon Jul 02 14:25:00 2001  Len Josephs
 *   #0000137 Add ability to fill with JobInfo from an array.
 *
 * Sun Apr 22 12:45:06 2001  John E. Collins
 *   #0000026 convert to singleton to allow protocol to contain keys
 *   rather than instances of JobInfo.
 *
 * 18Apr2001 DFK - Bug #0000025: added list() to return list of
 *                  currently-known JobInfo entries.
 *
 * Sun Mar 25 12:59:51 2001  John Collins
 *   Bug #0000013: Don't create default jobinfo type automatically.
 *
 * Sun Feb 25 20:05:15 2001  John Collins
 *   Initial Entry, moved static stuff out of protocol.cust.JobInfo.
 */

package edu.umn.magnet.client.plan;

import edu.umn.magnet.protocol.JobInfo;
import edu.umn.magnet.protocol.AbstractJobInfoFactory;
import edu.umn.magnet.util.rvars.UniformVariable;
import edu.umn.magnet.util.rvars.RVFManager;

import org.apache.log4j.Category;

import java.util.HashMap;
import java.util.TreeMap;   // dfk
import java.util.StringTokenizer;
import java.util.Properties;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.FileInputStream;

/**
 * A factory for generating random task types, useful in generating
 * random plans for testing purposes.  This is a singleton class,
 * intended to be served up by the inherited static
 * <tt>getInstance()</tt> method.
 * @author John Collins
 * @version $Revision: 1.9 $
 */
public class JobInfoFactory extends AbstractJobInfoFactory
{
  // log category
  private static Category log =
    Category.getInstance(JobInfoFactory.class.getName());

  // One hour
  private static final double HOUR = 3600000.0;

  // Random var for generating random job types
  private  UniformVariable s_randomVar
    = RVFManager.getRVF("PlanSeed").createUniformVariable(0.0, 1.0);
  // Sum of frequency data - make sure this gets set to 0.0 if any
  // frequency is ever changed.
  private  double s_totalFreq = 0.0;

  // ------------------  Methods ------------------------
  /**
   * default constructor, private to enforce singleton status
   */
  private JobInfoFactory ()
  {
    super();
  }

  /**
   * Instantiates the singleton.  You need to call this on a subclass,
   * to avoid the error you'll get if you call it on this abstract
   * superclass.  Note that if you call this a second time, the
   * original singleton will get replaced with a new one.
   * <p>
   * This method must be implemented appropriately in each subclass.
   * </p>
   */
  public static void instantiate ()
  {
    setInstance(new JobInfoFactory());
  }

  /**
   * Creates a new JobInfo instance and stashes it in an index under
   * the job type name.  This is typically done at
   * startup/configuration time, by reading a file or a database
   * record describing a particular task type.
   * @param props A Properties instance that contains the data to
   *   populate this thing. A null value gives you a default JobInfo.
   * @return The new JobInfo instance (may be recycled if the same
   *   name is reused). Null is returned if one of the property values
   *   is badly formatted. See the log for details.
   */
  public JobInfo create (Properties props, double freq)
  {
    // Make a default one just in case it hasn't been made already
    JobInfo defaultJi = new JobInfo();
    if (props == null) {
      if (getJobInfo("default") == null) {
	addJobInfo(defaultJi);
      }
      return defaultJi;
    }
    else {
      String ttName = props.getProperty("name");
      String description = props.getProperty("description");
      double expectedDuration = 0.0;
      double durSigma = 0.0;
      double expectedPrice = 0.0;
      double priceSigma = 0.0;
      double resourceAvail = 0.0;
      double supplierAvail = 0.0;
      try {
	expectedDuration = (getDouble(props,
				      "expected-duration",
				      defaultJi.getExpectedDuration())
			    * HOUR); // convert to milliseconds
	durSigma = (getDouble(props,
			      "duration-var",
			      (defaultJi.getDurationSigma()
			       / defaultJi.getExpectedDuration()))
		    * expectedDuration);
	expectedPrice = getDouble(props,
				  "expected-price",
				  defaultJi.getExpectedPrice());
	priceSigma = (getDouble(props,
				"price-var",
				(defaultJi.getPriceSigma()
				 / defaultJi.getExpectedPrice()))
		      * expectedPrice);
	resourceAvail = getDouble(props,
				  "resource-availability",
				  defaultJi.getResourceAvailability());
	supplierAvail = getDouble(props,
				  "supplier-availability",
				  defaultJi.getSupplierAvailability());
      }
      catch (NumberFormatException nfe) {
	log.error("** Bad JobInfo property value " + ttName);
	return null;
      }

      JobInfo ji = getJobInfo(ttName);
      if (ji != null) {
	log.error("redefine " + ttName);
      }
      ji = new JobInfo(ttName, description,
		       expectedDuration, durSigma,
		       expectedPrice, priceSigma,
		       resourceAvail, supplierAvail, freq);
      addJobInfo(ji);
      return ji;
    }
  }

  // Helper for create
  private double getDouble (Properties props, String pn,
				   double defaultVal)
    throws NumberFormatException
  {
    String pv = props.getProperty(pn);
    if (pv != null)
      return Double.valueOf(pv).doubleValue();
    else
      return defaultVal;
  }

  /**
   * Creates a set of these from a spec giving file name - frequency
   * data. The format is a string containing a space-separated list of
   * filename:frequency pairs. We expect to find a file with the given
   * name and a type of "props".  If any of them fail for some reason,
   * we just emit a message and proceed on to the next one.
   * <p>
   * <i>NOTE</i> This operation clears out all the existing JobInfo
   * instances and creates a completely new set.  Otherwise the
   * frequency data would make no sense.
   * @param jobSpec A String of the form [name:frequency ]*.
   */
  public void createFromSpec (String jobSpec)
  {
    log.debug("Creating task types from <" + jobSpec + ">");
    StringTokenizer tokens = new StringTokenizer(jobSpec, " :", true);
    Properties props = new Properties();
    String name, delim, freq;
    clear();
    s_totalFreq = 0.0;
    try {
      while (tokens.hasMoreTokens()) {
	//pick off the type name, delimited by a colon
	name = tokens.nextToken();
	if (!name.equals(" ")) {
	  if (name.equals(":")) {
	    // out of sync - complain and eat up one more token
	    log.error("createFromSpec - unexpected :");
	    name = tokens.nextToken();
	  }
	  else {
	    delim = tokens.nextToken();
	    if (delim.equals(":")) {
	      // this is good
	      freq = tokens.nextToken();
	      double dfreq = 0.0;
	      try {
		dfreq = Double.valueOf(freq).doubleValue();
	      }
	      catch (NumberFormatException nfe) {
		log.error("createFromSpec - bad number "
			     + freq);
		continue;
	      }
	      // now we read the props file and instantiate the
	      // JobInfo
	      props.clear();
	      log.debug("Opening task type file " + name + ".props");
	      try {
		FileInputStream stream = new FileInputStream(name + ".props");
		props.load(stream);
	      }
	      catch (FileNotFoundException fnf) {
		log.error("couldn't find task type file " + name
			  + ".props");
		continue;
	      }
	      catch (IOException ioe) {
		log.error("error loading task type file " + name
			  + ".props");
		continue;
	      }
	      JobInfo ji = create(props, dfreq);
	    }
	  }
	}
      }
    }
    catch (NoSuchElementException nse) {
      log.error("createFromSpec - premature end");
    }
  }

  /**
   * Leave off the argument, and you get a random JobInfo from the
   * internal stash.
   */
  public JobInfo find ()
  {
    JobInfo result = null;
    HashMap jobTypes = list();
    if (jobTypes.size() == 0) {
      return create((Properties)null, 1.0);
    }
    Iterator jobs = jobTypes.values().iterator();
    for (double entry = s_randomVar.nextDouble() * getTotalFreq();
	 (entry >= 0.0) && jobs.hasNext(); ) {
      result = (JobInfo)(jobs.next());
      entry -= result.getFrequency();
    }
    log.debug("find() returns " + (result.getTaskType()));
    return result;
  }

  // Lazy evaluation helper for random find.
  private double getTotalFreq ()
  {
    if (s_totalFreq == 0.0)
      for (Iterator jis = list().values().iterator();
	   jis.hasNext(); ) {
	JobInfo ji = (JobInfo)jis.next();
	s_totalFreq += ji.getFrequency();
      }
    return s_totalFreq;
  }

  /** Return the currently-known set of JobInfo entries, as a TreeMap
   * (ordered list).
   */
  public TreeMap sortedList()
  {
    return new TreeMap(list());
  }

  /**
   * Creates a set of JobInfo from a passed-in array of JobInfo.
   * <i>NOTE</i> This operation clears out all the existing JobInfo
   * instances and creates a completely new set.
   * @param jobInfos An array of JobInfos to put in the factory.
   */
  public void setFromArray (JobInfo[] jobInfos) {
    log.debug("Filling JobInfoFactory with jobInfo[] data.");
    clear();
    if (jobInfos == null)
      return;
    for (int i=0; i < jobInfos.length; i++) {
      addJobInfo(jobInfos[i]);
    }
  }

 /**
   * Applies the same scale factors on the attributes of the
   * JobInfos in the passed-in array as is applied when create
   * is called.  This allows JobInfo arrays to match file
   * specifications (e.g. the *.props files).  In other words,
   * the JobInfos in the passed-in array are treated as
   * JobInfo specifications and the JobInfoFactory will contain
   * usable JobInfos.
   * <i>NOTE</i> This operation will update the passed-in array.
   * @param jobInfos An array of JobInfos to scale.
   */
  public void scaleAndAddFromArray (JobInfo[] jobInfos) {
    log.debug("Scaling jobInfo[] data and adding to JobInfoFactory.");
    if (jobInfos == null)
      return;
    for (int i=0; i < jobInfos.length; i++) {
      jobInfos[i].setExpectedDuration(
        jobInfos[i].getExpectedDuration()*HOUR); // convert to milliseconds
      jobInfos[i].setDurationSigma(
        jobInfos[i].getDurationSigma()*
        jobInfos[i].getExpectedDuration());
      // Price not scaled.
      jobInfos[i].setPriceSigma(
      	jobInfos[i].getPriceSigma()*jobInfos[i].getExpectedPrice());
      // Don't scale resourceAvail, supplierAvail, or freq.

      // Finally, add the JobInfo.
      addJobInfo(jobInfos[i]);
    }
  }
}
