/*------------------------------------------------------------------------
 * MAGNET supplier simulation components
 * $Id: BidGenerator.java,v 1.4.2.1 2002/05/05 19:41:54 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Thu Jan 11 13:33:39 2001  John Collins
 *   use new logging system
 * 
 * Sat Dec 16 18:10:05 2000  John Collins
 *   Eliminate bid breakdowns
 * 
 * Fri Oct  6 10:04:01 2000  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.client.sup;

import edu.umn.magnet.util.rvars.UniformVariable;

import edu.umn.magnet.protocol.cust.RFQ;
import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.InvalidDataException;

import org.apache.log4j.Category;

import java.util.Collection;
import java.util.Iterator;
import java.util.Date;
import java.util.Properties;
import java.util.HashMap;

/**
 * Instances are used to generate individual bids according to some scheme,
 * as determined by subclass.  This abstract superclass of all the actual
 * bid generators provides a set of useful common attributes and processing
 * methods, especially as related to accessing the incoming RFQ and its plan.
 * @author John Collins
 * @version $Revision: 1.4.2.1 $
 */
public abstract class BidGenerator
{
  static Category log = Category.getInstance(BidGenerator.class.getName());

  private STBidGenerator m_stBidGenerator = null;
  private RFQ m_cfb = null;
  private Properties m_props = null;
  private SupplyMonster m_sm = null;
  private UniformVariable m_uVariable = null;
  private double m_maxDiscountPercent;

  /**
   * Every subclass must implement the protected init method, and must
   * include a call to super.init(props, sm)
   */
  protected void init (Properties props, SupplyMonster sm)
  {
    m_props = props;
    m_sm = sm;
    m_uVariable = sm.getRVF().createUniformVariable(0,1);
    m_maxDiscountPercent
	= sm.getPropertyDouble("Sup.MaxDiscountPercent", 10.0);
  }

  /**
   * Gets the configuration Properties object
   */
  protected Properties getProps ()
  {
    return m_props;
  }

  /**
   * Gets the reference to the overlying SupplyMonster.
   */
  protected SupplyMonster getSupplyMonster ()
  {
    return m_sm;
  }

  /**
   * Gets the TaskPlan instance for which we generate bids
   */
  protected TaskPlan getTaskPlan()
  {
    return m_sm.getPlan();
  }

  /**
   * This is the core method for bid generators.  Given a RFQ,
   * generate a Bid. Subclasses are distinguished by
   * how they generate the subtask bids, in generateSTBids().
   * @param cfb  The incoming RFQ instance
   * @param seed A SubtaskRequest that will definitely be included in
   *   the bid, if possible.  Otherwise the result will be null.
   */
  Bid generateBid (RFQ cfb, SubtaskRequest seed)
  {
    m_cfb = cfb;
    Bid aBid = new Bid();
    try {
      aBid.setId(m_sm.createUniqueId("B"));
      aBid.setRFQId(cfb.getId());
      aBid.setAgentId(m_sm.createUniqueId("S"));
    }
    catch (InvalidDataException ide) {
      ide.printStackTrace();
    }
    log.debug("bidId = " + aBid.getId());
    
    // Generate the subtask bids - actually a collection of TaskSpec
    // instances.  Abort if there aren't any.
    Collection stbids = generateSTBids(seed);
    if (stbids == null || stbids.isEmpty())
      return null;

    try {
      for (Iterator ts = stbids.iterator();
	   ts.hasNext(); ) {
	TaskSpec spec = (TaskSpec)ts.next();
	aBid.addTask(spec.getTaskId(),
		     spec.getEarlyStart(),
		     spec.getLateStart(),
		     spec.getDuration());
      }
    }
    catch (InvalidDataException ide) {
      ide.printStackTrace();
    }
    
    // find the total cost, total duration, early start and late finish.
    // as defined by subtasks
    long totalCost = 0L;
    Date earlyStart = new Date(Long.MAX_VALUE);
    Date lateFinish = new Date(0L);
    long earlyFinish = 0L;

    for (Iterator ts = stbids.iterator();
	 ts.hasNext(); ) {
      TaskSpec spec = (TaskSpec)ts.next();
      totalCost += spec.getCost();
      if (spec.getEarlyStart().getTime() < earlyStart.getTime()) {
	earlyStart = spec.getEarlyStart();
      }
      long ef = spec.getEarlyStart().getTime() + spec.getDuration();
      if (ef > earlyFinish)
	earlyFinish = ef;
      if (spec.getLateFinish().getTime() > lateFinish.getTime()) {
	lateFinish = spec.getLateFinish();
      }
    }
    
    // discount the entire bid
    int discount = (int)((m_uVariable.nextDouble()
			  * m_maxDiscountPercent
			  * totalCost) / 100.0 );
    log.debug("STATISTIC: discount for " + aBid.getId() + " = " + discount);
    totalCost -= discount;
    
    // add this bid if we made a bid on it's subtasks
    aBid.setCost(totalCost);
    aBid.setDuration(earlyFinish - earlyStart.getTime());
    aBid.setEarlyStart(earlyStart);
    aBid.setLateStart(new Date(earlyStart.getTime()
			       + (lateFinish.getTime() - earlyFinish)));
    return aBid;
  }

  /**
   * Generates a set of SubtaskBids.  Each subclass must implement
   * this method.  It is this implementation that primarily
   * distinguishes subclasses.
   */
  protected abstract Collection generateSTBids (SubtaskRequest seed);

  // =============== Property access ================
  
  /**
   * Gets the STBidGenerator used to generate subtask bids.
   */
  protected STBidGenerator getSTBidGenerator ()
  {
    return m_stBidGenerator;
  }

  /**
   * Sets the STBidGenerator.
   */
  private void setSTBidGenerator (STBidGenerator newGen)
  {
    m_stBidGenerator = newGen;
  }

  /**
   * Returns the RFQ being worked.
   */
  protected RFQ getRFQ ()
  {
    return m_cfb;
  }

  public double getMaxDiscountPercent() 
  {
    return m_maxDiscountPercent;
  }

  // ================== Time Window computations ===================

  /**
   * Figure the earliest time a bid for a task could start.  This is
   * the latest of the SubtaskRequest early start time and the early
   * finish times from the bids of all the predecessors that are
   * already included in the bid.
   */
  protected Date earlyStart (SubtaskRequest st, HashMap stBids)
  {
    long start = st.getEarliestTimeToStart().getTime();
    for (Iterator preds = getTaskPlan().getPredecessors(st);
	 preds.hasNext(); ) {
      SubtaskRequest pred = (SubtaskRequest)preds.next();
      TaskSpec spec = (TaskSpec)stBids.get(pred);
      if (spec != null) {
	long earlyFinish = (spec.getEarlyStart().getTime()
			    + spec.getDuration());
	if (earlyFinish > start) {
	  start = earlyFinish;
	}
      }
    }
    return new Date(start);
  }

  /**
   * Figure the latest time a bid for a task could finish.  This is
   * the earliest of the late finish time on the SubtaskRequest and
   * the late start times from the bids of all successors that are
   * already included in the bid.
   */
  protected Date lateFinish (SubtaskRequest st, HashMap stBids)
  {
    long finish = st.getDeadline().getTime();
    for (Iterator succs = getTaskPlan().getSuccessors(st);
	 succs.hasNext(); ) {
      SubtaskRequest succ = (SubtaskRequest)succs.next();
      TaskSpec spec = (TaskSpec)stBids.get(succ);
      if (spec != null) {
	long lateStart = spec.getLateStart().getTime();
	if (lateStart < finish) {
	  finish = lateStart;
	}
      }
    }
    return new Date(finish);
  }

  // ======================= The Factory =======================
  /**
   * This is the basic factory method for bid generators.  The
   * supplied Properties are used to determine what kind of generator
   * to build.
   */
  static BidGenerator create (Properties props, SupplyMonster sm)
  {
    String bgType = props.getProperty("Sup.BidMethod", null);
    BidGenerator result = null;
    log.info("prop: BidGenerator type = " + bgType);
    if (bgType == null) {
      log.warn("BidGenerator type not specified. Using default");
      result = new ContiguousBidder();
    }
    else {
      try {
	String classname = "edu.umn.magnet.client.sup." + bgType;
	Class target= Class.forName(classname);
	result = (BidGenerator)target.newInstance();
      }
      catch (ClassNotFoundException cnf) {
	log.error("specified BidGenerator not found.  Using default");
	result = new ContiguousBidder();
      }
      catch (InstantiationException ie) {
	log.error("specified BidGenerator can't be instantiated.  Using default");
	result = new ContiguousBidder();
      }
      catch (IllegalAccessException iae) {
	log.error("BidGenerator.create: Illegal access exception.  Using default");
	result = new ContiguousBidder();
      }
    }
    result.init(props, sm);
    result.setSTBidGenerator(STBidGenerator.create(props, sm));
    return result;
  }
}
