/*------------------------------------------------------------------------
 * MAGNET supplier simulation components
 * $Id: STBidGenerator.java,v 1.4.2.1 2002/05/05 19:41:54 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Fri Dec 15 10:13:51 2000  John Collins
 *   Eliminate subtask bids, move to new log scheme
 * 
 * Thu Oct  5 22:34:52 2000  John Collins
 *   Initial Entry, split out from XSupplyMonster
 */

package edu.umn.magnet.client.sup;

import edu.umn.magnet.util.rvars.NormalVariable;
import edu.umn.magnet.util.rvars.UniformVariable;
import edu.umn.magnet.util.time.AgentTime;

import edu.umn.magnet.protocol.JobInfo;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.InvalidDataException;

import org.apache.log4j.Category;

import java.util.Date;
import java.util.Properties;

/**
 * <class description>
 * @author John Collins
 */
public class STBidGenerator
{
  // Static logging category
  private static Category log
    = Category.getInstance(STBidGenerator.class.getName());

  private Properties m_props = null;
  private SupplyMonster m_sm = null;
  private UniformVariable m_uVariable = null;
  private NormalVariable m_normalVariable = null;

  /**
   * Just a default constructor.  The static factory does the work.
   */
  public STBidGenerator ()
  {
    super();
  }

  protected void init (Properties props, SupplyMonster sm)
  {
    m_props = props;
    m_sm = sm;
    m_uVariable = sm.getRVF().createUniformVariable(0,1);
    m_normalVariable = sm.getRVF().createNormalVariable(0,1);
  }

  TaskSpec generateBid (SubtaskRequest task,
			Date minEarlyStart,
			Date maxLateFinish)
  {
    log.debug("generateBid for " + task.getId());
    // generate the duration
    long duration = chooseDuration(task);

    // choose es and lf and validate them
    Date earlyStart = AgentTime.getDate();
    Date lateFinish = AgentTime.getDate();
    if (! chooseEsAndLf(earlyStart, minEarlyStart,
			lateFinish, maxLateFinish,
			task, duration)) {
      log.debug("CSM es or lf not feasible");
      return null;
    }

    int price = choosePrice(task);
    log.debug("original price = " + price);

    // fill the TaskSpec
    TaskSpec spec = new TaskSpec();
    spec.setTaskId(task.getId());
    spec.setEarlyStart(earlyStart);
    spec.setLateFinish(lateFinish);
    spec.setDuration(duration);
    spec.setCost(price);
    return spec;
  }
  
  private long chooseDuration(SubtaskRequest task) {
    double expDur = task.getJobInfo().getExpectedDuration();
    double sigma = task.getJobInfo().getDurationSigma();
    float var = m_normalVariable.nextFloat();
    long duration  = (long)(expDur + sigma * var);
    if (duration < 0) duration = 0l;
    log.debug("STB:chooseDuration: " + task.getId()
	      + ": " + duration);
    return duration;
  }

  private boolean chooseEsAndLf(Date earlyStart, Date minEarlyStart,
				Date lateFinish, Date maxLateFinish,
				SubtaskRequest task,
				long duration) 
  {
    // First, we need to find out whether resources are available...
    long minES = minEarlyStart.getTime();
    long maxLF = maxLateFinish.getTime();
    long totalWindow = maxLF - minES;  
    long startWindow = totalWindow - duration;
    double avail = task.getJobInfo().getResourceAvailability();
    log.debug("STB:chooseEsLf: task " + task.getId()
	      + ", tw=" + totalWindow + ", dur=" + duration);
    if ((totalWindow >= duration)
	&& ((1.0 - Math.pow((1.0 - avail),
			    ((double)totalWindow / (double)duration)))
	    >= m_uVariable.nextDouble())) {
      // Resource is available
      // calculate the expected time to start
      long expectedStart = ((long)(startWindow * m_uVariable.nextDouble()))
	  + minES;
      earlyStart.setTime(expectedStart);

      // late finish is chosen randomly between maxLF and expected
      // finish time
      long expectedFinish = expectedStart + duration;
      long lf = (((long)(m_uVariable.nextDouble()
			 * (maxLF - expectedFinish)))
		 + expectedFinish);
      lateFinish.setTime(lf);
      log.debug("STB:chooseEsLf: es = " + expectedStart
		+ ", lf = " + lf);
      return true;
    }
    else {
      // duration too long, or resource not available
      return false;
    }
  }

  private int choosePrice(SubtaskRequest task) 
  {
    JobInfo info = task.getJobInfo();
    log.debug("STB.choosePrice: " + info.getTaskType()
	      + ", expected price = " + info.getExpectedPrice()); 
    int result = (int) Math.round((m_normalVariable.nextDouble()
				   * info.getPriceSigma())
				  + info.getExpectedPrice());
    if (result < 0) {
      log.warn("STB: negative price chosen: " + result);
      result = 1;
    }
    return result;
  }
    
  private int chooseFlexibilityDiscount(long duration,
					int price,
					SubtaskRequest task) 
  {
    long cfbEarlyStart = task.getEarliestTimeToStart().getTime();
    long cfbLateFinish = task.getDeadline().getTime();
    double ratio = ((double) (cfbLateFinish - cfbEarlyStart))/duration;
    
    double discountPercent = ratio - 1.0;

    // can't have discount percent > 100%
    if (discountPercent > 1.0)
      discountPercent = 1.0;
    
    // it's hard to see how this could happen
    if (discountPercent < 0.0) {
      log.warn("STB:chooseFlexibilityDiscount(): discountPercent < 0");
      return 0;
    }
    else {
      int maxDiscount = (int) (.20 * price);
      return (int) (maxDiscount * discountPercent);
    }
  }
    
  private int choosePremium(long supplierWindow,
			    long duration,
			    int price) 
  {
    log.debug("Premium: supplierWindow = " + supplierWindow);
    log.debug("Premium: duration = " + duration);
    double ratio = ((double)supplierWindow) / (double)duration;
    double premiumPercent = ratio - 1.0;

    if (premiumPercent > 1.0)
      premiumPercent = 1.0;

    if (premiumPercent < 0.0) {
      log.warn("STB:choosePremium(): premiumPercent negative"
	       + premiumPercent);
      return 0;
    }
    else {
      int maxPremium = (int) (.20 * price);
      return (int) (maxPremium * premiumPercent);
    }
  }

  /**
   * Static instance creation method - can be extended to generate
   * subtypes as determined by properties.
   */
  static STBidGenerator create (Properties props, SupplyMonster sm)
  {
    STBidGenerator result = new STBidGenerator();
    result.init(props, sm);
    return result;
  }
}
