/*------------------------------------------------------------------------
 * MAGNET supplier simulation components
 * $Id: SupplyMonster.java,v 1.6.2.1 2002/05/05 19:41:55 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Sat Dec  8 12:48:03 2001  John Collins
 *   Move GuiFrame into this package.
 * 
 * Sat Dec 16 18:00:46 2000  John Collins
 *   Eliminate bid breakdowns
 * 
 * Thu Oct  5 17:21:55 2000  John Collins
 *   Move to the client.sup package, split out the actual bid
 *   generation, implement alternate bid generation strategies.  This
 *   class is now the abstract uber SupplyMonster, mainly contributing
 *   a create method that uses the properties to decide exactly what
 *   to create.
 * 
 * Tue Nov 23 12:13:57 1999  John Collins
 *   Implement run-run variability in the link-following probability
 * 
 * Sat Nov 20 18:40:13 1999  John E. Collins
 *   Update for new JobInfo format
 * 02Jun2001 JCollins - do initGuiMaybe() to load from props strings.
 */

package edu.umn.magnet.client.sup;

import edu.umn.magnet.util.rvars.RandomVariableFactory;
import edu.umn.magnet.util.rvars.RVFManager;

import edu.umn.magnet.protocol.cust.RFQ;
import edu.umn.magnet.protocol.cust.Bid;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.InvalidDataException;

import org.apache.log4j.Category;

import java.util.Collection;
import java.util.Iterator;
import java.util.Properties;

/**
 *  The SupplyMonster provides a way of imitating a set of suppliers in a 
 *  marketplace.
 *  
 *  Before gimmeTheBids() is called the following must be called
 *  to init the SupplyMonster:
 *  
 *  setRFQ( RFQ cfb)
 *  
 *  @author Erik S. Steinmetz, John Collins
 */
public abstract class SupplyMonster 
{
  // Set up log category
  static Category log = Category.getInstance(SupplyMonster.class.getName());

  private Properties m_props;
  private int m_idCounter = 0;
  private RFQ m_callForBids = null;
  private TaskPlan m_plan = null;
  private RandomVariableFactory m_randomVariableFactory = null;
  private BidGenerator m_bidGenerator = null;
  private GuiFrame m_frame = null;

  // ============ Constructors ================
  /**
   * Default constructor.  Subclasses that implement this must call
   * super();
   */  
  public SupplyMonster ()
  {
    super();
  }

  /**
   * Initializes the instance.  All subclasses must implement this
   * method, and each implementation must call super.init(props).  It
   * is expected that the 
   */
  protected void init (Properties props)
  {
    m_props = props;
    m_randomVariableFactory = RVFManager.getRVF("BidSeed");
    initGuiMaybe();
  }

  /**
   * Initializes the GUI if called for by the property settings
   */
  protected void initGuiMaybe ()
  {
    String prop = m_props.getProperty("Sup.UseGUI", "false");
    if (prop.equals("true")) {
      m_frame = GuiFrame.getInstance();
    }
  }

  // ================ Bid Generation =================

  /**
   * Supplies a Collection of bids.  The number of bids is determined
   * by a configuration property.  Subclasses must override this
   * method, including a call to super.gimmeTheBids(cfb).
   */
  public Collection gimmeTheBids (RFQ cfb)
  {
    setRFQ(cfb);
    m_plan = cfb.getTaskPlan();
    return null;
  }

  // ================ Access to instance properties ==============
  /**
   * Gets the call for bids that is being used to generate
   * bids.
   * @return current call for bids
   */
  RFQ getRFQ () 
  { 
    return (m_callForBids);
  }

  /**
   * Sets the RFQ instance, possibly displaying it in the GUI
   */
  void setRFQ (RFQ cfb)
  {
    this.m_callForBids = cfb;
    if (this.m_frame != null)
      this.m_frame.setRFQId(cfb.getId());
  }

  /**
   * Gets the task plan contained in the CFB
   */
  protected TaskPlan getPlan ()
  {
    return m_plan;
  }

  /**
   * Returns a RandomVariableFactory for bid generation.
   */
  RandomVariableFactory getRVF ()
  {
    return m_randomVariableFactory;
  }

  /**
   * Gets the specified bid generator
   */
  protected BidGenerator getBidGenerator ()
  {
    return m_bidGenerator;
  }

  /**
   * Sets the bid generator.  This is private so only the create
   * method can access it.
   */
  private void setBidGenerator (BidGenerator bg)
  {
    m_bidGenerator = bg;
  }

  // ================== Record-keeping ===================
  /*
   * Determine the following statistics in reponse to
   * each call for bids:
   *   o number of bids
   *   o subtask/bid
   *   o ave subtask cost
   *   o ave subtask duration
   *   o ave bid cost
   *   o ave bid duration
   */
  protected void statistics(Collection bids) 
  {
    log.debug("SM: CFB " + 
	      m_callForBids.getId()+
	      ": bid count = " + bids.size());

    Bid bid;
    int totalCostOfBids = 0;
    long totalDurationOfBids = 0;
    int totalSubtaskCount = 0;

    for (Iterator bi = bids.iterator();
	 bi.hasNext(); ) {
      bid = (Bid) bi.next();
      totalSubtaskCount += bid.size();
    }
    
    if (bids.size() != 0) {
      log.info("SM:" + m_callForBids.getId()
	       + " bids = " + bids.size()
	       + ", stBids = " + totalSubtaskCount
	       + ", tasks = " + m_plan.size()
	       + ", stBids/task = " + totalSubtaskCount/m_plan.size());
      log.debug("SM: ave bid size = " +
		((double) totalSubtaskCount)/bids.size());
    }

    if (m_frame != null)
      m_frame.setBidsSubmitted(bids.size());
  }

  String createUniqueId (String prefix) 
  {
    return (prefix + m_idCounter++);
  }

  /**
   * Returns the named property as a double.
   */
  protected double getPropertyDouble (String key, double df)
  {
    String prop = m_props.getProperty(key);
    if(prop == null) {
      return df;
    }
    double result = df;
    try {
      result = Double.valueOf(prop).doubleValue();
      log.debug("SupplyMonster: " + key + " = " + result);
    }
    catch (NumberFormatException nfe) {
      log.error("SM:Bad property spec " + prop
                + " for " + key +".  Using " + result);
    }
    return result;
  }

  // ==================== Factory ==================

  /**
   * This is the basic factory method for the whole bid monster
   * component.  We build a SupplyMonster, a BidGenerator, and a
   * STBidGenerator, in accordance with configuration specifications.
   */
  public static SupplyMonster create (Properties props)
  {
    String smType = props.getProperty("Sup.Method", null);
    SupplyMonster result = null;
    log.info("prop: SupplyMonster type = " + smType);
    if (smType == null) {
      log.warn("SupplyMonster type not specified. Using default ClassicSM");
      result = new ClassicSM();
    }
    else {
      try {
	String classname = "edu.umn.magnet.client.sup." + smType;
	Class target= Class.forName(classname);
	result = (SupplyMonster)target.newInstance();
      }
      catch (ClassNotFoundException cnf) {
	log.error("specified SupplyMonster not found.  Using default");
	result = new ClassicSM();
      }
      catch (InstantiationException ie) {
	log.error("specified SupplyMonster can't be instantiated.  Using default");
	result = new ClassicSM();
      }
      catch (IllegalAccessException iae) {
	log.error("SupplyMonster.create: Illegal access exception.  Using default");
	result = new ClassicSM();
      }
    }
    result.init(props);
    result.setBidGenerator(BidGenerator.create(props, result));
    return result;
  }
}
