/* --------------------------------------------------------------
 * Magnet Protocol elements
 * $Id: Bid.java,v 1.7.2.1 2002/05/05 19:41:55 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 * --------------------------------------------------------------
 * Revision History:
 * 
 * Mon Apr  9 11:43:06 2001  John Collins
 *   Use AgentTime rather than the system as the source of time
 *   information.
 * 
 * Tue Jan  2 21:19:04 2001  John Collins
 *   Correct handling of duration on non-milestone tasks, add early
 *   finish accessor.
 * 
 * Mon Dec 11 19:10:37 2000  John Collins
 *   Collapse with Generic Bid as part of the elimination of bid
 *   breakdowns.  This route was chosen, rather than just re-purposing
 *   SubtaskBid, to better support the by-task indexing of temporal
 *   data in this class.
 * 
 * Tue Oct  3 15:00:26 2000  John Collins
 *   Add findSubtaskBid.
 * 
 * Sun Dec  5 09:27:04 1999  John E. Collins
 *   Major cleanup, make all subtask bids refer to the bid, eliminate
 *   properties that collide with inherited properties from
 *   GenericBid.
 *   Get rid of serialization signature.  It's the wrong way to solve
 *   that particular problem.
 * 
 * Sat Jan 9 12:45:30 1999 Rashmi Sundareswara
 *   Methods setDiscount and getDiscount added
 */
 
package edu.umn.magnet.protocol.cust;

import edu.umn.magnet.protocol.ProtocolElement;
import edu.umn.magnet.protocol.InvalidDataException;

import edu.umn.magnet.util.time.AgentTime;

import org.apache.log4j.Category;

import java.io.Serializable;
import java.util.Date;
import java.util.Set;
import java.util.HashMap;
import java.util.Iterator;

/**
 * This class represents a bid that is offered by a supplier
 * during a <code>Session</code>.  A Bid may specify one or more
 * tasks, an overall price, an overall duration, early and late start
 * times, and optionally, durations and early and late start times for
 * the individual tasks.  A bid also specifies a latest-acceptance
 * time.
 * <p>
 * For each task, the bid may specify early and late start times and
 * duration.  Alternatively, multiple tasks may be "collapsed" into a
 * single work unit, with timing data specified only for the group as
 * a whole.  When this is done, the implication is that any precedence
 * relations in the TaskPlan, that must be satisfied by activities
 * external to this Bid, must be completed before the whole task set
 * (when a predecessor task is external), or after the set is
 * complete (for an external successor task). </p>
 * <p>
 * Here are the rules:
 * <ul>
 * <li>
 *   Tasks are collapsed by grouping them together.  You can do this
 *   in one of two ways:
 *   <ol>
 *   <li>
 *     If all tasks in the bid are a single set, then you only need to
 *     specify the timing data at the level of the bid.  In this case,
 *     all tasks must be added using the <pre>
 *    addTask(taskId)
 *     </pre> method.
 *   <li>
 *     Otherwise, you must specify one or more "milestone" tasks (see
 *     below).  Other tasks may then be grouped with one of the
 *     milestone tasks by using the <pre>
 *    addTask(taskId, milestoneId)
 *     </pre> method.  When you do this, all the non-milestone tasks
 *     will take their start and duration data from their respective
 *     milestone tasks.
 *   </ol>
 * <li>
 *   There are two kinds of milestone tasks.  The first specifies
 *   early start, late start, and duration, and is added using the
 *   <pre>
 *    addTask(taskId, es, ls, dur)
 *   </pre> method.  
 * <li>
 *   The second type of milestone task specifies only duration, which
 *   must be less than or equal to the duration of the bid.  If
 *   you add a task using <pre>
 *    addTask(taskId, duration)
 *   </pre> then the task is assumed to start when the bid starts,
 *   and if you add a task using <pre>
 *    addTask(taskId, milestoneId, duration)
 *   </pre> then the task is assumed to start immediately after the
 *   task or group of tasks specified by <tt>milestoneId</tt>.
 * <li>
 *   No milestone task can end later than the bid late start time plus
 *   the bid duration. 
 * <li>
 *   No milestone task can have early start or late start times
 *   earlier than the respective time values for the bid as a whole.
 * </ul>
 * @author John Collins
 */
public class Bid
  extends ProtocolElement
  implements Serializable
{
  static Category log = Category.getInstance(Bid.class.getName());
  
  // private data members
  private Date m_earlyStart = AgentTime.getDate();
  private Date m_lateStart = AgentTime.getDate();
  private long m_duration = 0L;
  private long  m_cost = 0L;
  private String m_callForBidsId = null;
  private Date m_latestAccept = AgentTime.getDate();
  // taskData maps task ID strings to TaskData instances
  private HashMap taskData = new HashMap();
  private TaskDataRef context = new TaskDataRef()
      {
	public TaskData getTD (String taskId)
	{
	  return (TaskData)taskData.get(taskId);
	}
	public Date getBidEarlyStart ()
	{
	  return m_earlyStart;
	}
      };

  /**
   * Default Constructor, inits everything to default
   */
  public Bid()
  {
    super();
  }

  /**
   * Returns the Id of the <code>RFQ</code> that this
   * <code>Bid</code> is associated with
   *
   * @return a <code>String</code> which is the ID for this <code>Bid</code>'s
   *		<code>RFQ</code>
   */
  public String getRFQId() {
    return m_callForBidsId;
  }
  
  /**
   * Sets the ID of the associated <tt>RFQ</tt>
   * 
   * @param newRFQId gives this <code>Bid</code> a new
   * 		<code>RFQ</code> to be associated with
   */
  public void setRFQId(String newRFQId) {
    m_callForBidsId = newRFQId;
  }

  // -------- latestAccept --------  
  /**
   * Returns the latest-acceptance deadline
   * @return	a <code>Date</code> which is the earliest that 
   */
  public Date getLatestAccept ()
  {
    return m_latestAccept;
  }
  
  /**
   * Sets the latest time this Bid may be accepted by the customer.
   * If this attribute is left null, (the default) then the Bid is
   * valid forever.
   */
  public void setLatestAccept (Date newLatestAccept)
  {
    m_latestAccept = newLatestAccept;
  }

  // -------- task list --------
  /**
   * Returns the TaskData instance, if any, for the given task ID.
   */
  protected TaskData getTD (String taskId)
  {
    return context.getTD(taskId);
  }

  /**
   * Adds a task to this bid without specifying any timing data.
   * @exception InvalidDataException if the task is already
   *   part of the bid.
   */
  public void addTask (String taskId)
    throws InvalidDataException
  {
    if (getTD(taskId) == null)
      taskData.put(taskId, new TaskData(taskId));
    else
      throw new InvalidDataException("task already in bid");
  }

  /**
   * Adds a task to one of the task groups in this bid, without
   * specifying any timing data.
   * @exception InvalidDataException if the task is already
   *   part of the bid, or if the specified milestone task is not part
   *   of the bid.
   */
  public void addTask (String taskId, String milestoneId)
    throws InvalidDataException
  {
    if (getTD(taskId) == null) {
      TaskData mstone = getTD(milestoneId);
      if (mstone == null)
	throw new InvalidDataException("invalid milestone spec");
      taskData.put(taskId, new TaskData(taskId, milestoneId));
    }
    else
      throw new InvalidDataException("task already in bid");
  }

  /**
   * Adds a milestone task to this bid, specifying all timing
   * parameters. 
   * @exception InvalidDataException if the task is already
   *   part of the bid.
   */
  public void addTask (String taskId,
		       Date earlyStart,
		       Date lateStart,
		       long duration)
    throws InvalidDataException
  {
    if (getTD(taskId) == null) {
      taskData.put(taskId,
		   new TaskData(taskId, earlyStart, lateStart, duration));
    }
    else {
      log.error("Duplicate task " + taskId);
      throw new InvalidDataException("task already in bid");
    }
  }

  /**
   * Adds a task to this bid, specifying only duration.  This task is
   * assumed to begin when work on the bid as a whole begins.
   * @exception InvalidDataException if the task is already
   *   part of the bid.
   */
  public void addTask (String taskId, long duration)
    throws InvalidDataException
  {
    if (getTD(taskId) == null)
      taskData.put(taskId, new TaskData(taskId, duration));
    else
      throw new InvalidDataException("task already in bid");
  }

  /**
   * Adds a task to this bid, specifying duration and a preceding
   * milestone.  The milestone task must already be in the bid.
   * @exception InvalidDataException if the task is already
   *   part of the bid, or if the specified milestone is not part of
   *   the bid.
   */
  public void addTask (String taskId,
		       String milestoneId,
		       long duration)
    throws InvalidDataException
  {
    if (getTD(taskId) == null) {
      TaskData mstone = getTD(milestoneId);
      if (mstone == null)
	throw new InvalidDataException("invalid milestone spec");
      taskData.put(taskId, new TaskData(taskId, duration, milestoneId));
    }
    else
      throw new InvalidDataException("task already in bid");
  }

  /**
   * Returns the set of task IDs for this bid
   */
  public Set getTaskSet ()
  {
    return taskData.keySet();
  }

  /**
   * Returns the number of tasks in this bid
   */
  public int size ()
  {
    return getTaskSet().size();
  }

  /**
   * Returns true just in case the specified task is already part of
   * this bid.
   */
  public boolean hasTask (String taskId)
  {
    if (getTD(taskId) != null)
      return true;
    else
      return false;
  }

  /**
   * Returns true just in case the specified task is a milestone task.
   */
  public boolean isMilestone (String taskId)
  {
    TaskData ms = getTD(taskId);
    return ((ms != null)
	    && ms.isMilestone());
  }

  /**
   * Returns the Id of the milestone task, if any, associated with
   * this task.
   */
  public String getMilestone (String succeedingTask)
  {
    TaskData succ = getTD(succeedingTask);
    if (succ != null)
      return succ.getPredecessor();
    else
      return null;
  }

  // -------- cost --------
  /**
   * Tells the caller the <code>cost</code> that has been bid
   * @return the <code>Cost</code> that is Bid by the
   *         <code>Supplier</code>
   */
  public long getCost () {
    return m_cost;
  }
  
  /**
   * Assigns this <code>Bid</code> a different value
   * @param newCost is the amount that the Supplier is bidding
   */
  public void setCost(long newCost) {
    m_cost = newCost;
  }

  // -------- earlyStart --------
  /**
   * Returns the earliest <code>Date</code> at which the tasks
   * specified in this Bid can be started.
   */
  public Date getEarlyStart ()
  {
    return m_earlyStart;
  }

  /**
   * Returns the early start date for a particluar task.  Result is
   * null if the task is not part of this bid, or if the early start
   * date can't be determined (in which case this bid is invalid).
   */
  public Date getEarlyStart (String taskId)
  {
    TaskData data = getTD(taskId);
    if (data != null)
      return data.getEarlyStart(context);
    else
      return null;
  }

  /**
   * Assigns a new early start date for this bid.
   */
  public void setEarlyStart (Date newEarlyStart)
  {
    m_earlyStart = newEarlyStart;
  }
 
  /**
   * Returns to the user the <code>Date</code> at which the job will be
   * started at.
   * @return  a <code>Date</code> which tells the caller when the
   *   task offered will be started at
   * @deprecated use getEarlyStart instead.
   */
  public Date getStartBy()
  {
    return m_earlyStart;
  }
  
  /**
   * Assigns a new value for when the <code>Bid</code> will be started
   * @param newStartBy specifies the new <code>Date</code> for the
   *   task to be started at
   * @deprecated use setEarlyStart instead
   */
  public void setStartBy(Date newEarlyStart)
  {
    m_earlyStart = newEarlyStart;
  }

  // -------- lateStart --------
  /**
   * Returns the latest time by which work on this bid must be
   * started. 
   */
  public Date getLateStart ()
  {
    return m_lateStart;
  }

  /**
   * Returns the late start date for a particluar task.  Result is
   * null if the task is not part of this bid, or if the late start
   * date can't be determined (in which case this bid is invalid).
   */
  public Date getLateStart (String taskId)
  {
    TaskData data = getTD(taskId);
    if (data != null)
      return data.getLateStart(context);
    else
      return null;
  }

  /**
   * Sets the late start date.
   */
  public void setLateStart (Date newValue)
  {
    m_lateStart = newValue;
  }

  // -------- earlyFinish --------
  /**
   * Returns the computed early finish time for the bid as a whole.
   */
  public Date getEarlyFinish ()
  {
    return new Date(m_earlyStart.getTime() + m_duration);
  }

  /**
   * Returns the computed early finish time for a particular task in
   * this bid.
   */
  public Date getEarlyFinish (String taskId)
  {
    TaskData data = getTD(taskId);
    if (data != null)
      return data.getEarlyFinish(context);
    else
      return null;
  }

  // -------- lateFinish --------
  /**
   * Returns the computed late finish time for the bid as a whole.
   */
  public Date getLateFinish ()
  {
    return new Date(m_lateStart.getTime() + m_duration);
  }

  /**
   * Returns the computed late finish time for a particular task in
   * this bid.
   */
  public Date getLateFinish (String taskId)
  {
    TaskData data = getTD(taskId);
    if (data != null)
      return data.getLateFinish(context);
    else
      return null;
  }

  /**
   * Returns a <code>Date</code> which will tell the user at what date the
   *  task associated with the RFQ will be completed
   * @return a <code>Date</code> which tells the user at what
   *       <code>Date</code> the task associated with the RFQ
   *       will be completed
   * @deprecated
   */
  public Date getEndBy()
  {
    return new Date(m_lateStart.getTime() + m_duration);
  }
  
  /**
   * Assigns a new amount of time for the Task associated with this
   *  <code>Bid</code> to be completed
   * @param newEndBy specifies the amount of time it will take
   *     for the task with the associated RFQ will take to complete
   * @deprecated
   */
  public void setEndBy(Date newEndBy)
  {
    m_lateStart = new Date(newEndBy.getTime() - m_duration);
  }

  // -------- duration --------
  /**
   * Returns the planned duration for the entire bid
   */
  public long getDuration ()
  {
    return m_duration;
  }

  /**
   * Gets the duration for a particular task in this bid.  If the task
   * is not in the bid, the value is 0.
   */
  public long getDuration (String taskId)
  {
    TaskData data = getTD(taskId);
    if (data == null)
      return 0L;
    else
      return data.getDuration(context);
  }

  /**
   * Sets the duration for the entire bid
   */
  public void setDuration (long newValue)
  {
    m_duration = newValue;
  }

  /**
   * Returns the amount of time to complete the task specified in the
   *  corresponding <code>RFQ</code>
   * @return  a <code>Date</code> which tells the caller what amount of time
   *   it will take to get the task associated with the bid done
   * @deprecated use getDuration instead.
   */
  public  long getTimeToDo()
  {
    return m_duration;
  }
  
  /**
   * Assigns a new amount of time for the task associated with the
   *  callForBids to be completed
   * @param newduration tells this <code>Bid</code> that it now takes the
   *   specified amount of time to complete the task associated with the
   *   callForBidsId
   * @deprecated use setDuration instead
   */
  public void setTimeToDo(long newduration)
  {
    m_duration = newduration;
  }

  // -------- validity testing --------
  /**
   * Returns true just in case this bid is internally consistent.  A
   * valid bid meets the following tests:
   * <ul>
   * <li>There is at least one task specified.
   * <li>Cost, early start, late start, and duration are defined for the bid
   * as a whole.
   * <li>The RFQ ID is specified.
   * <li>Late start is later than early start.
   * <li>All the internal TaskData instances are valid.
   * <li>No task specifies an early start earlier than the bid early
   * start, nor a late start earlier than the bid late start.
   * <li>No task has a combination of late start and duration that
   * would allow it to finish later than the bid late finish time.
   * </ul>
   */
  public boolean isValid ()
  {
    return true;
  }

  /**
   * Returns a BidExaminer for this bid.  You have to give it the task
   * plan so it can follow precedence relations.
   */
  public BidExaminer createBidExaminer ()
  {
    return new BidExaminerImpl(this);
  }

  /**
   * Returns a readable representation of this bid
   */
  public String toString()
  {
    StringBuffer rs = new StringBuffer();
    rs.append("Bid ");
    rs.append(getId());
    rs.append(": cost=");
    rs.append(m_cost);
    rs.append(", tasks=(");
    String sep = "";
    for (Iterator tds = taskData.keySet().iterator();
	 tds.hasNext(); ) {
      rs.append(sep);
      rs.append((String)tds.next());
      //rs.append(((TaskData)tds.next()).taskId);
      sep = " ";
    }
    rs.append(")");
    return rs.toString();
  }

  /**
   * Prints full details to a log category at info level
   */
  public void logVerbose (Category log)
  {
    log.info("Bid " + getId() + ": cost=" + m_cost
	     + ", td(" + m_earlyStart + ", " + m_lateStart
	     + ", " + m_duration + ")");
    for (Iterator tds = taskData.values().iterator();
	 tds.hasNext(); ) {
      TaskData td = (TaskData)tds.next();
      log.info("..." + td.taskId
	       + ((td.predecessor != null)?(":" + td.predecessor):"")
	       + " td(" + td.earlyStart + ", " + td.lateStart
	       + ", " + td.duration + ")");
    }
  }

  // ---------------------- TaskData -----------------------------
  /**
   * Reference interface to allow TaskData instances to refer to each
   * other in different contexts.
   */
  protected interface TaskDataRef extends Serializable
  {
    /**
     * Returns the TaskData instance, if any, from the current context
     * with the given task ID.
     */
    public TaskData getTD (String taskId);
    public Date getBidEarlyStart ();
  }

  /**
   * Inner class for managing per-task data.  A task can be a type-1
   * milestone with independent start and duration data, a type-2
   * milestone with specified duration only (directly follows
   * another task), or a "hidden" task, taking timing data from its
   * preceding task, or from the bid itself if no preceding task is
   * specified.
   */
  protected class TaskData implements Serializable
  {
    String taskId;
    Date earlyStart = null;
    Date lateStart = null;
    Long duration = null;
    String predecessor;

    // -------- constructors --------
    /**
     * Specifies a type-1 milestone task with its own timing data
     */
    TaskData (String taskId, Date earlyStart, Date lateStart, long duration)
    {
      super();
      this.taskId = taskId;
      this.duration = new Long(duration);
      this.earlyStart = earlyStart;
      this.lateStart = lateStart;
    }

    /**
     * Specifies a type-2 milestone task that directly follows a
     * preceeding task.
     */
    TaskData (String taskId, long duration, String precedingTask)
    {
      super();
      this.taskId = taskId;
      this.duration = new Long(duration);
      this.predecessor = precedingTask;
    }

    /**
     * Specifies a type-2 milestone task that starts when the bid
     * starts.
     */
    TaskData (String taskId, long duration)
    {
      super();
      this.taskId = taskId;
      this.duration = new Long(duration);
    }

    /**
     * This constructor only identifies a task, without specifying any
     * timing data.  The early start and late finish times will be
     * taken from its milestone task.
     */
    TaskData (String taskId, String milestone)
    {
      super();
      this.taskId = taskId;
      this.predecessor = milestone;
    }

    /**
     * This constructor only identifies a task, without specifying any
     * timing data.  The early start and late finish times will be
     * taken from the bid itself.
     */
    TaskData (String taskId)
    {
      super();
      this.taskId = taskId;
    }

    /**
     * This constructor simply makes a copy of a TaskData instance
     */
    TaskData (TaskData original)
    {
      super();
      this.taskId = original.taskId;
      this.earlyStart = original.earlyStart;
      this.lateStart = original.lateStart;
      this.duration = original.duration;
      this.predecessor = original.predecessor;
    }

    // -------- early start, early finish --------
    /**
     * Returns the specified or derived early start time for this
     * task.  If the early start has been set, return that.
     * Otherwise, return the late finish time of the
     * predecessor task.  Failing both tests, return the early start
     * from the bid.
     */ 
    Date getEarlyStart (TaskDataRef context)
    {
      if (earlyStart != null)
	return earlyStart;
      else if (predecessor != null)
	if (duration != null)
	  return context.getTD(predecessor).getEarlyFinish(context);
	else
	  return context.getTD(predecessor).getEarlyStart(context);
      else
	return context.getBidEarlyStart();
    }

    /**
     * Sets the early start date for a task
     */
    void setEarlyStart (Date newValue)
    {
      this.earlyStart = newValue;
    }
    
    /**
     * Returns the early finish time, which is the early start time
     * plus the duration.
     */
    private Date getEarlyFinish (TaskDataRef context)
    {
      if (duration != null)
	return new Date(getEarlyStart(context).getTime()
			+ duration.longValue());
      else if (predecessor != null)
	return context.getTD(predecessor).getEarlyFinish(context);
      else 
	return new Date(context.getBidEarlyStart().getTime()
			+ m_duration);
    }

    // -------- late start --------
    /**
     * Returns the specified or derived late start time for this
     * task.  If the late start has been set, return that.
     * Otherwise, return the maximum of the late finish times of the
     * predecessor tasks.  Failing both tests, return null.
     */ 
    Date getLateStart (TaskDataRef context)
    {
      if (lateStart != null)
	return lateStart;
      else if (predecessor != null)
	if (duration != null)
	  return context.getTD(predecessor).getLateFinish(context);
	else
	  return context.getTD(predecessor).getLateStart(context);
      else
	return m_lateStart;
    }

    /**
     * Sets the late start date for this task
     */
    void setLateStart (Date newValue)
    {
      this.lateStart = newValue;
    }
    
    /**
     * Returns the late finish time, which is the late start time
     * plus the duration.
     */
    private Date getLateFinish (TaskDataRef context)
    {
      if (duration != null)
	return new Date(getLateStart(context).getTime() + duration.longValue());
      else if (predecessor != null)
	return context.getTD(predecessor).getLateFinish(context);
      else
	return new Date(m_lateStart.getTime() + m_duration);
    }

    // -------- duration --------
    /**
     * Returns the duration for this task
     */
    long getDuration (TaskDataRef context)
    {
      if (duration != null)
	return duration.longValue();
      else if (predecessor != null)
	return context.getTD(predecessor).getDuration(context);
      else
	return m_duration;
    }

    /**
     * Sets the duration for this task
     */
    void setDuration (long newValue)
    {
      this.duration = new Long(newValue);
    }

    // -------- predecessor --------
    /**
     * Milestone test
     */
    boolean isMilestone ()
    {
      return ((duration != null)
	      || (earlyStart != null));
    }

    /**
     * Milestone access.  Note that this is a more complex test than
     * the getPredecessor method below.
     */
    String getMilestone ()
    {
      if (duration != null)
	return this.taskId;
      else if (earlyStart != null)
	// this case isn't supposed to happen...
	return this.taskId;
      else
	return predecessor;
    }

    /**
     * Gets the ID of the preceding task
     */
    String getPredecessor ()
    {
      return predecessor;
    }
  }

  // -------- Bid Examiner Implementation --------

  /**
   * Implementation of the BidExaminer interface.  It lives here to
   * allow access to internal bid data.
   * <p>
   * <strong>Design Note</strong>
   * This implementation is intended to allow the agent to examine the
   * impact on bid data when the task start and finish timing data is
   * externally constrained.  This is necessary because of the opacity
   * of data about individual tasks within the bid (or alternatively,
   * to preserve that opacity).</p>
   * <p>
   * The basic design approach is to use the logic built into the
   * TaskData instances to propagate start and finish time data from
   * one task to another.  This is necessary for two reasons: (1) the
   * bid may contain precedence relations not contained in the
   * TaskPlan, as represented by milestone relations, and (2)
   * changing the start or finish time for one task can change timing
   * data for other tasks, because of milestone relations.  When the
   * agent wishes to examine the impact of changing the timing data
   * for some task, we create a copy of the TaskData instance for that
   * task and store a copy within the examiner.  We then use that one
   * instead of the one in Bid when we do timing calculations.</p>
   * <p>
   * The real implementation problem is that the TaskData instances
   * contain direct references to each other.  This is useful in the
   * implementation of Bid, but less helpful in this context.  I see
   * three alternatives (jec, 4 Jan 2001):
   * <ol>
   * <li>
   *   Make a local copy of the Bid's taskData map, and copy each of
   *   the TaskData instances as well.  Then go through and update the
   *   pointers in the new copy.  This approach seems less than ideal
   *   for performance reasons.  We expect these things to be used
   *   frequently during bid evaluation.  As a last resort, we could
   *   implement a resource pool of re-usable TaskData instances.
   * <li>
   *   Get rid of the direct references in TaskData.  Have the methods
   *   on TaskData resolve references through a map.  This sounds
   *   good, but I don't see any easy way to let the TaskData instances
   *   know which map to use.  If this were Lisp, we would use a
   *   special variable for this (dynamic scope).  The only thing in
   *   Java that comes close is thread-local variables.  We would have
   *   to use a thread-local map of Bid -> TDmap pairs.
   * <li>
   *   Just replicate (and map locally the changed TaskData instances,
   *   and leave the rest in the Bid's map.  This seems like the
   *   easiest approach, except that we will have to replicate all the
   *   TaskData logic that uses the pointers.
   * <li>
   *   Same as #2, except that TaskData instances resolve references
   *   among themselves using a Visitor that is pre-loaded with the
   *   correct map.  The visitor would have to be passed with every
   *   method call.  The local interface TaskDataRef is such a
   *   visitor.
   * </ol> 
   */
  private class BidExaminerImpl implements BidExaminer
  {
    private Bid bid = null;
    private TaskPlan plan;
    private Date bidEarlyStart = null;

    private HashMap taskData = new HashMap();
    private TaskDataRef context
      = new TaskDataRef()
      {
	public TaskData getTD (String taskId)
	{
	  TaskData result = (TaskData)taskData.get(taskId);
	  if (result != null)
	    return result;
	  else
	    return bid.getTD(taskId);
	}
	public Date getBidEarlyStart ()
	{
	  if (bidEarlyStart != null)
	    return bidEarlyStart;
	  else
	    return bid.getEarlyStart();
	}
      };

    BidExaminerImpl (Bid theBid)
    {
      super();
      this.bid = theBid;
    }

    /**
     * Returns a reference to the Bid upon which this thing is built.
     */
    public Bid getBid ()
    {
      return bid;
    }

    /**
     * Returns the TaskData instance for a given taskId.  If there's
     * one locally, we get that one, otherwise we get the one from the
     * Bid.  This is what we use in place of direct references between
     * TaskData instances.
     */
    private TaskData getTD (String taskId)
    {
      return context.getTD(taskId);
    }
    
    /**
     * Returns the early start date for the bid as a whole.  This
     * could be modified by setting the early start time for some task
     * that runs concurrently with the bid.
     */
    public Date getEarlyStart ()
    {
      if (bidEarlyStart == null)
	return bid.getEarlyStart();
      else
	return bidEarlyStart;
    }

    /**
     * Returns the early start date for a particular task.  This may be
     * different from the early start date specified by the bid in case
     * this or some other task has had its early start date changed.
     */
    public Date getEarlyStart (String taskId)
    {
      TaskData data = getTD(taskId);
      if (data != null)
	return data.getEarlyStart(context);
      else 
	return null;
    }

    /**
     * Sets the early start date for some task.  This may affect dates
     * for other tasks in the bid.  An InvalidDataException is thrown if
     * this change would cause an internal infeasibility in the bid.
     * This could happen, for example, if there were a subsequent task
     * (as determined by the TaskPlan) with a late start time that
     * couldn't be met.
     */
    public void setEarlyStart (String taskId, Date newValue)
      throws InvalidDataException
    {
      TaskData td = getTD(taskId);

      // There are several possible cases, assuming the taskId is
      // valid:
      // 1. If the new early start date is earlier than the existing
      //    early start, then we don't actually do anything.
      if (newValue.before(td.getEarlyStart(context)))
	return;

      // 2. If the new early start date is later than the specified or
      //    computed late start date, it's an error.
      if (newValue.after(td.getLateStart(context)))
	throw new InvalidDataException("early start later than late start");

      // 3. We already have a local copy of this task.  We just use
      //    that. 
      if (isLocal(td)) {
	td.setEarlyStart(newValue);
	return;
      }

      // 3. The task is a full milestone, with early and late start
      //    dates.  If the new early start date is later than the late
      //    start date, then this isn't allowed.  Otherwise we just
      //    clone the TaskData and set the new date.
      if (td.earlyStart != null) {
	td = ensureLocal(td);
	td.setEarlyStart(newValue);
	return;
      }

      // 4. The task is a duration-only milestone.  If the new early
      //    start time is later than the original computed early
      //    start, we clone it and turn it into a full milestone.
      //    Otherwise nothing happens (we can't move the early start
      //    up in time).
      if (td.duration != null) {
	Date originalLS = td.getLateStart(context);
	td = ensureLocal(td);
	td.earlyStart = newValue;
	td.lateStart = originalLS;
	return;
      }

      // 5. The task was not a milestone.  We assume that the
      //    semantics of the bid means that the task is grouped with
      //    its milestone, and that changes to the task are applied to
      //    its milestone.  This is done with a recursive call.
      String milestoneId = td.getMilestone();
      if (milestoneId != null) {
	setEarlyStart(milestoneId, newValue);
	return;
      }

      // 6. The task was not a milestone, and it is not associated
      // with any milestone task.  Its early start data comes from the
      // bid itself. 
      bidEarlyStart = newValue;
    }

    /**
     * Returns the early finish date for the bid as a whole.
     */
    public Date getEarlyFinish ()
    {
      if (bidEarlyStart == null)
	return bid.getEarlyFinish();
      else
	return new Date(bidEarlyStart.getTime()
			+ bid.getDuration());
    }

    /**
     * Returns the early finish date for a particular task, possibly
     * modified by setting the early start date for some task.
     */
    public Date getEarlyFinish (String taskId)
    {
      TaskData data = getTD(taskId);
      if (data != null)
	return data.getEarlyFinish(context);
      else 
	return null;
    }

    /**
     * Returns true just in case the specified td is a local copy.
     */
    private boolean isLocal (TaskData td)
    {
      return isLocal(td.taskId);
    }

    private boolean isLocal (String taskId)
    {
      return (taskData.get(taskId) != null);
    }

    /**
     * Ensures that we have a local copy of a td, making a copy if
     * necessary. 
     */
    private TaskData ensureLocal (TaskData original)
    {
      if (isLocal(original))
	return original;
      else {
	TaskData result = new TaskData(original);
	taskData.put(original.taskId, result);
	return result;
      }
    }
  }
}
