/*------------------------------------------------------------------------
 * MAGNET utility test components
 * $Id: BidExaminerTest.java,v 1.2.2.1 2002/05/05 19:41:57 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Sun Nov 19 13:51:10 2000  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.protocol.cust;

import edu.umn.magnet.protocol.InvalidDataException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;

import java.util.Date;

/**
 * This is a tester for edu.umn.magnet.client.Something
 * @author John Collins
 */
public class BidExaminerTest extends TestCase
{
  static final long BASE = 1000000000000L;
  static final long HOUR = 3600000L;

  private Bid theBid;
  
  public BidExaminerTest (String testName)
  {
    super(testName);
  }

  /**
   * You run the main method to exercise this thing.  There are no
   * command-line args.
   */
  public static void main (String[] args)
  {
    junit.textui.TestRunner.run(suite());
  }

  public static Test suite ()
  {
    return new TestSuite(BidExaminerTest.class);
  }

  // ---------- basic data access ----------
  
  // basic bid setup
  private void setup1 ()
  {
    theBid = new Bid();
    theBid.setRFQId("RFQ1");
    theBid.setLatestAccept(new Date(BASE));
    theBid.setCost(100000L);
    theBid.setEarlyStart(new Date(BASE + HOUR));
    theBid.setLateStart(new Date(BASE + HOUR * 3));
    theBid.setDuration(HOUR * 24);
  }

  public void testCreate ()
  {
    setup1();
    BidExaminer bx = theBid.createBidExaminer();
    assertEquals("Correct bid", theBid, bx.getBid());
    assertEquals("Bid ES",
		 BASE + HOUR, bx.getEarlyStart().getTime());
    assertEquals("Bid EF",
		 BASE + HOUR * 25, bx.getEarlyFinish().getTime());
  }

  // add two non-milestone tasks that depend on the bid
  private void setup2 ()
  {
    setup1();
    try {
      theBid.addTask("N1");
      theBid.addTask("N2");
    }
    catch (InvalidDataException ide) {
      fail("Failed to add task");
    }
  }

  public void testNonMilestone1 ()
  {
    setup2();
    BidExaminer bx = theBid.createBidExaminer();
    assertEquals("N1 ES",
		 BASE + HOUR, bx.getEarlyStart("N1").getTime());
    assertEquals("N1 EF",
		 BASE + HOUR * 25, bx.getEarlyFinish("N1").getTime());
    assertEquals("N2 ES",
		 BASE + HOUR, bx.getEarlyStart("N1").getTime());
    assertEquals("N2 EF",
		 BASE + HOUR * 25, bx.getEarlyFinish("N1").getTime());
  }

  // ---------- simple milestone tests ----------

  // single full milestone task
  private void msetup1 ()
  {
    setup1();
    try {
      theBid.addTask("M1",
		     new Date(BASE + HOUR * 2),
		     new Date(BASE + HOUR * 4),
		     HOUR * 6);
    }
    catch (InvalidDataException ide) {
      fail("Failed to add task");
    }
  }

  public void testMilestone1 ()
  {
    msetup1();
    BidExaminer bx = theBid.createBidExaminer();
    // check before making a change
    assertEquals("M1 ES",
		 BASE + HOUR * 2, bx.getEarlyStart("M1").getTime());
    assertEquals("M1 EF",
		 BASE + HOUR * 8, bx.getEarlyFinish("M1").getTime());
    // make a change and check again
    try {
      bx.setEarlyStart("M1", new Date(BASE + HOUR * 3));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("new M1 ES",
		 BASE + HOUR * 3, bx.getEarlyStart("M1").getTime());
    assertEquals("new M1 EF",
		 BASE + HOUR * 9, bx.getEarlyFinish("M1").getTime());
    // make sure the original data is still available through the bid
    assertEquals("bid M1 ES",
		 BASE + HOUR * 2, theBid.getEarlyStart("M1").getTime());
    assertEquals("bid M1 EF",
		 BASE + HOUR * 8, theBid.getEarlyFinish("M1").getTime());
  }

  // single duration-only milestone task
  private void msetup2 ()
  {
    setup1();
    try {
      theBid.addTask("D1", HOUR * 6);
    }
    catch (InvalidDataException ide) {
      fail("Failed to add task");
    }
  }

  public void testMilestone2 ()
  {
    msetup2();
    BidExaminer bx = theBid.createBidExaminer();
    // check before
    assertEquals("D1 ES",
		 BASE + HOUR, bx.getEarlyStart("D1").getTime());
    assertEquals("D1 EF",
		 BASE + HOUR * 7, bx.getEarlyFinish("D1").getTime());
    // change and check after
    try {
      bx.setEarlyStart("D1", new Date(BASE + HOUR * 2));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("new D1 ES",
		 BASE + HOUR * 2, bx.getEarlyStart("D1").getTime());
    assertEquals("new D1 EF",
		 BASE + HOUR * 8, bx.getEarlyFinish("D1").getTime());
    // check bid independence
    assertEquals("bid D1 ES",
		 BASE + HOUR, theBid.getEarlyStart("D1").getTime());
    assertEquals("bid D1 EF",
		 BASE + HOUR * 7, theBid.getEarlyFinish("D1").getTime());
  }

  // ---------- propagation among milestones ----------

  // single full milestone task, followed by two duration milestones
  private void msetup3 ()
  {
    setup1();
    try {
      theBid.addTask("M1",
		     new Date(BASE + HOUR * 2),
		     new Date(BASE + HOUR * 4),
		     HOUR * 6);
      theBid.addTask("D1", "M1", HOUR * 4);
      theBid.addTask("D2", "D1", HOUR * 5);
    }
    catch (InvalidDataException ide) {
      fail("Failed to add task");
    }
  }

  public void testMilestone3 ()
  {
    msetup3();
    BidExaminer bx = theBid.createBidExaminer();
    // check before
    assertEquals("D1 ES",
		 BASE + HOUR * 8, bx.getEarlyStart("D1").getTime());
    assertEquals("D1 EF",
		 BASE + HOUR * 12, bx.getEarlyFinish("D1").getTime());
    assertEquals("D2 ES",
		 BASE + HOUR * 12, bx.getEarlyStart("D2").getTime());
    assertEquals("D2 EF",
		 BASE + HOUR * 17, bx.getEarlyFinish("D2").getTime());
    // change and check after
    try {
      bx.setEarlyStart("M1", new Date(BASE + HOUR * 3));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("new D1 ES",
		 BASE + HOUR * 9, bx.getEarlyStart("D1").getTime());
    assertEquals("new D1 EF",
		 BASE + HOUR * 13, bx.getEarlyFinish("D1").getTime());
    assertEquals("new D2 ES",
		 BASE + HOUR * 13, bx.getEarlyStart("D2").getTime());
    assertEquals("new D2 EF",
		 BASE + HOUR * 18, bx.getEarlyFinish("D2").getTime());
    // check bid independence
    assertEquals("bid D1 ES",
		 BASE + HOUR * 8, theBid.getEarlyStart("D1").getTime());
    assertEquals("bid D1 EF",
		 BASE + HOUR * 12, theBid.getEarlyFinish("D1").getTime());
    assertEquals("bid D2 ES",
		 BASE + HOUR * 12, theBid.getEarlyStart("D2").getTime());
    assertEquals("bid D2 EF",
		 BASE + HOUR * 17, theBid.getEarlyFinish("D2").getTime());
  }

  public void testMilestone3a ()
  {
    msetup3();
    BidExaminer bx = theBid.createBidExaminer();
    // check before
    assertEquals("D1 ES",
		 BASE + HOUR * 8, bx.getEarlyStart("D1").getTime());
    assertEquals("D1 EF",
		 BASE + HOUR * 12, bx.getEarlyFinish("D1").getTime());
    assertEquals("D2 ES",
		 BASE + HOUR * 12, bx.getEarlyStart("D2").getTime());
    assertEquals("D2 EF",
		 BASE + HOUR * 17, bx.getEarlyFinish("D2").getTime());
    // change and check after
    try {
      bx.setEarlyStart("D1", new Date(BASE + HOUR * 9));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("new D1 ES",
		 BASE + HOUR * 9, bx.getEarlyStart("D1").getTime());
    assertEquals("new D1 EF",
		 BASE + HOUR * 13, bx.getEarlyFinish("D1").getTime());
    assertEquals("new D2 ES",
		 BASE + HOUR * 13, bx.getEarlyStart("D2").getTime());
    assertEquals("new D2 EF",
		 BASE + HOUR * 18, bx.getEarlyFinish("D2").getTime());
    // check bid independence
    assertEquals("bid D1 ES",
		 BASE + HOUR * 8, theBid.getEarlyStart("D1").getTime());
    assertEquals("bid D1 EF",
		 BASE + HOUR * 12, theBid.getEarlyFinish("D1").getTime());
    assertEquals("bid D2 ES",
		 BASE + HOUR * 12, theBid.getEarlyStart("D2").getTime());
    assertEquals("bid D2 EF",
		 BASE + HOUR * 17, theBid.getEarlyFinish("D2").getTime());
  }

  // ---------- independence of parallel branches ----------

  // single full milestone task, followed by two duration milestones
  private void msetup4 ()
  {
    setup1();
    try {
      theBid.addTask("M1",
		     new Date(BASE + HOUR * 2),
		     new Date(BASE + HOUR * 4),
		     HOUR * 6);
      theBid.addTask("D1", "M1", HOUR * 4);
      theBid.addTask("D2", "D1", HOUR * 5);
      theBid.addTask("D3", "M1", HOUR * 6);
    }
    catch (InvalidDataException ide) {
      fail("Failed to add task");
    }
  }

  public void testMilestone4 ()
  {
    msetup4();
    BidExaminer bx = theBid.createBidExaminer();
    // check before
    assertEquals("D1 ES",
		 BASE + HOUR * 8, bx.getEarlyStart("D1").getTime());
    assertEquals("D1 EF",
		 BASE + HOUR * 12, bx.getEarlyFinish("D1").getTime());
    assertEquals("D2 ES",
		 BASE + HOUR * 12, bx.getEarlyStart("D2").getTime());
    assertEquals("D2 EF",
		 BASE + HOUR * 17, bx.getEarlyFinish("D2").getTime());
    assertEquals("D3 ES",
		 BASE + HOUR * 8, bx.getEarlyStart("D3").getTime());
    assertEquals("D3 EF",
		 BASE + HOUR * 14, bx.getEarlyFinish("D3").getTime());
    // change M1 and check after
    try {
      bx.setEarlyStart("M1", new Date(BASE + HOUR * 3));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("new D1 ES",
		 BASE + HOUR * 9, bx.getEarlyStart("D1").getTime());
    assertEquals("new D1 EF",
		 BASE + HOUR * 13, bx.getEarlyFinish("D1").getTime());
    assertEquals("new D2 ES",
		 BASE + HOUR * 13, bx.getEarlyStart("D2").getTime());
    assertEquals("new D2 EF",
		 BASE + HOUR * 18, bx.getEarlyFinish("D2").getTime());
    assertEquals("D3 ES",
		 BASE + HOUR * 9, bx.getEarlyStart("D3").getTime());
    assertEquals("D3 EF",
		 BASE + HOUR * 15, bx.getEarlyFinish("D3").getTime());
  }

  public void testMilestone4a ()
  {
    msetup4();
    BidExaminer bx = theBid.createBidExaminer();
    // change D1 and check after
    try {
      bx.setEarlyStart("D1", new Date(BASE + HOUR * 9));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("new D1 ES",
		 BASE + HOUR * 9, bx.getEarlyStart("D1").getTime());
    assertEquals("new D1 EF",
		 BASE + HOUR * 13, bx.getEarlyFinish("D1").getTime());
    assertEquals("new D2 ES",
		 BASE + HOUR * 13, bx.getEarlyStart("D2").getTime());
    assertEquals("new D2 EF",
		 BASE + HOUR * 18, bx.getEarlyFinish("D2").getTime());
    assertEquals("D3 ES",
		 BASE + HOUR * 8, bx.getEarlyStart("D3").getTime());
    assertEquals("D3 EF",
		 BASE + HOUR * 14, bx.getEarlyFinish("D3").getTime());
  }

  // ---------- propagation from non-milestone -> milestone ----------

  // single full milestone task, followed by two duration milestones,
  // plus some attached non-milestone tasks.
  private void nmsetup1 ()
  {
    setup1();
    try {
      theBid.addTask("M1",
		     new Date(BASE + HOUR * 2),
		     new Date(BASE + HOUR * 4),
		     HOUR * 6);
      theBid.addTask("N1", "M1");
      theBid.addTask("D1", "M1", HOUR * 4);
      theBid.addTask("N2", "D1");
      theBid.addTask("N3", "D1");
      theBid.addTask("D2", "D1", HOUR * 5);
      theBid.addTask("D3", "M1", HOUR * 6);
    }
    catch (InvalidDataException ide) {
      fail("Failed to add task");
    }
  }

  // change the milestone connected to M1
  public void testNMilestone1 ()
  {
    nmsetup1();
    BidExaminer bx = theBid.createBidExaminer();
    // check the non-milestones first
    assertEquals("N1 ES",
		 BASE + HOUR * 2, bx.getEarlyStart("N1").getTime());
    assertEquals("N1 EF",
		 BASE + HOUR * 8, bx.getEarlyFinish("N1").getTime());
    assertEquals("N2 ES",
		 BASE + HOUR * 8, bx.getEarlyStart("N2").getTime());
    assertEquals("N2 EF",
		 BASE + HOUR * 12, bx.getEarlyFinish("N2").getTime());
    assertEquals("N3 ES",
		 BASE + HOUR * 8, bx.getEarlyStart("N3").getTime());
    assertEquals("N3 EF",
		 BASE + HOUR * 12, bx.getEarlyFinish("N3").getTime());
    // change N1 and check
    try {
      bx.setEarlyStart("N1", new Date(BASE + HOUR * 3));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("new M1 ES",
		 BASE + HOUR * 3, bx.getEarlyStart("M1").getTime());
    assertEquals("new M1 EF",
		 BASE + HOUR * 9, bx.getEarlyFinish("M1").getTime());
    assertEquals("new N1 ES",
		 BASE + HOUR * 3, bx.getEarlyStart("N1").getTime());
    assertEquals("new N1 EF",
		 BASE + HOUR * 9, bx.getEarlyFinish("N1").getTime());
    assertEquals("new D1 ES",
		 BASE + HOUR * 9, bx.getEarlyStart("D1").getTime());
    assertEquals("new D1 EF",
		 BASE + HOUR * 13, bx.getEarlyFinish("D1").getTime());
    assertEquals("new N2 ES",
		 BASE + HOUR * 9, bx.getEarlyStart("N2").getTime());
    assertEquals("new N2 EF",
		 BASE + HOUR * 13, bx.getEarlyFinish("N2").getTime());
    assertEquals("new D2 ES",
		 BASE + HOUR * 13, bx.getEarlyStart("D2").getTime());
    assertEquals("new D2 EF",
		 BASE + HOUR * 18, bx.getEarlyFinish("D2").getTime());
    assertEquals("new D3 ES",
		 BASE + HOUR * 9, bx.getEarlyStart("D3").getTime());
    assertEquals("new D3 EF",
		 BASE + HOUR * 15, bx.getEarlyFinish("D3").getTime());
    // make sure N1 and N2 are still correct in the Bid
    assertEquals("bid N1 ES",
		 BASE + HOUR * 2, theBid.getEarlyStart("N1").getTime());
    assertEquals("bid N1 EF",
		 BASE + HOUR * 8, theBid.getEarlyFinish("N1").getTime());
    assertEquals("bid N2 ES",
		 BASE + HOUR * 8, theBid.getEarlyStart("N2").getTime());
    assertEquals("bid N2 EF",
		 BASE + HOUR * 12, theBid.getEarlyFinish("N2").getTime());
  }

  // change the milestone connected to D1
  public void testNMilestone1a ()
  {
    nmsetup1();
    BidExaminer bx = theBid.createBidExaminer();
    try {
      bx.setEarlyStart("N2", new Date(BASE + HOUR * 9));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("new D1 ES",
		 BASE + HOUR * 9, bx.getEarlyStart("D1").getTime());
    assertEquals("new D1 EF",
		 BASE + HOUR * 13, bx.getEarlyFinish("D1").getTime());
    assertEquals("new D2 ES",
		 BASE + HOUR * 13, bx.getEarlyStart("D2").getTime());
    assertEquals("new D2 EF",
		 BASE + HOUR * 18, bx.getEarlyFinish("D2").getTime());
    assertEquals("new N2 ES",
		 BASE + HOUR * 9, bx.getEarlyStart("N2").getTime());
    assertEquals("new N2 EF",
		 BASE + HOUR * 13, bx.getEarlyFinish("N2").getTime());
    assertEquals("new N3 ES",
		 BASE + HOUR * 9, bx.getEarlyStart("N3").getTime());
    assertEquals("new N3 EF",
		 BASE + HOUR * 13, bx.getEarlyFinish("N3").getTime());
  }

  // ---------- propagation from non-milestone -> bid ----------

  private void nbsetup1 ()
  {
    setup1();
    try {
      theBid.addTask("N1");
      theBid.addTask("N2");
      theBid.addTask("D1", HOUR * 4);
      theBid.addTask("D2", "D1", HOUR * 5);
    }
    catch (InvalidDataException ide) {
      fail("Failed to add task");
    }
  }

  public void testNBMilestone1 ()
  {
    nbsetup1();
    BidExaminer bx = theBid.createBidExaminer();
    // change N1 and see how it propagates
    try {
      bx.setEarlyStart("N1", new Date(BASE + HOUR * 2));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("N1 ES",
		 BASE + HOUR * 2, bx.getEarlyStart("N1").getTime());
    assertEquals("N1 EF",
		 BASE + HOUR * 26, bx.getEarlyFinish("N1").getTime());
    assertEquals("N2 ES",
		 BASE + HOUR * 2, bx.getEarlyStart("N2").getTime());
    assertEquals("N2 EF",
		 BASE + HOUR * 26, bx.getEarlyFinish("N2").getTime());
    assertEquals("D1 ES",
		 BASE + HOUR * 2, bx.getEarlyStart("D1").getTime());
    assertEquals("D1 EF",
		 BASE + HOUR * 6, bx.getEarlyFinish("D1").getTime());
    assertEquals("D2 ES",
		 BASE + HOUR * 6, bx.getEarlyStart("D2").getTime());
    assertEquals("D2 EF",
		 BASE + HOUR * 11, bx.getEarlyFinish("D2").getTime());
  }

  // ---------- out-of-range & boundary conditions ----------

  public void testOR1 ()
  {
    nmsetup1();
    BidExaminer bx = theBid.createBidExaminer();
    // change M1 too far back.  Nothing should happen.
    try {
      bx.setEarlyStart("M1", new Date(BASE));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("new M1 ES",
		 BASE + HOUR * 2, bx.getEarlyStart("M1").getTime());
    assertEquals("new M1 EF",
		 BASE + HOUR * 8, bx.getEarlyFinish("M1").getTime());
    // change M1 too far forward, by 1 msec. Should get an exception.
    try {
      bx.setEarlyStart("M1", new Date(BASE + HOUR * 4 + 1));
    }
    catch (InvalidDataException ide) {
      assertEquals("old M1 ES",
		   BASE + HOUR * 2, bx.getEarlyStart("M1").getTime());
      assertEquals("old M1 EF",
		   BASE + HOUR * 8, bx.getEarlyFinish("M1").getTime());
      return;
    }
    fail("Can't set early start");
  }

  public void testOR1a ()
  {
    nmsetup1();
    BidExaminer bx = theBid.createBidExaminer();
    // change D1 too far back.  Nothing should happen.
    try {
      bx.setEarlyStart("D1", new Date(BASE));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("D1 ES",
		 BASE + HOUR * 8, bx.getEarlyStart("D1").getTime());
    assertEquals("D1 EF",
		 BASE + HOUR * 12, bx.getEarlyFinish("D1").getTime());
    // change D1 too far forward, by 1 msec. Should get an exception.
    try {
      bx.setEarlyStart("D1", new Date(BASE + HOUR * 10 + 1));
    }
    catch (InvalidDataException ide) {
      assertEquals("new D1 ES",
		   BASE + HOUR * 8, bx.getEarlyStart("D1").getTime());
      assertEquals("new D1 EF",
		   BASE + HOUR * 12, bx.getEarlyFinish("D1").getTime());
      return;
    }
    fail("Can't set early start");
  }

  public void testOR2 ()
  {
    nbsetup1();
    BidExaminer bx = theBid.createBidExaminer();
    // change N1 too far back.  Nothing should happen
    try {
      bx.setEarlyStart("N1", new Date(BASE));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("N1 ES",
		 BASE + HOUR, bx.getEarlyStart("N1").getTime());
    assertEquals("N1 EF",
		 BASE + HOUR * 25, bx.getEarlyFinish("N1").getTime());
    // change N1 too far forward, by 1 msec. Should get an exception.
    try {
      bx.setEarlyStart("N1", new Date(BASE + HOUR * 3 + 1));
    }
    catch (InvalidDataException ide) {
      assertEquals("new N1 ES",
		   BASE + HOUR, bx.getEarlyStart("N1").getTime());
      assertEquals("new N1 EF",
		   BASE + HOUR * 25, bx.getEarlyFinish("N1").getTime());
      return;
    }
    fail("Can't set early start");
  }

  public void testLim1 ()
  {
    nmsetup1();
    BidExaminer bx = theBid.createBidExaminer();
    // change M1 to the limit.
    try {
      bx.setEarlyStart("M1", new Date(BASE + HOUR * 4));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("new M1 ES",
		 BASE + HOUR * 4, bx.getEarlyStart("M1").getTime());
    assertEquals("new M1 EF",
		 BASE + HOUR * 10, bx.getEarlyFinish("M1").getTime());
  }

  public void testLim1a ()
  {
    nmsetup1();
    BidExaminer bx = theBid.createBidExaminer();
    // change D1 to the limit
    try {
      bx.setEarlyStart("D1", new Date(BASE + HOUR * 10));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("D1 ES",
		 BASE + HOUR * 10, bx.getEarlyStart("D1").getTime());
    assertEquals("D1 EF",
		 BASE + HOUR * 14, bx.getEarlyFinish("D1").getTime());
  }

  public void testLim2 ()
  {
    nbsetup1();
    BidExaminer bx = theBid.createBidExaminer();
    // change N1 to the limit
    try {
      bx.setEarlyStart("N1", new Date(BASE + HOUR * 3));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("N1 ES",
		 BASE + HOUR * 3, bx.getEarlyStart("N1").getTime());
    assertEquals("N1 EF",
		 BASE + HOUR * 27, bx.getEarlyFinish("N1").getTime());
  }

  // ---------- independence of multiple bid examiners ----------

  public void testMB1 ()
  {
    nmsetup1();
    BidExaminer bx1 = theBid.createBidExaminer();
    BidExaminer bx2 = theBid.createBidExaminer();
    try {
      bx1.setEarlyStart("N2", new Date(BASE + HOUR * 9));
      bx2.setEarlyStart("N2", new Date(BASE + HOUR * 10));
    }
    catch (InvalidDataException ide) {
      fail("Can't set early start");
    }
    assertEquals("new D1 ES",
		 BASE + HOUR * 9, bx1.getEarlyStart("D1").getTime());
    assertEquals("new D1 EF",
		 BASE + HOUR * 13, bx1.getEarlyFinish("D1").getTime());
    assertEquals("new D2 ES",
		 BASE + HOUR * 13, bx1.getEarlyStart("D2").getTime());
    assertEquals("new D2 EF",
		 BASE + HOUR * 18, bx1.getEarlyFinish("D2").getTime());
    assertEquals("new N2 ES",
		 BASE + HOUR * 9, bx1.getEarlyStart("N2").getTime());
    assertEquals("new N2 EF",
		 BASE + HOUR * 13, bx1.getEarlyFinish("N2").getTime());
    assertEquals("new N3 ES",
		 BASE + HOUR * 9, bx1.getEarlyStart("N3").getTime());
    assertEquals("new N3 EF",
		 BASE + HOUR * 13, bx1.getEarlyFinish("N3").getTime());
    assertEquals("new D1 ES",
		 BASE + HOUR * 10, bx2.getEarlyStart("D1").getTime());
    assertEquals("new D1 EF",
		 BASE + HOUR * 14, bx2.getEarlyFinish("D1").getTime());
    assertEquals("new D2 ES",
		 BASE + HOUR * 14, bx2.getEarlyStart("D2").getTime());
    assertEquals("new D2 EF",
		 BASE + HOUR * 19, bx2.getEarlyFinish("D2").getTime());
    assertEquals("new N2 ES",
		 BASE + HOUR * 10, bx2.getEarlyStart("N2").getTime());
    assertEquals("new N2 EF",
		 BASE + HOUR * 14, bx2.getEarlyFinish("N2").getTime());
    assertEquals("new N3 ES",
		 BASE + HOUR * 10, bx2.getEarlyStart("N3").getTime());
    assertEquals("new N3 EF",
		 BASE + HOUR * 14, bx2.getEarlyFinish("N3").getTime());
  }
}
