/*------------------------------------------------------------------------
 * MAGNET utility test components
 * $Id: BidTest.java,v 1.2.2.1 2002/05/05 19:41:57 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Sun Nov 19 13:51:10 2000  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.protocol.cust;

import edu.umn.magnet.protocol.InvalidDataException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;

import java.util.Date;
import java.util.Set;

/**
 * This is a tester for edu.umn.magnet.client.Something
 * @author John Collins
 */
public class BidTest extends TestCase
{
  private Bid theBid;
  static final long BASE = 1000000000000L;
  static final long HOUR = 3600000L;

  public BidTest (String testName)
  {
    super(testName);
  }

  /**
   * You run the main method to exercise this thing.  There are no
   * command-line args.
   */
  public static void main (String[] args)
  {
    junit.textui.TestRunner.run(suite());
  }

  public static Test suite ()
  {
    return new TestSuite(BidTest.class);
  }

  // make sure you can create a bid.  
  public void testCreate ()
  {
    Bid newBid = new Bid();
    assertEquals("No tasks", 0, newBid.size());
  }

  // -------- test per-bid data --------
  private void setup1 ()
  {
    theBid = new Bid();
    theBid.setRFQId("RFQ1");
    theBid.setLatestAccept(new Date(BASE));
    theBid.setCost(100000L);
    theBid.setEarlyStart(new Date(BASE + HOUR));
    theBid.setLateStart(new Date(BASE + HOUR * 2));
    theBid.setDuration(HOUR * 24);
  }

  // check basic data elements
  public void testData ()
  {
    setup1();
    assert((theBid != null));
    assertEquals("RFQ1", theBid.getRFQId());
    assertEquals(BASE, theBid.getLatestAccept().getTime());
    assertEquals(100000L, theBid.getCost());
    assertEquals(BASE + HOUR, theBid.getEarlyStart().getTime());
    assertEquals(BASE + HOUR * 2, theBid.getLateStart().getTime());
    assertEquals(HOUR * 24, theBid.getDuration());
    assertEquals(BASE + HOUR * 2 + HOUR * 24,
		 theBid.getLateFinish().getTime());
  }

  // -------- test some non-milestone tasks --------
  private void setup2 ()
  {
    setup1();
    try {
      theBid.addTask("t1");
      theBid.addTask("t2");
    }
    catch (InvalidDataException ide) {
      fail("Failed to add task");
    }
  }

  // retrieve task ids
  public void testNonMilestone1 ()
  {
    setup2();
    assertEquals(2, theBid.size());
    assert(theBid.hasTask("t1"));
    assert(theBid.hasTask("t2"));
    assert(!theBid.hasTask("t3"));
  }
  
  // check task set
  public void testTaskSet ()
  {
    setup2();
    java.util.Set tasks = theBid.getTaskSet();
    assert(tasks != null);
    assertEquals(2, tasks.size());
    assert(tasks.contains("t1"));
    assert(tasks.contains("t2"));
    assert(!tasks.contains("t3"));
  }

  // check task-already-exists exception
  public void testInvalidData1 ()
  {
    setup2();
    try {
      theBid.addTask("t1");
    }
    catch (InvalidDataException ide) {
      return;
    }
    fail("Failed to reject duplicate task");
  }

  // check per-task time data, no precedence relations
  public void testNonMilestone2 ()
  {
    setup2();
    assertEquals("early start",
		 BASE + HOUR, theBid.getEarlyStart("t1").getTime());
    assertEquals("late start",
		 BASE + HOUR * 2, theBid.getLateStart("t1").getTime());
    assertEquals("early finish",
		 BASE + HOUR * 25, theBid.getEarlyFinish("t1").getTime());
    assertEquals("late finish",
		 BASE + HOUR * 26, theBid.getLateFinish("t1").getTime());
    assertEquals("duration", HOUR * 24, theBid.getDuration("t1"));
  }

  public void testNonMilestone2a ()
  {
    setup2();
    assertEquals("early start",
		 BASE + HOUR, theBid.getEarlyStart("t2").getTime());
    assertEquals("late start",
		 BASE + HOUR * 2, theBid.getLateStart("t2").getTime());
    assertEquals("early finish",
		 BASE + HOUR * 25, theBid.getEarlyFinish("t2").getTime());
    assertEquals("late finish",
		 BASE + HOUR * 26, theBid.getLateFinish("t2").getTime());
    assertEquals("duration",
		 HOUR * 24, theBid.getDuration("t2"));
  }

  // Add in a full milestone task and a couple more dependents
  private void setup3 ()
  {
    setup2();
    try {
      theBid.addTask("M1",
		     new Date(BASE + HOUR * 4),
		     new Date(BASE + HOUR * 6),
		     HOUR * 10);
      theBid.addTask("t3", "M1");
      theBid.addTask("t4", "M1");
    }
    catch (InvalidDataException ide) {
      fail("Failed to add task");
    }
  }

  // retrieve task ids
  public void testMilestone1 ()
  {
    setup3();
    assertEquals(5, theBid.size());
    assert(theBid.hasTask("t1"));
    assert(theBid.hasTask("t2"));
    assert(theBid.hasTask("t3"));
    assert(theBid.hasTask("t4"));
    assert(theBid.hasTask("M1"));
    assert(!theBid.hasTask("T3"));
  }
  
  // check task set
  public void testTaskSet2 ()
  {
    setup3();
    java.util.Set tasks = theBid.getTaskSet();
    assert(tasks != null);
    assertEquals(5, tasks.size());
    assert(tasks.contains("t1"));
    assert(tasks.contains("t2"));
    assert(tasks.contains("t3"));
    assert(tasks.contains("t4"));
    assert(!tasks.contains("t6"));
    assert(tasks.contains("M1"));
  }

  // check per-task time data for t1 and t2
  public void testMilestone2 ()
  {
    setup3();
    assertEquals("t1 early start",
		 BASE + HOUR, theBid.getEarlyStart("t1").getTime());
    assertEquals("t1 late start",
		 BASE + HOUR * 2, theBid.getLateStart("t1").getTime());
    assertEquals("t1 early finish",
		 BASE + HOUR * 25, theBid.getEarlyFinish("t1").getTime());
    assertEquals("t1 late finish",
		 BASE + HOUR * 26, theBid.getLateFinish("t1").getTime());
    assertEquals("t1 duration", HOUR * 24, theBid.getDuration("t1"));

    assertEquals("t2 early start",
		 BASE + HOUR, theBid.getEarlyStart("t2").getTime());
    assertEquals("t2 late start",
		 BASE + HOUR * 2, theBid.getLateStart("t2").getTime());
    assertEquals("t2 early finish",
		 BASE + HOUR * 25, theBid.getEarlyFinish("t2").getTime());
    assertEquals("t2 late finish",
		 BASE + HOUR * 26, theBid.getLateFinish("t2").getTime());
    assertEquals("t2 duration",
		 HOUR * 24, theBid.getDuration("t2"));
  }

  // check per-task time data for M1, t3, t4
  public void testMilestone3 ()
  {
    setup3();
    assertEquals("M1 early start",
		 BASE + HOUR * 4, theBid.getEarlyStart("M1").getTime());
    assertEquals("M1 late start",
		 BASE + HOUR * 6, theBid.getLateStart("M1").getTime());
    assertEquals("M1 early finish",
		 BASE + HOUR * 14, theBid.getEarlyFinish("M1").getTime());
    assertEquals("M1 late finish",
		 BASE + HOUR * 16, theBid.getLateFinish("M1").getTime());
    assertEquals("M1 duration",
		 HOUR * 10, theBid.getDuration("M1"));

    assertEquals("t3 early start",
		 BASE + HOUR * 4, theBid.getEarlyStart("t3").getTime());
    assertEquals("t3 late start",
		 BASE + HOUR * 6, theBid.getLateStart("t3").getTime());
    assertEquals("t3 early finish",
		 BASE + HOUR * 14, theBid.getEarlyFinish("t3").getTime());
    assertEquals("t3 late finish",
		 BASE + HOUR * 16, theBid.getLateFinish("t3").getTime());
    assertEquals("t3 duration",
		 HOUR * 10, theBid.getDuration("t3"));

    assertEquals("t4 early start",
		 BASE + HOUR * 4, theBid.getEarlyStart("t4").getTime());
    assertEquals("t4 late start",
		 BASE + HOUR * 6, theBid.getLateStart("t4").getTime());
    assertEquals("t4 early finish",
		 BASE + HOUR * 14, theBid.getEarlyFinish("t4").getTime());
    assertEquals("t4 late finish",
		 BASE + HOUR * 16, theBid.getLateFinish("t4").getTime());
    assertEquals("t4 duration",
		 HOUR * 10, theBid.getDuration("t4"));
  }

  // Add a duration-only milestone after M1
  private void setup4 ()
  {
    setup3();
    try {
      theBid.addTask("D2", "M1", HOUR * 4);
      theBid.addTask("t5", "D2");
    }
    catch (InvalidDataException ide) {
      fail("Failed to add task");
    }
  }

  public void testMilestone4 ()
  {
    setup4();
    assertEquals("t1 early start",
		 BASE + HOUR, theBid.getEarlyStart("t1").getTime());
    assertEquals("t1 late start",
		 BASE + HOUR * 2, theBid.getLateStart("t1").getTime());
    assertEquals("t1 duration",
		 HOUR * 24, theBid.getDuration("t1"));

    assertEquals("t2 early start",
		 BASE + HOUR, theBid.getEarlyStart("t2").getTime());
    assertEquals("t2 late start",
		 BASE + HOUR * 2, theBid.getLateStart("t2").getTime());
    assertEquals("t2 duration",
		 HOUR * 24, theBid.getDuration("t2"));

    assertEquals("M1 early start",
		 BASE + HOUR * 4, theBid.getEarlyStart("M1").getTime());
    assertEquals("M1 late start",
		 BASE + HOUR * 6, theBid.getLateStart("M1").getTime());
    assertEquals("M1 duration",
		 HOUR * 10, theBid.getDuration("M1"));

    assertEquals("D2 early start",
		 BASE + HOUR * 14, theBid.getEarlyStart("D2").getTime());
    assertEquals("D2 late start",
		 BASE + HOUR * 16, theBid.getLateStart("D2").getTime());
    assertEquals("D2 early finish",
		 BASE + HOUR * 18, theBid.getEarlyFinish("D2").getTime());
    assertEquals("D2 late finish",
		 BASE + HOUR * 20, theBid.getLateFinish("D2").getTime());
    assertEquals("D2 duration",
		 HOUR * 4, theBid.getDuration("D2"));

    assertEquals("t5 early start",
		 BASE + HOUR * 14, theBid.getEarlyStart("t5").getTime());
    assertEquals("t5 late start",
		 BASE + HOUR * 16, theBid.getLateStart("t5").getTime());
    assertEquals("t5 early finish",
		 BASE + HOUR * 18, theBid.getEarlyFinish("t5").getTime());
    assertEquals("t5 late finish",
		 BASE + HOUR * 20, theBid.getLateFinish("t5").getTime());
    assertEquals("t5 duration",
		 HOUR * 4, theBid.getDuration("t5"));
  }
}
