/*-------------------------------------------------------------------------
 * Magnet unit tests
 * $Id: TaskPlanTest.java,v 1.2.2.1 2002/05/05 19:41:58 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *-------------------------------------------------------------------------
 * Revision History:
 * 
 * Fri Dec  1 21:24:04 2000  John Collins
 *   Re-factor as a JUnit test
 * 
 * Wed Sep 16 14:23:19 1998  John E. Collins
 *   Handle InvalidTaskException when creating precedence links
 * 
 * Wed Sep 16 11:35:48 1998  John E. Collins
 *   Update exception types, add new header, change task accessor
 *   getID -> getId
 */

package edu.umn.magnet.protocol.cust;

import edu.umn.magnet.util.time.AgentTime;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestResult;
import junit.framework.TestSuite;

import java.util.Vector;
import java.util.Iterator;
import java.util.Date;

public class TaskPlanTest extends TestCase
{
  private TaskPlan plan;
  private SubtaskRequest subtask1;
  private SubtaskRequest subtask2;

  public TaskPlanTest(String testName)
  {
    super(testName);
  }

  /**
   * You run the main method to exercise this thing.  There are no
   * command-line args.
   */
  public static void main (String[] args)
  {
    junit.textui.TestRunner.run(suite());
  }

  /**
   * Here's the suite that includes all these tests
   */
  public static Test suite ()
  {
    return new TestSuite(TaskPlanTest.class);
  }

  // --------------------------------------------------------------------
  // 1.0 Create a new TaskPlan object and check it out.
  public void testCreate ()
  {
    plan = new TaskPlan();
    assertEquals("Size should be 0", 0, plan.size());

    Iterator enum = plan.iterator();
    assert("Plan should be empty", !enum.hasNext());
  }

  // --------------------------------------------------------------------
  // 2.0 Add some tasks and make sure they stay in there.
  private void setup1 ()
  {
    plan = new TaskPlan();
    subtask1 = new SubtaskRequest("T1");
    subtask2 = new SubtaskRequest("T2");
    try {
      plan.addTask(subtask1);
      plan.addTask(subtask2);
    }
    catch (Exception e) {
      fail("Exception adding task");
    }
  }

  public void testAdd ()
  {
    setup1();
    assertEquals("Check plan size", 2, plan.size());
    assertSame("Check task lookup 1", subtask1, plan.getTaskData("T1"));
    assertSame("Check task lookup 2", subtask2, plan.getTaskData("T2"));
  }

  public void testSuccessors ()
  {
    setup1();
    assert("Should have no successors",
	   !(plan.getSuccessors(subtask1)).hasNext());
    assert("Should have no successors", 
	   !(plan.getSuccessors("T2")).hasNext());
    assert("Should have no predecessors",
	   !(plan.getPredecessors(subtask1)).hasNext());
    assert("Should have no predecessors",
	   !(plan.getPredecessors("T2")).hasNext());
  }

  public void testRoots ()
  {
    setup1();
    Vector vector = enumToVector(plan.getSuccessorRoots());
    assertEquals("Should be 2 roots", 2, vector.size());
    assert("Task 1 should be a root", (vector.indexOf(subtask1) != -1));
    assert("Task 2 should be a root", (vector.indexOf(subtask2) != -1));
  }
  
  // --------------------------------------------------------------------
  // 3.0 Check handling of duplicate and non-existent tasks
  public void testNonexistentTask ()
  {
    setup1();
    assert("Should be no task named T3", 
	   (plan.getTaskData("T3") == null));
  }

  public void testDuplicate ()
  {
    setup1();
    try {
      // Try adding a second task called "T1"
      plan.addTask(new SubtaskRequest("T1"));
    }
    catch (InvalidTaskException e) {
      return;
    }
    fail("Expected duplicate task exception");
  }

  public void testBogusRemove ()
  {
    setup1();
    try {
      // Try removing a task called "T2" that is different from the "T2" in
      // the plan
      plan.removeTask(new SubtaskRequest("T2"));
    }
    catch (InvalidTaskException e) {
      return;
    }
    fail("Expected exception removing invalid task");
  }

  // --------------------------------------------------------------------
  // 4.0 Check out precedence relationships
  private void setup2 ()
  {
    setup1();
    // First, we make T2 be a successor of T1
    try {
      plan.addSuccessor("T1", "T2");
    }
    catch (Exception e) {
      fail("Unexpected exception" + e);
    }
  }

  public void testPrecedence ()
  {
    setup2();
    // make sure T2 is the only successor
    Vector vector = enumToVector(plan.getSuccessors("T1"));
    assertEquals("Check number of successors", 1, vector.size());
    assertEquals("Check identity of successor",
		 subtask2,
		 vector.elementAt(0));
  }

  public void testCircularDependency ()
  {
    setup2();
    try {
      // Try making T1 be a successor of T2
      plan.addSuccessor("T2", "T1");
    }
    catch (CircularDependencyException e) {
      return; // this is what we expect
    }
    catch (InvalidTaskException e) {
      fail("Unexpected exception" + e);
    }
    fail("Failed to detect circular dependency");
  }

  // --------------------------------------------------------------------
  // 5.0 Check out transitive precedence relationships
  private void setup3 ()
  {
    setup2();
    try {
      plan.addTask(new SubtaskRequest("T3"));
    }
    catch (InvalidTaskException e) {
      fail("Bogus InvalidTaskException");
    }
    try {
      plan.addSuccessor("T1", "T3");
    }
    catch (CircularDependencyException e) {
      fail("T3 should not cause circular dependency");
    }
    catch (InvalidTaskException e) {
      fail("T3 should not be an invalid task");
    }
  }
  
  public void testMultiplePrecedence ()
  {
    setup3();
    Iterator enum = plan.getSuccessors("T1");
    Vector vector = new Vector();
    while (enum.hasNext()) {
      vector.addElement(((SubtaskRequest)enum.next()).getId());
      }
    assertEquals("Should be 2 succesors", 2, vector.size());
    assert("T2 should be a successor", (vector.indexOf("T2") != -1));
    assert("T3 should be a successor", (vector.indexOf("T3") != -1));
  }

  public void testIsInSuccessors ()
  {
    setup3();
    assert("T2 is successor of T1", 
	   plan.isInSuccessors(subtask1, subtask2));
    assert("T3 is successor of T1",
	   plan.isInSuccessors("T1", "T3"));
  }

  public void testRootsAgain ()
  {
    setup3();
    // Check that T1 is now the only root
    Vector vector = enumToVector(plan.getSuccessorRoots());
    assertEquals("Only 1 root", 1, vector.size());
    assertEquals("T1 is only root",
		 subtask1,
		 (SubtaskRequest)vector.elementAt(0));
  }

  // Add 3 more successors to form a diamond and check results
  private void setup4 ()
  {
    setup3();
    // -- T4 --
    try {
      plan.addTask(new SubtaskRequest("T4"));
    }
    catch (InvalidTaskException e) {
      fail("T4 should not be an invalid task");
    }
    try {
      plan.addSuccessor("T2", "T4");
    }
    catch (CircularDependencyException e) {
      fail("T2 -> T4 should not cause circular dependency");
    }
    catch (InvalidTaskException e) {
      fail("Unexpected " + e);
    }
    // -- T5 --
    try {
      plan.addTask(new SubtaskRequest("T5"));
    }
    catch (InvalidTaskException e) {
      fail("T5 should not be an invalid task");
    }
    try {
      plan.addSuccessor("T3", "T5");
    }
    catch (CircularDependencyException e) {
      fail("T3 -> T5 should not cause circular dependency");
    }
    catch (InvalidTaskException e) {
      fail("Unexpected " + e);
    }
    // -- T6 --
    try {
      plan.addTask(new SubtaskRequest("T6"));
    }
    catch (InvalidTaskException e) {
      fail("T6 should not be an invalid task");
    }
    try {
      plan.addSuccessor("T4", "T6");
    }
    catch (CircularDependencyException e) {
      fail("T4 -> T6 should not cause circular dependency");
    }
    catch (InvalidTaskException e) {
      fail("Unexpected " + e);
    }
    try {
      plan.addSuccessor("T3", "T6");
    }
    catch (CircularDependencyException e) {
      fail("T3 -> T6 should not cause circular dependency");
    }
    catch (InvalidTaskException e) {
      fail("Unexpected " + e);
    }
    // We now have
    //   2-4-6
    //  /   /
    // 1   /
    //  \ /
    //   3-5
  }

  public void testSuccessors2 ()
  {
    setup4();
    // Make sure T1 still has 2 successors, and now has 5 in allSuccessors
    Vector vector = enumToVector(plan.getSuccessors("T1"));
    assertEquals("T1 has 2 successors",
		 2, vector.size());

    vector = enumToVector(plan.getAllSuccessors(subtask1));
    assertEquals("T1 has 5 tasks in allSuccessors",
		 5, vector.size());
    assert("T6 is in transitive closure of successors of T1",
	   plan.isInSuccessors("T1", "T6"));
  }

    // --------------------------------------------------------------------
    // 6.0 Check out bogus task error handling
  public void testBogusTask1 ()
  {
    setup4();
    // Create a bogus task, try putting it before and after good tasks
    SubtaskRequest bogusTask = new SubtaskRequest("bogus");
    try {
      plan.addSuccessor("bogus", "T6");
    }
    catch (CircularDependencyException e) {
      fail("this should not cause circular dependency");
    }
    catch (InvalidTaskException e) {
      return; // this is what we expect
    }
    fail("Failed to throw InvalidTaskException");
  }

  public void testBogusTask2 ()
  {
    setup4();
    SubtaskRequest bogusTask = new SubtaskRequest("bogus");
    try {
      plan.addSuccessor("T3", "bogus");
    }
    catch (CircularDependencyException e) {
      fail("this should not cause circular dependency");
    }
    catch (InvalidTaskException e) {
      return; // this is what we expect
    }
    fail("Failed to throw InvalidTaskException");
  }

  public void testBogusTask3 ()
  {
    setup4();
    SubtaskRequest bogusTask = new SubtaskRequest("bogus");
    try {
      plan.addSuccessor(plan.getTaskData("T3"), bogusTask);
    }
    catch (CircularDependencyException e) {
      fail("this should not cause circular dependency");
    }
    catch (InvalidTaskException e) {
      return; // this is what we expect
    }
    fail("Failed to throw InvalidTaskException");
  }

  private Vector enumToVector (Iterator input)
  {
    Vector result = new Vector();
    while (input.hasNext()) {
      result.addElement(input.next());
    }
    return result;
  }

  // ------------ Static TaskPlan generator for other tests ------------

  private static final long HOUR = 3600000L;

  public static TaskPlan generateTaskPlan ()
  {
    TaskPlan plan = new TaskPlan();
    // populate the plan with 6 tasks
    try {
      plan.addTask(new SubtaskRequest("T1"));
      plan.addTask(new SubtaskRequest("T2"));
      plan.addSuccessor("T1", "T2");
      plan.addTask(new SubtaskRequest("T3"));
      plan.addSuccessor("T1", "T3");
      plan.addTask(new SubtaskRequest("T4"));
      plan.addSuccessor("T2", "T4");
      plan.addTask(new SubtaskRequest("T5"));
      plan.addSuccessor("T3", "T5");
      plan.addTask(new SubtaskRequest("T6"));
      plan.addSuccessor("T4", "T6");
      plan.addSuccessor("T3", "T6");
    }
    catch (Exception e) {
      fail("Exception creating plan: " + e);
    }
    // We now have
    //   2-4-6
    //  /   /
    // 1   /
    //  \ /
    //   3-5
    //System.out.println("Subtask Tree:");
    //printSuccessorTree(plan, plan.getSuccessorRoots(), 1);
    // put timings on the tasks
    long baseTime = (AgentTime.getDate().getTime()) + HOUR * 12;
    try {
      plan.getTaskData("T1").setEarliestTimeToStart(new Date(baseTime));
      plan.getTaskData("T1").setDeadline(new Date(baseTime + HOUR * 175));
      plan.getTaskData("T2").setEarliestTimeToStart
	                       (new Date(baseTime + HOUR * 75));
      plan.getTaskData("T2").setDeadline(new Date(baseTime + HOUR * 250));
      plan.getTaskData("T3").setEarliestTimeToStart
	                       (new Date(baseTime + HOUR * 75));
      plan.getTaskData("T3").setDeadline(new Date(baseTime + HOUR * 325));
      plan.getTaskData("T4").setEarliestTimeToStart
	                       (new Date(baseTime + HOUR * 150));
      plan.getTaskData("T4").setDeadline(new Date(baseTime + HOUR * 325));
      plan.getTaskData("T5").setEarliestTimeToStart
	                       (new Date(baseTime + HOUR * 150));
      plan.getTaskData("T5").setDeadline(new Date(baseTime + HOUR * 400));
      plan.getTaskData("T6").setEarliestTimeToStart
	                       (new Date(baseTime + HOUR * 225));
      plan.getTaskData("T6").setDeadline(new Date(baseTime + HOUR * 400));
    }
    catch (Exception e) {
      fail("Exception setting task timings: " + e);
    }
    return plan;
  }
}
