/*-----------------------------------------------------------------------
 * Magnet Protocol Element
 * $Id: JobInfo.java,v 1.5 2001/07/03 00:21:25 josephs Exp $
 * Copyright (c) 1998 University of Minesota
 *-----------------------------------------------------------------------
 * Revision History:
 *
 * Fri Jun 29 18:30:00 2001  Len Josephs
 *   Added "setters"/mutators needed to make this JavaBean serializable
 *   by SOAP.
 *
 * 15May2001 DFK - Bug #0000073: javadoc: expected duration is in msec.
 * Tue Mar  6 12:31:37 2001  John Collins
 *   Put back in the protocol package, get rid of failed attempt to
 *   split out data elements, move static factory stuff out.
 *
 * Wed Feb 21 14:36:01 2001  John Collins
 *   Moved to customerProtocol component, split apart from data holder
 *   class JobInfoData.  Removed dependency on tools.Log in favor of
 *   log4j.
 *
 * Tue Nov  9 09:46:43 1999  John Collins
 *   Completely rewritten, into a class, to support job type data
 *
 * Sat Jun 27 10:49:43 1998  John E. Collins
 *   Move to protocol package
 */

package edu.umn.magnet.protocol;

/**
 * Detailed information about different job types, including
 * duration, cost, description, etc.  This is a customer-side wrapper
 * class for the JobInfoData protocol class.  It contains static
 * behavior for generating random task types, useful in generating
 * random plans for testing purposes.
 * @author John Collins
 * @version $Revision: 1.5 $
 */
public class JobInfo implements java.io.Serializable
{
  /**
   * A String naming this job type.
   */
  protected String m_taskType = "default";

  /**
   * A String describing this job type in reasonable detail.
   */
  protected String m_description = "general labor";

  /**
   * A double that gives the expected duration for tasks of this
   * type, in milliseconds.  The default value is one hour.
   */
  protected double m_expectedDuration = 3600000.0; // 1 hour

  /**
   * A double giving the standard deviation of the duration value for
   * tasks of this type, in milliseconds.  Default value is 12
   * minutes (20%).
   */
  protected double m_durSigma = 720000.0; // 20%

  /**
   * A double giving the average price for tasks of this type.
   * Default value is 1000 units.
   */
  protected double m_expectedPrice = 1000.0;

  /**
   * A double giving the standard deviation for tasks of this type.
   * Default value is 200 units, or 20% of the average value.
   */
  protected double m_priceSigma = 200.0;

  /**
   * A double giving the resource availability number for tasks of
   * this type.  (need to come up with a precise definition of what
   * this means).  Default value is 0.5.
   */
  protected double m_resourceAvail = 0.5;

  /**
   * A double giving the supplier availability for this type of task.
   * The number is equivalent to the probability that a given supplier
   * will bid on this type of task, assuming that resources are
   * available.  Default value is 0.5.  Note that it isn't clear how a
   * Server could come up with the resourceAvail number and the
   * supplierAvail number independently.
   */
  protected double m_supplierAvail = 0.5;

  /**
   * A double giving the probability that this type of task will occur
   * in plans in a given market.  A value of 1.0 means that every
   * single plan seen has included this type of task.  Default value
   * is 1.0.
   */
  protected double m_frequency = 1.0;

  // ---------- Constructors ----------

  /**
   * Default constructor for serializability
   */
  public JobInfo ()
  {
    super();
  }

  /**
   * To make it easier to populate a set of these at start-up time, we
   * give you a constructor that takes all the data at once.
   */
  public JobInfo (String tt, String desc,
		  double expectedDur, double durSigma,
		  double expectedPrice, double priceSigma,
		  double resourceAvail, double supplierAvail,
		  double frequency)
  {
    super();
    m_taskType = tt;
    m_description = desc;
    m_expectedDuration = expectedDur;
    m_durSigma = durSigma;
    m_expectedPrice = expectedPrice;
    m_priceSigma = priceSigma;
    m_resourceAvail = resourceAvail;
    m_supplierAvail = supplierAvail;
    m_frequency = frequency;
  }

  // --------------- Instance Methods -----------------

  /**
   * Returns the task type string
   */
  public String getTaskType ()
  {
    return m_taskType;
  }

  /**
   * Returns the description string
   */
  public String getDescription()
  {
    return m_description;
  }

  /**
   * Returns the expected duration in milliseconds
   */
  public double getExpectedDuration ()
  {
    return m_expectedDuration;
  }

  /**
   * Returns the standard deviation of the duration in milliseconds
   */
  public double getDurationSigma ()
  {
    return m_durSigma;
  }

  /**
   * Returns the expected price
   */
  public double getExpectedPrice ()
  {
    return m_expectedPrice;
  }

  /**
   * Returns the standard deviation of the price
   */
  public double getPriceSigma ()
  {
    return m_priceSigma;
  }

  /**
   * Returns the probability that resources will be available in a
   * window equal in length to the actual duration.
   */
  public double getResourceAvailability ()
  {
    return m_resourceAvail;
  }

  /**
   * Returns the probability that resources will be available in a
   * window equal in length to the actual duration.
   */
  public double getSupplierAvailability ()
  {
    return m_resourceAvail;
  }

  /**
   * Returns the relative frequency of this task type.  There is no
   * guarantee that all the frequencies add to 1.0 -- you have to add
   * 'em all up and divide by the total.
   */
  public double getFrequency ()
  {
    return m_frequency;
  }

  // Setters / mutators.
  public void setTaskType(String taskType){m_taskType = taskType;}
  public void setDescription(String description){m_description = description; }
  public void setExpectedDuration(double expectedDuration){
    m_expectedDuration = expectedDuration;
  }
  public void setDurationSigma(double durationSigma){ m_durSigma = durationSigma; }
  public void setExpectedPrice(double expectedPrice){
    m_expectedPrice = expectedPrice;
  }
  public void setPriceSigma(double priceSigma){m_priceSigma = priceSigma; }
  public void setResourceAvailability(double resourceAvailability){
    m_resourceAvail = resourceAvailability;
  }
  public void setSupplierAvailability(double supplierAvailability){
    m_supplierAvail = supplierAvailability;
  }
  public void setFrequency(double frequency){ m_frequency = frequency; }

  public String toString() {
    return "TaskType: "+m_taskType+", Description: "+m_description+
      ", ExpectedDuration: "+m_expectedDuration+
      ", DurationSigma: "+m_durSigma+
      ", Expected Price: "+m_expectedPrice+
      ", Price Sigma: "+m_priceSigma+
      ", Resource Avail: "+m_resourceAvail+
      ", Supplier Avail: "+m_supplierAvail+
      ", Frequency: "+m_frequency+".";
  }
}
