/*------------------------------------------------------------------------
 * MAGNET Manager User Interface component
 * $Id: BidEditorPanel.java,v 1.24.2.1 2002/05/05 19:41:59 ashutosh Exp $
 * Copyright (c) 2001 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Mon Jan  7 17:13:34 2002  John Collins  #0000153
 *   Check for coverage, provide dialog detailing coverage problems.
 *
 * 14Apr2001 PBG Initial Entry
 * 27Apr2001 PBG Incorporated the Bid Table
 * 06May2001 PBG Updated for for bid evaluation activity, buttons
 * 07May2001 PBG Disable Award, Evaluate bids if no bids in table
 * 11May2001 PBG Added Initial Timing information
 * 13May2001 DFK Bug #0000053: bidTable row-selection listener now highlights
 *                corresponding taskTable rows, overlays Gantt objs in GanttPanel.
 * 16May2001 PBG Bug #0000061: Made Progress Bar work
 * 19May2001 PBG - Bug #0000091, #0000101: Keep flow similar to that of agent
 * 20May2001 PBG - Bug #0000046, #0000104: More flow work
 * 23May2001 PBG - Bug #0000059, #0000060: Force Include/Exclude Bids
 * 25May2001 PBG - Bug #0000106: RFQ should expire if deadline passes
 */

package edu.umn.magnet.client.userif;

import java.util.*;
import java.text.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.swing.table.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.font.*;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

import edu.umn.magnet.protocol.cust.*;
import edu.umn.magnet.protocol.JobInfo;
import edu.umn.magnet.client.status.*; // for the status objects

import edu.umn.magnet.util.*; // for testing
import edu.umn.magnet.util.time.AgentTime;

import org.apache.log4j.Category;  // for the logging

/** This is the main class for the Magnet Client Bid Interface
 *
 *  This will contain the entry point to the GUI portion that will
 *  examine and work with bids.  This will allow you to view bids,
 *  evaluate bids, and award bids.  This will contain a main for testing
 *  purposes - but it is expected to really be a components of the
 *  MagnetClientMain Frame.
 *
 * GUI details:
 * This panel is split into 3 different panels.
 *  - the upper left panel is really a scroll pane.  This container will
 *    only hold the Bid Table
 *  - the lower left panel contains the buttons that are used to control
 *    the items/actions/model of the table
 *  - the right hand side panel contains the general controls over the
 *    entire bid process.  This includes the time bars and the buttons to
 *    evaluate and award bids
 */
public class BidEditorPanel extends JPanel
                            implements ActionListener, ListSelectionListener
{
  /** Private Variables */
  //log category information
  private static Category log =
                  Category.getInstance(BidEditorPanel.class.getName());

  private TPTableModel _tpTableModel = null;
  private TPGanttModel _tpGanttModel = null;

  /** Public Variables */
  // The 3 main components
  JScrollPane _bpBidTableScroll; // Put table in scroll
  JPanel _bpTableButtonPanel = new JPanel(new FlowLayout());    // for table buttons
  JPanel _bpTimeBarPanel = new JPanel(new GridBagLayout());      // for time bars and eval/award

  // The Table
  JTable _bpTable = new JTable();
  BidTableModel _bpModel = null;
  RFQStatus _rfqStatus = null;


  // The side components
  JLabel _bpRFQIdLabel = new JLabel();
  JLabel _bpStatusLabel = new JLabel();
  JLabel _bpStatusFieldLabel = new JLabel();

  /** Timer for updating ProgressBar */
  javax.swing.Timer _timer = null;
  Date _startDate = null;
  JLabel _bpProgressLabel = new JLabel();
  JProgressBar _bpBidProgress = new JProgressBar();
  JButton _bpCoverageButton = new JButton();
  JButton _bpEvalButton = new JButton();
  JButton _bpAwardButton = new JButton();

  // The table control components
  JButton _bpViewButton = new JButton();
  JButton _bpDeleteButton = new JButton();
  JButton _bpClearButton = new JButton();

  private static int ONE_SECOND=1000;
  
  //The variables required to remove the extra EvalWindows
  private static int MAXIMUM_EVAL_WINDOWS=20;
  private AwardDialog[] _evalWindows=new AwardDialog[MAXIMUM_EVAL_WINDOWS];
  private int evalWindowsCounter=0;

  /** This is the base and only constructor for the class
   *
   *  This requires the RFQStatus object in order to know about the bids
   *  This requires the TPTableModel object in order to adjust the table given
   *      certain bids being selected
   *  This requires the TPGanntModel object in order to overlay the Gannt Table
   */
  public BidEditorPanel(RFQStatus rfqInfo,
			TPTableModel tptm,
			TPGanttModel tpgm) {
    // Add components to panel
    _tpTableModel = tptm;
    _tpGanttModel = tpgm;
    addComponentsToPanel(rfqInfo, tptm, tpgm);
  }

   /** Add the components to the panel
    *  This will create and display all of the components of the BidPanel
    */
  public void addComponentsToPanel(RFQStatus rfqinfo, TPTableModel tptm,
                                    TPGanttModel tpgm) {
    // Create/add Views
    // add the 3 bigger containers  to this main panel
    // will help simplify grid bag layout

    // This is where all of the Bid Table Scroll Panel stuff goes
    _rfqStatus = rfqinfo;
    _bpModel = new BidTableModel(_rfqStatus, tptm, tpgm, BidEditorPanel.this);
    setTableModels(_bpTable, _bpModel);
    listenForTableChangeEvents(_bpTable);
    _bpTable.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
    _bpBidTableScroll = new JScrollPane(_bpTable);


    // This is where the buttons for the bid table go
    _bpViewButton.setText("View Details");
    _bpViewButton.setMnemonic('v');
    ActionListener val = new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        new BidDetailDialog(GuiUtil.getFrame(BidEditorPanel.this),
                              _bpModel.getBidFromRow(_bpTable.getSelectedRow()));
      }
    };
//    if (_bpTable.getRowCount() == 0) _bpViewButton.setEnabled(false);
    _bpViewButton.addActionListener(val);

    _bpDeleteButton.setText("Delete Bid");
    _bpDeleteButton.setMnemonic('d');
    ActionListener dal = new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        _bpModel.deleteBid(_bpTable.getSelectedRow());
      }
    };
    if (_bpTable.getRowCount() == 0) _bpDeleteButton.setEnabled(false);
    _bpDeleteButton.addActionListener(dal);

    _bpClearButton.setText("Clear Constraints");
    _bpClearButton.setMnemonic('c');
    ActionListener cal = new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        _bpModel.clearBidConstraints();
      }
    };
    if (_bpTable.getRowCount() == 0) _bpClearButton.setEnabled(false);
    _bpClearButton.addActionListener(cal);

    _bpTableButtonPanel.add(_bpViewButton);
    _bpTableButtonPanel.add(_bpDeleteButton);
    _bpTableButtonPanel.add(_bpClearButton);


    // This is where the buttons and time bar for the whole thing go
    if (_rfqStatus == null) {
      _bpRFQIdLabel.setText("RFQ ID:  " + "None");
    } else {
      _bpRFQIdLabel.setText("RFQ ID:  " + _rfqStatus.getRFQ().getId());
    }

    // Now add the phase and duration information
    _bpStatusLabel.setText("Current Phase of RFQ:");
    setStateMonitors();  // initializes this field and progress bar

    /** Need to add a listener to be able to update this as it changes */
    if (_rfqStatus != null) {
      _rfqStatus.addListener(new PropertyChangeListener() {
	public void propertyChange (PropertyChangeEvent pce) {
	  if (pce instanceof RFQStatusEvent) {
            handleRFQStatusEvent((RFQStatusEvent) pce);
	  }
	}
      });

    } // end opf if _rfqStatus is not null

    _bpProgressLabel.setText("Time Remaining in State:");

    _bpCoverageButton.setText("Coverage OK");
    ActionListener cval = new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        showCoverageDialog();
      }
    };
    _bpCoverageButton.setEnabled(false);
    _bpCoverageButton.addActionListener(cval);

    _bpEvalButton.setText("Evaluate Bids");
    _bpEvalButton.setMnemonic('e');
    ActionListener eal = new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        _bpModel.evaluateBids();
      }
    };
    _bpEvalButton.setEnabled(false);
    _bpEvalButton.addActionListener(eal);

    _bpAwardButton.setText("Award Bids");
    _bpAwardButton.setMnemonic('a');
    ActionListener aal = new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        _bpModel.awardBids();
      }
    };
    // Don't let 'em use this button
    _bpAwardButton.setEnabled(false);

    if (_bpTable.getRowCount() == 0) _bpAwardButton.setEnabled(false);
    _bpAwardButton.addActionListener(aal);

     GuiUtil.constrain(_bpTimeBarPanel, _bpRFQIdLabel,
                      0,0,  // Grid  x & Y
                      2,1,  // Grid Width & Height
                      GridBagConstraints.HORIZONTAL,  // fill
                      GridBagConstraints.WEST,   // anchor
                      0.0, 0.0,   // weight x & y
                      GuiUtil.LAYOUT_MARGIN, GuiUtil.LAYOUT_MARGIN,
                      GuiUtil.LAYOUT_MARGIN,GuiUtil.LAYOUT_MARGIN);

    GuiUtil.constrain(_bpTimeBarPanel, _bpStatusLabel,
                      0,2,  // Grid  x & Y
                      2,1,  // Grid Width & Height
                      GridBagConstraints.HORIZONTAL,  // fill
                      GridBagConstraints.WEST,   // anchor
                      0.0, 0.0,   // weight x & y
                      GuiUtil.LAYOUT_MARGIN, GuiUtil.LAYOUT_MARGIN,
                      GuiUtil.LAYOUT_NOMARGIN,GuiUtil.LAYOUT_MARGIN);

    GuiUtil.constrain(_bpTimeBarPanel, _bpStatusFieldLabel,
                      0,3,  // Grid  x & Y
                      2,1,  // Grid Width & Height
                      GridBagConstraints.NONE,  // fill
                      GridBagConstraints.CENTER,   // anchor
                      0.0, 0.0,   // weight x & y
                      GuiUtil.LAYOUT_NOMARGIN, GuiUtil.LAYOUT_MARGIN,
                      GuiUtil.LAYOUT_EXTRAMARGIN,GuiUtil.LAYOUT_MARGIN);

    GuiUtil.constrain(_bpTimeBarPanel, _bpProgressLabel,
                      0,4,  // Grid  x & Y
                      2,1,  // Grid Width & Height
                      GridBagConstraints.NONE,  // fill
                      GridBagConstraints.WEST,   // anchor
                      0.0, 0.0,   // weight x & y
                      GuiUtil.LAYOUT_EXTRAMARGIN, GuiUtil.LAYOUT_MARGIN,
                      GuiUtil.LAYOUT_NOMARGIN,GuiUtil.LAYOUT_MARGIN);

    GuiUtil.constrain(_bpTimeBarPanel, _bpBidProgress,
                      0,5,  // Grid  x & Y
                      2,1,  // Grid Width & Height
                      GridBagConstraints.HORIZONTAL,  // fill
                      GridBagConstraints.CENTER,   // anchor
                      0.5, 0.0,   // weight x & y
                      GuiUtil.LAYOUT_NOMARGIN, GuiUtil.LAYOUT_MARGIN,
                      GuiUtil.LAYOUT_EXTRAMARGIN,GuiUtil.LAYOUT_MARGIN);

    GuiUtil.constrain(_bpTimeBarPanel, _bpCoverageButton,
                      0,6,  // Grid  x & Y
                      2,1,  // Grid Width & Height
                      GridBagConstraints.HORIZONTAL,  // fill
                      GridBagConstraints.CENTER,   // anchor
                      0.0, 0.0,   // weight x & y
                      GuiUtil.LAYOUT_EXTRAMARGIN, GuiUtil.LAYOUT_MARGIN,
                      GuiUtil.LAYOUT_MARGIN,GuiUtil.LAYOUT_MARGIN);

    GuiUtil.constrain(_bpTimeBarPanel, _bpEvalButton,
                      0,7,  // Grid  x & Y
                      2,1,  // Grid Width & Height
                      GridBagConstraints.HORIZONTAL,  // fill
                      GridBagConstraints.CENTER,   // anchor
                      0.0, 0.0,   // weight x & y
                      GuiUtil.LAYOUT_EXTRAMARGIN, GuiUtil.LAYOUT_MARGIN,
                      GuiUtil.LAYOUT_MARGIN,GuiUtil.LAYOUT_MARGIN);

    GuiUtil.constrain(_bpTimeBarPanel, _bpAwardButton,
                      0,8,  // Grid  x & Y
                      2,1,  // Grid Width & Height
                      GridBagConstraints.HORIZONTAL,  // fill
                      GridBagConstraints.CENTER,   // anchor
                      0.0, 0.0,   // weight x & y
                      GuiUtil.LAYOUT_MARGIN, GuiUtil.LAYOUT_MARGIN,
                      GuiUtil.LAYOUT_MARGIN,GuiUtil.LAYOUT_MARGIN);


    // Now add the above panels to this main panel
    this.setLayout(new GridBagLayout());
    GuiUtil.constrain(this, _bpBidTableScroll,
                      0,0,  // Grid  x & Y
                      6,6,  // Grid Width & Height
                      GridBagConstraints.BOTH,  // fill
                      GridBagConstraints.NORTHWEST,   // anchor
                      0.5, 0.5,   // weight x & y
                      GuiUtil.LAYOUT_MARGIN, GuiUtil.LAYOUT_MARGIN,
                      GuiUtil.LAYOUT_MARGIN,GuiUtil.LAYOUT_MARGIN);

    GuiUtil.constrain(this, _bpTableButtonPanel, 0,6,6,1,
                      GridBagConstraints.HORIZONTAL,  // fill
                      GridBagConstraints.SOUTHWEST,   // anchor
                      0.5, 0.0,   // weight x & y
                      GuiUtil.LAYOUT_MARGIN, GuiUtil.LAYOUT_MARGIN,
                      GuiUtil.LAYOUT_MARGIN,GuiUtil.LAYOUT_MARGIN);

    GuiUtil.constrain(this, _bpTimeBarPanel,6,0,1,7,
                      GridBagConstraints.VERTICAL,
                      GridBagConstraints.NORTHEAST,
                      0.0, 0.5,
                      GuiUtil.LAYOUT_MARGIN, GuiUtil.LAYOUT_MARGIN,
                      GuiUtil.LAYOUT_MARGIN,GuiUtil.LAYOUT_MARGIN);

  }



  /** A Generic action listener to listen for button press events */
  public void actionPerformed(ActionEvent e)
  {
      log.info("BidPanel.actionPerformed: Unknown source "
	       + e.getSource());
  }

  /**
   * Shows coverage problems in dialog form
   */
  public void showCoverageDialog ()
  {
    StringBuffer msg = new StringBuffer();
    String delimiter = "";
    for (Iterator tns = _bpModel.unCoveredTasks().iterator();
	 tns.hasNext(); ) {
      msg.append(delimiter).append((String)tns.next());
      delimiter = "\n";
    }
    GuiUtil.inform(this, msg.toString(), "Tasks without bids:");
  }

  /** The following method will attach the {@link BidTableModel} to this table,
   *  it will also then attach a new {@link TableColumnModel} that has the
   *  column titles that will be used.
   *  <p>
   *  Note, much of this was borrowed from MagnetMain
   */
  public void setTableModels(JTable table, BidTableModel bidModel)
  {
    if (null==table  ||  null==bidModel) {  return;  }
    table.setModel(bidModel);
    bidModel.setView(table);
    table.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
    table.createDefaultColumnsFromModel();

    TableColumnModel tcm = table.getColumnModel();
    for (int i=0; i < BidTableModel.COL_MAX; i++) {
      TableColumn column;
      int width = 0;

      column = tcm.getColumn(i);
      if ((i == BidTableModel.COL_BID_IS_IN) ||
          (i == BidTableModel.COL_BID_IS_OUT)) {
          TableCellRenderer renderer = new CheckCellRenderer();
          column.setCellRenderer(renderer);
      }

      switch (i) {
        case BidTableModel.COL_BID_IS_SINGLETON:
        case BidTableModel.COL_BID_IS_IN:
        case BidTableModel.COL_BID_IS_OUT:
            width = 100;
            break;
        case BidTableModel.COL_BID_NAME:
        case BidTableModel.COL_BID_PRICE:
            width = 300;
            break;
        case BidTableModel.COL_BID_TASK_SET:
            width = 500;
            break;
        default:
            width = 200;
            break;
      } // end of switch

      column.setWidth(width);
      column.setMinWidth(width);
      column.setMaxWidth(width);
      column.setPreferredWidth(width);

      /** Now set up the table cell editor */
      if ((i == BidTableModel.COL_BID_IS_IN) ||
          (i == BidTableModel.COL_BID_IS_OUT)) {
          TableCellEditor editor = new DefaultCellEditor(new JCheckBox());
          column.setCellEditor(editor);
      }

      /** finally set up the column headers */
      column.setHeaderRenderer(new JComponentCellRenderer());
      String title = bidModel.getColumnName(i);
      column.setIdentifier(title);
      column.setHeaderValue(title);

    } // end of for loop

    table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
    bidModel.fireTableStructureChanged();
  }

  // -------------------------------
  /** Hook up a change listener to the JTable, and populate fields as necessary. */
  public void listenForTableChangeEvents(JTable table) {
    if (null==table) {  return;  }
    ListSelectionModel lsm = table.getSelectionModel();
    if (null==lsm) {
      table.setSelectionModel(lsm = new DefaultListSelectionModel());
    }
    lsm.addListSelectionListener(this);
  }

  /** Listen to row selection events from the bid table. */
  public void valueChanged(ListSelectionEvent e) {
    log.info("BEP.valueChanged(): e="+e.toString());
    // Only consider non-changing indices; bail otherwise.
    if (e.getValueIsAdjusting()==true) { return; }
    /** The event's index info is FLAKEY, so let's ask the selection model directly. */
    if (e.getSource() instanceof DefaultListSelectionModel) {
      ListSelectionModel m = (ListSelectionModel)e.getSource();
      /** @todo Should keep a "editor dirty" flag so we prompt before blasting any
       *  in-progress hand-edits..  For now, just blast 'em..
       */
//       foobar --  highlight tasks when bid is lit
      
      //clear list of highlighted rows in the task table
      TPTableModel tptm = _bpModel.getTPTableModel();
      tptm.unhighlightAllRows();

      if (null!=_bpModel) {
        if (!m.isSelectionEmpty()) {
          /** Yay! I support multiple row selections!! */
	  //get array of all selected rows in the table
	  int[] selectedRows = _bpTable.getSelectedRows();
	  //highlight each corresponding row in the task table
	  for (int i = 0; i < selectedRows.length; i++) {
	    loadBidDataFromTable(_bpModel, selectedRows[i]);
	  }          
        }
      }
    } else {
      System.err.println("BEP.valueChanged(): error: Unexpected event src "+e.getSource());
    }
  }

  /** getTimeDifference just calculates the time difference between
   *  where the current phase is at and when it will end
   */
  private int getTimeDifference() {

    long endDate;

    if (_rfqStatus == null) {
      _startDate = null;
      return 0;
    }

    _startDate = AgentTime.getDate();
    String stateString = _rfqStatus.getState();
    if (stateString.equals(_rfqStatus.SUBMITTED))  {
      /** The next state change is not until BidDeadline */
      endDate = _rfqStatus.getBidDeadlineTME().getExpirationTime();
    } else if ((stateString.equals(_rfqStatus.BIDS_ACQUIRED)) ||
              (stateString.equals(_rfqStatus.BIDDING_CLOSED))) {
      /** for now have to combine - bidding closed is really brief */
      /** the next state change is not until Earliest Offer */
      endDate = _rfqStatus.getEarliestOfferTME().getExpirationTime();
    } else if ((stateString.equals(_rfqStatus.AWARD_READY))) {
      /* next state change is start? */
      endDate = _tpTableModel.getTaskPlan().getPlanStartDate().getTime();
    } else if (stateString.equals(_rfqStatus.AWARDED)) {
      /** next state change is when they are done */
      endDate = _tpTableModel.getTaskPlan().getPlanDeadline().getTime();
    } else {
      /** These states we don't monitor */
      _startDate = null;
      return 0;
    }
    // Now take end_date - start_date
    return ((int) (endDate - _startDate.getTime()));

  } // end of getTImeDifference


  /** A bid is highlighted; mark the TPTable tasks and draw Gantt bid objs.
   *  <b>NOTE:</b> I'm only supporting SINGLE bid selection!!
   *  @param m the BidTableModel representing the table of bids
   *  @param selectedRow The selected row
   *  @todo Consider supporting highlighting multiple bids!
   */
  private void loadBidDataFromTable(BidTableModel m, int selectedRow) {
    if (null==m) {  return;  }
    TPTableModel tptm = m.getTPTableModel();
    if (null==tptm) {  return;  }
    Bid b = m.getBidFromRow(selectedRow);
    if (null==b) {  return;  }
    // add to list of highlighted rows
    m.highlightBid(b, MagnetClientConstants.H_TYPE0);
    
  } // endof loadBidDataFromTable

  /** This method will handle any updates from our RFQ Status Object
   *  The updates that we care about are:
   *  <ul>
   *  <li> RFQStatus.State -- state change in the RFQStatus Object
   *  <li> RFQStatus.New_Bid -- a new bid was received
   *  </ul>
   */
  public void handleRFQStatusEvent(RFQStatusEvent rfqse)
  {
    if (rfqse.getPropertyName() == null) {
      log.info("RFQStatusEvent set with no value");
      return;
    } else if (rfqse.getPropertyName().equals(RFQStatus.STATE)) {
      // state change, update label and progress bar
      setStateMonitors();
      if (_rfqStatus.getState().equals(RFQStatus.BIDS_ACQUIRED)) {
	// check for covering bid set
	updateCoverage();
      }
      return;
    } else if (rfqse.getPropertyName().equals(RFQStatus.NEW_BID)) {
      // a new bid has arrived - update the table
      _bpModel.addBid((Bid) rfqse.getNewValue());
      return;
    } else {
      return;  // we don't care
    }
  } // end of handleRfqStatusEvent

  /** setStateMonitors just clears out and resets the state label and the
   *  progress bar data
   */
  void setStateMonitors()
  {
    if (_timer != null) _timer.stop();
    _timer = null;
    
    _startDate = null;
    if (_rfqStatus == null) {
      _bpStatusFieldLabel.setText("No Current RFQ");
      
    } else {
      _bpStatusFieldLabel.setText(_rfqStatus.getStateName());
      
    } // end of if _rfqStatus is null
    
    /** These are all _rfqStatus being null proof */
    _bpBidProgress.setMaximum(getTimeDifference());
    _bpBidProgress.setValue(0);
    _bpBidProgress.setStringPainted(true);
    
    // need to create a timer so that we can wake up and update
    // Need to update using agent time
    _timer = new javax.swing.Timer(ONE_SECOND, new ActionListener() {
	public void actionPerformed(ActionEvent evt) {
	  if (_startDate != null)
	    _bpBidProgress.setValue((int)
				    (AgentTime.getDate().getTime() -
				     _startDate.getTime()));
	  checkIfTooLate();
	}
      });
    _timer.start();
    
  } // end of setStateMonitors

  /**
   * Runs coverage check and optionally enables controls if coverage
   * is achieved
   */
  public void updateCoverage ()
  {
    setCovered(_rfqStatus.isBidCoverage());
  }

  /**
   * Indicates that coverage is achieved or not. This is done by
   * enabling the evaluate button and disabling the coverage button or
   * vice versa.
   */
  public void setCovered (boolean isCovered)
  {
    //System.out.println("setCovered(" + isCovered + ")");
    if (isCovered) {
      _bpEvalButton.setEnabled(true);
      _bpCoverageButton.setText("Coverage OK");
      _bpCoverageButton.setEnabled(false);
    }
    else {
      _bpEvalButton.setEnabled(false);
      _bpCoverageButton.setText("No Coverage");
      _bpCoverageButton.setEnabled(true);
    }
  }

  /** checkIfTooLate verifies that the plan can still be awarded.
   *
   *  - if the plan start date is here and bid is not awarded
   */
  protected void checkIfTooLate()
  {
    if (_rfqStatus == null) return;  // no plan to invalidate

    if (_rfqStatus.getState() == _rfqStatus.AWARD_READY) {
      //verify not too late to award
      java.util.Date curDate = AgentTime.getDate();
      if (curDate.after
	  (_rfqStatus.getRFQ().getTaskPlan().getPlanStartDate())) {
        // too late
        GuiUtil.inform(BidEditorPanel.this,
		       "Unable to award bid as Plan Start Date has passed",
		       "RFQ/Plan Must be Abandoned");
        abandonPlan();
        return;
      } // end of if it is too late
    } // end of if it is award ready
    return;
  } // end of checkIfTooLate

  /** abandonPlan will take the steps necessary to abandon the current plan
   *  These include: <ul>
   *  <li> Canceling the RFQ
   *  <li> Disabling the panel
   *  </ul>
   */
  protected void abandonPlan() {

    if (_rfqStatus == null) return;  // nothing to abandon
    _rfqStatus.cancel();
    setEditable(false);

    return;
  } // end of abandonPlan

  /** setEditable(boolean) -- this will either make all of the bid editing
   *  and awarding editable if boolean is true and not editable if boolean is
   *  false
   */
  public void setEditable(boolean val)
  {
    //_bpAwardButton.setEnabled(val);
    _bpClearButton.setEnabled(val);
    _bpDeleteButton.setEnabled(val);
    //_bpEvalButton.setEnabled(val);
    return;
  } // end of setBidsEditable

          
    public void updateEvaluationWindows(AwardDialog ad){
        if(evalWindowsCounter<MAXIMUM_EVAL_WINDOWS){
	    _evalWindows[evalWindowsCounter++]=ad;
	}     
    }
 
    public void removeEvaluationWindows(){
        int startingPoint=0;

        while(startingPoint<MAXIMUM_EVAL_WINDOWS && _evalWindows[startingPoint]!=null){
	    _evalWindows[startingPoint].disposeAwardWindow();//remove the windows 
	    _evalWindows[startingPoint++]=null;
	   }  
	evalWindowsCounter=0;
    }
}//the BidEditorPanel class


/** This class is borrowed from some public domain java examples
 *
 */
class CheckCellRenderer extends JCheckBox implements TableCellRenderer
{
  protected static Border m_noFocusBorder;

  public CheckCellRenderer() {
    super();
    m_noFocusBorder = new EmptyBorder(1, 2, 1, 2);
    setOpaque(true);
    setBorder(m_noFocusBorder);
  }

  public Component getTableCellRendererComponent(JTable table,
   Object value, boolean isSelected, boolean hasFocus,
   int row, int column)
  {
    if (value instanceof Boolean) {
      Boolean b = (Boolean)value;
      setSelected(b.booleanValue());
    }

    setBackground(isSelected && !hasFocus ?
      table.getSelectionBackground() : table.getBackground());
    setForeground(isSelected && !hasFocus ?
      table.getSelectionForeground() : table.getForeground());

    setFont(table.getFont());
    setBorder(hasFocus ? UIManager.getBorder(
      "Table.focusCellHighlightBorder") : m_noFocusBorder);

    return this;
  }
        
    
}
