/*------------------------------------------------------------------------
 * MAGNET Bid Manager User Interface component
 * $Id: BidTableModel.java,v 1.28.2.1 2002/05/05 19:41:59 ashutosh Exp $
 * Copyright (c) 2001 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 *
 * 27Apr2001 PBG Initial Entry
 * 05May2001 PBG Updated to be interactive
 * 06May2001 PBG Updated to do evaluation
 * 07May2001 PBG Made Yes,No strings public
 * 13May2001 DFK Bug #0000053: uncommented L1 bid test data; added getTPTableModel().
 * 13May2001 PBG Bug #0000056: Added so doAward displays results
 * 20May2001 PBG Bug #0000046: Bids display automagically as they arrive
 * 23May2001 PBG - Bug #0000059, #0000060: Force Include/Exclude Bids
 * 25May2001 PBG - Bug #0000129: Have auto Force excludes go away when not needed
 * 02Jun2001 JCollins - Bug #0000131: for now, put forceIncludeExclude dialogs
 *				out to logfile instead.
 */

package edu.umn.magnet.client.userif;

import javax.swing.table.*;
import javax.swing.JTable;  // for javadoc
import javax.swing.*;
import java.awt.Component;
import java.awt.Rectangle;
import java.util.*;
import java.text.DateFormat;
import javax.swing.border.*;

import edu.umn.magnet.client.status.*; // for rfq,bid info, status objects
import edu.umn.magnet.protocol.cust.Bid;     // for Bid data structure
import edu.umn.magnet.protocol.cust.SubtaskRequest; // for Subtask // Structure
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.client.eval.TaskBidPlan;  // for the TaskBidPlan
import edu.umn.magnet.client.eval.BidEvaluator; // for Bid Evaluation

import org.apache.log4j.Category;


/**
 * This model provides a layer of abstraction and grouping of the {@link Bid}
 * data object.  The {@link Bid} data object stores data on a particular bid.
 * However, the {@link JTable} will need to be able to work with a set of
 * {@link Bid} objects - it gets this information via the
 * {@link RFQStatus} object. This will also allow a {@link JTable} to have
 * easy access to this
 * data.
 *
 * <p>
 * The following data items are provided in this model.  Some are configurable
 * items, other are translated from the {@link Bid} data class.
 * Note, that this is not always a one for one translation.  This could be
 * modified to do this at a later date if needed.
 * <ul>
 *  <li> Bid ID
 *  <li> Include in Award
 *  <li> Exclude from Award
 *  <li> Bid Is Singleton
 *  <li> Bid Name
 *  <li> Bid Supplier ID
 *  <li> Bid Price
 *  <li> Bid Task Set
 *  </ul>
 *
 *  <p>
 *  Additional data stores that are used and referenced internally include:
 *  <ul>
 *    <li> RFQ - this is an indicator as to what Bid Set is being looked at
 *    <li> TaskPlan information - this allows easy access back to the
 *                Task Plan data as needed for visual representation.
 *  </ul>
 *
 * <p>
 * Much of this is borrowed from the TaskPlanModel.  This is for ease of
 * maintainability moving forward.  So that two distinctive Table Modeling
 * methods do not have to be picked up on in order to support the work.
 *
 */
public class BidTableModel extends AbstractTableModel
                          //implements StatusListener
{
  /** Defining some enums for the columns */
  /** Enum for the column indicating bid contains a singleton */
  public static final int COL_BID_IS_SINGLETON              = 0;
 /** Enum for column used to check that bid is in award */
  public static final int COL_BID_IS_IN               = 1;
  /** Enum for column used to check that bid is out of award */
  public static final int COL_BID_IS_OUT             = 2;
  /** Enum for bid name column (supplier ID + Bid ID) */
  public static final int COL_BID_NAME              = 3;
  /** Enum for bid price column */
  public static final int COL_BID_PRICE         = 4;
  /** Enum for bid Task Set Column */
  public static final int COL_BID_TASK_SET      = 5;
  /* Max num of columns */
  public static final int COL_MAX                   = 6;

  public static final String BPYES_STRING = "YES";
  public static final String BPNO_STRING = "NO";
  public static final String BPAUTONO_STRING = "ANO";
  private static final boolean doL1 = false;

  /** The RFQStatus object - we need this so the model can listen for events */
  private RFQStatus _rfqStatus = null;

  /** The TaskPlanTableModel - need this in order to update the Gannt values */
  private TPTableModel _tpTable = null;
  private TPGanttModel _tpGantt = null;

  /** Our data that we are wrapping */
  //private java.util.Collection _bidCollection = null;
  private java.util.Vector _bidCollection = null;
  // This will be Yes String, No String, or Empty String
  //- will correspond value to rownum
  private Vector toBeIncluded = new Vector();
  // This will be a Yes string or a No string
  //-  based on if the corresponding bid is a singleton
  private Vector singletons = new Vector();

  /** Maintains a vector list of column titles.
   *  @todo Stolen from DefaultTableModel; Untested as of 2/24/01; dfk
   */
  private Vector columnIdentifiers = null;

  /** Keep a handle to the current table view. */
  /*private*/ JTable _table = null;

  /** Need a handle back to the panel so can disable buttons, etc */
  BidEditorPanel _parentPanel = null;

  private static Category log =
    Category.getInstance(BidTableModel.class.getName());

  /** Empty constructor.
   * Create a new instance of {@link BidTableModel} which
   *  doesn't yet wrap a {@link Bid}, and supports a
   *  table of {@link #COL_MAX} columns, 0 rows.
   */
  public BidTableModel() {
    setColumnIdentifiers();
  }

  /** Create a new instance of {@link BidTableModel} which
   *  wraps the specified {@link Bid} collection.
   */
  public BidTableModel(RFQStatus rfqInfo, TPTableModel tptm, TPGanttModel tpgm,
                        BidEditorPanel parent) {
  _rfqStatus = rfqInfo;
  _tpTable = tptm;
  _tpGantt = tpgm;
  _parentPanel = parent;

  if (_rfqStatus == null) {
    _bidCollection = null;
  } else {
    _bidCollection = new Vector(_rfqStatus.getBids());
    //if ((_bidCollection.isEmpty()) && doL1) {
    //  generateBids(_rfqStatus);
    //}

   if (_bidCollection.isEmpty()) {
    _parentPanel.setEditable(false);
   } else {
      toBeIncluded = new Vector();
      singletons = new Vector();
      toBeIncluded.setSize(_bidCollection.size());
      singletons.setSize(_bidCollection.size());
      for (int i=0; i < _bidCollection.size(); i++) {
        toBeIncluded.set(i, new String(""));
      } // end of for loop
      setSingletons();
   } // end of if bidCollection is empty
  } // end of if RFQ is valid

  setColumnIdentifiers();
  //fireTableDataChanged(); // don't need to call as setColumnIdentifiers calls
 } // end of BidTableModel

  /** getColumnCount is a required method to override from the AbstractTableModel
   *  class.  Since this is constant, it just returns COL_MAX
   */
  public int getColumnCount() {
    return COL_MAX;
  }

 /** getValueAt is a required method to override from the AbstractTableModel
  *  class.  This takes two integers and returns an object (a string or
  *  Boolean) that
  *  is contained in that cell.  If the values are invalid or that cell is
  *  empty, then the empty string or false will be returned.
  */
  public Object getValueAt(int row, int col) {
    int i = 0;
    Bid _bid = null;

    if (null==_bidCollection) { return new String(""); }
    if (_bidCollection.isEmpty()) {return new String("");}
    if (row < 0) { return new String(""); }
    if (col > getColumnCount()) {return new String("");}
    if (row > getRowCount()) {return new String("");}

    // step 1: Get the Bid that is requested
    _bid = getBidFromRow(row);
    if (_bid == null) {
      return new String("");
    }


    // step 2: Get the correct field for that bid
    switch (col) {
      case COL_BID_IS_SINGLETON:
          // This returns empty string, * if true
          if (isSingleton(row)) {
            return new String("*");
          } else {
            return new String("");
          }
      case COL_BID_IS_IN:
          // Returns false if bid is not in, otherwise true
          if (((String) toBeIncluded.get(row)).equals(BPYES_STRING)) {
            return new Boolean(true);
          } else {
            return new Boolean(false);
          }
      case COL_BID_IS_OUT:
          // Returns false if bid is in, else true
          if (((String) toBeIncluded.get(row)).equals(BPNO_STRING) ||
              ((String) toBeIncluded.get(row)).equals(BPAUTONO_STRING)) {
            return new Boolean(true);
          } else {
            return new Boolean(false);
          }
      case COL_BID_NAME:
          //Need to return string which is the Bid ID
          return new String(_bid.getId());
      case COL_BID_PRICE:
          // Need to return string indicating the cost of the bid
          String priceString = new String("");
          priceString += _bid.getCost();
          return priceString;
      case COL_BID_TASK_SET:
          // Need to return a sting indicating what the task set is
          String taskString = new String("");
          java.util.Set taskSet = null;
          taskSet = _bid.getTaskSet();
          boolean isFirst = true;
          if (taskSet.isEmpty()) return taskString;
          // iterate through set putting numbers to string
          for (Iterator tsi = taskSet.iterator(); tsi.hasNext();) {
            if (isFirst) {
              isFirst = false;
            } else {
              taskString += ", ";
            }
            String taskId = (String) tsi.next();
            taskString += taskId;
          } // end of for loop
          return taskString;
      default:
          return new String("");
    }//switch
    // should never get here
  }

 /**
  * Provides for editing of the include/exclude checkboxes. This
  * overrides from the AbstractTableModel 
  * class.  This takes two integers and  an object (for our purposes only a
  * Boolean) that is contained in that cell (only 2 cells are editable).
  * If the values are invalid than an error is displayed and nothing is set
  */
  public void setValueAt(Object value, int row, int col)
  {
    if ((row < 0) || ( row >= getRowCount())) return;
    if (col == COL_BID_IS_IN) {
      // set the value for the bid being in
      if (((Boolean)value).booleanValue() == false) {
        toBeIncluded.set(row, new String(""));
        setAutoExcludesOff();
        // Generate notification
        markChange();
        return;
      }
      // OK we need to include it
      else if (toBeIncluded.get(row).equals(BPNO_STRING)) {
	GuiUtil.inform(_parentPanel,
		       "You must first de-select this bid from Force Exclude",
		       "Error Forcing Bid Include");
	return;
      }
      else if (toBeIncluded.get(row).equals(BPAUTONO_STRING)) {
	GuiUtil.inform(_parentPanel,
		       "This bid cannot be included as it conflicts with another Force Included bid",
		       "Error Forcing Bid Include");
	return;
      } // done error checking
      /** OK to force include it */
      setIncluded(row);
      // Generate notification
      markChange();
      return;
    }
    else if (col == COL_BID_IS_OUT) {
      if (((Boolean)value).booleanValue() == false) {
        if (toBeIncluded.get(row).equals(BPAUTONO_STRING)) {
              GuiUtil.inform(_parentPanel,
              "This bid cannot be de-selected as it conflicts with another Force Included bid",
              "Error De-Selecting Force Exclude");
              return;
          } // done error checking
        toBeIncluded.set(row, new String(""));
        // Generate notification
        markChange();
        return;
      }
      // set the value for the bid being out
      if (toBeIncluded.get(row).equals(BPYES_STRING)) {
          GuiUtil.inform(_parentPanel,
                          "You must first de-select this bid from Force Include",
                          "Error Forcing Bid Exclude");
          return;
      }
      /** OK to force exclude it */
      setExcluded(row);
    } // end of if to set the value

    // Generate notification
    markChange();
    return;
  } // end of setValueAt

  private void markChange ()
  {
    _parentPanel.setCovered(unCoveredTasks().size() == 0);
    fireTableStructureChanged();
  }

   /** getRowCount is a required method to override from the AbstractTableModel
   *  class.  It returns the number of Bid Entries in the table - if there are
   *  no bid entries or the BidCollection is null - it will return 0
   */
  public int getRowCount()
  {
    if (null==_bidCollection) {
      return 0;
    } else {
      return _bidCollection.size();
    }
  }

  /** This method returns a string name based on the column ID
   *  This is being overridden from the AbstractTableModel Class
   *  as that class only provides the basic spread sheet default
   *  column names of A,B,C,etc
   */
  public String getColumnName(int column) {
    if (0>column  ||  COL_MAX<=column) {
      return new String("");
    }
    switch (column) {
      case COL_BID_IS_SINGLETON:
          return new String("Must Include");
      case COL_BID_IS_IN:
          return new String( "Include" );
      case COL_BID_IS_OUT:
          return new String( "Exclude" );
      case COL_BID_NAME:
          return new String("Bid Identifier");
      case COL_BID_PRICE:
          return new String("Bid Price");
      case COL_BID_TASK_SET:
          return new String("Tasks included in Bid");
      default:
          return new String("");
    }//switch
  }

  /** Set the column titles. */
  /** This is from the TaskPlanModel */
  public void setColumnIdentifiers() {
    Vector vColTitles = new Vector();
    for (int i=0; i < COL_MAX; i++) {
      vColTitles.add(getColumnName(i));
    }
    setColumnIdentifiers(vColTitles);
  }

    /**
     * Replaces the column identifiers in the model.  If the number of
     * <code>newIdentifier</code>s is greater than the current number
     * of columns, new columns are added to the end of each row in the model.
     * If the number of <code>newIdentifier</code>s is less than the current
     * number of columns, all the extra columns at the end of a row are
     * discarded. <p>
     *
     * @param   newIdentifiers  vector of column identifiers.  If
     *				<code>null</code>, set the model
     *                          to zero columns
     * @see "javax.swing.table.DefaultTableModel#setRowCount(int)"
     * @todo Stolen from DefaultTableModel.  Untested as of 2/24/01; dfk
     * Stolen from the TaskPlanModel stuff
     */
    public void setColumnIdentifiers(Vector newIdentifiers) {
        if (newIdentifiers != null) {
            columnIdentifiers = newIdentifiers;
        }
        else {
            columnIdentifiers = new Vector();
        }

        // Generate notification
        fireTableStructureChanged();
    }

    /** This is being overridden from the Abstract Class
     */
     public Class getColumnClass(int ColumnIndex) {
      if ((ColumnIndex == COL_BID_IS_IN) || (ColumnIndex == COL_BID_IS_OUT)) {
        return new Boolean(true).getClass();
      } else {
        return new String().getClass();
      }
     }

  /** This is being overridden from the Abstract Class
   *  This will return the column number based on the column title
   */
  public int findColumn(String columnName) {
    for (int i=0; i < COL_MAX; i++) {
      if (getColumnName(i).equals(columnName)) {
        return i;
      }
    }
    // if we got here no match
    return -1;
  }


    /** This is required for the table to be viewed properly */
   public void setView(JTable aTable) {
    _table = aTable;
  }

  /** This is required for the table to be viewed properly */
  public JTable getView() {
    return _table;
  }

  /**
   * Deletes a bid from our Bid Collection.
   *  This will be a permanent deletion.
   */
  public void deleteBid (int row)
  {
    Bid delBid = null;
    int index = -1;
    
    // Need to find the Bid in the collection that was deleted
    delBid = getBidFromRow(row);
    if (delBid == null) {
      GuiUtil.inform(_parentPanel,
		     "Unable to delete Bid, Null Bid selected",
		     "Error Deleting Bid");
      return;
    }
    
    // Now, verify that this bid can be deleted
    if (isSingleton(row)) {
      GuiUtil.inform(_parentPanel, "Unable to delete bid as it is a singleton",
		     "Error Deleting Bid");
      return;
    } // end of singleton check

    if (toBeIncluded.get(row).equals(BPYES_STRING)) {
      GuiUtil.inform(_parentPanel, "This Bid is currently Force Included, unable to delete",
		     "Error Deleting Bid");
      return;
    } // end of if it is force included
    
    // OK to remove the bid from this collection
    index = _bidCollection.indexOf(delBid);
    if (index < 0) return;

    if (_bidCollection.remove(delBid) == false) {
      GuiUtil.inform(_parentPanel,
		     "Unable to remove Bid, internal error",
		     "Error Deleting Bid" );
      return;
    } else {
      // Now, remove this entry from the vectors
      singletons.removeElementAt(index);
      toBeIncluded.removeElementAt(index);
      //_rfqStatus.deleteBid(delBid);

      // Need to recalculate singletons
      setSingletons();
    }
    // Generate notification
    markChange();
  } // end of deleteBid
  
  /** This method will remove all of the constraints from the current
   *  BidModel and Table
   */
  public void clearBidConstraints()
  {
    // walk through Bid Vectors of In,Out - set to Null
    for (int i=0; i < toBeIncluded.size(); i++) {
      toBeIncluded.set(i, new String(""));
    }
    
    // fire action
    markChange();
  } // end of clearBidCOnstraints


  /** This method will display the bid evaluation dialog.  This will allow the
   *  user to define how they want to evaluate the bids.
   */
   public void evaluateBids() {
    EvalDialog ed = new EvalDialog(GuiUtil.getFrame(this.getView()),
                      toBeIncluded, _bidCollection,
                      _tpTable.getTaskPlan(), _rfqStatus, _parentPanel);
    return;
   } // end of evaluateBids

  /**
   * Awards the bids that are set to for include on
   * the bid table.  If this is not a complete solution, an error will be
   * displayed to the user
   */
  public void awardBids()
  {
    BidEvaluator be = null;
    PlanStatus ps = null;
    boolean addedBid = false;
    
    // Make sure that the time to accept bids has arrived
    if (!(_rfqStatus.getState().equals(_rfqStatus.AWARD_READY))) {
      GuiUtil.inform(_parentPanel,
		     "It is too early to award bids, the state must be Award Ready",
		     "Error Awarding Bid");
      return;
    }
    ps = GuiUtil.getPlanStatus(AgentStatus.getInstance(),
			       _tpTable.getTaskPlan());
    be = GuiEvalUtil.createBidEvaluator(ps,
					_rfqStatus,
					_tpTable.getTaskPlan());
    if (be == null) {
      GuiUtil.inform(_parentPanel,
		     "Internal error, unable to create Bid Evaluator",
		     "Error Awarding Bids");
      return;
    }
    
    //Identify the bid set for the award
    for (int i=0; i <toBeIncluded.size(); i++)  {
      if (((String) toBeIncluded.get(i)).equals(BPYES_STRING)) {
	// add this to the collection of bids to be included
	Bid b = getBidFromRow(i);
	if (b != null) {
	  //tbp.addBid(b);
	  be.addBid(b);
	  addedBid = true;
	} // end of if can add bid
      } // end of if it equals it
    } // end of the for loop
    
    if (addedBid == false) {
      GuiUtil.inform(_parentPanel,
		     "You must first Force Include Bids to be awarded, or else run the Evaluator",
		     "Error Awarding Bids");
      return;
    }
    
    //Verify that coverage has been achieved
    // We do this just by running the search on the existing, this
    // should be fast as there should only be one bid per task
    TaskBidPlan tbp = be.evaluateBids();
    if (tbp == null) {
      /** Invalid Plan */
      
      if (!(be.getInfeasibleSets().isEmpty())) {
	// There were infeasible sets
	String bidString = new String("");
	java.util.Set infeasible = be.getInfeasibleSets();
	for (Iterator iter=infeasible.iterator(); iter.hasNext();) {
	  String tempString = new String("");
	  java.util.Set bidSet = (java.util.Set) iter.next();
	  for (Iterator iter2 = bidSet.iterator(); iter2.hasNext();) {
	    Bid b = (Bid)iter2.next();
	    tempString+=" Bid "+b.getId();
	  } // end of for the bids
	  bidString+="\nConflicting Bids:"+tempString;
	} // end of for Bid Sets

	GuiUtil.inform(_parentPanel,
		       "Conflicting Bids Contained as follows:"+bidString,
		       "Error Awarding Bids");
      } else {
	// The plan did not have coverage
	String taskString = new String("");
	java.util.Collection tasks = be.getTaskPlan().getTasks();
	
	for (Iterator iter= tasks.iterator(); iter.hasNext();) {
	  SubtaskRequest stask = (SubtaskRequest)iter.next();
	  int numFound = 0;
	  for (int i=0; i < _bidCollection.size(); i++) {
	    if (toBeIncluded.get(i).equals(BPYES_STRING)) {
	      // was included see if it matches this task
	      if (((Bid)_bidCollection.get(i)).getTaskSet()
		                   .contains(stask.getId())) {
		numFound = 1;
		break;
	      }
	    } // if it was included
	  } // end of for loop
	  if (numFound == 0) {
	    // This task doesn't have a bid
	    taskString +="\nTask ID: " + stask.getId() + " Not Covered";
	  }
	} // end of iterator over tasks
	
	GuiUtil.inform(_parentPanel,
		       "Coverage is not obtained:" + taskString,
		       "Error Awarding Bids");
	
      } // end of if-else for why it failed
      
    } else {
      GuiEvalUtil.doAward(tbp,ps,GuiUtil.getFrame(this.getView()),
			  _parentPanel);
    } // end of if can award
    return;
  } // end of awardBids

  /**
   * Finds a Bid object from the collection based on the row
   * number that is passed into the method.  If it cannot find a bid object
   * it returns null
   */
  public Bid getBidFromRow(int row)
  {
    int i = 0;
    Bid _bid = null;
    if ((row < 0) || (row >= getRowCount())) return null;
    if (_bidCollection == null) return null;
    return ((Bid) _bidCollection.get(row));
  } // end of getBidFromRow

  /** This method is needed as some cells are editable and others are not
   *  This is called on a per cell basis
   */
   public boolean isCellEditable(int row, int col) {
      if ((row >= 0) && (row < getRowCount())) {
        if (isSingleton(row)) return false; // can't edit singletons
        if (col == COL_BID_IS_IN) {
//      This needs to be editable so the user can see why they can't edit it
//          if ((toBeIncluded.get(row).equals(BPNO_STRING)) ||
//              (toBeIncluded.get(row).equals(BPAUTONO_STRING))) {
//              return false;
//          } else {
//              return true;
//          } // end of if can edit the include bid
        return true;
        } // end of if bid is in
        if (col == COL_BID_IS_OUT) {
//      This needs to be editable so the user can see why they can't edit it
//          if (toBeIncluded.get(row).equals(BPYES_STRING)) {
//            return false;
//          } else {
//            return true;
//          }
        return true;
        } // end of if it is the bid is out
      } // end of funky if statements
      return false;
  } // end of isCellEditable

  /** THis method determines if a particular row can be active/edited
   *  This is necessary as some bids must be disregarded if others are
   *  accepted
   */
   public boolean isRowEditable(int row) {

    if ((row < 0) || (row >= getRowCount())) return false;
    if (isSingleton(row)) return false;
//  This needs to be editable so that the user can see why they can't edit it
//  if (toBeIncluded.get(row).equals(BPAUTONO_STRING)) return false;
    return true;
  } // end of isRowEditable

  /**
   * The following method takes the current bidcollection and taskPlan and
   * determines which bids are singletons - if any
   *
   * This returns true if the bid is a singleton, otherwise false
   * This is more complicated than just calling evaluation - due to the fact
   * that we could be looking at bids prior to all bids received - so a plan
   * may or may not be complete
   */
  protected void setSingletons ()
  {
    int index = 0;
    int savedIndex = 0;
    int numFound = 0;
    String taskId = null;

    // reset all of singletons to false & set Includes due to singletons
    for (int i=0; i < singletons.size(); i++) {
      if (singletons.get(i).equals(BPYES_STRING))
          toBeIncluded.set(i, new String(""));
      singletons.set(i, new String(""));
      if (toBeIncluded.get(i).equals(BPAUTONO_STRING))
          toBeIncluded.set(i, new String(""));

    }

    // First, get a list of all tasks for this plan
    java.util.Collection tc = _tpTable.getTaskPlan().getTasks();
    for (Iterator iter = tc.iterator(); iter.hasNext();) {
      SubtaskRequest stask = (SubtaskRequest)iter.next();
      taskId = stask.getId();
      index = 0; numFound = 0; savedIndex = -1;
      // for this task - loop through the bids to see if any have this task
      for (int i=0; i < _bidCollection.size(); i++) {
        Bid b = (Bid) _bidCollection.get(i);
        if (b.hasTask(taskId)) {
          // this bid contains this task
          if (numFound == 1) {
            // more than one bid with this task
            numFound++;
            break;
          } else {
            numFound++;
            savedIndex = index;
          }
        } // end of if this bid has this task
        index++;
      } // end of the for loop
      if ((numFound == 1) && (savedIndex >= 0)) {
        // This task only has one bid
        singletons.set(savedIndex, new String(BPYES_STRING));
        // This must be included in the bid
        setIncluded(savedIndex);
      }
    } // end of looping through the tasks
  } // end of setSingletons


  /**
   * The following method sets the bid at a given row index to be INCLUDED
   *
   * This will verify that the bid is appropriate to add
   * This will set all other bids to "exclude" that contradict this bid
   * This will NOT recalculate singletons - the singleton is a pure value
   *    reflected from the bids that were received
   */
  private void setIncluded(int row) {

    Vector backup = null;
    Bid b = null;
    java.util.Set taskList = null;

    // First, let's do some error checking and make a copy for easy backup
    backup = (Vector) toBeIncluded.clone();

    // Now set the value that we are changing to
    toBeIncluded.set(row, BPYES_STRING);
    b = (Bid) _bidCollection.get(row);
    taskList = b.getTaskSet();

    /** Now loop through other bids */
    for (int i=0; i < _bidCollection.size(); i++) {
      if (i != row) {
        Bid compBid = (Bid) _bidCollection.get(i);
        java.util.Set compList = compBid.getTaskSet();
        for (Iterator iter=taskList.iterator(); iter.hasNext();) {
          if (compList.contains(iter.next())) {
            // this bid contains an identical task
            if (toBeIncluded.get(i).equals(BPYES_STRING)) {
              // Yowsers can't include due to conflicts
	      log.info("Cannot force Include "
		       + b.getId() + ", conflicts with Bid "
		       + compBid.getId());
              toBeIncluded = backup;
              toBeIncluded.set(row, BPAUTONO_STRING);
              return;
            } else if (toBeIncluded.get(i).equals(BPNO_STRING)) {
              // Don't do anything as NO is stronger than AUTO_NO
            } else {
              // first verify that it is not a singleton
              if (isSingleton(i)) {
		log.info("Bid " + b.getId()
			 + "cannot be force Included, it conflicts with Bid "
			 + compBid.getId());
                toBeIncluded = backup;
                toBeIncluded.set(row, BPAUTONO_STRING);
                return;
              } // end of verifying comBid is not singleton
              // set that we auto-determined it can't be included
              toBeIncluded.set(i, BPAUTONO_STRING);
            } // end of setting toBeIncluded
          } // end of if this current bid contains this current task
        } // end of looping through tasks
      } // end of if this isn't the row of the bid we care about
    } // end of looping through bids
    //_parentPanel.setCovered(unCoveredTasks().size() == 0);
    return;
  } // end of setIncluded

  /**
   * Unsets any auto-excludes that are no longer needed.
   * This will NOT recalculate singletons - the singleton is a pure value
   * reflected from the bids that were received.
   */
  private void setAutoExcludesOff()
  {
    java.util.Set taskList = null;

    // We need to create a set of all tasks currently included
    for (int i=0; i < toBeIncluded.size(); i++) {
      if (toBeIncluded.get(i).equals(BPYES_STRING)) {
        // add these tasks to the master set
        Bid b = (Bid) _bidCollection.get(i);
        if (taskList == null) {
          // This is the first one
          taskList = b.getTaskSet();
        } else {
         taskList.addAll(b.getTaskSet());
        } // end of if there is a task List
      }
    }

    if (taskList == null) {
      // just set them all to off
      for (int i = 0; i < toBeIncluded.size(); i++) {
        if (toBeIncluded.get(i).equals(BPAUTONO_STRING)) {
          toBeIncluded.set(i, new String(""));
        }
      } // end of for loop
      //_parentPanel.setCovered(unCoveredTasks().size() == 0);
      return;
    } // end of if no Yeses

    // Now, we need to go through and verify whether we can unset any tasks
    for (int i=0; i< toBeIncluded.size(); i++) {
      if (toBeIncluded.get(i).equals(BPAUTONO_STRING)) {
        // verify still needs to be auto off
        boolean keepAsIs = false;

        Bid b = (Bid) _bidCollection.get(i);
        java.util.Set compList = b.getTaskSet();
        // Need to loop through this bids tasks - should be shorter than master
        for (Iterator iter=compList.iterator(); iter.hasNext();) {
          if (taskList.contains(iter.next())) {
            // this still needs to be auto-off
            keepAsIs = true;
            break;
          }
        } // end of for loop
        if (keepAsIs == false) {
          // reset
          toBeIncluded.set(i, new String(""));
        }
      } // end of if toBeIncluded
    } // end of for loop
    //_parentPanel.setCovered(unCoveredTasks().size() == 0);
    return;
  } // end of setAutoExcludes

  /**
   * Sets the bid at a given row index to be EXCLUDED.
   * This will verify that the bid is appropriate to exclude
   */
  private void setExcluded(int row)
  {
    /** first check to make sure that this is not a singleton */
    if (isSingleton(row)) {
      GuiUtil.inform(_parentPanel,
        "This Bid cannot be force Excluded as it is a singleton.",
        "Error Forcing Exclude");
      return;
    }
    // OK to set to force exclude
    toBeIncluded.set(row, BPNO_STRING);
    //_parentPanel.setCovered(unCoveredTasks().size() == 0);
    return;
  } // end of setExcluded

  /**
   * Returns true just in case the bid in the indicated row is a singleton.
   */
  private boolean isSingleton(int row)
  {
    if (singletons.get(row).equals(BPYES_STRING)) {
      return true;
    }
    return false;
  } // end of isSingleton

  /** 
   * Adds this particular bid to the table.
   * It is assumed that this is a new Bid and that this bid is not forced to
   * be included or excluded.  After the bid is added, the singleton
   * check will have to be re-done.
   */
  public void addBid(Bid b)
  {
    if (b == null) return;
    _parentPanel.setEditable(true);
    int numRows = getRowCount();

    if (_bidCollection.add(b) == false) {
      GuiUtil.inform(_parentPanel,
		     "Internal Java Error, unable to add Bid to collection",
		     "Error adding Bid");
      return;
    }
    else {
      // Now, add this entry to the vectors
      singletons.add(new String(""));
      toBeIncluded.add(new String(""));
      
      // finally, now re-calculate singletons
      setSingletons();
    }
    // Generate notification
    fireTableStructureChanged();
    return;
  } // end of addBid

  /**
   * Returns the set of task names of the tasks in the underlying RFQ
   * that are not covered, given the bids and "exclude bid" choices
   * made by the user.
   */
  public Collection unCoveredTasks ()
  {
    // We do this by making a list of the tasks, then removing the
    // covered ones. If any remain, we're not covered.
    TaskPlan thePlan = _rfqStatus.getRFQ().getTaskPlan();
    HashSet theTasks = new HashSet(thePlan.size());
    for (Iterator tasks = thePlan.iterator(); tasks.hasNext(); ) {
      theTasks.add(((SubtaskRequest)tasks.next()).getId());
    }
    for (int i=0; i < toBeIncluded.size(); i++) {
      if (toBeIncluded.get(i).equals(BPYES_STRING)
	  || toBeIncluded.get(i).equals("")) {
        // add these tasks to the master set
        Bid b = (Bid)_bidCollection.get(i);
	for (Iterator btasks = b.getTaskSet().iterator();
	     btasks.hasNext(); ) {
	  theTasks.remove((String)btasks.next());
	}
      }
    }
    //System.out.println("Uncovered tasks: " + theTasks.size());
    return theTasks;
  }

  ///** This is a dorky L1 call that will generate bids for a particular RFQ -
  // *  once the agent has been integrated with the bidsupplier - then this
  // *  can go away.  In the mean time...
  // *
  // */
  // /** @todo: Remove GenerateBids */
//     public void generateBids(RFQStatus _rs) {
//      java.util.Collection bc = null;
//      Bid _bid = null;

//      if (_rs == null) return;

//      /*bc = _rs.getBids();
//      if (bc != null) {
//        // delete all old bids first
//        for (Iterator iter = bc.iterator(); iter.hasNext();) {
//            _bid = (Bid)(iter.next());
//            _rs.deleteBid(_bid);
//        }

//      }*/

//      bc = _rs.getBids();
//      if (!bc.isEmpty()) {
//        return;
//      }

//      // Create some Bids against the rfq
//      Bid b1 = new Bid();
//      Bid b2 = new Bid();
//      Bid b3 = new Bid();
//      Bid b4 = new Bid();
//      Bid b5 = new Bid();
//      long baseTime = 30000000L;
//      long s_msph = 36000001;
//      try {
//      b1.addTask("TaskID2",
//                new Date(baseTime + s_msph * 80),
//                 new Date(baseTime + s_msph * 200),
//                 s_msph * 80);
//      b1.addTask("TaskID4",
//                  new Date(baseTime + s_msph * 160),
//                  new Date(baseTime + s_msph * 280),
//                  s_msph * 110);
//        b1.setId("B1");
//        b1.setCost(210);
//        _rs.addBid(b1);

//        b2.addTask("TaskID1",
//                    new Date(baseTime + s_msph * 20),
//                    new Date(baseTime + s_msph * 100),
//                    s_msph * 60);
//        b2.addTask("TaskID3",
//                   new Date(baseTime + s_msph * 90),
//                   new Date(baseTime + s_msph * 200),
//                   s_msph * 80);
//        b2.setId("B2");
//        b2.setCost(240);
//        _rs.addBid(b2);

//        b3.addTask("TaskID1",
//                    new Date(baseTime + s_msph * 20),
//                    new Date(baseTime + s_msph * 150),
//                    s_msph * 70);
//        b3.addTask("TaskID3",
//                    new Date(baseTime + s_msph * 130),
//                    new Date(baseTime + s_msph * 250),
//                    s_msph * 100);
//        b3.setId("B3");
//        b3.setCost(170);
//        _rs.addBid(b3);

//        b4.addTask("TaskID1",
//                    new Date(baseTime + s_msph * 20),
//                    new Date(baseTime + s_msph * 120),
//                    s_msph * 80);
//        b4.addTask("TaskID2",
//                    new Date(baseTime + s_msph * 120),
//                    new Date(baseTime + s_msph * 240),
//                    s_msph * 110);
//        b4.addTask("TaskID3",
//                    new Date(baseTime + s_msph * 240),
//                    new Date(baseTime + s_msph * 325),
//                    s_msph * 80);
//        b4.setId("B4");
//        b4.setCost(300);
//        _rs.addBid(b4);
//      }
//      catch (Exception e) {

//        log.info("Exception creating bids: " + e);
//      }

//    } // end of generateBids


  /** Return the current TPTableModel, which may be null */
  public TPTableModel getTPTableModel() {
    return _tpTable;
  }

  /** A bid is highlighted; mark the TPTable tasks and draw Gantt bid objs.
   *  Parameters include: Bid to highlight, Color to Highlight
   */
  public void highlightBid(Bid b, int highlightColor) {
    TPTableModel tptm = getTPTableModel();
    if (null==tptm) {  return;  }
    if (null==b) {  return;  }

    // Get list of taskIDs for this bid.
    Set sTaskIDs = b.getTaskSet();
    if (!sTaskIDs.isEmpty()) {
      // Add each task row with highlight style highlightColor
      for (Iterator i=sTaskIDs.iterator(); i.hasNext(); ) {
        String taskID = (String)i.next();
        if (null==taskID  ||  0==taskID.length()) {
          continue;   // skip it
        }
        tptm.highlightRow(taskID, highlightColor);
        GanttPanel gpView = _tpGantt.getView();
        if (null!=gpView) {  gpView.addBid(b);  }
      }//iter i
    }
  } // highlightBid
}
