/*------------------------------------------------------------------------
 * MAGNET User Interface components
 * $Id: TaskComparator.java,v 1.2.2.1 2002/05/05 19:42:02 ashutosh Exp $
 * Copyright (c) 2000 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 *
 * 18Feb2001 DFK - Initial Entry
 * 23Apr2001 DFK - Bug #0000009: cleanup some javadocs
 */

package edu.umn.magnet.client.userif;

import java.util.Comparator;
import java.util.Date;

import edu.umn.magnet.protocol.JobInfo;
import edu.umn.magnet.protocol.cust.SubtaskRequest;
import edu.umn.magnet.protocol.cust.TaskPlan;
import edu.umn.magnet.protocol.cust.TaskWithPrecedence;


/** A comparator for comparing various MAGNET Task objects.
 *  Sorting is based on EarlyStart, then Deadline, then Alphabetical
 *  taskID and Description, then ??.
 *  Tasks can be of types {@link TaskWithPrecedence} or {@link SubtaskRequest}.
 * <p>
 *  <b>NOTE:</b>  For now, all comparisons will be SubtaskRequest-based;
 *  if one or both objects are TaskWithPrecedence, I'll just call
 *  {@link TaskWithPrecedence#getTaskData()} to get the associated SubtaskRequest.
 * @todo Add getters/setters for TaskPlan (if a TaskPlan's ever necessary..)
 */
 /*
 *  <p>
 *  <font size=-1>
 *  The following would <b>ONLY</b> apply if I need to compare based on TWP attributes,
 *  instead of just harvesting the SubtaskRequest object living within the TWP:
 *  <i>Since a SubtaskRequest alone has no sense of hierarchy, you will first
 *  need to specify the {@link TaskPlan} in which the SubtaskRequest resides
 *  (via method
 *  {@link TaskComparator#setTaskPlan(TaskPlan)}.</i> </font>
 *
 */
public class TaskComparator implements Comparator {

  /** Used in {@link TaskComparator#equals(Object)} to confirm compatibility
   *  of 2 different comparators.
   *  @todo Need to do something more robust than this!!
   */
  private static String cvsID = "$Id: TaskComparator.java,v 1.2.2.1 2002/05/05 19:42:02 ashutosh Exp $";

  public TaskComparator() {
  }

  /** Compare 2 tasks for sorting.
   *  Task objs may be either {@link TaskWithPrecedence} or
   *  {@link SubtaskRequest}.
   *  Comparisons are based on earlier StartTime, then earlier Deadline,
   *    then alphabetically earlier taskId, then alphabetically earlier
   *    Description (based on task's
   *    {@link JobInfo#getDescription()} string), then ??.
   *  @return Similar to {@link java.lang.String#compareTo(String)}, it returns
   *          <tt>-1</tt> if <tt>o1 &lt; o2</tt>, <tt>0</tt> if <tt>o1 == o2</tt>,
   *          or </tt>1</tt> if <tt>o1 &gt; o2</tt>
   */
  public int compare(Object o1, Object o2) {
    SubtaskRequest t1=null, t2=null;
    /** @todo Allow null args?  If so, how to compare? */
    if (null==o1  ||  null==o2) {
      throw new java.lang.UnsupportedOperationException(
        "TaskComparator.compare(o1,o2): illegal null inputs");
    }
    if (!(o1 instanceof SubtaskRequest  ||  o1 instanceof TaskWithPrecedence)  &&
        !(o2 instanceof SubtaskRequest  ||  o2 instanceof TaskWithPrecedence)) {
      throw new java.lang.UnsupportedOperationException(
        "TaskComparator.compare(o1,o2): Unknown object types [o1="+o1+",o2="+o2+"]");
    }
    if (o1 instanceof TaskWithPrecedence) {
      t1 = ((TaskWithPrecedence)o1).getTaskData();
    } else {
      t1 = (SubtaskRequest)o1;
    }
    if (o2 instanceof TaskWithPrecedence) {
      t2 = ((TaskWithPrecedence)o2).getTaskData();
    } else {
      t2 = (SubtaskRequest)o2;
    }

    // Who has an earlier Start time?
    Date d1 = t1.getEarliestTimeToStart();
    Date d2 = t2.getEarliestTimeToStart();
    if (d1.getTime() < d2.getTime()) {
      return -1;
    } else if (d1.getTime() > d2.getTime()) {
      return 1;
    }

    // Who has an earlier Deadline?
    d1 = t1.getDeadline();
    d2 = t2.getDeadline();
    if (d1.getTime() < d2.getTime()) {
      return -1;
    } else if (d1.getTime() > d2.getTime()) {
      return 1;
    }

    // Whose taskID occurs alphabetically earlier?
    String s1 = t1.getId();
    String s2 = t2.getId();
    int idResult = s1.compareTo(s2);
    if (0!=idResult) {
      return idResult;
    }

    // Whose descrip occurs alphabetically earlier?
    /*String*/ s1 = t1.getJobInfo().getDescription();
    /*String*/ s2 = t2.getJobInfo().getDescription();
    int descripResult = s1.compareTo(s2);
    if (0!=descripResult) {
      return descripResult;
    }

    // Any more checks?   Otherwise, return <tt>0</tt> for equality.
    return 0;
  }


  /** Verify that this instance of a {@link Comparator} is compatible w/ another.
   *  This is cheezy, but for now just compare CVS Id's.
   *  @see java.util.Comparator#equals(Object)
   */
  public boolean equals(Object obj) {
    /**@todo: Compare values, not just checking for matching taskID strings! */
    TaskComparator o2=null;
    if (null==obj) {
      throw new java.lang.UnsupportedOperationException("TaskComparator.equals(o): null param");
    }
    if (!(obj instanceof TaskComparator)) {
      return false;
    }
    o2 = (TaskComparator)obj;
    String s2 = o2.cvsID;
    if (!this.cvsID.equals(s2)) {
      return false;
    }
    return true;
  }
}