/*------------------------------------------------------------------------
 * Magnet client components
 * $Id: RVFManager.java,v 1.3.2.3 2002/05/05 19:42:04 ashutosh Exp $
 * Copyright (c) 1999 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Mon Feb 25 14:19:29 2002  John Collins  #0000198
 *   Add ability to cycle seeds
 * 
 * Fri Apr 20 23:43:12 2001  John Collins
 *   Eliminate use of System properties
 * 
 * Fri Feb 23 21:41:42 2001  John Collins
 *   Move into util.rvars package as part of componentization.
 * 
 * Sat Jan 16 12:06:58 1999  John E. Collins
 *   Randomize default value.
 *   Force creation of factories when requested.
 *
 * Tue Jan 12 14:06:54 1999
 *   Changed factories to be accessed through Hashtable s_factories,
 *   instead of using private static variables for each factory.
 *   added addFactory method to implement this new design.
 *
 * Tue Jan 12 9:19:13 1999  Tim Lee
 *   Added this header
 *
 * Mon Jan 11 15:19:13 1999  Tim Lee
 *   Checked in class
 */

package edu.umn.magnet.util.rvars;

import edu.umn.magnet.util.MagnetProperties;

import org.apache.log4j.Category;
import org.apache.log4j.Priority;

import java.util.Iterator;
import java.util.Random;
import java.util.HashMap;

/**
 * This class manages multiple RandomVariableFactories, each with a seed read
 * in from a system property. They are stored in HashMap s_factories.
 * the factory with hash key "ProblemSeed" is for random numbers needed
 * in the bid-creation code, and "SearchSeed" is for the search code.
 * They are referenced by symbols.
 *
 * @author Tim Lee
 */


public class RVFManager
{
  /** Amount to increment random number seeds when cycling */
  public static final long SEED_INCREMENT = 11L;

  private static Category log =
    Category.getInstance(RVFManager.class.getName());

  private static HashMap s_factories=new HashMap();

  /**
   * Creates and returns a new {@link RandomVariableFactory
   * RandomVariableFactory} with the given name.
   */
  public static RandomVariableFactory addFactory(String s)
  {
    long l = 100L;

    String prop = MagnetProperties.getProperty(s);
    if(prop == null)
    {
      Random ran = new Random();
      prop = String.valueOf(ran.nextLong());
    }

    try
    {
      double value = Double.valueOf(prop).doubleValue();
      l = Math.round(value);
      log.debug("RVFManager: " + s + " " + l);
    }
    catch (NumberFormatException nfe)
    {
      log.error("Bad property spec " + prop
                + " for " + s +".  Using "
                + l);
    }

    RandomVariableFactory rvf = new RandomVariableFactory(l);
    s_factories.put(s, rvf);
    return rvf;
  }

  /**
   * Returns the {@link RandomVariableFactory RandomVariableFactory}
   * with the given name. If no such instance exists, then a new one
   * is created.
   */
  public static RandomVariableFactory getRVF(String key)
  {
    RandomVariableFactory rvf = (RandomVariableFactory)s_factories.get(key);
    if (rvf != null)
      return rvf;
    else
      return addFactory(key);
  }

  /**
   * Increments the seeds of all existing {@link RandomVariableFactory
   * RandomVariableFactory} instances by SEED_INCREMENT.
   */
  public static void cycleSeeds ()
  {
    for (Iterator rvfi = s_factories.keySet().iterator();
	 rvfi.hasNext(); ) {
      String name = (String)rvfi.next();
      RandomVariableFactory rvf = (RandomVariableFactory)s_factories.get(name);
      long newSeed = rvf.getSeed() + SEED_INCREMENT;
      s_factories.put(name, new RandomVariableFactory(newSeed));
    }
  }

  /**
   * Causes the seeds of all known {@link RandomVariableFactory
   * RandomVariableFactory} instances to be logged along with their
   * names. The <tt>level</tt> argument must be one of the recognized
   * logging levels (see org.apache.log4j.Level).
   */
  public static void logSeeds (Priority level)
  {
    for (Iterator rvfi = s_factories.keySet().iterator();
	 rvfi.hasNext(); ) {
      String name = (String)rvfi.next();
      RandomVariableFactory rvf = (RandomVariableFactory)s_factories.get(name);
      log.log(level, "Seed for " + name + ": " + rvf.getSeed());
    }
  }
}
