/*------------------------------------------------------------------------
 * Magnet utility components
 * $Id: RandomVariableFactory.java,v 1.1.2.1 2002/05/05 19:42:04 ashutosh Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Fri Sep 25 20:48:10 1998  John E. Collins
 *   Fix syntax error
 * 
 * Wed Sep 23 09:02:28 1998  John E. Collins
 *   Initial Entry, modified from version in jec.util
 */

package edu.umn.magnet.util.rvars;

import java.util.Random;

/**
   This class serves as a container for a set of random variables.
   The general usage pattern is to create the factory with a seed,
   then create the variables needed by your application.  That way, if
   you use the same seed and create the same set of variables, you
   will get repeatable sequences of values.
   @author John Collins
*/
public class RandomVariableFactory 
{
  // Local variables
  private Random m_rvar = new Random();
  private long m_seed;

  public RandomVariableFactory () {
    super();
    m_seed = m_rvar.nextLong();
  }

  public RandomVariableFactory (long seed) {
    super();
    m_rvar.setSeed(seed);
    m_seed = seed;
  }

  /**
     Return the value of the seed.
  */
  public long getSeed () {
    return m_seed;
  }

  /**
     Change the value of the seed.
  */
  public void setSeed (long newSeed) {
    m_seed = newSeed;
  }

  /**
     Create and return a new random variable with a uniform
     distribution between offset and offset + range.
  */
  public UniformVariable createUniformVariable
      (double offset, double range) 
    {
      return new UniformVariable(this, offset, range);
    }

  /**
     Create and return a new random variable with an exponential
     distribution and a specified mean.
  */
  public ExponentialVariable createExponentialVariable
      (double mean)
    {
      return new ExponentialVariable(this, mean);
    }

  /**
     Create and return a new random variable with an m-Erlang
     distribution, having the specified shape and mean.
  */
  public MErlangVariable createMErlangVariable
      (double mean, int shape)
    {
      return new MErlangVariable(this, mean, shape);
    }

  /**
     Create and return a new random variable with a normal
     distribution and the specified mean and standard deviation.
  */
  public NormalVariable createNormalVariable
      (double mean, double sigma)
    {
      return new NormalVariable(this, mean, sigma);
    }

  protected double nextDouble () {
    return m_rvar.nextDouble();
  }
}
