/*------------------------------------------------------------------------
 * Magnet search utility components
 * $Id: BFSearchQueue.java,v 1.9.2.4 2002/05/08 00:03:35 jcollins Exp $
 * Copyright (c) 1998 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 * $Id: BFSearchQueue.java,v 1.9.2.4 2002/05/08 00:03:35 jcollins Exp $
 * Copyright (c) 1998 University of Minnesota
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Tue Mar  5 13:36:25 2002  John Collins
 *   Add ability to trim beyond a bound as given by a predicate.
 * 
 * Thu Jan 24 18:04:31 2002  John Collins  #0000174
 *   Handle search nodes with equal values
 * 
 * Wed Sep  5 16:29:00 2001  John Collins  #0000158
 *   Convert to B-tree version
 * 
 * Tue Sep  4 11:10:11 2001  John Collins - bug #0000158
 *   Initial Entry
 */

package edu.umn.magnet.util.search;

import edu.umn.magnet.util.rvars.RandomVariableFactory;
import edu.umn.magnet.util.rvars.RVFManager;
import edu.umn.magnet.util.rvars.UniformVariable;

import org.apache.log4j.Category;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Iterator;

/**
 * A BFSearchQueue is a {@link SearchQueue SearchQueue} that supports
 * a Best First queue discipline.  This is implemented as a b-tree
 * with maximum bucket size <tt>MAX_BUCKET_SIZE</tt>.
 */
public class BFSearchQueue extends SearchQueue
{
  private static Category log =
    Category.getInstance(SearchQueue.class.getName());

  private static final int MAX_BUCKET_SIZE = 8;

  private Comparator m_comparator = new MinimumIsBest();

  private NBucket topBucket = new NBucket();
  private int m_nodeCount = 0;

  // This supports random selection among equal nodes
  private RandomVariableFactory rvf
    = RVFManager.getRVF("SearchSeed");
  private UniformVariable uv = rvf.createUniformVariable(0.0, 1.0);
  
  /**
   * Default constructor and standard accessors makes this a Bean, but
   * you have to create one using the static <tt>{@link SearchQueue#create
   * SearchQueue.create()}</tt> method, which will return a
   * SearchQueue of an appropriate class to handle the indicated
   * search discipline. 
   */
  protected BFSearchQueue ()
  {
    super();
    m_searchDiscipline = BEST_FIRST;
  }

  // ----------------- Queue control attributes -------------------
 
  /**
   * Return the Comparator used to order nodes for best-first search.
   */
  public Comparator getComparator ()
  {
    return m_comparator;
  }

  /**
   * Set the Comparator to be used to order nodes.  This will have an
   * effect only on newly-inserted nodes, and will not re-order nodes
   * already in the queue.
   * @param v  Value to assign to comparator.
   */
  public void setComparator (Comparator v)
  {
    m_comparator = v;
  }

  // ------------------ Queue operations -------------------

  /**
   * Add a node to the queue according to the current search
   * discipline.
   */
  public void insert (SearchNode newNode)
  {
    //log.debug("insert " + newNode.value());
    NBucket newBucket = (NBucket)topBucket.insert(newNode);
    if (newBucket != null) {
      // two top-level buckets - need to add a new top-level to the
      // hierarchy
      //System.out.println("adding new top-level bucket");
      NBucket newTop = new NBucket(topBucket, newBucket);
      topBucket = newTop;      
    } // if ()
    
    // increment the count and trim if necessary
    m_nodeCount++;
    if (m_maxSize != 0
	&& m_nodeCount > (m_maxSize + m_sizeTolerance))
      trimNodes();
  }

  /**
   * Remove the next node from the queue according to the current
   * search discipline.
   */
  public SearchNode extract ()
  {
    if (m_nodeCount == 0) {
      return null;
    } // if ()
    // the queue is ordered, so we just extract and return the first
    // one from topBucket
    m_nodeCount--;
    return topBucket.extract();
  }

  /**
   * Access the next node on the queue without removing it.
   */
  public SearchNode select ()
  {
    // the queue is ordered, so we just return the first one.
    return topBucket.select();
  }

  /**
   * Access a node by position.  The parameter is multiplied by the
   * current node count to determine the position in the queue to return.
   */
  public SearchNode selectByPosition (double mult)
  {
    //log.debug("selectByPosition " + mult);
    int position = (int)Math.round(m_nodeCount * mult);
    return topBucket.selectByPosition(position);
  }

  /**
   * Access a node by predicate.  The queue is walked from the
   * beginning, and the last node for which the predicate's test
   * method returns true is returned. 
   * <p>
   * This admittedly odd
   * specification is intended to allow a finite
   * probability that the first node will be returned if the test is
   * based on a comparison within the value range of the queue.  As a
   * consequence of this twist, the first node is actually not tested;
   * instead, if the second node fails the test, the first node is
   * returned.</p>
   * <p>
   * Note that this is the only selection method that will return a
   * random result from a group of equal nodes.</p>
   */
  public SearchNode selectByPredicate (Tester predicate)
  {
    SearchNode result = topBucket.selectByPredicate(predicate);
    //if (result != null)
    //  log.debug("selectByPredicate returns " + result.value());
    //else
    //  log.debug("selectByPredicate returns null");
    return result;
  }

  /**
   * Removes the given node from the queue.  This is done by a linear
   * walk down the queue.
   */
  public void remove (SearchNode node)
  {
    //log.debug("remove " + node.value());
    if (topBucket.remove(node) != null) {
      m_nodeCount = topBucket.size();
    }
    else {
      log.warn("Remove failed: " + node.value());
      log.warn(toString());
    }
  }

  /**
   * Trims the tail of the queue, starting with the first node that
   * satisfies the predicte.
   */
  public void trimByPredicate (Tester predicate)
  {
    int removed = topBucket.trimByPredicate(predicate);
    m_nodeCount -= removed;
    if (m_nodeCount != topBucket.size())
      throw new Error("Node count error");
  }

  /**
   * Clears out the entire queue by dropping all the existing
   * structure on the floor.
   */
  public void clear ()
  {
    log.debug("clear");
    topBucket = new NBucket();
    m_nodeCount = 0;
  }

  // --------------- derived attributes ------------------
  /**
   * Returns the value of the first item in the queue.
   */
  public double firstValue ()
  {
    SearchNode first = topBucket.select();
    if (first != null)
      return first.value();
    else
      return 0.0;
  }

  /**
   * Returns the value of the last item in the queue.
   */
  public double lastValue ()
  {
    SearchNode last = topBucket.selectLast();
    if (last != null)
      return last.value();
    else {
      log.info("Null last value:" + toString());
      return 0.0;
    }
  }

  /**
   * Returns the size of the queue
   */
  public int nodeCount ()
  {
    return m_nodeCount;
  }

  /**
   * True if the count is zero
   */
  public boolean isEmpty ()
  {
    return (m_nodeCount == 0);
  }

  /**
   * Returns the size of the queue
   */
  public int size ()
  {
    return m_nodeCount;
  }

  public String toString ()
  {
    return "BFSQ:" + topBucket.toString();
  }

  // --------------- local methods ---------------------

  /**
   * Cut down the queue
   */
  protected void trimNodes()
  {
    //log.debug("Trimming");
    topBucket.trim(m_nodeCount - m_maxSize);
    m_nodeCount = topBucket.size();
    setWasTrimmed();
  }

  // --------------------------------------------------------------
  /**
   * This is the leaf-level B-tree node.  Entries are kept in an
   * ArrayList.  When the bucket size limit is reached, the bucket is
   * split in two. Entries can either be normal SearchNodes, or they
   * can be EBucket instances that contain a set of nodes with equal
   * values.
   */
  protected class Bucket implements SearchNode
  {
    protected ArrayList entries = new ArrayList(MAX_BUCKET_SIZE);
    protected int count = 0;

    protected Bucket ()
    {
      super();
    }

    public double value ()
    {
      if (entries.size() > 0) {
	SearchNode item = (SearchNode)entries.get(0);
	if (item instanceof EBucket)
	  return ((EBucket)item).value();
	else
	  return item.value();
      } // if ()
      return 0.0;
    }

    /**
     * Inserts a new node, splitting the node and returning the new
     * node if necessary.
     */
    protected Bucket insert (SearchNode newNode)
    {
      if (entries.size() == 0) {
	entries.add(newNode);
      }
      else {
	int comp = (m_comparator
		    .compare(newNode,
			     (SearchNode)entries.get(entries.size() - 1)));
	if (comp > 0) {
	  // new node comes after all nodes in this bucket
	  entries.add(entries.size(), newNode);
	}
	else {
	  // new node comes before (or in) last node in bucket
	  for (int i = 0; i < entries.size(); i++) {
	    SearchNode entry = (SearchNode)entries.get(i);
	    int c1 = (m_comparator.compare(newNode, entry));
	    if (c1 < 0) {
	      // new node goes in front of this one
	      entries.add(i, newNode);
	      break;
	    }
	    else if (c1 == 0) {
	      // EBucket needed here
	      if (entry instanceof EBucket) {
		((EBucket)entry).insert(newNode);
	      }
	      else {
		EBucket newEntry = new EBucket(entry);
		newEntry.insert(newNode);
		entries.set(i, newEntry);
	      }
	      break;
	    } // if (c1 == 0)
	  } // for ()
	} // if (comp < 0)
      } // if (entries.size() == 0)
      count += 1;
      if (entries.size() >= MAX_BUCKET_SIZE) {
	return split();
      } // if ()
      return null;
    }

    /**
     * Extracts the first node, removing it.
     */
    protected SearchNode extract ()
    {
      if (entries.size() == 0)
	return null;

      count -= 1;
      SearchNode sn = (SearchNode)entries.get(0);
      if (sn instanceof EBucket) {
	SearchNode result = ((EBucket)sn).extract();
	if (((EBucket)sn).size() == 0) {
	  //log.debug("extract: empty EBucket");
	  entries.remove(0);
	}
	return result;
      }
      else {
	entries.remove(0);
	return sn;
      }
    }

    /**
     * Returns the first node without removing it.
     */
    protected SearchNode select ()
    {
      if (entries.size() == 0)
	return null;
      else
	return ensureSN((SearchNode)entries.get(0));
    }

    /**
     * Returns the last node, without removing it
     */
    protected SearchNode selectLast ()
    {
      if (entries.size() == 0)
	return null;
      else
	return ensureSN((SearchNode)entries.get(entries.size() - 1));
    }

    /**
     * Selects the node given by <tt>position</tt>.
     */
    protected SearchNode selectByPosition (int position)
    {
      if (entries.size() == 0)
	return null;
      else if (position >= entries.size() - 1)
	return selectLast();
      else
	return ensureSN((SearchNode)entries.get(position));
    }

    /**
     * Selects the last node before the one that fails to satisfy the
     * predicate. This assumes that the predicate only tests the value
     * of a node.
     */
    public SearchNode selectByPredicate (Tester predicate)
    {
      if (entries.size() == 0)
	return null;

      int i = 0;
      while (i < (entries.size() - 1)
	     && predicate.test((SearchNode)entries.get(i + 1))) {
	i += 1;
      }
      SearchNode sn = (SearchNode)entries.get(i);
      if (sn instanceof EBucket)
	return (SearchNode)((EBucket)sn).selectRandom();
      else
	return sn;
    }
    
    /**
     * Removes the given node. Note that the node could be in an
     * EBucket. 
     */
    protected Object remove (SearchNode node)
    {
      int i = entries.indexOf(node);
      if (i >= 0) {
	count -= 1;
	return entries.remove(i);
      }
      else {
	// Could be that the requested node is in an embedded
	// EBucket. 
	int j = 0;
	while (j < (entries.size())
	       && (m_comparator.compare((SearchNode)entries.get(j),
					node) != 0)) {
	  j += 1;
	} // while ()
	if (j < entries.size()) {
	  SearchNode entry = (SearchNode)entries.get(j);
	  if (entry instanceof EBucket) {
	    count -= 1;
	    SearchNode result = (SearchNode)((EBucket)entry).remove(node);
	    if (((EBucket)entry).size() == 0) {
	      entries.remove(j);
	    }
	    return result;
	  }
	  else {
	    log.error("Entry " + entry + " should be EBucket");
	    return null;
	  }
	}
	else
	  return null;
      } // if (i >= 0)
    }

    private SearchNode ensureSN (SearchNode sn)
    {
      if (sn instanceof EBucket)
	return (SearchNode)((EBucket)sn).select();
      else
	return sn;
    }

    /**
     * Trims the specified number of nodes from the end of the queue.
     */
    protected void trim (int extras)
    {
      if (extras >= count) {
	// Empty out the whole bucket
	//log.debug("Bucket.trim: empty");
	entries.clear();
	count = 0;
      }
      else {
	//log.debug("Bucket.trim " + extras);
	int rem = extras;
	for (int i = entries.size() - 1;
	     (i >= 0) && (rem > 0);
	     i--) {
	  SearchNode entry = (SearchNode)entries.get(i);
	  if (entry instanceof EBucket) {
	    // take what we can out of this EBucket
	    //if (log.isDebugEnabled())
	    //  log.debug("Trimming " + rem
	    //	+ " from EBucket of size "
	    //	+ ((EBucket)entry).size());
	    int es = ((EBucket)entry).size();
	    if (es > rem) {
	      // there's more than enough here
	      ((EBucket)entry).trim(rem);
	      rem = 0;
	    }
	    else {
	      // This EBucket is to be entirely removed
	      rem -= es;
	      entries.remove(i);
	    } // if (es > rem)
	  }
	  else {
	    // Just remove this SearchNode
	    rem -= 1;
	    //log.debug("Trim " + ((SearchNode)entries.get(i)).value());
	    entries.remove(i);
	  } // if (entry...)
	} // for ()
	count -= (extras - rem);
      } // if (extras >= count)
    }

    /**
     * Trims the tail of a bucket by looking for the first instance
     * that satisfies the predicate and then trimming it and all the
     * rest. We assume that all subsequent nodes also satisfy the
     * predicate.
     */
    protected int trimByPredicate (Tester pred)
    {
      if (!pred.test(selectLast()))
	return 0;
      boolean flag = false;
      int index = 0;
      int result = 0;
      while (index < entries.size() && !flag) {
	if (pred.test((SearchNode)entries.get(index)))
	  flag = true;
	else
	  index += 1;
      }
      while (index < entries.size()) {
	SearchNode node = (SearchNode)entries.get(index);
	if (node instanceof EBucket) {
	  result += ((EBucket)node).size();
	}
	else {
	  result += 1;
	}
	entries.remove(index);
      }
      count -= result;
      return result;
    }

    /**
     * Size is just the count, if we've been tracking it correctly.
     */
    public int size ()
    {
      return count;
    }

    /**
     * Splits this bucket in two pieces
     */
    protected Bucket split ()
    {
      Bucket result = new Bucket();
      for (int i = entries.size() - 1; i >= (MAX_BUCKET_SIZE / 2); i--) {
	SearchNode sn = (SearchNode)entries.remove(i);
	int sz = 1;
	result.entries.add(0, sn);
	if (sn instanceof EBucket)
	  sz = ((EBucket)sn).size();
	count -= sz;
	result.count += sz;
      } // for ()
      return result;
    }

    public String toString ()
    {
      StringBuffer sb = new StringBuffer();
      sb.append("B(");
      String sep = "";
      for (int i = 0; i < entries.size(); i++) {
	sb.append(sep);
	SearchNode node = (SearchNode)entries.get(i);
	if (node instanceof EBucket)
	  sb.append(((EBucket)node).toString());
	else
	  sb.append(((SearchNode)entries.get(i)).value());
	sep = ",";
      } // for ()
      sb.append(")");
      return sb.toString();
    }
  }

  // ---------------------------------------------------------------------
  /**
   * This is the non-leaf bucket
   */
  protected class NBucket extends Bucket
  {
    private int count = 0;

    /**
     * Standard constructor: we assume that this constructor is used
     * only for creating first-level NBucket instances, whose entries
     * are Buckets.  Therefore we initialize the entries list with a
     * single Bucket.
     */
    protected NBucket ()
    {
      super();
    }

    /**
     * Construct a new top-level bucket containing two existing
     * buckets.
     */
    protected NBucket (NBucket first, NBucket second)
    {
      super();
      entries.add(first);
      entries.add(second);
      count = first.count + second.count;
    }

    /**
     * Inserts a new node in the correct next-level bucket.
     */
    protected Bucket insert (SearchNode newNode)
    {
      if (entries.size() == 0)
	entries.add(new Bucket());
      for (int i = 0; i < entries.size(); i++) {
	if (i >= entries.size() - 1
	    || (m_comparator.compare (newNode,
				      (SearchNode)entries.get(i + 1))
		< 0)) {
	  // if newNode compares with successive node, then it goes in
	  // the current node.
	  Bucket splitResult = ((Bucket)entries.get(i)).insert(newNode);
	  if (splitResult != null) {
	    // node got split, new node goes after current
	    entries.add(i + 1, splitResult);
	  } // if ()
	  break;
	} // if ()
      } // for ()
      count += 1;
      if (entries.size() >= MAX_BUCKET_SIZE) {
	return split();
      } // if ()
      return null;
    }

    /**
     * Extracts the first node from the first bucket.  If the first
     * bucket ends up empty, deletes the bucket.
     */
    protected SearchNode extract ()
    {
      Bucket first = (Bucket)entries.get(0);
      SearchNode result = first.extract();
      count -= 1;
      if (first.size() == 0) {
	// need to drop the first bucket
	entries.remove(0);
      } // if ()
      if (result == null) {
	log.error("Null result: size = " + count);
	throw new Error("size = " + count);
      }
      return result;
    }

    /**
     * Selects the first element of the first element, without
     * removing it.
     */
    protected SearchNode select ()
    {
      if (entries.size() > 0) {
	return ((Bucket)entries.get(0)).select();
      } // if ()
      return null;
    }

    /**
     * Returns the last node, without removing it
     */
    protected SearchNode selectLast ()
    {
      if (entries.size() > 0) {
	return ((Bucket)entries.get(entries.size() - 1)).selectLast();
      } // if ()
      return null;
    }

    /**
     * Selects the node given by <tt>mult</tt>, a number between 0.0
     * and 1.0.
     */
    protected SearchNode selectByPosition (int position)
    {
      if (position <= 0)
	return select();
      else if (position >= count - 1)
	return selectLast();
      else {
	int i = -1;
	int newPosn = position;
	do {
	  i += 1;
	  position = newPosn;
	  newPosn -= ((Bucket)entries.get(i)).size();
	} while (newPosn > 0);
	return ((Bucket)entries.get(i)).selectByPosition(position);
      }
    }

    /**
     * Selects the last node before the one that satisfies the
     * predicate. 
     */
    public SearchNode selectByPredicate (Tester predicate)
    {
      if (size() == 0
	  || !predicate.test(select()))
	return null;
      int i = 0;
      while (i < (entries.size() - 1)
	     && predicate.test(((Bucket)entries.get(i + 1)).select())) {
	i += 1;
      } // while ()
      return ((Bucket)entries.get(i)).selectByPredicate(predicate);
    }

    /**
     * Removes the given node from the queue.  This is done by a linear
     * walk down the queue.
     */
    public Object remove (SearchNode node)
    {
      Object result = null;
      int i = 0;
      while (i < (entries.size() - 1)
	     && (m_comparator.compare(node,
				      ((Bucket)entries.get(i + 1))
				      .select())
		 >= 0)) {
	i += 1;
      } // while ()
      result = ((Bucket)entries.get(i)).remove(node);
      if (result != null) {
	count -= 1;
	if (((Bucket)entries.get(i)).size() == 0)
	  entries.remove(i);
      }
      else if (i > 0) {
	result = ((Bucket)entries.get(i - 1)).remove(node);
	if (result != null) {
	  count -= 1;
	  if (((Bucket)entries.get(i - 1)).size() == 0)
	    entries.remove(i - 1);
	}
      }
      return result;
    }

    /**
     * Splits a bucket of buckets.  The only difference between this
     * and the leaf-level split is that here we need to update the
     * counts.
     */
    protected Bucket split ()
    {
      NBucket result = new NBucket();
      for (int i = entries.size() - 1;
	   i >= (MAX_BUCKET_SIZE / 2);
	   i--) {
	result.entries.add(0, entries.remove(i));
      }
      this.recount();
      result.recount();
      return result;
    }

    /**
     * Recomputes node count after some sort of restructuring
     */
    protected void recount ()
    {
      count = 0;
      for (Iterator nodes = entries.iterator(); nodes.hasNext(); ) {
	count += ((Bucket)nodes.next()).size();
      } // for ()
    }

    /**
     * Size of one of these is the same as count.
     */
    public int size ()
    {
      return count;
    }

    /**
     * Trims the specified number of nodes from the end of the queue.
     */
    protected void trim (int extras)
    {
      if (extras > count)
	extras = count;
      count -= extras;
      while (extras > 0) {
	Bucket last = (Bucket)entries.get(entries.size() - 1);
	int lastSize = last.size();
	//System.out.println("Trim " + extras + " from " + lastSize);
	last.trim(extras);
	if (last.size() == 0) {
	  entries.remove(entries.size() - 1);
	} // if ()
	extras -= lastSize;
      } // while ()
    }

    /**
     * Trims the first node that satisfied the predicate, and all
     * succeeding nodes.
     * @returns the number of nodes removed by the trimming.
     */
    protected int trimByPredicate (Tester pred)
    {
      if (!pred.test(selectLast()))
	return 0;
      boolean flag = false;
      int index = 0;
      int result = 0;
      // walk the list until we find the first node that triggers
      while (index < entries.size() && !flag) {
	int removed
	    = ((Bucket)entries.get(index)).trimByPredicate(pred);
	if (removed > 0) {
	  // this is the first bucket that had a match.
	  flag = true;
	  result += removed;
	}
	if (flag && ((Bucket)entries.get(index)).size() == 0)
	  entries.remove(index);
	else
	  index += 1;
      }
      // now walk the rest of the list, removing them
      // all and counting the results
      while (index < entries.size()) {
	result += ((Bucket)entries.get(index)).size();
	entries.remove(index);
      }
      count -= result;
      return result;
    }

    public String toString ()
    {
      StringBuffer sb = new StringBuffer();
      sb.append("NB{");
      String sep = "";
      for (Iterator ents = entries.iterator(); ents.hasNext(); ) {
	sb.append(sep).append(ents.next().toString());
	sep = ";";
      } // for ()
      sb.append("}");
      return sb.toString();
    }
  }

  // ---------------------------------------------------------------------
  /**
   * This is the equal-value bucket
   */
  protected class EBucket extends Bucket
  {
    private ArrayList entries = new ArrayList();

    /**
     * You have to have a SearchNode to create one of these
     */
    protected EBucket (SearchNode initialNode)
    {
      super();
      //log.debug("Creating new EBucket, value = " + initialNode.value());
      entries.add(initialNode);
    }

    public double value ()
    {
      if (entries.size() > 0)
	return ((SearchNode)entries.get(0)).value();
      else
	return 0.0;
    }

    /**
     * Adds a new node to the end of this bucket. We assume without
     * checking that all entries in this bucket have the same value.
     */
    protected Bucket insert (SearchNode newNode)
    {
      entries.add(newNode);
      return null;
    }

    /**
     * Extracts the first node from the this bucket.
     */
    protected SearchNode extract ()
    {
      return (SearchNode)entries.remove(0);
    }

    /**
     * Selects the first element, without removing it.
     */
    protected SearchNode select ()
    {
      return (SearchNode)entries.get(0);
    }

    /**
     * Selects a random element, without removing it.
     */
    protected SearchNode selectRandom ()
    {
      int index = (int)Math.floor(uv.nextDouble() * entries.size());
      if (index == entries.size())
	index -= 1;
      return (SearchNode)entries.get(index);
    }

    /**
     * Removes the given node from this bucket, if it exists.
     */
    public Object remove (SearchNode node)
    {
      if (entries.remove(node))
	return node;
      else
	return null;
    }

    /**
     * Size of one of these is the same as count.
     */
    public int size ()
    {
      return entries.size();
    }

    /**
     * Trims the specified number of nodes from the end of the queue.
     */
    protected void trim (int extras)
    {
      if (extras >= entries.size())
	entries.clear();
      else {
	int osize = entries.size();
	for (int i = 1; i <= extras; i++) {
	  entries.remove(osize - i);
	}
      }
    }

    public String toString ()
    {
      StringBuffer sb = new StringBuffer();
      sb.append("EB(");
      String sep = "";
      for (int i = 0; i < entries.size(); i++) {
	sb.append(sep).append(((SearchNode)entries.get(i)).value());
	sep = ",";
      } // for ()
      sb.append(")");
      return sb.toString();
    }
  }
}
