/*------------------------------------------------------------------------
 * MAGNET utility component
 * $Id: SilkHook.java,v 1.5.2.1 2002/05/05 19:42:05 ashutosh Exp $
 * Copyright (c) 2001 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Wed Mar 21 19:37:06 2001  John Collins
 *   Add loadFileOrResource
 * 
 * Mon Mar 19 09:23:08 2001  John Collins
 *   Initial Entry
 */

package edu.umn.magnet.util.silk;

import silk.Scheme;
import silk.SI;
import silk.InputPort;

import org.apache.log4j.Category;

/**
 * This is the Java side of the "hook" mechanism for extending MAGNET
 * agents with Scheme code.  As with most Silk stuff, this is
 * completely static - you don't create instances of this class.
 * @author John Collins
 * @version $Revision: 1.5.2.1 $
 */
public class SilkHook
{
  static Category log = Category.getInstance(SilkHook.class.getName());
  static boolean initialized = false;

  /**
   * Runs all the hook procedures that are registered under
   * <tt>name</tt>.  You create a hook in Scheme as follows:
   * <pre>
   *   (define (<i>fn-name</i> arg)
   *     ...scheme code here...)
   *   (add-hook <i>'hook-name fn-name</i>)
   * </pre>
   * where fn-name is the name of a scheme function, and hook-name is
   * the name of the hook as defined in your Java code.  Remember that
   * the <tt>hook-name</tt> is a String in Java, and a symbol in
   * Scheme.
   * @param name  a String naming your hook.
   * @param thing a Java Object that is passed to your hook
   *   function.  This will show up in the above Scheme code as the
   *   value of <tt>arg</tt>.
   */
  static public void runHook (String name, Object thing)
  {
    ensureInit();
    SI.call("run-hook", name, thing);
  }

  /**
   * Loads the scheme setup files.
   * Two files are loaded here.  Other files may be loaded as needed
   * by other methods.
   * <ol>
   * <li>"hooks.scm" is loaded as a Java resource.  That means it is
   * package along with this class in the jar file.  It is not good if
   * this file is not found.
   * <li>"init.scm" is loaded from the current runtime environment.
   * It's no big deal if this file is not found.
   * </ol>
   */
  static public void init ()
  {
    java.io.InputStream in =
	SilkHook.class.getResourceAsStream("hooks.scm");
    if (in != null) {
      Scheme.load(new InputPort(in));
    }
    else {
      System.out.println("+++ hook init file not found +++");
      log.error("Scheme init file hooks.scm not found");
    }
    try {
      SI.load(new java.io.FileReader(new java.io.File("init.scm")));
    }
    catch (java.io.FileNotFoundException fnf) {
    }
    // We set the initialized flag regardless of success just to keep
    // the auto-initialization function from getting annoying.
    initialized = true;
  }

  static private void ensureInit ()
  {
    if (!initialized) {
      init();
    }
  }

  /**
   * Loads a Scheme script from the named file, or from a resource.
   * The idea is that you provide a default version of the script as a
   * resource, tied to a particular class.  When you attempt to load,
   * this looks first in the current environment for a file with the
   * given name, and failing that, attempts to load the resource by
   * the same name.  A WARN entry is made in the log if the resource
   * gets loaded, and an ERROR entry is made if neither file was loaded.
   * @param name  The name of the file or resource to load.
   * @param class  The class to be used to look up the resource.  This
   *     means that the resource needs to be packaged in the same jar
   *     file, in the same package, as the given class.
   */
  static public void loadFileOrResource (String name, Class cls)
  {
    try {
      // try loading the file first
      SI.load(new java.io.FileReader(new java.io.File(name)));
    }
    catch (java.io.FileNotFoundException fnf) {
      // if that fails, try to load the resource
      java.io.InputStream in = cls.getResourceAsStream(name);
      if (in != null) {
	Scheme.load(new InputPort(in));
	log.warn("defaulted to resource file for " + name);
      }
      else {
	// Here, we've got a problem.  Couldn't find either one.
	log.error("failed to find either file or resource for " + name);
      }
    }
  }
}
