/*------------------------------------------------------------------------
 * MAGNET agent status component
 * $Id: AgentTime.java,v 1.5.2.1 2002/05/05 19:42:06 ashutosh Exp $
 * Copyright (c) 2001 Regents of the University of Minnesota.
 * All Rights Reserved by the Regents of the University of Minnesota.
 * This file contains Original Code and/or Modifications of Original
 * Code as defined in the MAGNET Public Source License, which is
 * available at http://www.cs.umn.edu/magnet/source_license_v1.txt.
 * You may not use this file except in compliance with the License.
 *------------------------------------------------------------------------
 * Revision History:
 * 
 * Sat Sep  5 21:44:18 1998  John E. Collins
 *   Initial Entry
 */

package edu.umn.magnet.util.time;

import edu.umn.magnet.util.MagnetProperties;

import org.apache.log4j.Category;

import java.util.Date;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.text.ParsePosition;

/**
 * Static Time source for the agent as a whole.  Time may be
 * synchronized with an external source, may or may not agree with
 * "real time", and may pass at any desired multiple of "real time".
 * If this agent is running in a true multi-agent environment, then
 * all agents must be synchronized, and must somehow agree on the rate
 * at which time will pass.
 * <p>
 * Agent time is determined by three numbers: base, offset, and rate.
 * If <tt>offset=0</tt> and <tt>rate=1.0</tt>, then agent time will
 * match wall clock time.  Otherwise, if <i>W</i> is the current wall
 * time, then the agent time <i>A</i> is given by
 * <pre>
 *     <i>A</i> = base + offset + (rate * (<i>W</i> - base))
 * </pre>
 * </p>
 * @author John Collins
 * @version $Revision: 1.5.2.1 $ */
public class AgentTime
{
  private static Category log
    = Category.getInstance(AgentTime.class.getName());

  private static long base = new Date().getTime();
  private static long offset = 0L;
  private static double rate = 1.0;
  private static boolean initDone = false;

  /**
   * Sets the agent time by giving base, offset, and rate.
   */
  public static void setDate (long base, long offset, double rate)
  {
    AgentTime.base = base;
    AgentTime.offset = offset;
    AgentTime.rate = rate;
  }

  /**
   * Sets agent time by giving a Date.  Agent time will be set to that
   * time, and will pass at 1 second/second.
   */
  public static void setDate (Date newTime)
  {
    AgentTime.setDate(newTime, 1.0);
  }

  /**
   * Sets agent time by giving a Date and a rate.  Agent time will be
   * set to the given time, and will pass at rate seconds/second.
   */
  public static void setDate (Date newTime, double rate)
  {
    Date now = new Date();
    AgentTime.base = now.getTime();
    AgentTime.offset = newTime.getTime() - now.getTime();
    AgentTime.rate = rate;
  }

  /**
   * Returns the current agent time as a Date.
   */
  public static Date getDate ()
  {
    ensureInit();
    Date now = new Date();
    now.setTime(base + offset + (long)(rate * (now.getTime() - base)));
    return now;
  }

  /**
   * Returns the rate at which time is passing
   */
  public static double getRate ()
  {
    ensureInit();
    return AgentTime.rate;
  }

  /**
   * Returns the current offset value
   */
  public static long getOffset ()
  {
    ensureInit();
    return AgentTime.offset;
  }

  /**
   * Initializes the AgentTime from properties
   */
  private static void ensureInit()
  {
    if (!initDone) {
      String prop;
      Date newbase = null;
      long newoffset = 0L;
      double newrate = 1.0;

      // pick out the base
      prop = MagnetProperties.getProperty("AgentTime.Base");
      if (prop != null) {
	// we only worry about the AgentTime values if the Base is
	// specified.
	SimpleDateFormat format
	    = new SimpleDateFormat("MM/dd/yyyy HH:mm");
	newbase = format.parse(prop, new ParsePosition(0));
	if (newbase == null)
	  log.error("Date Format Parse Exception: " + prop);
      }
      prop = MagnetProperties.getProperty("AgentTime.Offset", "0");
      if (prop != null) {
	try {
	  newoffset = Long.valueOf(prop).longValue();
	  log.info("prop: AgentTime.Offset = " + newoffset);
	}
	catch (NumberFormatException nfe) {
	  log.error("Bad property spec " + prop
		    + " for AgentTime.Offset");
	}
      }
      prop = MagnetProperties.getProperty("AgentTime.Rate", "1.0");
      if (prop != null) {
	try {
	  newrate = Double.valueOf(prop).doubleValue();
	  log.info("prop: AgentTime.Rate = " + newrate);
	}
	catch (NumberFormatException nfe) {
	  log.error("Bad property spec " + prop
		    + " for AgentTime.Rate");
	}
      }
      if (newbase != null) {
	setDate(newbase.getTime(), newoffset, newrate);
      }
      initDone = true;
    }
  }

  /**
   * Waits the specified number of milliseconds, scaled by the agent's
   * clock rate.
   */
  public static void waitFor (long msec)
    throws InterruptedException
  {
    Thread.sleep((long)(msec / rate));
  }

  /**
   * Waits until the specified time, by the Agent's clock.
   */
  public static void waitUntil (Date target)
    throws InterruptedException
  {
    waitUntil(target.getTime());
  }

  /**
   * Waits until the specified time, by the Agent's clock.
   */
  public static void waitUntil (long target)
    throws InterruptedException
  {
    long interval = target - getDate().getTime();
    if (interval > 0L)
      waitFor(interval);
  }
}
